'use strict';
var context = require('context');
var router = require('router');
const PROGRAM = 'entry';

var serverModeTypes = {
	single: 0,
	multi: 1,
	parent: 2,
	child: 3,
};
var runningMode = serverModeTypes.parent;

function Extension() {
	this.robots = {};
	this.connectionState = new Buffer([0x01, 0x00, 0x00, 0x00]);
	this.masterRoomIds = [];
	this.connections = [];
	this.connectionSet = {};
	this.childServerList = {};
	this.clientTargetList = {};
}

Extension.prototype.getRobotByIndex = function(module, index) {
	var key = module + index;
	var robot = this.robots[key];
	if(!robot) {
		var component = context.getComponentByModule(PROGRAM, module);
		robot = component.createRobot(index);
		if(robot) {
			this.robots[key] = robot;
		}
	}
	return robot;
};

Extension.prototype.getRobotByRoute = function(route) {
	return this.getRobotByIndex(route.module, route.index);
};

Extension.prototype.start = function(server, connection) {
	var self = this;
	self.server = server;
	self.connection = connection;
	
	runningMode = serverModeTypes.parent;
	self.connectionSet[connection.id] = connection;
	self.connections.push(connection);
	
	var roomId = connection.handshake.query.roomId;
	if(connection.handshake.query.childServer === 'true') {
		self.childServerList[connection.id] = true;
	} else {
		connection.join(roomId);
		connection.roomId = roomId;
	}
	
	var childServerListCnt = Object.keys(self.childServerList).length;
	if(childServerListCnt > 0) {
		server.emit('mode', serverModeTypes.multi);
	} else {
		server.emit('mode', serverModeTypes.single);
	}
	
	connection.on('matchTarget', function(data) {
		if(connection.handshake.query.childServer === 'true' && data.roomId) {
			if(!connection.roomIds) {
				connection.roomIds = [];
			}
			if(connection.roomIds.indexOf(data.roomId) === -1) {
				connection.roomIds.push(data.roomId);
			}
			self.clientTargetList[data.roomId] = connection.id;
			server.to(data.roomId).emit('matched', connection.id);
		}
	});
	
	connection.on('message', function(message) {
		if(message.mode === serverModeTypes.single || self.masterRoomIds.indexOf(connection.roomId) >= 0 ) {
			if(message.type === 'utf8') {
				self.handleRemoteData(message.data);
			}
		} else {
			if(connection.handshake.query.childServer === 'true') {
				if(connection.roomIds && connection.roomIds.length > 0) {
					connection.roomIds.forEach(function(roomId) {
						server.to(roomId).emit('message', message);
					});
				}
			} else if(self.clientTargetList[connection.roomId]) {
				server.to(self.clientTargetList[connection.roomId]).emit('message', message);
			}
		}
	});
};

Extension.prototype.disconnected = function(connection) {
	this.__reset();
	//this.__closeSingleConnection(connection);
	if(connection.handshake.query.childServer === 'true') {
		if(connection.roomIds && connection.roomIds.length > 0) {
			connection.roomIds.forEach(function (roomId) {
				server.to(roomId).emit('matching');
			});
		}
		delete this.connectionSet[connection.id];
		delete this.childServerList[connection.id];

		var childServerListCnt = Object.keys(this.childServerList).length;
		if(childServerListCnt <= 0) {
			server.emit('mode', serverModeTypes.single);
		}
	} else {
		delete this.connectionSet[connection.id];
	}
};

Extension.prototype.closed = function(connection) {
	this.__reset();
	this.__closeSingleConnection(connection);
};

Extension.prototype.__closeSingleConnection = function(connection) {
	var connections = this.connections;
	var index = connections.indexOf(connection);
	if(index > -1) {
		this.connections.slice(index, 1);
		connection.disconnect(true);
	}
	this.connection = undefined;
};

Extension.prototype.close = function() {
	this.__reset();
	//this.send('disconnectHardware');
	this.__closeSingleConnection(this.connection);
};

Extension.prototype.__reset = function() {
	router.reset();
	this.robots = {};
};

Extension.prototype.stop = function() {
	this.__reset();
	this.connections = [];
};

Extension.prototype.send = function(data) {
	if(this.connection) {
		var self = this;
		var childServerListCnt = Object.keys(this.childServerList).length;
		if((runningMode === serverModeTypes.parent && childServerListCnt === 0) || runningMode === serverModeTypes.child) {
			if(this.connections.length !== 0) {
				this.connections.map(function(connection) {
					connection.emit('message', {data : data});
				});
			}
		} else if(self.masterRoomIds.length > 0) {
			self.masterRoomIds.forEach(function (masterRoomId) {
				self.server.to(masterRoomId).emit('message', {data : data});
			});
		}
	}
};

Extension.prototype.setState = function(route) {
	if(route == router.firstRoute && this.connection && this.connections.length) {
		var connectionState = this.connectionState;
		var state = route.state;
		if(state == 'connecting') {
			connectionState[3] = 0x01;
			this.send(connectionState);
		} else if(state == 'connected') {
			connectionState[3] = 0x02;
			this.send(connectionState);
		} else if(state == 'lost') {
			connectionState[3] = 0x03;
			this.send(connectionState);
		} else if(state == 'disconnected') {
			connectionState[3] = 0x04;
			this.send(connectionState);
		}
	}
};

Extension.prototype.setOption = function(option) {
};

Extension.prototype.handleRemoteData = function(utf8Data) {
	try {
		var received = JSON.parse(utf8Data);
		var data, module, robot, component;
		for(var i in received) {
			data = received[i];
			module = data.module;
			if(!module) module = 'hamster';
			robot = this.getRobotByIndex(module, data.index);
			if(robot) {
				component = context.getComponentByModule(PROGRAM, module);
				component.handleRemoteData(robot, data);
			}
		}
	} catch (e) {
	}
};

Extension.prototype.handleSensory = function(data, route) {
	var robot = this.getRobotByRoute(route);
	if(robot) {
		var component = context.getComponentByRoute(PROGRAM, route);
		var key = route.module + route.index;
		component.prepareSensoryMulti(robot.sensory, key, robot, data);
		robot.sensory.model = 0xFF;
		try {
			this.send(JSON.stringify(robot.sensory));
		} catch (e) {
		}
		component.updateSensory(robot);
	}
};

Extension.prototype.handleMotoring = function(data, route) {
	var robot = this.getRobotByRoute(route);
	if(robot) {
		var roboid = robot.roboid;
		roboid.handleMotoring(data);
		roboid.updateMotoring();
	}
};

Extension.prototype.requestMotoring = function(route) {
	var robot = this.getRobotByRoute(route);
	if(robot) {
		return robot.roboid.requestMotoring();
	}
};

Extension.prototype.reset = function(route) {
	var robot = this.getRobotByRoute(route);
	if(robot) {
		var component = context.getComponentByRoute(PROGRAM, route);
		component.reset(robot);
	}
};

module.exports.create = function() {
	return new Extension();
};