'use strict';
var context = require('context');
var router = require('router');
const PROGRAM = 'entry';

function Extension() {
	this.robots = {};
	this.connectionState = new Buffer([0x01, 0x00, 0x00, 0x00]);
}

Extension.prototype.getRobotByModule = function(module) {
	var robot = this.robots[module];
	if(!robot) {
		var component = context.getComponentByModule(PROGRAM, module);
		robot = component.createRobot(0);
		if(robot) {
			this.robots[module] = robot;
		}
	}
	return robot;
};

Extension.prototype.getRobotByRoute = function(route) {
	return this.getRobotByModule(route.module);
};

Extension.prototype.start = function(connection) {
	var self = this;
	self.connection = connection;
	connection.on('message', function(message) {
		if(message.type === 'utf8') {
			self.handleRemoteData(message.utf8Data);
		}
	});
};

Extension.prototype.stop = function() {
	this.connection = undefined;
	router.reset();
	this.robots = {};
};

Extension.prototype.send = function(data) {
	if(this.connection) {
		this.connection.send(data);
	}
};

Extension.prototype.setState = function(route) {
	if(route == router.selectedRoute && this.connection) {
		var connectionState = this.connectionState;
		var state = route.state;
		if(state == 'connecting') {
			connectionState[3] = 0x01;
			this.send(connectionState);
		} else if(state == 'connected') {
			connectionState[3] = 0x02;
			this.send(connectionState);
		} else if(state == 'lost') {
			connectionState[3] = 0x03;
			this.send(connectionState);
		} else if(state == 'disconnected') {
			connectionState[3] = 0x04;
			this.send(connectionState);
		}
	}
};

Extension.prototype.setOption = function(option) {
};

Extension.prototype.handleRemoteData = function(utf8Data) {
	var selectedRoute = router.selectedRoute;
	if(selectedRoute) {
		try {
			var data = JSON.parse(utf8Data);
			var module = data.module;
			if(!module) module = 'hamster';
			if(module == selectedRoute.module) {
				var robot = this.getRobotByModule(module);
				if(robot) {
					var component = context.getComponentByModule(PROGRAM, module);
					component.handleRemoteData(robot, data);
				}
			}
		} catch (e) {
		}
	}
};

Extension.prototype.handleSensory = function(data, route) {
	var robot = this.getRobotByRoute(route);
	if(robot) {
		var component = context.getComponentByRoute(PROGRAM, route);
		component.prepareSensory(robot, data);
		try {
			this.send(JSON.stringify(robot.sensory));
		} catch (e) {
		}
		component.updateSensory(robot);
	}
};

Extension.prototype.handleMotoring = function(data, route) {
	var robot = this.getRobotByRoute(route);
	if(robot) {
		var roboid = robot.roboid;
		roboid.handleMotoring(data);
		roboid.updateMotoring();
	}
};

Extension.prototype.requestMotoring = function(route) {
	var robot = this.getRobotByRoute(route);
	if(robot) {
		return robot.roboid.requestMotoring();
	}
};

Extension.prototype.reset = function(route) {
	if(route == router.selectedRoute) {
		var robot = this.getRobotByRoute(route);
		if(robot) {
			var component = context.getComponentByRoute(PROGRAM, route);
			component.reset(robot);
		}
	}
};

module.exports.create = function() {
	return new Extension();
};