'use strict';
var context = require('context');
var router = require('router');
var translator = require('translator');
const PROGRAM = 'scratch';
const HAMSTER = 'hamster';
const TURTLE = 'turtle';

const HEADER = { 'Content-Type': 'text/plain' };

function Extension() {
	this.robotsByModule = {};
	this.robotsByGroup = {};
}

Extension.prototype.getComponents = function() {
	if(!this.components) {
		this.components = context.getComponents(PROGRAM);
	}
	return this.components;
};

Extension.prototype.getRobotByIndex = function(group, index) {
	if(index >= 0) {
		var robot = this.robotsByGroup[group + index];
		if(!robot) {
			var route = router.findRouteByIndex(group, index);
			if(route) {
				robot = this.getRobotByRoute(route);
			}
		}
		return robot;
	}
};

Extension.prototype.getRobotByRoute = function(route) {
	var index = route.index;
	if(index >= 0) {
		var key = route.module + index;
		var robot = this.robotsByModule[key];
		if(!robot) {
			var component = context.getComponentByModule(PROGRAM, route.module);
			robot = component.createRobot(index);
			if(robot) {
				this.robotsByModule[key] = robot;
			}
		}
		this.robotsByGroup[route.group + index] = robot;
		return robot;
	}
};

Extension.prototype.start = function(express) {
	var self = this;
	var components = self.getComponents();
	var robotsByModule = self.robotsByModule;
	var robotsByGroup = self.robotsByGroup;
	var problemMessage = '';
	
	self.running = true;
	for(var i in robotsByModule) {
		robotsByModule[i].clearWaits();
	}
	var setup = function() {
		for(var i in components) {
			components[i].updateLang(translator);
		}
		problemMessage = '_problem ' + translator.translate('First robot is not connected.') + '\n';
	};
	translator.on('all-scratch-multi', function() {
		setup();
	});
	setup();
	
	if(express) {
		express.get('/reset_all', function(request, response) {
			response.writeHead(200, HEADER);
			var robot;
			for(var i in robotsByModule) {
				robot = robotsByModule[i];
				robot.clearWaits();
				robot.clearTimeouts();
				robot.reset();
			}
			router.reset();
			response.end();
		});
		
		express.get('/poll', function(request, response) {
			if(!self.running) {
				response.set('Connection', 'close');
			}
			response.writeHead(200, HEADER);
			var robot, waits, busy = '';
			for(var i in robotsByGroup) {
				robot = robotsByGroup[i];
				if(robot) {
					waits = robot.getWaits();
					if(waits.length > 0) {
						busy += ' ' + waits.join(' ');
					}
				}
			}
			if(busy.length > 0) {
				response.write('_busy' + busy + '\n');
			}
			for(var i in robotsByGroup) {
				robot = robotsByGroup[i];
				if(robot) {
					robot.writeMultiSensory(response);
				}
			}
			if(self.state != 'connected') {
				response.write(problemMessage);
			}
			response.end();
		});
		
		self.startHamster(express);
		self.startHamsterS(express);
		self.startTurtle(express);
	}
};

Extension.prototype.startHamster = function(express) {
	var self = this;
	express.get('/boardMoveForward/:id/:index', function(request, response) {
		response.writeHead(200, HEADER);
		var robot = self.getRobotByIndex(HAMSTER, parseInt(request.params.index));
		if(robot) {
			robot.boardMoveForward(request);
		}
		response.end();
	});
	
	express.get('/boardTurn/:id/:index/:direction', function(request, response) {
		response.writeHead(200, HEADER);
		var robot = self.getRobotByIndex(HAMSTER, parseInt(request.params.index));
		if(robot) {
			robot.boardTurn(request);
		}
		response.end();
	});
	
	express.get('/moveForward/:id/:index', function(request, response) {
		response.writeHead(200, HEADER);
		var robot = self.getRobotByIndex(HAMSTER, parseInt(request.params.index));
		if(robot) {
			robot.moveForward(request);
		}
		response.end();
	});
	
	express.get('/moveBackward/:id/:index', function(request, response) {
		response.writeHead(200, HEADER);
		var robot = self.getRobotByIndex(HAMSTER, parseInt(request.params.index));
		if(robot) {
			robot.moveBackward(request);
		}
		response.end();
	});
	
	express.get('/turn/:id/:index/:direction', function(request, response) {
		response.writeHead(200, HEADER);
		var robot = self.getRobotByIndex(HAMSTER, parseInt(request.params.index));
		if(robot) {
			robot.turn(request);
		}
		response.end();
	});
	
	express.get('/moveForwardForSecs/:id/:index/:sec', function(request, response) {
		response.writeHead(200, HEADER);
		var robot = self.getRobotByIndex(HAMSTER, parseInt(request.params.index));
		if(robot) {
			robot.moveForwardForSecs(request);
		}
		response.end();
	});
	
	express.get('/moveBackwardForSecs/:id/:index/:sec', function(request, response) {
		response.writeHead(200, HEADER);
		var robot = self.getRobotByIndex(HAMSTER, parseInt(request.params.index));
		if(robot) {
			robot.moveBackwardForSecs(request);
		}
		response.end();
	});
	
	express.get('/turnForSecs/:id/:index/:direction/:sec', function(request, response) {
		response.writeHead(200, HEADER);
		var robot = self.getRobotByIndex(HAMSTER, parseInt(request.params.index));
		if(robot) {
			robot.turnForSecs(request);
		}
		response.end();
	});
	
	express.get('/changeBothWheelsBy/:index/:left/:right', function(request, response) {
		response.writeHead(200, HEADER);
		var robot = self.getRobotByIndex(HAMSTER, parseInt(request.params.index));
		if(robot) {
			robot.changeBothWheelsBy(request);
		}
		response.end();
	});
	
	express.get('/setBothWheelsTo/:index/:left/:right', function(request, response) {
		response.writeHead(200, HEADER);
		var robot = self.getRobotByIndex(HAMSTER, parseInt(request.params.index));
		if(robot) {
			robot.setBothWheelsTo(request);
		}
		response.end();
	});
	
	express.get('/changeWheelBy/:index/:which/:speed', function(request, response) {
		response.writeHead(200, HEADER);
		var robot = self.getRobotByIndex(HAMSTER, parseInt(request.params.index));
		if(robot) {
			robot.changeWheelBy(request);
		}
		response.end();
	});
	
	express.get('/setWheelTo/:index/:which/:speed', function(request, response) {
		response.writeHead(200, HEADER);
		var robot = self.getRobotByIndex(HAMSTER, parseInt(request.params.index));
		if(robot) {
			robot.setWheelTo(request);
		}
		response.end();
	});
	
	express.get('/followLineUsingFloorSensor/:index/:color/:which', function(request, response) {
		response.writeHead(200, HEADER);
		var robot = self.getRobotByIndex(HAMSTER, parseInt(request.params.index));
		if(robot) {
			robot.followLineUsingFloorSensor(request);
		}
		response.end();
	});
	
	express.get('/followLineUntilIntersection/:id/:index/:color/:which', function(request, response) {
		response.writeHead(200, HEADER);
		var robot = self.getRobotByIndex(HAMSTER, parseInt(request.params.index));
		if(robot) {
			robot.followLineUntilIntersection(request);
		}
		response.end();
	});
	
	express.get('/setFollowingSpeedTo/:index/:speed', function(request, response) {
		response.writeHead(200, HEADER);
		var robot = self.getRobotByIndex(HAMSTER, parseInt(request.params.index));
		if(robot) {
			robot.setFollowingSpeedTo(request);
		}
		response.end();
	});
	
	express.get('/stop/:index', function(request, response) {
		response.writeHead(200, HEADER);
		var robot = self.getRobotByIndex(HAMSTER, parseInt(request.params.index));
		if(robot) {
			robot.stop(request);
		}
		response.end();
	});
	
	express.get('/setLedTo/:index/:which/:color', function(request, response) {
		response.writeHead(200, HEADER);
		var robot = self.getRobotByIndex(HAMSTER, parseInt(request.params.index));
		if(robot) {
			robot.setLedTo(request);
		}
		response.end();
	});
	
	express.get('/clearLed/:index/:which', function(request, response) {
		response.writeHead(200, HEADER);
		var robot = self.getRobotByIndex(HAMSTER, parseInt(request.params.index));
		if(robot) {
			robot.clearLed(request);
		}
		response.end();
	});
	
	express.get('/beep/:id/:index', function(request, response) {
		response.writeHead(200, HEADER);
		var robot = self.getRobotByIndex(HAMSTER, parseInt(request.params.index));
		if(robot) {
			robot.beep(request);
		}
		response.end();
	});
	
	express.get('/changeBuzzerBy/:index/:value', function(request, response) {
		response.writeHead(200, HEADER);
		var robot = self.getRobotByIndex(HAMSTER, parseInt(request.params.index));
		if(robot) {
			robot.changeBuzzerBy(request);
		}
		response.end();
	});
	
	express.get('/setBuzzerTo/:index/:value', function(request, response) {
		response.writeHead(200, HEADER);
		var robot = self.getRobotByIndex(HAMSTER, parseInt(request.params.index));
		if(robot) {
			robot.setBuzzerTo(request);
		}
		response.end();
	});
	
	express.get('/clearBuzzer/:index', function(request, response) {
		response.writeHead(200, HEADER);
		var robot = self.getRobotByIndex(HAMSTER, parseInt(request.params.index));
		if(robot) {
			robot.clearBuzzer(request);
		}
		response.end();
	});
	
	express.get('/playNote/:index/:note/:octave', function(request, response) {
		response.writeHead(200, HEADER);
		var robot = self.getRobotByIndex(HAMSTER, parseInt(request.params.index));
		if(robot) {
			robot.playNote(request);
		}
		response.end();
	});
	
	express.get('/playNoteFor/:id/:index/:note/:octave/:beat', function(request, response) {
		response.writeHead(200, HEADER);
		var robot = self.getRobotByIndex(HAMSTER, parseInt(request.params.index));
		if(robot) {
			robot.playNoteFor(request);
		}
		response.end();
	});
	
	express.get('/restFor/:id/:index/:beat', function(request, response) {
		response.writeHead(200, HEADER);
		var robot = self.getRobotByIndex(HAMSTER, parseInt(request.params.index));
		if(robot) {
			robot.restFor(request);
		}
		response.end();
	});
	
	express.get('/changeTempoBy/:index/:value', function(request, response) {
		response.writeHead(200, HEADER);
		var robot = self.getRobotByIndex(HAMSTER, parseInt(request.params.index));
		if(robot) {
			robot.changeTempoBy(request);
		}
		response.end();
	});
	
	express.get('/setTempoTo/:index/:value', function(request, response) {
		response.writeHead(200, HEADER);
		var robot = self.getRobotByIndex(HAMSTER, parseInt(request.params.index));
		if(robot) {
			robot.setTempoTo(request);
		}
		response.end();
	});
	
	express.get('/setPortTo/:index/:which/:value', function(request, response) {
		response.writeHead(200, HEADER);
		var robot = self.getRobotByIndex(HAMSTER, parseInt(request.params.index));
		if(robot) {
			robot.setPortTo(request);
		}
		response.end();
	});
	
	express.get('/changeOutputBy/:index/:which/:value', function(request, response) {
		response.writeHead(200, HEADER);
		var robot = self.getRobotByIndex(HAMSTER, parseInt(request.params.index));
		if(robot) {
			robot.changeOutputBy(request);
		}
		response.end();
	});
	
	express.get('/setOutputTo/:index/:which/:value', function(request, response) {
		response.writeHead(200, HEADER);
		var robot = self.getRobotByIndex(HAMSTER, parseInt(request.params.index));
		if(robot) {
			robot.setOutputTo(request);
		}
		response.end();
	});
	
	express.get('/gripper/:id/:index/:action', function(request, response) {
		response.writeHead(200, HEADER);
		var robot = self.getRobotByIndex(HAMSTER, parseInt(request.params.index));
		if(robot) {
			robot.gripper(request);
		}
		response.end();
	});
	
	express.get('/releaseGripper/:index', function(request, response) {
		response.writeHead(200, HEADER);
		var robot = self.getRobotByIndex(HAMSTER, parseInt(request.params.index));
		if(robot) {
			robot.releaseGripper(request);
		}
		response.end();
	});
};

Extension.prototype.startHamsterS = function(express) {
	var self = this;
	express.get('/sBoardMoveForward/:id/:index', function(request, response) {
		response.writeHead(200, HEADER);
		var robot = self.getRobotByIndex(HAMSTER, parseInt(request.params.index));
		if(robot) {
			robot.boardMoveForward(request);
		}
		response.end();
	});
	
	express.get('/sBoardTurn/:id/:index/:direction', function(request, response) {
		response.writeHead(200, HEADER);
		var robot = self.getRobotByIndex(HAMSTER, parseInt(request.params.index));
		if(robot) {
			robot.boardTurn(request);
		}
		response.end();
	});
	
	express.get('/sMoveForward/:id/:index', function(request, response) {
		response.writeHead(200, HEADER);
		var robot = self.getRobotByIndex(HAMSTER, parseInt(request.params.index));
		if(robot) {
			robot.moveForward(request);
		}
		response.end();
	});
	
	express.get('/sMoveBackward/:id/:index', function(request, response) {
		response.writeHead(200, HEADER);
		var robot = self.getRobotByIndex(HAMSTER, parseInt(request.params.index));
		if(robot) {
			robot.moveBackward(request);
		}
		response.end();
	});
	
	express.get('/sTurn/:id/:index/:direction', function(request, response) {
		response.writeHead(200, HEADER);
		var robot = self.getRobotByIndex(HAMSTER, parseInt(request.params.index));
		if(robot) {
			robot.turn(request);
		}
		response.end();
	});
	
	express.get('/sMoveForwardUnit/:id/:index/:value/:unit', function(request, response) {
		response.writeHead(200, HEADER);
		var robot = self.getRobotByIndex(HAMSTER, parseInt(request.params.index));
		if(robot) {
			robot.moveForwardUnit(request);
		}
		response.end();
	});

	express.get('/sMoveBackwardUnit/:id/:index/:value/:unit', function(request, response) {
		response.writeHead(200, HEADER);
		var robot = self.getRobotByIndex(HAMSTER, parseInt(request.params.index));
		if(robot) {
			robot.moveBackwardUnit(request);
		}
		response.end();
	});

	express.get('/sTurnUnitInPlace/:id/:index/:direction/:value/:unit', function(request, response) {
		response.writeHead(200, HEADER);
		var robot = self.getRobotByIndex(HAMSTER, parseInt(request.params.index));
		if(robot) {
			robot.turnUnitInPlace(request);
		}
		response.end();
	});
	
	express.get('/sPivotAroundWheelUnitInDirection/:id/:index/:wheel/:value/:unit/:forward', function(request, response) {
		response.writeHead(200, HEADER);
		var robot = self.getRobotByIndex(HAMSTER, parseInt(request.params.index));
		if(robot) {
			robot.pivotAroundWheelUnitInDirection(request);
		}
		response.end();
	});
	
	express.get('/sTurnBodyUnitWithRadiusInDirection/:id/:index/:direction/:value/:unit/:radius/:forward', function(request, response) {
		response.writeHead(200, HEADER);
		var robot = self.getRobotByIndex(HAMSTER, parseInt(request.params.index));
		if(robot) {
			robot.turnBodyUnitWithRadiusInDirection(request);
		}
		response.end();
	});
	
	express.get('/sPivotAroundUnitInDirection/:id/:index/:part/:value/:unit/:forward', function(request, response) {
		response.writeHead(200, HEADER);
		var robot = self.getRobotByIndex(HAMSTER, parseInt(request.params.index));
		if(robot) {
			robot.pivotAroundUnitInDirection(request);
		}
		response.end();
	});
	
	express.get('/sTurnUnitWithRadiusInDirection/:id/:index/:part/:direction/:value/:unit/:radius/:forward', function(request, response) {
		response.writeHead(200, HEADER);
		var robot = self.getRobotByIndex(HAMSTER, parseInt(request.params.index));
		if(robot) {
			robot.turnUnitWithRadiusInDirection(request);
		}
		response.end();
	});
	
	express.get('/sChangeBothWheelsBy/:index/:left/:right', function(request, response) {
		response.writeHead(200, HEADER);
		var robot = self.getRobotByIndex(HAMSTER, parseInt(request.params.index));
		if(robot) {
			robot.changeBothWheelsBy(request);
		}
		response.end();
	});
	
	express.get('/sSetBothWheelsTo/:index/:left/:right', function(request, response) {
		response.writeHead(200, HEADER);
		var robot = self.getRobotByIndex(HAMSTER, parseInt(request.params.index));
		if(robot) {
			robot.setBothWheelsTo(request);
		}
		response.end();
	});
	
	express.get('/sChangeWheelBy/:index/:which/:speed', function(request, response) {
		response.writeHead(200, HEADER);
		var robot = self.getRobotByIndex(HAMSTER, parseInt(request.params.index));
		if(robot) {
			robot.changeWheelBy(request);
		}
		response.end();
	});
	
	express.get('/sSetWheelTo/:index/:which/:speed', function(request, response) {
		response.writeHead(200, HEADER);
		var robot = self.getRobotByIndex(HAMSTER, parseInt(request.params.index));
		if(robot) {
			robot.setWheelTo(request);
		}
		response.end();
	});
	
	express.get('/sFollowLineUsingFloorSensor/:index/:color/:which', function(request, response) {
		response.writeHead(200, HEADER);
		var robot = self.getRobotByIndex(HAMSTER, parseInt(request.params.index));
		if(robot) {
			robot.followLineUsingFloorSensor(request);
		}
		response.end();
	});
	
	express.get('/sFollowLineUntilIntersection/:id/:index/:color/:which', function(request, response) {
		response.writeHead(200, HEADER);
		var robot = self.getRobotByIndex(HAMSTER, parseInt(request.params.index));
		if(robot) {
			robot.followLineUntilIntersection(request);
		}
		response.end();
	});
	
	express.get('/sSetFollowingSpeedTo/:index/:speed', function(request, response) {
		response.writeHead(200, HEADER);
		var robot = self.getRobotByIndex(HAMSTER, parseInt(request.params.index));
		if(robot) {
			robot.setFollowingSpeedTo(request);
		}
		response.end();
	});
	
	express.get('/sSetFollowingGainTo/:index/:gain', function(request, response) {
		response.writeHead(200, HEADER);
		var robot = self.getRobotByIndex(HAMSTER, parseInt(request.params.index));
		if(robot) {
			robot.setFollowingGainTo(request);
		}
		response.end();
	});
	
	express.get('/sStop/:index', function(request, response) {
		response.writeHead(200, HEADER);
		var robot = self.getRobotByIndex(HAMSTER, parseInt(request.params.index));
		if(robot) {
			robot.stop(request);
		}
		response.end();
	});
	
	express.get('/sSetLedTo/:index/:which/:color', function(request, response) {
		response.writeHead(200, HEADER);
		var robot = self.getRobotByIndex(HAMSTER, parseInt(request.params.index));
		if(robot) {
			robot.setLedTo(request);
		}
		response.end();
	});
	
	express.get('/sChangeLedByRGB/:index/:which/:red/:green/:blue', function(request, response) {
		response.writeHead(200, HEADER);
		var robot = self.getRobotByIndex(HAMSTER, parseInt(request.params.index));
		if(robot) {
			robot.changeLedByRGB(request);
		}
		response.end();
	});
	
	express.get('/sSetLedToRGB/:index/:which/:red/:green/:blue', function(request, response) {
		response.writeHead(200, HEADER);
		var robot = self.getRobotByIndex(HAMSTER, parseInt(request.params.index));
		if(robot) {
			robot.setLedToRGB(request);
		}
		response.end();
	});
	
	express.get('/sClearLed/:index/:which', function(request, response) {
		response.writeHead(200, HEADER);
		var robot = self.getRobotByIndex(HAMSTER, parseInt(request.params.index));
		if(robot) {
			robot.clearLed(request);
		}
		response.end();
	});
	
	express.get('/sPlaySound/:index/:sound', function(request, response) {
		response.writeHead(200, HEADER);
		var robot = self.getRobotByIndex(HAMSTER, parseInt(request.params.index));
		if(robot) {
			robot.playSound(request);
		}
		response.end();
	});
	
	express.get('/sPlaySoundTimes/:index/:sound/:count', function(request, response) {
		response.writeHead(200, HEADER);
		var robot = self.getRobotByIndex(HAMSTER, parseInt(request.params.index));
		if(robot) {
			robot.playSoundTimes(request);
		}
		response.end();
	});
	
	express.get('/sPlaySoundTimesUntilDone/:id/:index/:sound/:count', function(request, response) {
		response.writeHead(200, HEADER);
		var robot = self.getRobotByIndex(HAMSTER, parseInt(request.params.index));
		if(robot) {
			robot.playSoundTimesUntilDone(request);
		}
		response.end();
	});
	
	express.get('/sChangeBuzzerBy/:index/:value', function(request, response) {
		response.writeHead(200, HEADER);
		var robot = self.getRobotByIndex(HAMSTER, parseInt(request.params.index));
		if(robot) {
			robot.changeBuzzerBy(request);
		}
		response.end();
	});
	
	express.get('/sSetBuzzerTo/:index/:value', function(request, response) {
		response.writeHead(200, HEADER);
		var robot = self.getRobotByIndex(HAMSTER, parseInt(request.params.index));
		if(robot) {
			robot.setBuzzerTo(request);
		}
		response.end();
	});
	
	express.get('/sClearSound/:index', function(request, response) {
		response.writeHead(200, HEADER);
		var robot = self.getRobotByIndex(HAMSTER, parseInt(request.params.index));
		if(robot) {
			robot.clearSound(request);
		}
		response.end();
	});
	
	express.get('/sPlayNote/:index/:note/:octave', function(request, response) {
		response.writeHead(200, HEADER);
		var robot = self.getRobotByIndex(HAMSTER, parseInt(request.params.index));
		if(robot) {
			robot.playNote(request);
		}
		response.end();
	});
	
	express.get('/sPlayNoteFor/:id/:index/:note/:octave/:beat', function(request, response) {
		response.writeHead(200, HEADER);
		var robot = self.getRobotByIndex(HAMSTER, parseInt(request.params.index));
		if(robot) {
			robot.playNoteFor(request);
		}
		response.end();
	});
	
	express.get('/sRestFor/:id/:index/:beat', function(request, response) {
		response.writeHead(200, HEADER);
		var robot = self.getRobotByIndex(HAMSTER, parseInt(request.params.index));
		if(robot) {
			robot.restFor(request);
		}
		response.end();
	});
	
	express.get('/sChangeTempoBy/:index/:value', function(request, response) {
		response.writeHead(200, HEADER);
		var robot = self.getRobotByIndex(HAMSTER, parseInt(request.params.index));
		if(robot) {
			robot.changeTempoBy(request);
		}
		response.end();
	});
	
	express.get('/sSetTempoTo/:index/:value', function(request, response) {
		response.writeHead(200, HEADER);
		var robot = self.getRobotByIndex(HAMSTER, parseInt(request.params.index));
		if(robot) {
			robot.setTempoTo(request);
		}
		response.end();
	});
	
	express.get('/sSetPortTo/:index/:which/:value', function(request, response) {
		response.writeHead(200, HEADER);
		var robot = self.getRobotByIndex(HAMSTER, parseInt(request.params.index));
		if(robot) {
			robot.setPortTo(request);
		}
		response.end();
	});
	
	express.get('/sChangeOutputBy/:index/:which/:value', function(request, response) {
		response.writeHead(200, HEADER);
		var robot = self.getRobotByIndex(HAMSTER, parseInt(request.params.index));
		if(robot) {
			robot.changeOutputBy(request);
		}
		response.end();
	});
	
	express.get('/sSetOutputTo/:index/:which/:value', function(request, response) {
		response.writeHead(200, HEADER);
		var robot = self.getRobotByIndex(HAMSTER, parseInt(request.params.index));
		if(robot) {
			robot.setOutputTo(request);
		}
		response.end();
	});
	
	express.get('/sGripper/:id/:index/:action', function(request, response) {
		response.writeHead(200, HEADER);
		var robot = self.getRobotByIndex(HAMSTER, parseInt(request.params.index));
		if(robot) {
			robot.gripper(request);
		}
		response.end();
	});
	
	express.get('/sReleaseGripper/:index', function(request, response) {
		response.writeHead(200, HEADER);
		var robot = self.getRobotByIndex(HAMSTER, parseInt(request.params.index));
		if(robot) {
			robot.releaseGripper(request);
		}
		response.end();
	});
	
	express.get('/sWriteSerial/:id/:index/:mode/:value', function(request, response) {
		response.writeHead(200, HEADER);
		var robot = self.getRobotByIndex(HAMSTER, parseInt(request.params.index));
		if(robot) {
			robot.writeSerial(request);
		}
		response.end();
	});
	
	express.get('/sReadSerialUntil/:id/:index/:delimiter', function(request, response) {
		response.writeHead(200, HEADER);
		var robot = self.getRobotByIndex(HAMSTER, parseInt(request.params.index));
		if(robot) {
			robot.readSerialUntil(request);
		}
		response.end();
	});
	
	express.get('/sSetSerialRateTo/:index/:baud', function(request, response) {
		response.writeHead(200, HEADER);
		var robot = self.getRobotByIndex(HAMSTER, parseInt(request.params.index));
		if(robot) {
			robot.setSerialRateTo(request);
		}
		response.end();
	});
};

Extension.prototype.startTurtle = function(express) {
	var self = this;
	express.get('/turtleMoveForward/:id/:index', function(request, response) {
		response.writeHead(200, HEADER);
		var robot = self.getRobotByIndex(TURTLE, parseInt(request.params.index));
		if(robot) {
			robot.moveForward(request);
		}
		response.end();
	});
	
	express.get('/turtleMoveBackward/:id/:index', function(request, response) {
		response.writeHead(200, HEADER);
		var robot = self.getRobotByIndex(TURTLE, parseInt(request.params.index));
		if(robot) {
			robot.moveBackward(request);
		}
		response.end();
	});
	
	express.get('/turtleTurn/:id/:index/:direction', function(request, response) {
		response.writeHead(200, HEADER);
		var robot = self.getRobotByIndex(TURTLE, parseInt(request.params.index));
		if(robot) {
			robot.turn(request);
		}
		response.end();
	});
	
	express.get('/turtleMoveForwardUnit/:id/:index/:value/:unit', function(request, response) {
		response.writeHead(200, HEADER);
		var robot = self.getRobotByIndex(TURTLE, parseInt(request.params.index));
		if(robot) {
			robot.moveForwardUnit(request);
		}
		response.end();
	});
	
	express.get('/turtleMoveBackwardUnit/:id/:index/:value/:unit', function(request, response) {
		response.writeHead(200, HEADER);
		var robot = self.getRobotByIndex(TURTLE, parseInt(request.params.index));
		if(robot) {
			robot.moveBackwardUnit(request);
		}
		response.end();
	});
	
	express.get('/turtleTurnUnitInPlace/:id/:index/:direction/:value/:unit', function(request, response) {
		response.writeHead(200, HEADER);
		var robot = self.getRobotByIndex(TURTLE, parseInt(request.params.index));
		if(robot) {
			robot.turnUnitInPlace(request);
		}
		response.end();
	});
	
	express.get('/turtleTurnUnitWithRadiusInDirection/:id/:index/:direction/:value/:unit/:radius/:head', function(request, response) {
		response.writeHead(200, HEADER);
		var robot = self.getRobotByIndex(TURTLE, parseInt(request.params.index));
		if(robot) {
			robot.turnUnitWithRadiusInDirection(request);
		}
		response.end();
	});
	
	express.get('/turtlePivotAroundWheelUnitInDirection/:id/:index/:wheel/:value/:unit/:head', function(request, response) {
		response.writeHead(200, HEADER);
		var robot = self.getRobotByIndex(TURTLE, parseInt(request.params.index));
		if(robot) {
			robot.pivotAroundWheelUnitInDirection(request);
		}
		response.end();
	});
	
	express.get('/turtleChangeWheelsByLeftRight/:index/:left/:right', function(request, response) {
		response.writeHead(200, HEADER);
		var robot = self.getRobotByIndex(TURTLE, parseInt(request.params.index));
		if(robot) {
			robot.changeWheelsByLeftRight(request);
		}
		response.end();
	});
	
	express.get('/turtleSetWheelsToLeftRight/:index/:left/:right', function(request, response) {
		response.writeHead(200, HEADER);
		var robot = self.getRobotByIndex(TURTLE, parseInt(request.params.index));
		if(robot) {
			robot.setWheelsToLeftRight(request);
		}
		response.end();
	});
	
	express.get('/turtleChangeWheelBy/:index/:wheel/:speed', function(request, response) {
		response.writeHead(200, HEADER);
		var robot = self.getRobotByIndex(TURTLE, parseInt(request.params.index));
		if(robot) {
			robot.changeWheelBy(request);
		}
		response.end();
	});
	
	express.get('/turtleSetWheelTo/:index/:wheel/:speed', function(request, response) {
		response.writeHead(200, HEADER);
		var robot = self.getRobotByIndex(TURTLE, parseInt(request.params.index));
		if(robot) {
			robot.setWheelTo(request);
		}
		response.end();
	});
	
	express.get('/turtleFollowLine/:index/:color', function(request, response) {
		response.writeHead(200, HEADER);
		var robot = self.getRobotByIndex(TURTLE, parseInt(request.params.index));
		if(robot) {
			robot.followLine(request);
		}
		response.end();
	});
	
	express.get('/turtleFollowLineUntil/:id/:index/:color', function(request, response) {
		response.writeHead(200, HEADER);
		var robot = self.getRobotByIndex(TURTLE, parseInt(request.params.index));
		if(robot) {
			robot.followLineUntil(request);
		}
		response.end();
	});
	
	express.get('/turtleFollowLineUntilBlack/:id/:index/:color', function(request, response) {
		response.writeHead(200, HEADER);
		var robot = self.getRobotByIndex(TURTLE, parseInt(request.params.index));
		if(robot) {
			robot.followLineUntilBlack(request);
		}
		response.end();
	});
	
	express.get('/turtleCrossIntersection/:id/:index', function(request, response) {
		response.writeHead(200, HEADER);
		var robot = self.getRobotByIndex(TURTLE, parseInt(request.params.index));
		if(robot) {
			robot.crossIntersection(request);
		}
		response.end();
	});
	
	express.get('/turtleTurnAtIntersection/:id/:index/:direction', function(request, response) {
		response.writeHead(200, HEADER);
		var robot = self.getRobotByIndex(TURTLE, parseInt(request.params.index));
		if(robot) {
			robot.turnAtIntersection(request);
		}
		response.end();
	});
	
	express.get('/turtleSetFollowingSpeedTo/:index/:speed', function(request, response) {
		response.writeHead(200, HEADER);
		var robot = self.getRobotByIndex(TURTLE, parseInt(request.params.index));
		if(robot) {
			robot.setFollowingSpeedTo(request);
		}
		response.end();
	});
	
	express.get('/turtleStop/:index', function(request, response) {
		response.writeHead(200, HEADER);
		var robot = self.getRobotByIndex(TURTLE, parseInt(request.params.index));
		if(robot) {
			robot.stop(request);
		}
		response.end();
	});
	
	express.get('/turtleSetHeadLedTo/:index/:color', function(request, response) {
		response.writeHead(200, HEADER);
		var robot = self.getRobotByIndex(TURTLE, parseInt(request.params.index));
		if(robot) {
			robot.setHeadLedTo(request);
		}
		response.end();
	});
	
	express.get('/turtleChangeHeadLedByRGB/:index/:red/:green/:blue', function(request, response) {
		response.writeHead(200, HEADER);
		var robot = self.getRobotByIndex(TURTLE, parseInt(request.params.index));
		if(robot) {
			robot.changeHeadLedByRGB(request);
		}
		response.end();
	});
	
	express.get('/turtleSetHeadLedToRGB/:index/:red/:green/:blue', function(request, response) {
		response.writeHead(200, HEADER);
		var robot = self.getRobotByIndex(TURTLE, parseInt(request.params.index));
		if(robot) {
			robot.setHeadLedToRGB(request);
		}
		response.end();
	});
	
	express.get('/turtleClearHeadLed/:index', function(request, response) {
		response.writeHead(200, HEADER);
		var robot = self.getRobotByIndex(TURTLE, parseInt(request.params.index));
		if(robot) {
			robot.clearHeadLed(request);
		}
		response.end();
	});
	
	express.get('/turtlePlaySound/:index/:sound', function(request, response) {
		response.writeHead(200, HEADER);
		var robot = self.getRobotByIndex(TURTLE, parseInt(request.params.index));
		if(robot) {
			robot.playSound(request);
		}
		response.end();
	});
	
	express.get('/turtlePlaySoundTimes/:index/:sound/:count', function(request, response) {
		response.writeHead(200, HEADER);
		var robot = self.getRobotByIndex(TURTLE, parseInt(request.params.index));
		if(robot) {
			robot.playSoundTimes(request);
		}
		response.end();
	});
	
	express.get('/turtlePlaySoundTimesUntilDone/:id/:index/:sound/:count', function(request, response) {
		response.writeHead(200, HEADER);
		var robot = self.getRobotByIndex(TURTLE, parseInt(request.params.index));
		if(robot) {
			robot.playSoundTimesUntilDone(request);
		}
		response.end();
	});
	
	express.get('/turtleChangeBuzzerBy/:index/:hz', function(request, response) {
		response.writeHead(200, HEADER);
		var robot = self.getRobotByIndex(TURTLE, parseInt(request.params.index));
		if(robot) {
			robot.changeBuzzerBy(request);
		}
		response.end();
	});
	
	express.get('/turtleSetBuzzerTo/:index/:hz', function(request, response) {
		response.writeHead(200, HEADER);
		var robot = self.getRobotByIndex(TURTLE, parseInt(request.params.index));
		if(robot) {
			robot.setBuzzerTo(request);
		}
		response.end();
	});
	
	express.get('/turtleClearSound/:index', function(request, response) {
		response.writeHead(200, HEADER);
		var robot = self.getRobotByIndex(TURTLE, parseInt(request.params.index));
		if(robot) {
			robot.clearSound(request);
		}
		response.end();
	});
	
	express.get('/turtlePlayNote/:index/:note/:octave', function(request, response) {
		response.writeHead(200, HEADER);
		var robot = self.getRobotByIndex(TURTLE, parseInt(request.params.index));
		if(robot) {
			robot.playNote(request);
		}
		response.end();
	});
	
	express.get('/turtlePlayNoteForBeats/:id/:index/:note/:octave/:beat', function(request, response) {
		response.writeHead(200, HEADER);
		var robot = self.getRobotByIndex(TURTLE, parseInt(request.params.index));
		if(robot) {
			robot.playNoteForBeats(request);
		}
		response.end();
	});
	
	express.get('/turtleRestForBeats/:id/:index/:beat', function(request, response) {
		response.writeHead(200, HEADER);
		var robot = self.getRobotByIndex(TURTLE, parseInt(request.params.index));
		if(robot) {
			robot.restForBeats(request);
		}
		response.end();
	});
	
	express.get('/turtleChangeTempoBy/:index/:bpm', function(request, response) {
		response.writeHead(200, HEADER);
		var robot = self.getRobotByIndex(TURTLE, parseInt(request.params.index));
		if(robot) {
			robot.changeTempoBy(request);
		}
		response.end();
	});
	
	express.get('/turtleSetTempoTo/:index/:bpm', function(request, response) {
		response.writeHead(200, HEADER);
		var robot = self.getRobotByIndex(TURTLE, parseInt(request.params.index));
		if(robot) {
			robot.setTempoTo(request);
		}
		response.end();
	});
};

Extension.prototype.stop = function() {
	this.running = false;
	translator.remove('all-scratch-multi');
	router.reset();
	this.robotsByModule = {};
	this.robotsByGroup = {};
};

Extension.prototype.setState = function(route) {
	if(route == router.firstRoute) {
		this.state = route.state;
	}
};

Extension.prototype.setOption = function(option) {
};

Extension.prototype.handleSensory = function(data, route) {
	var robot = this.getRobotByRoute(route);
	if(robot) {
		robot.handleSensory(data);
	}
};

Extension.prototype.handleMotoring = function(data, route) {
	var robot = this.getRobotByRoute(route);
	if(robot) {
		var roboid = robot.roboid;
		roboid.handleMotoring(data);
		roboid.updateMotoring();
	}
};

Extension.prototype.requestMotoring = function(route) {
	var robot = this.getRobotByRoute(route);
	if(robot) {
		return robot.roboid.requestMotoring();
	}
};

Extension.prototype.reset = function(route) {
	var robot = this.getRobotByRoute(route);
	if(robot) {
		robot.roboid.reset();
		robot.motoring.reset();
	}
};

module.exports.create = function() {
	return new Extension();
};