'use strict';
var context = require('context');
var router = require('router');
var translator = require('translator');
const PROGRAM = 'scratch';
const HAMSTER = 'hamster';
const TURTLE = 'turtle';

const HEADER = { 'Content-Type': 'text/plain' };

function Extension() {
	this.robotsByModule = {};
	this.robotsByGroup = {};
}

Extension.prototype.getComponents = function() {
	if(!this.components) {
		this.components = context.getComponents(PROGRAM);
	}
	return this.components;
};

Extension.prototype.getRobotByGroup = function(group) {
	return this.robotsByGroup[group];
};

Extension.prototype.getRobotByRoute = function(route) {
	var robot = this.robotsByModule[route.module];
	if(!robot) {
		var component = context.getComponentByModule(PROGRAM, route.module);
		robot = component.createRobot(0);
		if(robot) {
			this.robotsByModule[route.module] = robot;
		}
	}
	this.robotsByGroup[route.group] = robot;
	return robot;
};

Extension.prototype.start = function(express) {
	var self = this;
	var components = self.getComponents();
	var robotsByModule = self.robotsByModule;
	var robotsByGroup = self.robotsByGroup;
	var problemMessage = '';
	
	self.running = true;
	for(var i in robotsByModule) {
		robotsByModule[i].clearWaits();
	}
	var setup = function() {
		for(var i in components) {
			components[i].updateLang(translator);
		}
		problemMessage = '_problem ' + translator.translate('First robot is not connected.') + '\n';
	};
	translator.on('all-scratch', function() {
		setup();
	});
	setup();
	
	if(express) {
		express.get('/reset_all', function(request, response) {
			response.writeHead(200, HEADER);
			var robot;
			for(var i in robotsByModule) {
				robot = robotsByModule[i];
				robot.clearWaits();
				robot.clearTimeouts();
				robot.reset();
			}
			router.reset();
			response.end();
		});
		
		express.get('/poll', function(request, response) {
			if(!self.running) {
				response.set('Connection', 'close');
			}
			response.writeHead(200, HEADER);
			var robot, waits, busy = '';
			for(var i in robotsByGroup) {
				robot = robotsByGroup[i];
				if(robot) {
					waits = robot.getWaits();
					if(waits.length > 0) {
						busy += ' ' + waits.join(' ');
					}
				}
			}
			if(busy.length > 0) {
				response.write('_busy' + busy + '\n');
			}
			for(var i in robotsByGroup) {
				robot = robotsByGroup[i];
				if(robot) {
					robot.writeSingleSensory(response);
				}
			}
			if(self.state != 'connected') {
				response.write(problemMessage);
			}
			response.end();
		});
		
		self.startHamster(express);
		self.startHamsterS(express);
		self.startTurtle(express);
	}
};

Extension.prototype.startHamster = function(express) {
	var self = this;
	express.get('/boardMoveForward/:id', function(request, response) {
		response.writeHead(200, HEADER);
		var robot = self.getRobotByGroup(HAMSTER);
		if(robot) {
			robot.boardMoveForward(request);
		}
		response.end();
	});
	
	express.get('/boardTurn/:id/:direction', function(request, response) {
		response.writeHead(200, HEADER);
		var robot = self.getRobotByGroup(HAMSTER);
		if(robot) {
			robot.boardTurn(request);
		}
		response.end();
	});
	
	express.get('/moveForward/:id', function(request, response) {
		response.writeHead(200, HEADER);
		var robot = self.getRobotByGroup(HAMSTER);
		if(robot) {
			robot.moveForward(request);
		}
		response.end();
	});
	
	express.get('/moveBackward/:id', function(request, response) {
		response.writeHead(200, HEADER);
		var robot = self.getRobotByGroup(HAMSTER);
		if(robot) {
			robot.moveBackward(request);
		}
		response.end();
	});
	
	express.get('/turn/:id/:direction', function(request, response) {
		response.writeHead(200, HEADER);
		var robot = self.getRobotByGroup(HAMSTER);
		if(robot) {
			robot.turn(request);
		}
		response.end();
	});
	
	express.get('/moveForwardForSecs/:id/:sec', function(request, response) {
		response.writeHead(200, HEADER);
		var robot = self.getRobotByGroup(HAMSTER);
		if(robot) {
			robot.moveForwardForSecs(request);
		}
		response.end();
	});
	
	express.get('/moveBackwardForSecs/:id/:sec', function(request, response) {
		response.writeHead(200, HEADER);
		var robot = self.getRobotByGroup(HAMSTER);
		if(robot) {
			robot.moveBackwardForSecs(request);
		}
		response.end();
	});
	
	express.get('/turnForSecs/:id/:direction/:sec', function(request, response) {
		response.writeHead(200, HEADER);
		var robot = self.getRobotByGroup(HAMSTER);
		if(robot) {
			robot.turnForSecs(request);
		}
		response.end();
	});
	
	express.get('/changeBothWheelsBy/:left/:right', function(request, response) {
		response.writeHead(200, HEADER);
		var robot = self.getRobotByGroup(HAMSTER);
		if(robot) {
			robot.changeBothWheelsBy(request);
		}
		response.end();
	});
	
	express.get('/setBothWheelsTo/:left/:right', function(request, response) {
		response.writeHead(200, HEADER);
		var robot = self.getRobotByGroup(HAMSTER);
		if(robot) {
			robot.setBothWheelsTo(request);
		}
		response.end();
	});
	
	express.get('/changeWheelBy/:which/:speed', function(request, response) {
		response.writeHead(200, HEADER);
		var robot = self.getRobotByGroup(HAMSTER);
		if(robot) {
			robot.changeWheelBy(request);
		}
		response.end();
	});
	
	express.get('/setWheelTo/:which/:speed', function(request, response) {
		response.writeHead(200, HEADER);
		var robot = self.getRobotByGroup(HAMSTER);
		if(robot) {
			robot.setWheelTo(request);
		}
		response.end();
	});
	
	express.get('/followLineUsingFloorSensor/:color/:which', function(request, response) {
		response.writeHead(200, HEADER);
		var robot = self.getRobotByGroup(HAMSTER);
		if(robot) {
			robot.followLineUsingFloorSensor(request);
		}
		response.end();
	});
	
	express.get('/followLineUntilIntersection/:id/:color/:which', function(request, response) {
		response.writeHead(200, HEADER);
		var robot = self.getRobotByGroup(HAMSTER);
		if(robot) {
			robot.followLineUntilIntersection(request);
		}
		response.end();
	});
	
	express.get('/setFollowingSpeedTo/:speed', function(request, response) {
		response.writeHead(200, HEADER);
		var robot = self.getRobotByGroup(HAMSTER);
		if(robot) {
			robot.setFollowingSpeedTo(request);
		}
		response.end();
	});
	
	express.get('/stop', function(request, response) {
		response.writeHead(200, HEADER);
		var robot = self.getRobotByGroup(HAMSTER);
		if(robot) {
			robot.stop(request);
		}
		response.end();
	});
	
	express.get('/setLedTo/:which/:color', function(request, response) {
		response.writeHead(200, HEADER);
		var robot = self.getRobotByGroup(HAMSTER);
		if(robot) {
			robot.setLedTo(request);
		}
		response.end();
	});
	
	express.get('/clearLed/:which', function(request, response) {
		response.writeHead(200, HEADER);
		var robot = self.getRobotByGroup(HAMSTER);
		if(robot) {
			robot.clearLed(request);
		}
		response.end();
	});
	
	express.get('/beep/:id', function(request, response) {
		response.writeHead(200, HEADER);
		var robot = self.getRobotByGroup(HAMSTER);
		if(robot) {
			robot.beep(request);
		}
		response.end();
	});
	
	express.get('/changeBuzzerBy/:value', function(request, response) {
		response.writeHead(200, HEADER);
		var robot = self.getRobotByGroup(HAMSTER);
		if(robot) {
			robot.changeBuzzerBy(request);
		}
		response.end();
	});
	
	express.get('/setBuzzerTo/:value', function(request, response) {
		response.writeHead(200, HEADER);
		var robot = self.getRobotByGroup(HAMSTER);
		if(robot) {
			robot.setBuzzerTo(request);
		}
		response.end();
	});
	
	express.get('/clearBuzzer', function(request, response) {
		response.writeHead(200, HEADER);
		var robot = self.getRobotByGroup(HAMSTER);
		if(robot) {
			robot.clearBuzzer(request);
		}
		response.end();
	});
	
	express.get('/playNote/:note/:octave', function(request, response) {
		response.writeHead(200, HEADER);
		var robot = self.getRobotByGroup(HAMSTER);
		if(robot) {
			robot.playNote(request);
		}
		response.end();
	});
	
	express.get('/playNoteFor/:id/:note/:octave/:beat', function(request, response) {
		response.writeHead(200, HEADER);
		var robot = self.getRobotByGroup(HAMSTER);
		if(robot) {
			robot.playNoteFor(request);
		}
		response.end();
	});
	
	express.get('/restFor/:id/:beat', function(request, response) {
		response.writeHead(200, HEADER);
		var robot = self.getRobotByGroup(HAMSTER);
		if(robot) {
			robot.restFor(request);
		}
		response.end();
	});
	
	express.get('/changeTempoBy/:value', function(request, response) {
		response.writeHead(200, HEADER);
		var robot = self.getRobotByGroup(HAMSTER);
		if(robot) {
			robot.changeTempoBy(request);
		}
		response.end();
	});
	
	express.get('/setTempoTo/:value', function(request, response) {
		response.writeHead(200, HEADER);
		var robot = self.getRobotByGroup(HAMSTER);
		if(robot) {
			robot.setTempoTo(request);
		}
		response.end();
	});
	
	express.get('/setPortTo/:which/:value', function(request, response) {
		response.writeHead(200, HEADER);
		var robot = self.getRobotByGroup(HAMSTER);
		if(robot) {
			robot.setPortTo(request);
		}
		response.end();
	});
	
	express.get('/changeOutputBy/:which/:value', function(request, response) {
		response.writeHead(200, HEADER);
		var robot = self.getRobotByGroup(HAMSTER);
		if(robot) {
			robot.changeOutputBy(request);
		}
		response.end();
	});
	
	express.get('/setOutputTo/:which/:value', function(request, response) {
		response.writeHead(200, HEADER);
		var robot = self.getRobotByGroup(HAMSTER);
		if(robot) {
			robot.setOutputTo(request);
		}
		response.end();
	});
	
	express.get('/gripper/:id/:action', function(request, response) {
		response.writeHead(200, HEADER);
		var robot = self.getRobotByGroup(HAMSTER);
		if(robot) {
			robot.gripper(request);
		}
		response.end();
	});
	
	express.get('/releaseGripper', function(request, response) {
		response.writeHead(200, HEADER);
		var robot = self.getRobotByGroup(HAMSTER);
		if(robot) {
			robot.releaseGripper(request);
		}
		response.end();
	});
};

Extension.prototype.startHamsterS = function(express) {
	var self = this;
	express.get('/sBoardMoveForward/:id', function(request, response) {
		response.writeHead(200, HEADER);
		var robot = self.getRobotByGroup(HAMSTER);
		if(robot) {
			robot.boardMoveForward(request);
		}
		response.end();
	});
	
	express.get('/sBoardTurn/:id/:direction', function(request, response) {
		response.writeHead(200, HEADER);
		var robot = self.getRobotByGroup(HAMSTER);
		if(robot) {
			robot.boardTurn(request);
		}
		response.end();
	});
	
	express.get('/sMoveForward/:id', function(request, response) {
		response.writeHead(200, HEADER);
		var robot = self.getRobotByGroup(HAMSTER);
		if(robot) {
			robot.moveForward(request);
		}
		response.end();
	});
	
	express.get('/sMoveBackward/:id', function(request, response) {
		response.writeHead(200, HEADER);
		var robot = self.getRobotByGroup(HAMSTER);
		if(robot) {
			robot.moveBackward(request);
		}
		response.end();
	});
	
	express.get('/sTurn/:id/:direction', function(request, response) {
		response.writeHead(200, HEADER);
		var robot = self.getRobotByGroup(HAMSTER);
		if(robot) {
			robot.turn(request);
		}
		response.end();
	});
	
	express.get('/sMoveForwardUnit/:id/:value/:unit', function(request, response) {
		response.writeHead(200, HEADER);
		var robot = self.getRobotByGroup(HAMSTER);
		if(robot) {
			robot.moveForwardUnit(request);
		}
		response.end();
	});

	express.get('/sMoveBackwardUnit/:id/:value/:unit', function(request, response) {
		response.writeHead(200, HEADER);
		var robot = self.getRobotByGroup(HAMSTER);
		if(robot) {
			robot.moveBackwardUnit(request);
		}
		response.end();
	});

	express.get('/sTurnUnitInPlace/:id/:direction/:value/:unit', function(request, response) {
		response.writeHead(200, HEADER);
		var robot = self.getRobotByGroup(HAMSTER);
		if(robot) {
			robot.turnUnitInPlace(request);
		}
		response.end();
	});
	
	express.get('/sPivotAroundWheelUnitInDirection/:id/:wheel/:value/:unit/:forward', function(request, response) {
		response.writeHead(200, HEADER);
		var robot = self.getRobotByGroup(HAMSTER);
		if(robot) {
			robot.pivotAroundWheelUnitInDirection(request);
		}
		response.end();
	});
	
	express.get('/sTurnBodyUnitWithRadiusInDirection/:id/:direction/:value/:unit/:radius/:forward', function(request, response) {
		response.writeHead(200, HEADER);
		var robot = self.getRobotByGroup(HAMSTER);
		if(robot) {
			robot.turnBodyUnitWithRadiusInDirection(request);
		}
		response.end();
	});
	
	express.get('/sPivotAroundUnitInDirection/:id/:part/:value/:unit/:forward', function(request, response) {
		response.writeHead(200, HEADER);
		var robot = self.getRobotByGroup(HAMSTER);
		if(robot) {
			robot.pivotAroundUnitInDirection(request);
		}
		response.end();
	});
	
	express.get('/sTurnUnitWithRadiusInDirection/:id/:part/:direction/:value/:unit/:radius/:forward', function(request, response) {
		response.writeHead(200, HEADER);
		var robot = self.getRobotByGroup(HAMSTER);
		if(robot) {
			robot.turnUnitWithRadiusInDirection(request);
		}
		response.end();
	});
	
	express.get('/sChangeBothWheelsBy/:left/:right', function(request, response) {
		response.writeHead(200, HEADER);
		var robot = self.getRobotByGroup(HAMSTER);
		if(robot) {
			robot.changeBothWheelsBy(request);
		}
		response.end();
	});
	
	express.get('/sSetBothWheelsTo/:left/:right', function(request, response) {
		response.writeHead(200, HEADER);
		var robot = self.getRobotByGroup(HAMSTER);
		if(robot) {
			robot.setBothWheelsTo(request);
		}
		response.end();
	});
	
	express.get('/sChangeWheelBy/:which/:speed', function(request, response) {
		response.writeHead(200, HEADER);
		var robot = self.getRobotByGroup(HAMSTER);
		if(robot) {
			robot.changeWheelBy(request);
		}
		response.end();
	});
	
	express.get('/sSetWheelTo/:which/:speed', function(request, response) {
		response.writeHead(200, HEADER);
		var robot = self.getRobotByGroup(HAMSTER);
		if(robot) {
			robot.setWheelTo(request);
		}
		response.end();
	});
	
	express.get('/sFollowLineUsingFloorSensor/:color/:which', function(request, response) {
		response.writeHead(200, HEADER);
		var robot = self.getRobotByGroup(HAMSTER);
		if(robot) {
			robot.followLineUsingFloorSensor(request);
		}
		response.end();
	});
	
	express.get('/sFollowLineUntilIntersection/:id/:color/:which', function(request, response) {
		response.writeHead(200, HEADER);
		var robot = self.getRobotByGroup(HAMSTER);
		if(robot) {
			robot.followLineUntilIntersection(request);
		}
		response.end();
	});
	
	express.get('/sSetFollowingSpeedTo/:speed', function(request, response) {
		response.writeHead(200, HEADER);
		var robot = self.getRobotByGroup(HAMSTER);
		if(robot) {
			robot.setFollowingSpeedTo(request);
		}
		response.end();
	});
	
	express.get('/sSetFollowingGainTo/:gain', function(request, response) {
		response.writeHead(200, HEADER);
		var robot = self.getRobotByGroup(HAMSTER);
		if(robot) {
			robot.setFollowingGainTo(request);
		}
		response.end();
	});
	
	express.get('/sStop', function(request, response) {
		response.writeHead(200, HEADER);
		var robot = self.getRobotByGroup(HAMSTER);
		if(robot) {
			robot.stop(request);
		}
		response.end();
	});
	
	express.get('/sSetLedTo/:which/:color', function(request, response) {
		response.writeHead(200, HEADER);
		var robot = self.getRobotByGroup(HAMSTER);
		if(robot) {
			robot.setLedTo(request);
		}
		response.end();
	});
	
	express.get('/sChangeLedByRGB/:which/:red/:green/:blue', function(request, response) {
		response.writeHead(200, HEADER);
		var robot = self.getRobotByGroup(HAMSTER);
		if(robot) {
			robot.changeLedByRGB(request);
		}
		response.end();
	});
	
	express.get('/sSetLedToRGB/:which/:red/:green/:blue', function(request, response) {
		response.writeHead(200, HEADER);
		var robot = self.getRobotByGroup(HAMSTER);
		if(robot) {
			robot.setLedToRGB(request);
		}
		response.end();
	});
	
	express.get('/sClearLed/:which', function(request, response) {
		response.writeHead(200, HEADER);
		var robot = self.getRobotByGroup(HAMSTER);
		if(robot) {
			robot.clearLed(request);
		}
		response.end();
	});
	
	express.get('/sPlaySound/:sound', function(request, response) {
		response.writeHead(200, HEADER);
		var robot = self.getRobotByGroup(HAMSTER);
		if(robot) {
			robot.playSound(request);
		}
		response.end();
	});
	
	express.get('/sPlaySoundTimes/:sound/:count', function(request, response) {
		response.writeHead(200, HEADER);
		var robot = self.getRobotByGroup(HAMSTER);
		if(robot) {
			robot.playSoundTimes(request);
		}
		response.end();
	});
	
	express.get('/sPlaySoundTimesUntilDone/:id/:sound/:count', function(request, response) {
		response.writeHead(200, HEADER);
		var robot = self.getRobotByGroup(HAMSTER);
		if(robot) {
			robot.playSoundTimesUntilDone(request);
		}
		response.end();
	});
	
	express.get('/sChangeBuzzerBy/:value', function(request, response) {
		response.writeHead(200, HEADER);
		var robot = self.getRobotByGroup(HAMSTER);
		if(robot) {
			robot.changeBuzzerBy(request);
		}
		response.end();
	});
	
	express.get('/sSetBuzzerTo/:value', function(request, response) {
		response.writeHead(200, HEADER);
		var robot = self.getRobotByGroup(HAMSTER);
		if(robot) {
			robot.setBuzzerTo(request);
		}
		response.end();
	});
	
	express.get('/sClearSound', function(request, response) {
		response.writeHead(200, HEADER);
		var robot = self.getRobotByGroup(HAMSTER);
		if(robot) {
			robot.clearSound(request);
		}
		response.end();
	});
	
	express.get('/sPlayNote/:note/:octave', function(request, response) {
		response.writeHead(200, HEADER);
		var robot = self.getRobotByGroup(HAMSTER);
		if(robot) {
			robot.playNote(request);
		}
		response.end();
	});
	
	express.get('/sPlayNoteFor/:id/:note/:octave/:beat', function(request, response) {
		response.writeHead(200, HEADER);
		var robot = self.getRobotByGroup(HAMSTER);
		if(robot) {
			robot.playNoteFor(request);
		}
		response.end();
	});
	
	express.get('/sRestFor/:id/:beat', function(request, response) {
		response.writeHead(200, HEADER);
		var robot = self.getRobotByGroup(HAMSTER);
		if(robot) {
			robot.restFor(request);
		}
		response.end();
	});
	
	express.get('/sChangeTempoBy/:value', function(request, response) {
		response.writeHead(200, HEADER);
		var robot = self.getRobotByGroup(HAMSTER);
		if(robot) {
			robot.changeTempoBy(request);
		}
		response.end();
	});
	
	express.get('/sSetTempoTo/:value', function(request, response) {
		response.writeHead(200, HEADER);
		var robot = self.getRobotByGroup(HAMSTER);
		if(robot) {
			robot.setTempoTo(request);
		}
		response.end();
	});
	
	express.get('/sSetPortTo/:which/:value', function(request, response) {
		response.writeHead(200, HEADER);
		var robot = self.getRobotByGroup(HAMSTER);
		if(robot) {
			robot.setPortTo(request);
		}
		response.end();
	});
	
	express.get('/sChangeOutputBy/:which/:value', function(request, response) {
		response.writeHead(200, HEADER);
		var robot = self.getRobotByGroup(HAMSTER);
		if(robot) {
			robot.changeOutputBy(request);
		}
		response.end();
	});
	
	express.get('/sSetOutputTo/:which/:value', function(request, response) {
		response.writeHead(200, HEADER);
		var robot = self.getRobotByGroup(HAMSTER);
		if(robot) {
			robot.setOutputTo(request);
		}
		response.end();
	});
	
	express.get('/sGripper/:id/:action', function(request, response) {
		response.writeHead(200, HEADER);
		var robot = self.getRobotByGroup(HAMSTER);
		if(robot) {
			robot.gripper(request);
		}
		response.end();
	});
	
	express.get('/sReleaseGripper', function(request, response) {
		response.writeHead(200, HEADER);
		var robot = self.getRobotByGroup(HAMSTER);
		if(robot) {
			robot.releaseGripper(request);
		}
		response.end();
	});
	
	express.get('/sWriteSerial/:id/:mode/:value', function(request, response) {
		response.writeHead(200, HEADER);
		var robot = self.getRobotByGroup(HAMSTER);
		if(robot) {
			robot.writeSerial(request);
		}
		response.end();
	});
	
	express.get('/sReadSerialUntil/:id/:delimiter', function(request, response) {
		response.writeHead(200, HEADER);
		var robot = self.getRobotByGroup(HAMSTER);
		if(robot) {
			robot.readSerialUntil(request);
		}
		response.end();
	});
	
	express.get('/sSetSerialRateTo/:baud', function(request, response) {
		response.writeHead(200, HEADER);
		var robot = self.getRobotByGroup(HAMSTER);
		if(robot) {
			robot.setSerialRateTo(request);
		}
		response.end();
	});
};

Extension.prototype.startTurtle = function(express) {
	var self = this;
	express.get('/turtleMoveForward/:id', function(request, response) {
		response.writeHead(200, HEADER);
		var robot = self.getRobotByGroup(TURTLE);
		if(robot) {
			robot.moveForward(request);
		}
		response.end();
	});
	
	express.get('/turtleMoveBackward/:id', function(request, response) {
		response.writeHead(200, HEADER);
		var robot = self.getRobotByGroup(TURTLE);
		if(robot) {
			robot.moveBackward(request);
		}
		response.end();
	});
	
	express.get('/turtleTurn/:id/:direction', function(request, response) {
		response.writeHead(200, HEADER);
		var robot = self.getRobotByGroup(TURTLE);
		if(robot) {
			robot.turn(request);
		}
		response.end();
	});
	
	express.get('/turtleMoveForwardUnit/:id/:value/:unit', function(request, response) {
		response.writeHead(200, HEADER);
		var robot = self.getRobotByGroup(TURTLE);
		if(robot) {
			robot.moveForwardUnit(request);
		}
		response.end();
	});

	express.get('/turtleMoveBackwardUnit/:id/:value/:unit', function(request, response) {
		response.writeHead(200, HEADER);
		var robot = self.getRobotByGroup(TURTLE);
		if(robot) {
			robot.moveBackwardUnit(request);
		}
		response.end();
	});

	express.get('/turtleTurnUnitInPlace/:id/:direction/:value/:unit', function(request, response) {
		response.writeHead(200, HEADER);
		var robot = self.getRobotByGroup(TURTLE);
		if(robot) {
			robot.turnUnitInPlace(request);
		}
		response.end();
	});
	
	express.get('/turtleTurnUnitWithRadiusInDirection/:id/:direction/:value/:unit/:radius/:head', function(request, response) {
		response.writeHead(200, HEADER);
		var robot = self.getRobotByGroup(TURTLE);
		if(robot) {
			robot.turnUnitWithRadiusInDirection(request);
		}
		response.end();
	});
	
	express.get('/turtlePivotAroundWheelUnitInDirection/:id/:wheel/:value/:unit/:head', function(request, response) {
		response.writeHead(200, HEADER);
		var robot = self.getRobotByGroup(TURTLE);
		if(robot) {
			robot.pivotAroundWheelUnitInDirection(request);
		}
		response.end();
	});
	
	express.get('/turtleChangeWheelsByLeftRight/:left/:right', function(request, response) {
		response.writeHead(200, HEADER);
		var robot = self.getRobotByGroup(TURTLE);
		if(robot) {
			robot.changeWheelsByLeftRight(request);
		}
		response.end();
	});
	
	express.get('/turtleSetWheelsToLeftRight/:left/:right', function(request, response) {
		response.writeHead(200, HEADER);
		var robot = self.getRobotByGroup(TURTLE);
		if(robot) {
			robot.setWheelsToLeftRight(request);
		}
		response.end();
	});
	
	express.get('/turtleChangeWheelBy/:wheel/:speed', function(request, response) {
		response.writeHead(200, HEADER);
		var robot = self.getRobotByGroup(TURTLE);
		if(robot) {
			robot.changeWheelBy(request);
		}
		response.end();
	});
	
	express.get('/turtleSetWheelTo/:wheel/:speed', function(request, response) {
		response.writeHead(200, HEADER);
		var robot = self.getRobotByGroup(TURTLE);
		if(robot) {
			robot.setWheelTo(request);
		}
		response.end();
	});
	
	express.get('/turtleFollowLine/:color', function(request, response) {
		response.writeHead(200, HEADER);
		var robot = self.getRobotByGroup(TURTLE);
		if(robot) {
			robot.followLine(request);
		}
		response.end();
	});
	
	express.get('/turtleFollowLineUntil/:id/:color', function(request, response) {
		response.writeHead(200, HEADER);
		var robot = self.getRobotByGroup(TURTLE);
		if(robot) {
			robot.followLineUntil(request);
		}
		response.end();
	});
	
	express.get('/turtleFollowLineUntilBlack/:id/:color', function(request, response) {
		response.writeHead(200, HEADER);
		var robot = self.getRobotByGroup(TURTLE);
		if(robot) {
			robot.followLineUntilBlack(request);
		}
		response.end();
	});
	
	express.get('/turtleCrossIntersection/:id', function(request, response) {
		response.writeHead(200, HEADER);
		var robot = self.getRobotByGroup(TURTLE);
		if(robot) {
			robot.crossIntersection(request);
		}
		response.end();
	});
	
	express.get('/turtleTurnAtIntersection/:id/:direction', function(request, response) {
		response.writeHead(200, HEADER);
		var robot = self.getRobotByGroup(TURTLE);
		if(robot) {
			robot.turnAtIntersection(request);
		}
		response.end();
	});
	
	express.get('/turtleSetFollowingSpeedTo/:speed', function(request, response) {
		response.writeHead(200, HEADER);
		var robot = self.getRobotByGroup(TURTLE);
		if(robot) {
			robot.setFollowingSpeedTo(request);
		}
		response.end();
	});
	
	express.get('/turtleStop', function(request, response) {
		response.writeHead(200, HEADER);
		var robot = self.getRobotByGroup(TURTLE);
		if(robot) {
			robot.stop(request);
		}
		response.end();
	});
	
	express.get('/turtleSetHeadLedTo/:color', function(request, response) {
		response.writeHead(200, HEADER);
		var robot = self.getRobotByGroup(TURTLE);
		if(robot) {
			robot.setHeadLedTo(request);
		}
		response.end();
	});
	
	express.get('/turtleChangeHeadLedByRGB/:red/:green/:blue', function(request, response) {
		response.writeHead(200, HEADER);
		var robot = self.getRobotByGroup(TURTLE);
		if(robot) {
			robot.changeHeadLedByRGB(request);
		}
		response.end();
	});
	
	express.get('/turtleSetHeadLedToRGB/:red/:green/:blue', function(request, response) {
		response.writeHead(200, HEADER);
		var robot = self.getRobotByGroup(TURTLE);
		if(robot) {
			robot.setHeadLedToRGB(request);
		}
		response.end();
	});
	
	express.get('/turtleClearHeadLed', function(request, response) {
		response.writeHead(200, HEADER);
		var robot = self.getRobotByGroup(TURTLE);
		if(robot) {
			robot.clearHeadLed(request);
		}
		response.end();
	});
	
	express.get('/turtlePlaySound/:sound', function(request, response) {
		response.writeHead(200, HEADER);
		var robot = self.getRobotByGroup(TURTLE);
		if(robot) {
			robot.playSound(request);
		}
		response.end();
	});
	
	express.get('/turtlePlaySoundTimes/:sound/:count', function(request, response) {
		response.writeHead(200, HEADER);
		var robot = self.getRobotByGroup(TURTLE);
		if(robot) {
			robot.playSoundTimes(request);
		}
		response.end();
	});
	
	express.get('/turtlePlaySoundTimesUntilDone/:id/:sound/:count', function(request, response) {
		response.writeHead(200, HEADER);
		var robot = self.getRobotByGroup(TURTLE);
		if(robot) {
			robot.playSoundTimesUntilDone(request);
		}
		response.end();
	});
	
	express.get('/turtleChangeBuzzerBy/:hz', function(request, response) {
		response.writeHead(200, HEADER);
		var robot = self.getRobotByGroup(TURTLE);
		if(robot) {
			robot.changeBuzzerBy(request);
		}
		response.end();
	});
	
	express.get('/turtleSetBuzzerTo/:hz', function(request, response) {
		response.writeHead(200, HEADER);
		var robot = self.getRobotByGroup(TURTLE);
		if(robot) {
			robot.setBuzzerTo(request);
		}
		response.end();
	});
	
	express.get('/turtleClearSound', function(request, response) {
		response.writeHead(200, HEADER);
		var robot = self.getRobotByGroup(TURTLE);
		if(robot) {
			robot.clearSound(request);
		}
		response.end();
	});
	
	express.get('/turtlePlayNote/:note/:octave', function(request, response) {
		response.writeHead(200, HEADER);
		var robot = self.getRobotByGroup(TURTLE);
		if(robot) {
			robot.playNote(request);
		}
		response.end();
	});
	
	express.get('/turtlePlayNoteForBeats/:id/:note/:octave/:beat', function(request, response) {
		response.writeHead(200, HEADER);
		var robot = self.getRobotByGroup(TURTLE);
		if(robot) {
			robot.playNoteForBeats(request);
		}
		response.end();
	});
	
	express.get('/turtleRestForBeats/:id/:beat', function(request, response) {
		response.writeHead(200, HEADER);
		var robot = self.getRobotByGroup(TURTLE);
		if(robot) {
			robot.restForBeats(request);
		}
		response.end();
	});
	
	express.get('/turtleChangeTempoBy/:bpm', function(request, response) {
		response.writeHead(200, HEADER);
		var robot = self.getRobotByGroup(TURTLE);
		if(robot) {
			robot.changeTempoBy(request);
		}
		response.end();
	});
	
	express.get('/turtleSetTempoTo/:bpm', function(request, response) {
		response.writeHead(200, HEADER);
		var robot = self.getRobotByGroup(TURTLE);
		if(robot) {
			robot.setTempoTo(request);
		}
		response.end();
	});
};

Extension.prototype.stop = function() {
	this.running = false;
	translator.remove('all-scratch');
	router.reset();
	this.robotsByModule = {};
	this.robotsByGroup = {};
};

Extension.prototype.setState = function(route) {
	if(route == router.firstRoute) {
		this.state = route.state;
	}
};

Extension.prototype.setOption = function(option) {
};

Extension.prototype.handleSensory = function(data, route) {
	var robot = this.getRobotByRoute(route);
	if(robot) {
		robot.handleSensory(data);
	}
};

Extension.prototype.handleMotoring = function(data, route) {
	var robot = this.getRobotByRoute(route);
	if(robot) {
		var roboid = robot.roboid;
		roboid.handleMotoring(data);
		roboid.updateMotoring();
	}
};

Extension.prototype.requestMotoring = function(route) {
	var robot = this.getRobotByRoute(route);
	if(robot) {
		return robot.roboid.requestMotoring();
	}
};

Extension.prototype.reset = function(route) {
	var robot = this.getRobotByRoute(route);
	if(robot) {
		robot.roboid.reset();
		robot.motoring.reset();
	}
};

module.exports.create = function() {
	return new Extension();
};