'use strict';
var context = require('context');
var router = require('router');
const PROGRAM = 'scratch3';

function Extension() {
	this.robots = {};
	this.connectionPacket = {
		type: 0,
		states: {}
	};
	this.optionPacket = {
		type: 2,
		multi: false
	};
}

Extension.prototype.getRobotByIndex = function(module, index) {
	if(index >= 0) {
		var key = module + index;
		var robot = this.robots[key];
		if(!robot) {
			var component = context.getComponentByModule(PROGRAM, module);
			robot = component.createRobot(index);
			if(robot) {
				this.robots[key] = robot;
			}
		}
		return robot;
	}
};

Extension.prototype.getRobotByRoute = function(route) {
	if(router.multi) return this.getRobotByIndex(route.module, route.index);
	return this.getRobotByIndex(route.module, 0);
};

Extension.prototype.start = function(connection) {
	var self = this;
	self.connection = connection;
	connection.on('message', function(message) {
		if(message.type === 'utf8') {
			self.handleRemoteData(message.utf8Data);
		}
	});
};

Extension.prototype.stop = function() {
	this.connection = undefined;
	router.reset();
	this.robots = {};
};

Extension.prototype.setState = function(route) {
	if(this.connection) {
		var states = this.connectionPacket.states;
		var module, modules = context.getModules();
		if(router.multi) {
			for(var i in modules) {
				module = modules[i];
				states[module] = router.checkConnectionState(module);
			}
		} else {
			for(var i in modules) {
				module = modules[i];
				states[module] = (router.selectedRoute && module == router.selectedRoute.module) ? router.getConnectionState(router.selectedRoute.state) : -1;
			}
		}
		try {
			this.connection.send(JSON.stringify(this.connectionPacket));
		} catch (e) {
		}
	}
};

Extension.prototype.setOption = function(option) {
	if(this.connection) {
		var packet = this.optionPacket;
		var packetModules = packet.modules = {};
		var routerModules = router.getModules();
		var routes = router.getRoutes(), module;
		for(var i in routes) {
			module = routes[i].module;
			packetModules[module] = routerModules[module] || 0;
		}
		packet.multi = router.multi;
		packet.level = option.level;
		try {
			this.connection.send(JSON.stringify(packet));
		} catch (e) {
		}
	}
};

Extension.prototype.handleRemoteDataByModule = function(data) {
	var module = data.module;
	if(module) {
		var robot = router.multi ? this.getRobotByIndex(module, data.index) : this.getRobotByIndex(module, 0);
		if(robot) {
			var component = context.getComponentByModule(PROGRAM, module);
			component.handleRemoteData(robot, data);
		}
	}
};

Extension.prototype.handleRemoteData = function(utf8Data) {
	try {
		var received = JSON.parse(utf8Data);
		for(var i in received) {
			this.handleRemoteDataByModule(received[i]);
		}
	} catch (e) {
	}
};

Extension.prototype.handleSensory = function(data, route) {
	var robot = this.getRobotByRoute(route);
	if(robot) {
		var component = context.getComponentByRoute(PROGRAM, route);
		component.handleSensory(robot, data);
		var sensory = robot.sensory;
		sensory.type = 1;
		if(this.connection) {
			try {
				this.connection.send(JSON.stringify(sensory));
			} catch (e) {
			}
		}
	}
};

Extension.prototype.handleMotoring = function(data, route) {
	var robot = this.getRobotByRoute(route);
	if(robot) {
		var roboid = robot.roboid;
		roboid.handleMotoring(data);
		roboid.updateMotoring();
	}
};

Extension.prototype.requestMotoring = function(route) {
	var robot = this.getRobotByRoute(route);
	if(robot) {
		return robot.roboid.requestMotoring();
	}
};

Extension.prototype.reset = function(route) {
	var robot = this.getRobotByRoute(route);
	if(robot) {
		robot.roboid.reset();
	}
};

module.exports.create = function() {
	return new Extension();
};