'use strict';
var context = require('context');
var router = require('router');
const PROGRAM = 'websocket';

function Extension() {
	this.robots = {};
}

Extension.prototype.getRobotByModule = function(module, state) {
	var robot = this.robots[module];
	if(!robot) {
		var component = context.getComponentByModule(PROGRAM, module);
		robot = component.createRobot(0, state);
		if(robot) {
			this.robots[module] = robot;
		}
	}
	return robot;
};

Extension.prototype.getRobotByIndex = function(module, index) {
	if(index >= 0) {
		var key = module + index;
		var robot = this.robots[key];
		if(!robot) {
			var route = router.findRouteByIndex(module, index);
			if(route) {
				var component = context.getComponentByModule(PROGRAM, module);
				robot = component.createRobot(index, route.state);
				if(robot) {
					this.robots[key] = robot;
				}
			}
		}
		return robot;
	}
};

Extension.prototype.getRobotByRoute = function(route) {
	if(router.multi) {
		var key = route.module + route.index;
		var robot = this.robots[key];
		if(!robot) {
			var component = context.getComponentByRoute(PROGRAM, route);
			robot = component.createRobot(route.index, route.state);
			if(robot) {
				this.robots[key] = robot;
			}
		}
		return robot;
	} else {
		return this.getRobotByModule(route.module, route.state);
	}
};

Extension.prototype.start = function(connection) {
	var self = this;
	self.connection = connection;
	connection.on('message', function(message) {
		if(message.type === 'utf8') {
			self.handleRemoteData(message.utf8Data);
		}
	});
};

Extension.prototype.stop = function() {
	this.connection = undefined;
	router.reset();
	this.robots = {};
};

Extension.prototype.setState = function(route) {
	if(router.multi || route == router.selectedRoute) {
		var robot = this.getRobotByRoute(route);
		if(robot) {
			var component = context.getComponentByRoute(PROGRAM, route);
			component.setState(robot, route.state);
		}
	}
};

Extension.prototype.setOption = function(option) {
};

Extension.prototype.handleRemoteData = function(utf8Data) {
	if(router.multi) {
		try {
			var received = JSON.parse(utf8Data);
			var data, module, robot, component;
			for(var i in received) {
				data = received[i];
				module = context.model2module(data.model);
				robot = this.getRobotByIndex(module, data.index);
				if(robot) {
					component = context.getComponentByModule(PROGRAM, module);
					component.handleRemoteData(robot, data);
				}
			}
		} catch (e) {
		}
	} else {
		var selectedRoute = router.selectedRoute;
		if(selectedRoute) {
			try {
				var received = JSON.parse(utf8Data);
				var data, module, robot, component;
				for(var i in received) {
					data = received[i];
					module = context.model2module(data.model);
					if(selectedRoute.module == module && data.index == 0) {
						robot = this.getRobotByModule(module, selectedRoute.state);
						if(robot) {
							component = context.getComponentByModule(PROGRAM, module);
							component.handleRemoteData(robot, data);
						}
					}
				}
			} catch (e) {
			}
		}
	}
};

Extension.prototype.handleSensory = function(data, route) {
	var robot = this.getRobotByRoute(route);
	if(robot) {
		var component = context.getComponentByRoute(PROGRAM, route);
		component.handleSensory(robot, data, route.state, this.connection);
	}
};

Extension.prototype.handleMotoring = function(data, route) {
	var robot = this.getRobotByRoute(route);
	if(robot) {
		var roboid = robot.roboid;
		roboid.handleMotoring(data);
		roboid.updateMotoring();
	}
};

Extension.prototype.requestMotoring = function(route) {
	var robot = this.getRobotByRoute(route);
	if(robot) {
		return robot.roboid.requestMotoring();
	}
};

Extension.prototype.reset = function(route) {
	if(router.multi || route == router.selectedRoute) {
		var robot = this.getRobotByRoute(route);
		if(robot) {
			robot.roboid.reset();
		}
	}
};

module.exports.create = function() {
	return new Extension();
};