'use strict';
var translations = require('./translations');
var Cheese = require('../cheese');
var factory = require('../cheese-roboid');
var matrix = require('../led-matrix');
var neo = require('../neopixel');
var SerialQueue = require('../serial-queue');
const MODULE = 'cheese';

var NOTES = {};
var BEATS = { '¼': 0.25, '½': 0.5, '¾': 0.75, '1¼': 1.25, '1½': 1.5, '1¾': 1.75 };
var SOUNDS = {};
var RGB_COLORS = {};
var MODE_INPUTS_SL = {};
var SPECIAL_SENSORS = {};
var MODE_SOUNDS = {};
var ACTION_STEP_MOTORS = {};
var MODE_STEP_MOTORS = {};
var PORT_PWM_OUTPUTS = {};
var LED_TYPES = { '-RGB': 1, '-RBG': 2, '-GRB': 3, '-GBR': 4, '-BRG': 5, '-BGR': 6, '+RGB': 7, '+RBG': 8, '+GRB': 9, '+GBR': 10, '+BRG': 11, '+BGR': 12 };
var SHAPES = {};
var DIRECTIONS = {};
var SERIAL_MODES = {};
var SERIAL_DELIMITERS = {};
var NEOPIXEL_TYPES = {};
var VALUES = {};

var BOOLEAN_TAP = '/tap ';
var BOOLEAN_FREE_FALL = '/free%20fall ';
var BOOLEAN_BATTERY_NORMAL = '/battery%20normal ';
var BOOLEAN_BATTERY_LOW = '/battery%20low ';
var BOOLEAN_BATTERY_EMPTY = '/battery%20empty ';
var SENSOR_ACCELERATION_X = '/x%20acceleration ';
var SENSOR_ACCELERATION_Y = '/y%20acceleration ';
var SENSOR_ACCELERATION_Z = '/z%20acceleration ';
var SENSOR_SIGNAL_STRENGTH = '/signal%20strength%20(dBm) ';
var SENSOR_INPUT = '/input%20';
var SENSOR_ULTRASONIC = '/ultrasonic%20(cm) ';
var SENSOR_ENCODER = '/encoder ';
var SENSOR_TEMPERATURE = '/temperature%20(℃) ';
var SENSOR_HUMIDITY = '/humidity%20(%25RH) ';
var SENSOR_GESTURE = '/gesture ';
var TOUCH_BACKGROUND = '/background ';
var TOUCH_OTHER_SPRITE = '/other%20sprite ';
var TOUCH_LEFT_WALL = '/left%20wall ';
var TOUCH_RIGHT_WALL = '/right%20wall ';
var TOUCH_TOP_WALL = '/top%20wall ';
var TOUCH_BOTTOM_WALL = '/bottom%20wall ';
var TOUCH_ANY_WALL = '/any%20wall ';
var X_POSITION = '/x%20position ';
var Y_POSITION = '/y%20position ';

const ULTRASONIC = 1;
const ENCODER = 2;
const DS18B20 = 3;
const DHT11 = 4;
const DHT21 = 5;
const DHT22 = 6;
const GESTURE = 7;

const STOP = 1;
const TURN_OFF = 2;

const CRGB = 1;
const CRBG = 2;
const CGRB = 3;
const CGBR = 4;
const CBRG = 5;
const CBGR = 6;
const ARGB = 7;
const ARBG = 8;
const AGRB = 9;
const AGBR = 10;
const ABRG = 11;
const ABGR = 12;

const LEFT = 1;
const RIGHT = 2;
const TOP = 3;
const BOTTOM = 4;

const BACKGROUND = 1;
const ALL = 2;
const ON = 1;
const OFF = 2;
const INVERSE = 3;
const CLEAR = 1;
const SHOW = 2;
const HIDE = 3;
const POS_X = 1;
const POS_Y = 2;
const CLOCKWISE = 1;
const COUNTERCLOCKWISE = 2;
const LEFT_RIGHT = 1;
const UP_DOWN = 2;
const LANDSCAPE = 1;
const PORTRAIT = 2;
const AUTO = 1;
const MANUAL = 2;

const NEOPIXEL_GRB = 0;
const NEOPIXEL_GRBW = 1;

const SERIAL_STRING = 1;
const SERIAL_STRING_LINE = 2;
const SERIAL_NEW_LINE = 0x0D;
const SERIAL_COMMA = 0x2C;
const SERIAL_COLON = 0x3A;
const SERIAL_DOLLAR = 0x24;
const SERIAL_SHARP = 0x23;

function Robot(index) {
	var roboid = this.roboid = factory.create();
	this.index = index;
	this.waits = [];
	this.timeouts = [];
	this.sensory = {
		signalStrength: 0,
		accelerationX: 0,
		accelerationY: 0,
		accelerationZ: 0,
		inputSA: 0,
		inputSB: 0,
		inputSC: 0,
		inputLA: 0,
		inputLB: 0,
		inputLC: 0,
		ultrasonic: 0,
		encoder: 0,
		temperature: 0,
		humidity: 0,
		serial: '',
		stepCount: 0,
		batteryState: 2
	};
	this.motoring = {
		step: roboid.findDeviceById(Cheese.STEP),
		note: roboid.findDeviceById(Cheese.NOTE),
		sound: roboid.findDeviceById(Cheese.SOUND),
		modeSound: roboid.findDeviceById(Cheese.MODE_SOUND),
		mode: {
			Sa: roboid.findDeviceById(Cheese.MODE_SA),
			Sb: roboid.findDeviceById(Cheese.MODE_SB),
			Sc: roboid.findDeviceById(Cheese.MODE_SC),
			La: roboid.findDeviceById(Cheese.MODE_LA),
			Lb: roboid.findDeviceById(Cheese.MODE_LB),
			Lc: roboid.findDeviceById(Cheese.MODE_LC),
			M: roboid.findDeviceById(Cheese.MODE_M)
		},
		output: {},
		resetEncoder: roboid.findDeviceById(Cheese.RESET_ENCODER),
		writeI2c: roboid.findDeviceById(Cheese.WRITE_I2C),
		writeSerial: roboid.findDeviceById(Cheese.WRITE_SERIAL),
		neopixel: roboid.findDeviceById(Cheese.NEOPIXEL),
		reset: function() {
			this.output.Sa.data = 0;
			this.output.Sb.data = 0;
			this.output.Sc.data = 0;
			this.output.La.data = 0;
			this.output.Lb.data = 0;
			this.output.Lc.data = 0;
			this.output.Lx.data = 0;
			this.output.Ma.data = 0;
			this.output.Mb.data = 0;
			this.buzzer.data = 0;
			this.velocity.data = 0;
		}
	};
	this.registerOutput('Sa', Cheese.OUTPUT_SA);
	this.registerOutput('Sb', Cheese.OUTPUT_SB);
	this.registerOutput('Sc', Cheese.OUTPUT_SC);
	this.registerOutput('La', Cheese.OUTPUT_LA);
	this.registerOutput('Lb', Cheese.OUTPUT_LB);
	this.registerOutput('Lc', Cheese.OUTPUT_LC);
	this.registerOutput('Lx', Cheese.OUTPUT_LX);
	this.registerOutput('Ma', Cheese.OUTPUT_MA);
	this.registerOutput('Mb', Cheese.OUTPUT_MB);
	this.register('buzzer', Cheese.BUZZER);
	this.register('velocity', Cheese.VELOCITY);
	this.i2cWaits = [];
	this.serialWriteWaits = [];
	this.serialReadWaits = [];
	this.neopixelWaits = [];
	this.writeQueue = SerialQueue.createWriteQueue(64);
	this.readQueue = SerialQueue.createReadQueue(64);
	this.range = {
		Sa: { active: false, has_zero: false, zero: 127, in_lower: 0, in_upper: 255, out_lower: -100, out_upper: 100 },
		Sb: { active: false, has_zero: false, zero: 127, in_lower: 0, in_upper: 255, out_lower: -100, out_upper: 100 },
		Sc: { active: false, has_zero: false, zero: 127, in_lower: 0, in_upper: 255, out_lower: -100, out_upper: 100 },
		La: { active: false, has_zero: false, zero: 127, in_lower: 0, in_upper: 255, out_lower: -100, out_upper: 100 },
		Lb: { active: false, has_zero: false, zero: 127, in_lower: 0, in_upper: 255, out_lower: -100, out_upper: 100 },
		Lc: { active: false, has_zero: false, zero: 127, in_lower: 0, in_upper: 255, out_lower: -100, out_upper: 100 },
		reset: function() {
			this.Sa.active = false;
			this.Sa.has_zero = false;
			this.Sa.zero = 127;
			this.Sa.in_lower = 0;
			this.Sa.in_upper = 255;
			this.Sa.out_lower = -100;
			this.Sa.out_upper = 100;
			this.Sb.active = false;
			this.Sb.has_zero = false;
			this.Sb.zero = 127;
			this.Sb.in_lower = 0;
			this.Sb.in_upper = 255;
			this.Sb.out_lower = -100;
			this.Sb.out_upper = 100;
			this.Sc.active = false;
			this.Sc.has_zero = false;
			this.Sc.zero = 127;
			this.Sc.in_lower = 0;
			this.Sc.in_upper = 255;
			this.Sc.out_lower = -100;
			this.Sc.out_upper = 100;
			this.La.active = false;
			this.La.has_zero = false;
			this.La.zero = 127;
			this.La.in_lower = 0;
			this.La.in_upper = 255;
			this.La.out_lower = -100;
			this.La.out_upper = 100;
			this.Lb.active = false;
			this.Lb.has_zero = false;
			this.Lb.zero = 127;
			this.Lb.in_lower = 0;
			this.Lb.in_upper = 255;
			this.Lb.out_lower = -100;
			this.Lb.out_upper = 100;
			this.Lc.active = false;
			this.Lc.has_zero = false;
			this.Lc.zero = 127;
			this.Lc.in_lower = 0;
			this.Lc.in_upper = 255;
			this.Lc.out_lower = -100;
			this.Lc.out_upper = 100;
		}
	};
	this.matrix = matrix.create();
	this.neopixel = neo.create();
	this.motoring.mode['Ma'] = this.motoring.mode.M;
	this.motoring.mode['Mb'] = this.motoring.mode.M;
	this.motoring.mode['Mab'] = this.motoring.mode.M;
	this.motoring.output['Mab'] = this.motoring.output.Ma;
	this.tempo = 60;
	this.stepId = -1;
	this.stepMode = Cheese.MODE_M_STEP_WAVE;
	this.rgbPort = {};
	this.sound = 0;
	this.soundId = -1;
	this.soundRepeat = 1;
	this.serialDelimiter = 0;
	this.freeFall = false;
	this.tap = false;
	this.active = true;
	this.received = false;
	this.matrix.reset();
	this.neopixel.reset();
}

Robot.prototype.reset = function() {
	this.tempo = 60;
	this.stepId = -1;
	this.stepMode = Cheese.MODE_M_STEP_WAVE;
	this.rgbPort = {};
	this.sound = 0;
	this.soundId = -1;
	this.soundRepeat = 1;
	this.serialDelimiter = 0;
	this.freeFall = false;
	this.tap = false;
	this.roboid.reset();
	this.motoring.reset();
	this.range.reset();
	this.matrix.reset();
	this.neopixel.reset();
	this.writeQueue.reset();
	this.readQueue.reset();
};

Robot.prototype.register = function(name, deviceId) {
	var self = this;
	self.motoring[name] = {
		data: 0,
		device: self.roboid.findDeviceById(deviceId),
		write: function(value) {
			this.data = value;
			this.device.write(value);
		},
		read: function() {
			return this.data;
		}
	};
};

Robot.prototype.registerOutput = function(name, deviceId) {
	var self = this;
	self.motoring.output[name] = {
		data: 0,
		device: self.roboid.findDeviceById(deviceId),
		write: function(value) {
			this.data = value;
			this.device.write(value);
		},
		read: function() {
			return this.data;
		}
	};
};

Robot.prototype.getWaits = function() {
	return this.waits;
};

Robot.prototype.addI2cWait = function(id) {
	this.waits.push(id);
	this.i2cWaits.push(id);
};

Robot.prototype.addSerialWriteWait = function(id) {
	this.waits.push(id);
	this.serialWriteWaits.push(id);
};

Robot.prototype.addSerialReadWait = function(id) {
	this.waits.push(id);
	this.serialReadWaits.push(id);
};

Robot.prototype.addNeopixelWait = function(id) {
	this.waits.push(id);
	this.neopixelWaits.push(id);
};

Robot.prototype.removeWait = function(id) {
	var waits = this.waits;
	var index = waits.indexOf(id);
	if(index >= 0) {
		waits.splice(index, 1);
	}
};

Robot.prototype.removeI2cWaits = function() {
	var i2cWaits = this.i2cWaits;
	for(var i in i2cWaits) {
		this.removeWait(i2cWaits[i]);
	}
	this.i2cWaits = [];
};

Robot.prototype.removeSerialWriteWaits = function() {
	var serialWaits = this.serialWriteWaits;
	for(var i in serialWaits) {
		this.removeWait(serialWaits[i]);
	}
	this.serialWriteWaits = [];
};

Robot.prototype.removeSerialReadWaits = function() {
	var serialWaits = this.serialReadWaits;
	for(var i in serialWaits) {
		this.removeWait(serialWaits[i]);
	}
	this.serialReadWaits = [];
};

Robot.prototype.removeNeopixelWaits = function() {
	var neopixelWaits = this.neopixelWaits;
	for(var i in neopixelWaits) {
		this.removeWait(neopixelWaits[i]);
	}
	this.neopixelWaits = [];
};

Robot.prototype.clearWaits = function() {
	this.waits = [];
	this.i2cWaits = [];
	this.serialWriteWaits = [];
	this.serialReadWaits = [];
	this.neopixelWaits = [];
};

Robot.prototype.removeTimeout = function(id) {
	clearTimeout(id);
	var timeouts = this.timeouts;
	var index = timeouts.indexOf(id);
	if(index >= 0) {
		timeouts.splice(index, 1);
	}
};

Robot.prototype.clearTimeouts = function() {
	var timeouts = this.timeouts;
	for(var i in timeouts) {
		clearTimeout(timeouts[i]);
	}
	this.timeouts = [];
};

Robot.prototype.writeSingleSensory = function(response) {
	var sensory = this.sensory;
	response.write('cheeseTap ' + this.tap + '\n');
	response.write('cheeseBoolean' + BOOLEAN_TAP + this.tap + '\n');
	response.write('cheeseBoolean' + BOOLEAN_FREE_FALL + this.freeFall + '\n');
	response.write('cheeseBoolean' + BOOLEAN_BATTERY_NORMAL + (sensory.batteryState == 2) + '\n');
	response.write('cheeseBoolean' + BOOLEAN_BATTERY_LOW + (sensory.batteryState == 1) + '\n');
	response.write('cheeseBoolean' + BOOLEAN_BATTERY_EMPTY + (sensory.batteryState == 0) + '\n');
	
	response.write('cheeseSensor' + SENSOR_ACCELERATION_X + sensory.accelerationX + '\n');
	response.write('cheeseSensor' + SENSOR_ACCELERATION_Y + sensory.accelerationY + '\n');
	response.write('cheeseSensor' + SENSOR_ACCELERATION_Z + sensory.accelerationZ + '\n');
	response.write('cheeseSensor' + SENSOR_SIGNAL_STRENGTH + sensory.signalStrength + '\n');
	
	response.write('cheeseDigitalInput/Sa ' + (sensory.inputSA != 0) + '\n');
	response.write('cheeseDigitalInput/Sb ' + (sensory.inputSB != 0) + '\n');
	response.write('cheeseDigitalInput/Sc ' + (sensory.inputSC != 0) + '\n');
	response.write('cheeseDigitalInput/La ' + (sensory.inputLA != 0) + '\n');
	response.write('cheeseDigitalInput/Lb ' + (sensory.inputLB != 0) + '\n');
	response.write('cheeseDigitalInput/Lc ' + (sensory.inputLC != 0) + '\n');
	
	response.write('cheeseAnalogInput' + SENSOR_INPUT + 'Sa ' + sensory.inputSA + '\n');
	response.write('cheeseAnalogInput' + SENSOR_INPUT + 'Sb ' + sensory.inputSB + '\n');
	response.write('cheeseAnalogInput' + SENSOR_INPUT + 'Sc ' + sensory.inputSC + '\n');
	response.write('cheeseAnalogInput' + SENSOR_INPUT + 'La ' + sensory.inputLA + '\n');
	response.write('cheeseAnalogInput' + SENSOR_INPUT + 'Lb ' + sensory.inputLB + '\n');
	response.write('cheeseAnalogInput' + SENSOR_INPUT + 'Lc ' + sensory.inputLC + '\n');
	response.write('cheeseAnalogInput' + SENSOR_ULTRASONIC + sensory.ultrasonic + '\n');
	response.write('cheeseAnalogInput' + SENSOR_ENCODER + sensory.encoder + '\n');
	response.write('cheeseAnalogInput' + SENSOR_TEMPERATURE + sensory.temperature + '\n');
	response.write('cheeseAnalogInput' + SENSOR_HUMIDITY + sensory.humidity + '\n');
	
	//response.write('cheeseAnalogInput' + SENSOR_GESTURE + sensory.stepCount + '\n');
	
	response.write('cheeseStepNumber ' + sensory.stepCount + '\n');
	
	response.write('cheeseSerial ' + sensory.serial + '\n');
	
	var matrix = this.matrix;
	if(matrix.active) {
		var positions = matrix.getSpritePositions();
		var pos;
		for(var i in positions) {
			pos = positions[i];
			response.write('cheeseLedMatrixSpritePosition/' + i + X_POSITION + pos.x + '\n');
			response.write('cheeseLedMatrixSpritePosition/' + i + Y_POSITION + pos.y + '\n');
		}
		var touching = this.touching;
		if(touching) {
			var data = touching.background;
			for(var i in data) {
				response.write('cheeseLedMatrixTouching/' + i + TOUCH_BACKGROUND + data[i] + '\n');
			}
			data = touching.otherSprite;
			for(var i in data) {
				response.write('cheeseLedMatrixTouching/' + i + TOUCH_OTHER_SPRITE + data[i] + '\n');
			}
			data = touching.walls;
			var wall;
			for(var i in data) {
				wall = data[i];
				response.write('cheeseLedMatrixTouching/' + i + TOUCH_LEFT_WALL + wall.left + '\n');
				response.write('cheeseLedMatrixTouching/' + i + TOUCH_RIGHT_WALL + wall.right + '\n');
				response.write('cheeseLedMatrixTouching/' + i + TOUCH_TOP_WALL + wall.top + '\n');
				response.write('cheeseLedMatrixTouching/' + i + TOUCH_BOTTOM_WALL + wall.bottom + '\n');
				response.write('cheeseLedMatrixTouching/' + i + TOUCH_ANY_WALL + (wall.left || wall.right || wall.top || wall.bottom) + '\n');
			}
			data = touching.sprites;
			var cell;
			for(var i in data) {
				cell = data[i];
				for(var j in cell) {
					response.write('cheeseLedMatrixTouchingSprite/' + i + '/' + j + ' ' + cell[j] + '\n');
				}
			}
		}
	}
	this.freeFall = false;
	this.tap = false;
};

Robot.prototype.writeMultiSensory = function(response) {
	var sensory = this.sensory;
	if(this.active) {
		var index = this.index;
		response.write('cheeseTap/' + index + ' ' + this.tap + '\n');
		response.write('cheeseBoolean/' + index + BOOLEAN_TAP + this.tap + '\n');
		response.write('cheeseBoolean/' + index + BOOLEAN_FREE_FALL + this.freeFall + '\n');
		response.write('cheeseBoolean/' + index + BOOLEAN_BATTERY_NORMAL + (sensory.batteryState == 2) + '\n');
		response.write('cheeseBoolean/' + index + BOOLEAN_BATTERY_LOW + (sensory.batteryState == 1) + '\n');
		response.write('cheeseBoolean/' + index + BOOLEAN_BATTERY_EMPTY + (sensory.batteryState == 0) + '\n');
		
		response.write('cheeseSensor/' + index + SENSOR_ACCELERATION_X + sensory.accelerationX + '\n');
		response.write('cheeseSensor/' + index + SENSOR_ACCELERATION_Y + sensory.accelerationY + '\n');
		response.write('cheeseSensor/' + index + SENSOR_ACCELERATION_Z + sensory.accelerationZ + '\n');
		response.write('cheeseSensor/' + index + SENSOR_SIGNAL_STRENGTH + sensory.signalStrength + '\n');
		
		response.write('cheeseDigitalInput/' + index + '/Sa ' + (sensory.inputSA != 0) + '\n');
		response.write('cheeseDigitalInput/' + index + '/Sb ' + (sensory.inputSB != 0) + '\n');
		response.write('cheeseDigitalInput/' + index + '/Sc ' + (sensory.inputSC != 0) + '\n');
		response.write('cheeseDigitalInput/' + index + '/La ' + (sensory.inputLA != 0) + '\n');
		response.write('cheeseDigitalInput/' + index + '/Lb ' + (sensory.inputLB != 0) + '\n');
		response.write('cheeseDigitalInput/' + index + '/Lc ' + (sensory.inputLC != 0) + '\n');
		
		response.write('cheeseAnalogInput/' + index + SENSOR_INPUT + 'Sa ' + sensory.inputSA + '\n');
		response.write('cheeseAnalogInput/' + index + SENSOR_INPUT + 'Sb ' + sensory.inputSB + '\n');
		response.write('cheeseAnalogInput/' + index + SENSOR_INPUT + 'Sc ' + sensory.inputSC + '\n');
		response.write('cheeseAnalogInput/' + index + SENSOR_INPUT + 'La ' + sensory.inputLA + '\n');
		response.write('cheeseAnalogInput/' + index + SENSOR_INPUT + 'Lb ' + sensory.inputLB + '\n');
		response.write('cheeseAnalogInput/' + index + SENSOR_INPUT + 'Lc ' + sensory.inputLC + '\n');
		response.write('cheeseAnalogInput/' + index + SENSOR_ULTRASONIC + sensory.ultrasonic + '\n');
		response.write('cheeseAnalogInput/' + index + SENSOR_ENCODER + sensory.encoder + '\n');
		response.write('cheeseAnalogInput/' + index + SENSOR_TEMPERATURE + sensory.temperature + '\n');
		response.write('cheeseAnalogInput/' + index + SENSOR_HUMIDITY + sensory.humidity + '\n');
		
		//response.write('cheeseAnalogInput/' + index + SENSOR_GESTURE + sensory.stepCount + '\n');
		
		response.write('cheeseStepNumber/' + index + ' ' + sensory.stepCount + '\n');
		
		response.write('cheeseSerial/' + index + ' ' + sensory.serial + '\n');
		
		var matrix = this.matrix;
		if(matrix.active) {
			var positions = matrix.getSpritePositions();
			var tag = index + '/';
			var pos;
			for(var i in positions) {
				pos = positions[i];
				response.write('cheeseLedMatrixSpritePosition/' + tag + i + X_POSITION + pos.x + '\n');
				response.write('cheeseLedMatrixSpritePosition/' + tag + i + Y_POSITION + pos.y + '\n');
			}
			var touching = this.touching;
			if(touching) {
				var data = touching.background;
				for(var i in data) {
					response.write('cheeseLedMatrixTouching/' + tag + i + TOUCH_BACKGROUND + data[i] + '\n');
				}
				data = touching.otherSprite;
				for(var i in data) {
					response.write('cheeseLedMatrixTouching/' + tag + i + TOUCH_OTHER_SPRITE + data[i] + '\n');
				}
				data = touching.walls;
				var wall;
				for(var i in data) {
					wall = data[i];
					response.write('cheeseLedMatrixTouching/' + tag + i + TOUCH_LEFT_WALL + wall.left + '\n');
					response.write('cheeseLedMatrixTouching/' + tag + i + TOUCH_RIGHT_WALL + wall.right + '\n');
					response.write('cheeseLedMatrixTouching/' + tag + i + TOUCH_TOP_WALL + wall.top + '\n');
					response.write('cheeseLedMatrixTouching/' + tag + i + TOUCH_BOTTOM_WALL + wall.bottom + '\n');
					response.write('cheeseLedMatrixTouching/' + tag + i + TOUCH_ANY_WALL + (wall.left || wall.right || wall.top || wall.bottom) + '\n');
				}
				data = touching.sprites;
				var cell;
				for(var i in data) {
					cell = data[i];
					for(var j in cell) {
						response.write('cheeseLedMatrixTouchingSprite/' + tag + i + '/' + j + ' ' + cell[j] + '\n');
					}
				}
			}
		}
	}
	this.freeFall = false;
	this.tap = false;
	if(this.received == false) this.active = false;
};

Robot.prototype.setInputPortTo = function(request) {
	var value = MODE_INPUTS_SL[request.params.value];
	if(typeof value == 'number') {
		var motoring = this.motoring;
		var port = request.params.port;
		var mode = motoring.mode[port];
		if(mode) {
			if(port == 'Sa' || port == 'Sb' || port == 'Sc') {
				motoring.modeSound.write(Cheese.MODE_SOUND_INTERNAL);
			}
			mode.write(value);
		}
	}
};

Robot.prototype.setInputRangeTo = function(request) {
	var lower1 = parseFloat(request.params.lower1);
	var upper1 = parseFloat(request.params.upper1);
	var lower2 = parseFloat(request.params.lower2);
	var upper2 = parseFloat(request.params.upper2);
	if((typeof lower1 == 'number') && (typeof upper1 == 'number') && (typeof lower2 == 'number') && (typeof upper2 == 'number') && (lower1 != upper1)) {
		var range = this.range[request.params.port];
		if(range) {
			range.has_zero = false;
			range.in_lower = lower1;
			range.in_upper = upper1;
			range.out_lower = lower2;
			range.out_upper = upper2;
			range.active = true;
		}
	}
};

Robot.prototype.setInputZeroAndRangeTo = function(request) {
	var zero = parseFloat(request.params.zero);
	var lower1 = parseFloat(request.params.lower1);
	var upper1 = parseFloat(request.params.upper1);
	var lower2 = parseFloat(request.params.lower2);
	var upper2 = parseFloat(request.params.upper2);
	if((typeof zero == 'number') && (typeof lower1 == 'number') && (typeof upper1 == 'number') && (typeof lower2 == 'number') && (typeof upper2 == 'number') && (zero != lower1 || zero != upper1)) {
		var range = this.range[request.params.port];
		if(range) {
			if(zero == lower1 || zero == upper1) {
				range.has_zero = false;
			} else {
				range.has_zero = true;
			}
			range.zero = zero;
			range.in_lower = lower1;
			range.in_upper = upper1;
			range.out_lower = lower2;
			range.out_upper = upper2;
			range.active = true;
		}
	}
};

Robot.prototype.startSensor = function(request) {
	var sensor = SPECIAL_SENSORS[request.params.sensor];
	if(typeof sensor == 'number') {
		switch(sensor) {
			case ULTRASONIC:
				this.motoring.mode.La.write(Cheese.MODE_L_ULTRASONIC);
				break;
			case ENCODER:
				this.motoring.mode.La.write(Cheese.MODE_L_ENCODER);
				break;
			case DS18B20:
				this.motoring.mode.La.write(Cheese.MODE_L_DS18B20);
				break;
			case DHT11:
				this.motoring.mode.La.write(Cheese.MODE_L_DHT11);
				break;
			case DHT21:
				this.motoring.mode.La.write(Cheese.MODE_L_DHT21);
				break;
			case DHT22:
				this.motoring.mode.La.write(Cheese.MODE_L_DHT22);
				break;
			case GESTURE:
				break;
		}
	}
};

Robot.prototype.resetEncoder = function(request) {
	this.motoring.resetEncoder.write();
};

Robot.prototype.setSoundPortTo = function(request) {
	var mode = MODE_SOUNDS[request.params.port];
	if(typeof mode == 'number') {
		this.motoring.modeSound.write(mode);
	}
};

Robot.prototype.playSound = function(request) {
	var sound = SOUNDS[request.params.sound];
	var motoring = this.motoring;
	motoring.buzzer.write(0);
	motoring.note.write(0);
	if(sound) {
		this.runSound(sound);
	}
};

Robot.prototype.playSoundTimes = function(request) {
	var sound = SOUNDS[request.params.sound];
	var count = parseInt(request.params.count);
	var motoring = this.motoring;
	motoring.buzzer.write(0);
	motoring.note.write(0);
	if(sound && count) {
		this.runSound(sound, count);
	}
};

Robot.prototype.playSoundTimesUntilDone = function(request) {
	var sound = SOUNDS[request.params.sound];
	var count = parseInt(request.params.count);
	var motoring = this.motoring;
	motoring.buzzer.write(0);
	motoring.note.write(0);
	if(sound && count) {
		this.soundId = parseInt(request.params.id);
		this.waits.push(this.soundId);
		this.runSound(sound, count);
	}
};

Robot.prototype.changeBuzzerBy = function(request) {
	var hz = parseFloat(request.params.hz);
	var motoring = this.motoring;
	if(typeof hz == 'number') {
		hz += motoring.buzzer.read();
		motoring.buzzer.write(hz);
	}
	motoring.note.write(0);
	this.runSound(0);
};

Robot.prototype.setBuzzerTo = function(request) {
	var hz = parseFloat(request.params.hz);
	var motoring = this.motoring;
	if(typeof hz == 'number') {
		motoring.buzzer.write(hz);
	}
	motoring.note.write(0);
	this.runSound(0);
};

Robot.prototype.clearSound = function(request) {
	var motoring = this.motoring;
	motoring.buzzer.write(0);
	motoring.note.write(0);
	this.runSound(0);
};

Robot.prototype.playNote = function(request) {
	var note = NOTES[request.params.note];
	var octave = parseInt(request.params.octave);
	var motoring = this.motoring;
	motoring.buzzer.write(0);
	if(note && octave && octave > 0 && octave < 8) {
		note += (octave - 1) * 12;
		motoring.note.write(note);
	}
	this.runSound(0);
};

Robot.prototype.playNoteForBeats = function(request) {
	var self = this;
	var note = NOTES[request.params.note];
	var octave = parseInt(request.params.octave);
	var beat = BEATS[request.params.beat];
	if(!beat) {
		beat = parseFloat(request.params.beat);
	}
	var motoring = self.motoring;
	motoring.buzzer.write(0);
	if(note && octave && octave > 0 && octave < 8 && beat && beat > 0 && self.tempo > 0) {
		var id = parseInt(request.params.id);
		self.waits.push(id);
		note += (octave - 1) * 12;
		motoring.note.write(note);
		var timeout = beat * 60 * 1000 / self.tempo;
		var tail = 0;
		if(timeout > 100) {
			tail = 100;
		}
		if(tail > 0) {
			var timer1 = setTimeout(function() {
				motoring.note.write(0);
				self.removeTimeout(timer1);
			}, timeout - tail);
			self.timeouts.push(timer1);
		}
		var timer2 = setTimeout(function() {
			motoring.note.write(0);
			self.removeWait(id);
			self.removeTimeout(timer2);
		}, timeout);
		self.timeouts.push(timer2);
	}
	self.runSound(0);
};

Robot.prototype.restForBeats = function(request) {
	var self = this;
	var beat = BEATS[request.params.beat];
	if(!beat) {
		beat = parseFloat(request.params.beat);
	}
	var motoring = self.motoring;
	motoring.buzzer.write(0);
	motoring.note.write(0);
	self.runSound(0);
	if(beat && beat > 0 && self.tempo > 0) {
		var id = parseInt(request.params.id);
		self.waits.push(id);
		var timer = setTimeout(function() {
			self.removeWait(id);
			self.removeTimeout(timer);
		}, beat * 60 * 1000 / self.tempo);
		self.timeouts.push(timer);
	}
};

Robot.prototype.changeTempoBy = function(request) {
	var bpm = parseFloat(request.params.bpm);
	if(typeof bpm == 'number') {
		this.tempo += bpm;
		if(this.tempo < 1) this.tempo = 1;
	}
};

Robot.prototype.setTempoTo = function(request) {
	var bpm = parseFloat(request.params.bpm);
	if(typeof bpm == 'number') {
		this.tempo = bpm;
		if(this.tempo < 1) this.tempo = 1;
	}
};

Robot.prototype.runSound = function(sound, count) {
	if(typeof count != 'number') count = 1;
	if(count < 0) count = -1;
	if(count) {
		this.sound = sound;
		this.soundRepeat = count;
		this.motoring.sound.write(sound);
	}
};

Robot.prototype.changeServoMotorAngleBy = function(request) {
	var value = parseInt(request.params.value);
	if(typeof value == 'number') {
		var motoring = this.motoring;
		var port = request.params.port;
		var mode = motoring.mode[port];
		var output = motoring.output[port];
		if(mode && output) {
			if(port == 'Sa' || port == 'Sb' || port == 'Sc') {
				motoring.modeSound.write(Cheese.MODE_SOUND_INTERNAL);
				mode.write(Cheese.MODE_S_SERVO_OUTPUT);
			} else if(port == 'La' || port == 'Lb' || port == 'Lc') {
				mode.write(Cheese.MODE_L_SERVO_OUTPUT);
			} else if(port == 'Ma' || port == 'Mb') {
				mode.write(Cheese.MODE_M_SERVO_DUAL);
			} else if(port == 'Mab') {
				mode.write(Cheese.MODE_M_SERVO_MONO);
			}
			value += output.read();
			if(value < 1) value = 1;
			else if(value > 180) value = 180;
			output.write(value);
		}
	}
};

Robot.prototype.setServoMotorAngleTo = function(request) {
	var value = parseInt(request.params.value);
	if(typeof value == 'number') {
		var motoring = this.motoring;
		var port = request.params.port;
		var mode = motoring.mode[port];
		var output = motoring.output[port];
		if(mode && output) {
			if(port == 'Sa' || port == 'Sb' || port == 'Sc') {
				motoring.modeSound.write(Cheese.MODE_SOUND_INTERNAL);
				mode.write(Cheese.MODE_S_SERVO_OUTPUT);
			} else if(port == 'La' || port == 'Lb' || port == 'Lc') {
				mode.write(Cheese.MODE_L_SERVO_OUTPUT);
			} else if(port == 'Ma' || port == 'Mb') {
				mode.write(Cheese.MODE_M_SERVO_DUAL);
			} else if(port == 'Mab') {
				mode.write(Cheese.MODE_M_SERVO_MONO);
			}
			if(value < 1) value = 1;
			else if(value > 180) value = 180;
			output.write(value);
		}
	}
};

Robot.prototype.turnOffServoMotor = function(request) {
	var motoring = this.motoring;
	var port = request.params.port;
	var mode = motoring.mode[port];
	var output = motoring.output[port];
	if(mode && output) {
		if(port == 'Sa' || port == 'Sb' || port == 'Sc') {
			motoring.modeSound.write(Cheese.MODE_SOUND_INTERNAL);
			mode.write(Cheese.MODE_S_SERVO_OUTPUT);
		} else if(port == 'La' || port == 'Lb' || port == 'Lc') {
			mode.write(Cheese.MODE_L_SERVO_OUTPUT);
		} else if(port == 'Ma' || port == 'Mb') {
			mode.write(Cheese.MODE_M_SERVO_DUAL);
		} else if(port == 'Mab') {
			mode.write(Cheese.MODE_M_SERVO_MONO);
		}
		output.write(0);
	}
};

Robot.prototype.changeDcMotorVelocityBy = function(request) {
	var value = parseInt(request.params.value);
	if(typeof value == 'number') {
		var motoring = this.motoring;
		var port = request.params.port;
		var mode = motoring.mode[port];
		var output = motoring.output[port];
		if(mode && output) {
			mode.write(Cheese.MODE_M_DC_50);
			value += output.read();
			if(value < -100) value = -100;
			else if(value > 100) value = 100;
			output.write(value);
		}
	}
};

Robot.prototype.setDcMotorVelocityTo = function(request) {
	var value = parseInt(request.params.value);
	if(typeof value == 'number') {
		var motoring = this.motoring;
		var port = request.params.port;
		var mode = motoring.mode[port];
		var output = motoring.output[port];
		if(mode && output) {
			mode.write(Cheese.MODE_M_DC_50);
			if(value < -100) value = -100;
			else if(value > 100) value = 100;
			output.write(value);
		}
	}
};

Robot.prototype.stopDcMotor = function(request) {
	var motoring = this.motoring;
	var port = request.params.port;
	var mode = motoring.mode[port];
	var output = motoring.output[port];
	if(mode && output) {
		mode.write(Cheese.MODE_M_DC_50);
		output.write(0);
	}
};

Robot.prototype.rotateStepMotor = function(request) {
	var value = parseInt(request.params.value);
	var velocity = parseInt(request.params.velocity);
	if((typeof value == 'number') && (typeof velocity == 'number')) {
		var motoring = this.motoring;
		this.stepId = parseInt(request.params.id);
		this.waits.push(this.stepId);
		motoring.mode.M.write(this.stepMode);
		if(velocity < -1000) velocity = -1000;
		else if(velocity > 1000) velocity = 1000;
		motoring.velocity.write(velocity);
		motoring.step.write(value);
	}
};

Robot.prototype.changeStepMotorVelocityBy = function(request) {
	var value = parseInt(request.params.value);
	if(typeof value == 'number') {
		var motoring = this.motoring;
		value += motoring.velocity.read();
		if(value < -1000) value = -1000;
		else if(value > 1000) value = 1000;
		motoring.velocity.write(value);
	}
};

Robot.prototype.setStepMotorVelocityTo = function(request) {
	var value = parseInt(request.params.value);
	if(typeof value == 'number') {
		if(value < -1000) value = -1000;
		else if(value > 1000) value = 1000;
		this.motoring.velocity.write(value);
	}
};

Robot.prototype.stopOffStepMotor = function(request) {
	var action = ACTION_STEP_MOTORS[request.params.action];
	if(typeof action == 'number') {
		var motoring = this.motoring;
		if(action == STOP) {
			motoring.mode.M.write(this.stepMode);
			motoring.velocity.write(0);
			motoring.step.write(0);
		} else {
			motoring.mode.M.write(Cheese.MODE_M_STEP_OFF);
		}
	}
};

Robot.prototype.setStepMotorModeTo = function(request) {
	var value = MODE_STEP_MOTORS[request.params.value];
	if(typeof value == 'number') {
		this.stepMode = value;
		this.motoring.mode.M.write(value);
	}
};

Robot.prototype.setDigitalOutputTo = function(request) {
	var value = parseInt(request.params.value);
	if(typeof value == 'number') {
		var motoring = this.motoring;
		var port = request.params.port;
		var mode = motoring.mode[port];
		var output = motoring.output[port];
		if(mode && output) {
			if(port == 'Sa' || port == 'Sb' || port == 'Sc') {
				motoring.modeSound.write(Cheese.MODE_SOUND_INTERNAL);
				mode.write(Cheese.MODE_S_DIGITAL_OUTPUT);
			} else if(port == 'La' || port == 'Lb' || port == 'Lc') {
				mode.write(Cheese.MODE_L_DIGITAL_OUTPUT);
			}
			output.write(value);
		}
	}
};

Robot.prototype.changePwmOutputBy = function(request) {
	var value = parseInt(request.params.value);
	if(typeof value == 'number') {
		var motoring = this.motoring;
		var port = PORT_PWM_OUTPUTS[request.params.port];
		var mode = motoring.mode[port];
		var output = motoring.output[port];
		if(mode && output) {
			if(port == 'Sa' || port == 'Sb' || port == 'Sc') {
				motoring.modeSound.write(Cheese.MODE_SOUND_INTERNAL);
				mode.write(Cheese.MODE_S_PWM_OUTPUT);
			} else if(port == 'La' || port == 'Lb' || port == 'Lc') {
				mode.write(Cheese.MODE_L_PWM_OUTPUT);
			} else if(port == 'Ma' || port == 'Mb') {
				mode.write(Cheese.MODE_M_DC);
			}
			value += output.read();
			if(value < 0) value = 0;
			else if(value > 100) value = 100;
			output.write(value);
		}
	}
};

Robot.prototype.setPwmOutputTo = function(request) {
	var value = parseInt(request.params.value);
	if(typeof value == 'number') {
		var motoring = this.motoring;
		var port = PORT_PWM_OUTPUTS[request.params.port];
		var mode = motoring.mode[port];
		var output = motoring.output[port];
		if(mode && output) {
			if(port == 'Sa' || port == 'Sb' || port == 'Sc') {
				motoring.modeSound.write(Cheese.MODE_SOUND_INTERNAL);
				mode.write(Cheese.MODE_S_PWM_OUTPUT);
			} else if(port == 'La' || port == 'Lb' || port == 'Lc') {
				mode.write(Cheese.MODE_L_PWM_OUTPUT);
			} else if(port == 'Ma' || port == 'Mb') {
				mode.write(Cheese.MODE_M_DC);
			}
			if(value < 0) value = 0;
			else if(value > 100) value = 100;
			output.write(value);
		}
	}
};

Robot.prototype.setLedMode = function(portR, portG, portB) {
	var motoring = this.motoring;
	var modeR = motoring.mode[portR];
	var modeG = motoring.mode[portG];
	var modeB = motoring.mode[portB];
	if(modeR && modeG && modeB) {
		if(portR == 'Sa' || portR == 'Sb' || portR == 'Sc') {
			motoring.modeSound.write(Cheese.MODE_SOUND_INTERNAL);
			modeR.write(Cheese.MODE_S_PWM_OUTPUT);
		} else if(portR == 'La' || portR == 'Lb' || portR == 'Lc') {
			modeR.write(Cheese.MODE_L_PWM_OUTPUT);
		}
		if(portG == 'Sa' || portG == 'Sb' || portG == 'Sc') {
			motoring.modeSound.write(Cheese.MODE_SOUND_INTERNAL);
			modeG.write(Cheese.MODE_S_PWM_OUTPUT);
		} else if(portG == 'La' || portG == 'Lb' || portG == 'Lc') {
			modeG.write(Cheese.MODE_L_PWM_OUTPUT);
		}
		if(portB == 'Sa' || portB == 'Sb' || portB == 'Sc') {
			motoring.modeSound.write(Cheese.MODE_SOUND_INTERNAL);
			modeB.write(Cheese.MODE_S_PWM_OUTPUT);
		} else if(portB == 'La' || portB == 'Lb' || portB == 'Lc') {
			modeB.write(Cheese.MODE_L_PWM_OUTPUT);
		}
	}
};

Robot.prototype.setLedTypeTo = function(request) {
	var type = LED_TYPES[request.params.type];
	if(type) {
		var rgbPort = {};
		switch(type) {
			case ARGB:
			case ARBG:
			case AGRB:
			case AGBR:
			case ABRG:
			case ABGR: rgbPort.anode = true; break;
			default: rgbPort.anode = false; break;
		}
		if(request.params.port == 'L') {
			switch(type) {
				case CRGB: rgbPort.r = 'La'; rgbPort.g = 'Lb'; rgbPort.b = 'Lc'; break;
				case CRBG: rgbPort.r = 'La'; rgbPort.b = 'Lb'; rgbPort.g = 'Lc'; break;
				case CGRB: rgbPort.g = 'La'; rgbPort.r = 'Lb'; rgbPort.b = 'Lc'; break;
				case CGBR: rgbPort.g = 'La'; rgbPort.b = 'Lb'; rgbPort.r = 'Lc'; break;
				case CBRG: rgbPort.b = 'La'; rgbPort.r = 'Lb'; rgbPort.g = 'Lc'; break;
				case CBGR: rgbPort.b = 'La'; rgbPort.g = 'Lb'; rgbPort.r = 'Lc'; break;
				case ARGB: rgbPort.r = 'La'; rgbPort.g = 'Lb'; rgbPort.b = 'Lc'; break;
				case ARBG: rgbPort.r = 'La'; rgbPort.b = 'Lb'; rgbPort.g = 'Lc'; break;
				case AGRB: rgbPort.g = 'La'; rgbPort.r = 'Lb'; rgbPort.b = 'Lc'; break;
				case AGBR: rgbPort.g = 'La'; rgbPort.b = 'Lb'; rgbPort.r = 'Lc'; break;
				case ABRG: rgbPort.b = 'La'; rgbPort.r = 'Lb'; rgbPort.g = 'Lc'; break;
				case ABGR: rgbPort.b = 'La'; rgbPort.g = 'Lb'; rgbPort.r = 'Lc'; break;
				default: rgbPort.r = 'La'; rgbPort.g = 'Lb'; rgbPort.b = 'Lc'; break;
			}
		} else {
			switch(type) {
				case CRGB: rgbPort.r = 'Sa'; rgbPort.g = 'Sb'; rgbPort.b = 'Sc'; break;
				case CRBG: rgbPort.r = 'Sa'; rgbPort.b = 'Sb'; rgbPort.g = 'Sc'; break;
				case CGRB: rgbPort.g = 'Sa'; rgbPort.r = 'Sb'; rgbPort.b = 'Sc'; break;
				case CGBR: rgbPort.g = 'Sa'; rgbPort.b = 'Sb'; rgbPort.r = 'Sc'; break;
				case CBRG: rgbPort.b = 'Sa'; rgbPort.r = 'Sb'; rgbPort.g = 'Sc'; break;
				case CBGR: rgbPort.b = 'Sa'; rgbPort.g = 'Sb'; rgbPort.r = 'Sc'; break;
				case ARGB: rgbPort.r = 'Sa'; rgbPort.g = 'Sb'; rgbPort.b = 'Sc'; break;
				case ARBG: rgbPort.r = 'Sa'; rgbPort.b = 'Sb'; rgbPort.g = 'Sc'; break;
				case AGRB: rgbPort.g = 'Sa'; rgbPort.r = 'Sb'; rgbPort.b = 'Sc'; break;
				case AGBR: rgbPort.g = 'Sa'; rgbPort.b = 'Sb'; rgbPort.r = 'Sc'; break;
				case ABRG: rgbPort.b = 'Sa'; rgbPort.r = 'Sb'; rgbPort.g = 'Sc'; break;
				case ABGR: rgbPort.b = 'Sa'; rgbPort.g = 'Sb'; rgbPort.r = 'Sc'; break;
				default: rgbPort.r = 'Sa'; rgbPort.g = 'Sb'; rgbPort.b = 'Sc'; break;
			}
		}
		rgbPort.cr = 0;
		rgbPort.cg = 0;
		rgbPort.cb = 0;
		this.rgbPort[request.params.led] = rgbPort;
	}
};

Robot.prototype.setLedToColor = function(request) {
	var rgbPort = this.rgbPort[request.params.led];
	var color = RGB_COLORS[request.params.color];
	if(rgbPort && color) {
		var output = this.motoring.output;
		var outputR = output[rgbPort.r];
		var outputG = output[rgbPort.g];
		var outputB = output[rgbPort.b];
		if(outputR && outputG && outputB) {
			rgbPort.cr = color[0];
			rgbPort.cg = color[1];
			rgbPort.cb = color[2];
			var r = parseInt(color[0] * 100 / 255.0);
			var g = parseInt(color[1] * 100 / 255.0);
			var b = parseInt(color[2] * 100 / 255.0);
			if(r < 0) r = 0;
			else if(r > 100) r = 100;
			if(g < 0) g = 0;
			else if(g > 100) g = 100;
			if(b < 0) b = 0;
			else if(b > 100) b = 100;
			if(rgbPort.anode) {
				r = 100 - r;
				g = 100 - g;
				b = 100 - b;
			}
			this.setLedMode(rgbPort.r, rgbPort.g, rgbPort.b);
			outputR.write(r);
			outputG.write(g);
			outputB.write(b);
		}
	}
};

Robot.prototype.changeLedByRGB = function(request) {
	var rgbPort = this.rgbPort[request.params.led];
	var r = parseInt(request.params.r);
	var g = parseInt(request.params.g);
	var b = parseInt(request.params.b);
	if(rgbPort && (typeof r == 'number') && (typeof g == 'number') && (typeof b == 'number')) {
		var output = this.motoring.output;
		var outputR = output[rgbPort.r];
		var outputG = output[rgbPort.g];
		var outputB = output[rgbPort.b];
		if(outputR && outputG && outputB) {
			r += rgbPort.cr;
			g += rgbPort.cg;
			b += rgbPort.cb;
			if(r < 0) r = 0;
			else if(r > 255) r = 255;
			if(g < 0) g = 0;
			else if(g > 255) g = 255;
			if(b < 0) b = 0;
			else if(b > 255) b = 255;
			rgbPort.cr = r;
			rgbPort.cg = g;
			rgbPort.cb = b;
			r = parseInt(r * 100 / 255.0);
			g = parseInt(g * 100 / 255.0);
			b = parseInt(b * 100 / 255.0);
			if(r < 0) r = 0;
			else if(r > 100) r = 100;
			if(g < 0) g = 0;
			else if(g > 100) g = 100;
			if(b < 0) b = 0;
			else if(b > 100) b = 100;
			if(rgbPort.anode) {
				r = 100 - r;
				g = 100 - g;
				b = 100 - b;
			}
			this.setLedMode(rgbPort.r, rgbPort.g, rgbPort.b);
			outputR.write(r);
			outputG.write(g);
			outputB.write(b);
		}
	}
};

Robot.prototype.setLedToRGB = function(request) {
	var rgbPort = this.rgbPort[request.params.led];
	var r = parseInt(request.params.r);
	var g = parseInt(request.params.g);
	var b = parseInt(request.params.b);
	if(rgbPort && (typeof r == 'number') && (typeof g == 'number') && (typeof b == 'number')) {
		var output = this.motoring.output;
		var outputR = output[rgbPort.r];
		var outputG = output[rgbPort.g];
		var outputB = output[rgbPort.b];
		if(outputR && outputG && outputB) {
			if(r < 0) r = 0;
			else if(r > 255) r = 255;
			if(g < 0) g = 0;
			else if(g > 255) g = 255;
			if(b < 0) b = 0;
			else if(b > 255) b = 255;
			rgbPort.cr = r;
			rgbPort.cg = g;
			rgbPort.cb = b;
			r = parseInt(r * 100 / 255.0);
			g = parseInt(g * 100 / 255.0);
			b = parseInt(b * 100 / 255.0);
			if(r < 0) r = 0;
			else if(r > 100) r = 100;
			if(g < 0) g = 0;
			else if(g > 100) g = 100;
			if(b < 0) b = 0;
			else if(b > 100) b = 100;
			if(rgbPort.anode) {
				r = 100 - r;
				g = 100 - g;
				b = 100 - b;
			}
			this.setLedMode(rgbPort.r, rgbPort.g, rgbPort.b);
			outputR.write(r);
			outputG.write(g);
			outputB.write(b);
		}
	}
};

Robot.prototype.clearLed = function(request) {
	var rgbPort = this.rgbPort[request.params.led];
	if(rgbPort) {
		var output = this.motoring.output;
		var outputR = output[rgbPort.r];
		var outputG = output[rgbPort.g];
		var outputB = output[rgbPort.b];
		if(outputR && outputG && outputB) {
			rgbPort.cr = 0;
			rgbPort.cg = 0;
			rgbPort.cb = 0;
			this.setLedMode(rgbPort.r, rgbPort.g, rgbPort.b);
			if(rgbPort.anode) {
				outputR.write(100);
				outputG.write(100);
				outputB.write(100);
			} else {
				outputR.write(0);
				outputG.write(0);
				outputB.write(0);
			}
		}
	}
};

Robot.prototype.refreshLedMatrix = function(id) {
	this.matrix.active = true;
	if(this.matrix.auto) {
		this.addI2cWait(id);
		var packet = this.matrix.update();
		this.motoring.writeI2c.write(packet);
	}
};

Robot.prototype.clearLedMatrixBackground = function(request) {
	var id = parseInt(request.params.id);
	this.matrix.clearBackground();
	this.refreshLedMatrix(id);
};

Robot.prototype.clearLedMatrix = function(request) {
	var target = VALUES[request.params.target];
	if(typeof target == 'number') {
		var id = parseInt(request.params.id);
		if(target == BACKGROUND) this.matrix.clearBackground();
		else this.matrix.clearAll();
		this.refreshLedMatrix(id);
	}
};

Robot.prototype.turnLedMatrixOnOff = function(request) {
	var x = parseInt(request.params.x);
	var y = parseInt(request.params.y);
	var action = VALUES[request.params.action];
	if((typeof x == 'number') && (typeof y == 'number') && (typeof action == 'number')) {
		var id = parseInt(request.params.id);
		this.matrix.setBackgroundPixel(x, y, action);
		this.refreshLedMatrix(id);
	}
};

Robot.prototype.drawLedMatrixShape = function(request) {
	var shape = SHAPES[request.params.shape];
	if(shape && this.matrix.drawBackgroundShape(0, 0, shape)) {
		var id = parseInt(request.params.id);
		this.refreshLedMatrix(id);
	}
};

Robot.prototype.drawLedMatrixShapeAt = function(request) {
	var x = parseInt(request.params.x);
	var y = parseInt(request.params.y);
	if((typeof x == 'number') && (typeof y == 'number')) {
		var shape = SHAPES[request.params.shape];
		if(shape && this.matrix.drawBackgroundShape(x, y, shape)) {
			var id = parseInt(request.params.id);
			this.refreshLedMatrix(id);
		}
	}
};

Robot.prototype.drawLedMatrixStringAt = function(request) {
	var x = parseInt(request.params.x);
	var y = parseInt(request.params.y);
	if((typeof x == 'number') && (typeof y == 'number')) {
		if(this.matrix.drawBackgroundString(x, y, request.params.text)) {
			var id = parseInt(request.params.id);
			this.refreshLedMatrix(id);
		}
	}
};

Robot.prototype.drawLedMatrixBinaryAt = function(request) {
	var x = parseInt(request.params.x);
	var y = parseInt(request.params.y);
	if((typeof x == 'number') && (typeof y == 'number')) {
		if(this.matrix.drawBackgroundBinary(x, y, request.params.value)) {
			var id = parseInt(request.params.id);
			this.refreshLedMatrix(id);
		}
	}
};

Robot.prototype.moveLedMatrixTo = function(request) {
	var direction = DIRECTIONS[request.params.direction];
	if(typeof direction == 'number') {
		var id = parseInt(request.params.id);
		switch(direction) {
			case LEFT: this.matrix.moveBackgroundBy(-1, 0); break;
			case RIGHT: this.matrix.moveBackgroundBy(1, 0); break;
			case TOP: this.matrix.moveBackgroundBy(0, -1); break;
			case BOTTOM: this.matrix.moveBackgroundBy(0, 1); break;
		}
		this.refreshLedMatrix(id);
	}
};

Robot.prototype.moveLedMatrixBackgroundBy = function(request) {
	var x = parseInt(request.params.x);
	var y = parseInt(request.params.y);
	if((typeof x == 'number') && (typeof y == 'number')) {
		var id = parseInt(request.params.id);
		this.matrix.moveBackgroundBy(x, y);
		this.refreshLedMatrix(id);
	}
};

Robot.prototype.moveLedMatrixBy = function(request) {
	var target = VALUES[request.params.target];
	var x = parseInt(request.params.x);
	var y = parseInt(request.params.y);
	if((typeof target == 'number') && (typeof x == 'number') && (typeof y == 'number')) {
		var id = parseInt(request.params.id);
		if(target == BACKGROUND) this.matrix.moveBackgroundBy(x, y);
		else this.matrix.moveAllBy(x, y);
		this.refreshLedMatrix(id);
	}
};

Robot.prototype.clearShowHideLedMatrixSprite = function(request) {
	var sprite = parseInt(request.params.sprite);
	var action = VALUES[request.params.action];
	if((typeof sprite == 'number') && (typeof action == 'number')) {
		var id = parseInt(request.params.id);
		if(action == CLEAR) this.matrix.clearSprite(sprite);
		else if(action == SHOW) this.matrix.showSprite(sprite);
		else this.matrix.hideSprite(sprite);
		this.refreshLedMatrix(id);
	}
};

Robot.prototype.setLedMatrixSpriteShape = function(request) {
	var sprite = parseInt(request.params.sprite);
	var shape = SHAPES[request.params.shape];
	if(shape && (typeof sprite == 'number')) {
		if(this.matrix.setSpriteShape(sprite, shape)) {
			var id = parseInt(request.params.id);
			this.refreshLedMatrix(id);
		}
	}
};

Robot.prototype.setLedMatrixSpriteString = function(request) {
	var sprite = parseInt(request.params.sprite);
	if(typeof sprite == 'number') {
		if(this.matrix.setSpriteString(sprite, request.params.text)) {
			var id = parseInt(request.params.id);
			this.refreshLedMatrix(id);
		}
	}
};

Robot.prototype.setLedMatrixSpriteBinary = function(request) {
	var sprite = parseInt(request.params.sprite);
	if(typeof sprite == 'number') {
		if(this.matrix.setSpriteBinary(sprite, request.params.value)) {
			var id = parseInt(request.params.id);
			this.refreshLedMatrix(id);
		}
	}
};

Robot.prototype.changeLedMatrixSpritePositionsBy = function(request) {
	var sprite = parseInt(request.params.sprite);
	var x = parseInt(request.params.x);
	var y = parseInt(request.params.y);
	if((typeof sprite == 'number') && (typeof x == 'number') && (typeof y == 'number')) {
		var id = parseInt(request.params.id);
		this.matrix.changeSpritePositionsBy(sprite, x, y);
		this.refreshLedMatrix(id);
	}
};

Robot.prototype.setLedMatrixSpritePositionsTo = function(request) {
	var sprite = parseInt(request.params.sprite);
	var x = parseInt(request.params.x);
	var y = parseInt(request.params.y);
	if((typeof sprite == 'number') && (typeof x == 'number') && (typeof y == 'number')) {
		var id = parseInt(request.params.id);
		this.matrix.setSpritePositionsTo(sprite, x, y);
		this.refreshLedMatrix(id);
	}
};

Robot.prototype.setLedMatrixSpritePositionTo = function(request) {
	var sprite = parseInt(request.params.sprite);
	var position = VALUES[request.params.position];
	var value = parseInt(request.params.value);
	if((typeof sprite == 'number') && (typeof position == 'number') && (typeof value == 'number')) {
		var id = parseInt(request.params.id);
		this.matrix.setSpritePositionTo(sprite, position == POS_X, value);
		this.refreshLedMatrix(id);
	}
};

Robot.prototype.rotateLedMatrixSprite = function(request) {
	var sprite = parseInt(request.params.sprite);
	var direction = VALUES[request.params.direction];
	if((typeof sprite == 'number') && (typeof direction == 'number')) {
		var id = parseInt(request.params.id);
		this.matrix.rotateSprite(sprite, direction == CLOCKWISE);
		this.refreshLedMatrix(id);
	}
};

Robot.prototype.flipLedMatrixSprite = function(request) {
	var sprite = parseInt(request.params.sprite);
	var direction = VALUES[request.params.direction];
	if((typeof sprite == 'number') && (typeof direction == 'number')) {
		var id = parseInt(request.params.id);
		this.matrix.flipSprite(sprite, direction == LEFT_RIGHT);
		this.refreshLedMatrix(id);
	}
};

Robot.prototype.stampLedMatrixSprite = function(request) {
	var sprite = parseInt(request.params.sprite);
	if(typeof sprite == 'number') {
		var id = parseInt(request.params.id);
		this.matrix.stampSprite(sprite);
		this.refreshLedMatrix(id);
	}
};

Robot.prototype.changeLedMatrixBrightnessBy = function(request) {
	var value = parseInt(request.params.value);
	if(typeof value == 'number') {
		var id = parseInt(request.params.id);
		this.matrix.changeBrightnessBy(value);
		this.refreshLedMatrix(id);
	}
};

Robot.prototype.setLedMatrixBrightnessTo = function(request) {
	var value = parseInt(request.params.value);
	if(typeof value == 'number') {
		var id = parseInt(request.params.id);
		this.matrix.setBrightnessTo(value);
		this.refreshLedMatrix(id);
	}
};

Robot.prototype.setLedMatrixDirectionTo = function(request) {
	var direction = VALUES[request.params.direction];
	if(typeof direction == 'number') {
		var id = parseInt(request.params.id);
		this.matrix.setDirection(direction);
		this.refreshLedMatrix(id);
	}
};

Robot.prototype.setLedMatrixModeTo = function(request) {
	var auto = VALUES[request.params.auto];
	if(typeof auto == 'number') {
		var id = parseInt(request.params.id);
		this.matrix.setAuto(auto == AUTO);
		this.refreshLedMatrix(id);
	}
};

Robot.prototype.updateLedMatrix = function(request) {
	this.matrix.active = true;
	this.addI2cWait(parseInt(request.params.id));
	var packet = this.matrix.update();
	this.motoring.writeI2c.write(packet);
};

Robot.prototype.refreshNeopixel = function(id) {
	if(this.neopixel.auto) {
		this.addNeopixelWait(id);
		var packet = this.neopixel.update();
		if(packet) {
			this.motoring.neopixel.write(packet);
		}
	}
};

Robot.prototype.neopixelSetNumberAndTypeTo = function(request) {
	var num = parseInt(request.params.num);
	var type = NEOPIXEL_TYPES[request.params.type];
	if((typeof num == 'number') && (typeof type == 'number')) {
		var id = parseInt(request.params.id);
		this.neopixel.setNumLeds(num);
		this.neopixel.setType(type);
		this.motoring.mode.La.write(Cheese.MODE_L_NEOPIXEL_I2C);
	}
};

Robot.prototype.neopixelSetAllLedsToColor = function(request) {
	var color = RGB_COLORS[request.params.color];
	if(color) {
		var id = parseInt(request.params.id);
		this.neopixel.setAllRgb(color[0], color[1], color[2]);
		this.refreshNeopixel(id);
	}
};

Robot.prototype.neopixelChangeAllLedsByRGB = function(request) {
	var r = parseInt(request.params.r);
	var g = parseInt(request.params.g);
	var b = parseInt(request.params.b);
	if((typeof r == 'number') && (typeof g == 'number') && (typeof b == 'number')) {
		var id = parseInt(request.params.id);
		this.neopixel.changeAllRgb(r, g, b);
		this.refreshNeopixel(id);
	}
};

Robot.prototype.neopixelSetAllLedsToRGB = function(request) {
	var r = parseInt(request.params.r);
	var g = parseInt(request.params.g);
	var b = parseInt(request.params.b);
	if((typeof r == 'number') && (typeof g == 'number') && (typeof b == 'number')) {
		var id = parseInt(request.params.id);
		this.neopixel.setAllRgb(r, g, b);
		this.refreshNeopixel(id);
	}
};

Robot.prototype.neopixelClearAllLeds = function(request) {
	var id = parseInt(request.params.id);
	this.neopixel.clearAll();
	this.refreshNeopixel(id);
};

Robot.prototype.neopixelSetLedAtToColor = function(request) {
	var pos = parseInt(request.params.pos);
	var color = RGB_COLORS[request.params.color];
	if(color && (typeof pos == 'number')) {
		var id = parseInt(request.params.id);
		this.neopixel.setRgbAt(pos - 1, color[0], color[1], color[2]);
		this.refreshNeopixel(id);
	}
};

Robot.prototype.neopixelChangeLedAtByRGB = function(request) {
	var pos = parseInt(request.params.pos);
	var r = parseInt(request.params.r);
	var g = parseInt(request.params.g);
	var b = parseInt(request.params.b);
	if((typeof pos == 'number') && (typeof r == 'number') && (typeof g == 'number') && (typeof b == 'number')) {
		var id = parseInt(request.params.id);
		this.neopixel.changeRgbAt(pos - 1, r, g, b);
		this.refreshNeopixel(id);
	}
};

Robot.prototype.neopixelSetLedAtToRGB = function(request) {
	var pos = parseInt(request.params.pos);
	var r = parseInt(request.params.r);
	var g = parseInt(request.params.g);
	var b = parseInt(request.params.b);
	if((typeof pos == 'number') && (typeof r == 'number') && (typeof g == 'number') && (typeof b == 'number')) {
		var id = parseInt(request.params.id);
		this.neopixel.setRgbAt(pos - 1, r, g, b);
		this.refreshNeopixel(id);
	}
};

Robot.prototype.neopixelClearLedAt = function(request) {
	var pos = parseInt(request.params.pos);
	if(typeof pos == 'number') {
		var id = parseInt(request.params.id);
		this.neopixel.clearAt(pos - 1);
		this.refreshNeopixel(id);
	}
};

Robot.prototype.neopixelSetLedFromToColor = function(request) {
	var from = parseInt(request.params.from);
	var to = parseInt(request.params.to);
	var interval = parseInt(request.params.interval);
	var color = RGB_COLORS[request.params.color];
	if(color && (typeof from == 'number') && (typeof to == 'number') && (typeof interval == 'number')) {
		var id = parseInt(request.params.id);
		this.neopixel.setRgbFromTo(from - 1, to - 1, interval, color[0], color[1], color[2]);
		this.refreshNeopixel(id);
	}
};

Robot.prototype.neopixelChangeLedFromByRGB = function(request) {
	var from = parseInt(request.params.from);
	var to = parseInt(request.params.to);
	var interval = parseInt(request.params.interval);
	var r = parseInt(request.params.r);
	var g = parseInt(request.params.g);
	var b = parseInt(request.params.b);
	if((typeof from == 'number') && (typeof to == 'number') && (typeof interval == 'number') && (typeof r == 'number') && (typeof g == 'number') && (typeof b == 'number')) {
		var id = parseInt(request.params.id);
		this.neopixel.changeRgbFromTo(from - 1, to - 1, interval, r, g, b);
		this.refreshNeopixel(id);
	}
};

Robot.prototype.neopixelSetLedFromToRGB = function(request) {
	var from = parseInt(request.params.from);
	var to = parseInt(request.params.to);
	var interval = parseInt(request.params.interval);
	var r = parseInt(request.params.r);
	var g = parseInt(request.params.g);
	var b = parseInt(request.params.b);
	if((typeof from == 'number') && (typeof to == 'number') && (typeof interval == 'number') && (typeof r == 'number') && (typeof g == 'number') && (typeof b == 'number')) {
		var id = parseInt(request.params.id);
		this.neopixel.setRgbFromTo(from - 1, to - 1, interval, r, g, b);
		this.refreshNeopixel(id);
	}
};

Robot.prototype.neopixelClearLedFromTo = function(request) {
	var from = parseInt(request.params.from);
	var to = parseInt(request.params.to);
	var interval = parseInt(request.params.interval);
	if((typeof from == 'number') && (typeof to == 'number') && (typeof interval == 'number')) {
		var id = parseInt(request.params.id);
		this.neopixel.clearFromTo(from - 1, to - 1, interval);
		this.refreshNeopixel(id);
	}
};

Robot.prototype.neopixelChangeBrightnessBy = function(request) {
	var value = parseInt(request.params.value);
	if(typeof value == 'number') {
		var id = parseInt(request.params.id);
		this.neopixel.changeBrightnessBy(value);
		this.refreshNeopixel(id);
	}
};

Robot.prototype.neopixelSetBrightnessTo = function(request) {
	var value = parseInt(request.params.value);
	if(typeof value == 'number') {
		var id = parseInt(request.params.id);
		this.neopixel.setBrightnessTo(value);
		this.refreshNeopixel(id);
	}
};

Robot.prototype.neopixelSetModeTo = function(request) {
	var auto = VALUES[request.params.auto];
	if(typeof auto == 'number') {
		var id = parseInt(request.params.id);
		this.neopixel.setAuto(auto == AUTO);
		this.refreshNeopixel(id);
	}
};

Robot.prototype.neopixelUpdate = function(request) {
	this.addNeopixelWait(parseInt(request.params.id));
	var packet = this.neopixel.update();
	this.motoring.neopixel.write(packet);
};

Robot.prototype.writeSerial = function(request) {
	var motoring = this.motoring;
	motoring.mode.La.write(Cheese.MODE_L_SERIAL);
	var mode = SERIAL_MODES[request.params.mode];
	if(typeof mode == 'number') {
		var queue = this.writeQueue;
		queue.push(request.params.value, mode == SERIAL_STRING_LINE);
		var data = queue.pop();
		if(data) {
			this.addSerialWriteWait(parseInt(request.params.id));
			motoring.writeSerial.write(data);
		}
	}
};

Robot.prototype.readSerial = function(request) {
	this.motoring.mode.La.write(Cheese.MODE_L_SERIAL);
	this.serialDelimiter = 0;
	this.addSerialReadWait(parseInt(request.params.id));
};

Robot.prototype.readSerialUntil = function(request) {
	this.motoring.mode.La.write(Cheese.MODE_L_SERIAL);
	var delimiter = SERIAL_DELIMITERS[request.params.delimiter];
	if(typeof delimiter == 'number') {
		this.serialDelimiter = delimiter;
		this.addSerialReadWait(parseInt(request.params.id));
	}
};

Robot.prototype.mapRange = function(port, range, sensory, data) {
	if(range.active) {
		if(range.has_zero) {
			if(range.zero == range.in_upper || range.zero == range.in_lower) {
				if(range.in_lower == range.in_upper) {
					sensory[port] = data[port];
				} else {
					sensory[port] = parseInt((range.out_upper - range.out_lower) * (data[port] - range.in_lower) / (range.in_upper - range.in_lower) + range.out_lower);
				}
			} else {
				if(data[port] > range.zero) {
					sensory[port] = parseInt(range.out_upper * (data[port] - range.zero) / (range.in_upper - range.zero));
				} else {
					sensory[port] = parseInt(range.out_lower - range.out_lower * (data[port] - range.in_lower) / (range.zero - range.in_lower));
				}
			}
		} else {
			if(range.in_lower == range.in_upper) {
				sensory[port] = data[port];
			} else {
				sensory[port] = parseInt((range.out_upper - range.out_lower) * (data[port] - range.in_lower) / (range.in_upper - range.in_lower) + range.out_lower);
			}
		}
	} else {
		sensory[port] = data[port];
	}
};

Robot.prototype.handleSensory = function(data) {
	this.received = true;
	this.active = true;
	var motoring = this.motoring;
	var sensory = this.sensory;
	sensory.signalStrength = data.signalStrength;
	sensory.accelerationX = data.acceleration.x;
	sensory.accelerationY = data.acceleration.y;
	sensory.accelerationZ = data.acceleration.z;
	
	this.mapRange('inputSA', this.range.Sa, sensory, data);
	this.mapRange('inputSB', this.range.Sb, sensory, data);
	this.mapRange('inputSC', this.range.Sc, sensory, data);
	this.mapRange('inputLA', this.range.La, sensory, data);
	this.mapRange('inputLB', this.range.Lb, sensory, data);
	this.mapRange('inputLC', this.range.Lc, sensory, data);
	
	sensory.ultrasonic = data.ultrasonic;
	sensory.encoder = data.encoder;
	sensory.temperature = data.temperature;
	sensory.humidity = data.humidity;
	
	if(data.map & 0x00080000) {
		this.freeFall = true;
	}
	if(data.map & 0x00040000) {
		this.tap = true;
	}
	sensory.stepCount = data.stepCount;

	if(this.stepId >= 0) {
		if((data.map & 0x00004000) != 0 && data.stepState == 0) {
			motoring.velocity.write(0);
			this.removeWait(this.stepId);
			this.stepId = -1;
		}
	}
	if((data.map & 0x00002000) != 0 && data.soundState == 0) {
		if(this.sound > 0) {
			if(this.soundRepeat < 0) {
				this.runSound(this.sound, -1);
			} else if(this.soundRepeat > 1) {
				this.soundRepeat --;
				this.runSound(this.sound, this.soundRepeat);
			} else {
				this.sound = 0;
				this.soundRepeat = 1;
				if(this.soundId >= 0) this.removeWait(this.soundId);
				this.soundId = -1;
			}
		} else {
			this.sound = 0;
			this.soundRepeat = 1;
			if(this.soundId >= 0) this.removeWait(this.soundId);
			this.soundId = -1;
		}
	}
	sensory.batteryState = data.batteryState;

	if(data.map & 0x00010000) {
		if(data.serial) robot.readQueue.push(data.serial);
	}
	if(data.map & 0x00000800) {
		var matrix = this.matrix;
		if(matrix.active) {
			this.touching = matrix.getTouching();
		}
		this.removeI2cWaits();
	}
	if(data.map & 0x00000400) {
		var tmp = this.writeQueue.pop();
		if(tmp) {
			motoring.writeSerial.write(tmp);
		} else {
			this.removeSerialWriteWaits();
		}
	}
	if(data.map & 0x00000200) {
		var neopixel = this.neopixel;
		if(neopixel.hasNext()) {
			var packet = neopixel.update();
			if(packet) {
				motoring.neopixel.write(packet);
			}
		} else {
			this.removeNeopixelWaits();
		}
	}
	if(this.serialReadWaits.length > 0) {
		var tmp = this.readQueue.pop(this.serialDelimiter);
		if(tmp) {
			sensory.serial = tmp;
			this.removeSerialReadWaits();
		}
	}
};

function Component() {
	var setup = function(map, key, value) {
		var data = translations[key];
		if(data) {
			if(value === undefined) value = key;
			for(var i in data) {
				map[data[i]] = value;
			}
			map[key] = value;
		}
	};
	var setupIndex = function(map, key, index, value) {
		var data = translations[key];
		if(data) {
			if(value === undefined) value = key;
			for(var i in data) {
				map[data[i] + ' ' + index] = value;
			}
			map[key + ' ' + index] = value;
		}
	};
	setup(NOTES, 'C', 4);
	setup(NOTES, 'C♯ (D♭)', 5);
	setup(NOTES, 'D', 6);
	setup(NOTES, 'D♯ (E♭)', 7);
	setup(NOTES, 'E', 8);
	setup(NOTES, 'F', 9);
	setup(NOTES, 'F♯ (G♭)', 10);
	setup(NOTES, 'G', 11);
	setup(NOTES, 'G♯ (A♭)', 12);
	setup(NOTES, 'A', 13);
	setup(NOTES, 'A♯ (B♭)', 14);
	setup(NOTES, 'B', 15);
	setup(SOUNDS, 'beep', Cheese.SOUND_BEEP);
	setup(SOUNDS, 'random beep', Cheese.SOUND_RANDOM_BEEP);
	setup(SOUNDS, 'noise', Cheese.SOUND_NOISE);
	setup(SOUNDS, 'siren', Cheese.SOUND_SIREN);
	setup(SOUNDS, 'engine', Cheese.SOUND_ENGINE);
	setup(SOUNDS, 'chop', Cheese.SOUND_CHOP);
	setup(SOUNDS, 'robot', Cheese.SOUND_ROBOT);
	setup(SOUNDS, 'dibidibidip', Cheese.SOUND_DIBIDIBIDIP);
	setup(SOUNDS, 'good job', Cheese.SOUND_GOOD_JOB);
	setup(SOUNDS, 'happy', Cheese.SOUND_HAPPY);
	setup(SOUNDS, 'angry', Cheese.SOUND_ANGRY);
	setup(SOUNDS, 'sad', Cheese.SOUND_SAD);
	setup(SOUNDS, 'sleep', Cheese.SOUND_SLEEP);
	setup(SOUNDS, 'march', Cheese.SOUND_MARCH);
	setup(SOUNDS, 'birthday', Cheese.SOUND_BIRTHDAY);
	setup(RGB_COLORS, 'red', [255, 0, 0]);
	setup(RGB_COLORS, 'orange', [255, 63, 0]);
	setup(RGB_COLORS, 'yellow', [255, 255, 0]);
	setup(RGB_COLORS, 'green', [0, 255, 0]);
	setup(RGB_COLORS, 'sky blue', [0, 255, 255]);
	setup(RGB_COLORS, 'blue', [0, 0, 255]);
	setup(RGB_COLORS, 'violet', [63, 0, 255]);
	setup(RGB_COLORS, 'purple', [255, 0, 255]);
	setup(RGB_COLORS, 'white', [255, 255, 255]);
	setup(MODE_INPUTS_SL, 'digital', Cheese.MODE_S_DIGITAL_INPUT);
	setup(MODE_INPUTS_SL, 'analog', Cheese.MODE_S_ANALOG_INPUT);
	setup(MODE_INPUTS_SL, 'button', Cheese.MODE_S_BUTTON_INPUT);
	setup(SPECIAL_SENSORS, 'ultrasonic', ULTRASONIC);
	setup(SPECIAL_SENSORS, 'encoder', ENCODER);
	setup(SPECIAL_SENSORS, 'temperature DS18B20 (RW1820)', DS18B20);
	setup(SPECIAL_SENSORS, 'humidity & temperature DHT11 (RHT01)', DHT11);
	setup(SPECIAL_SENSORS, 'humidity & temperature DHT21 (RHT02, AM2301, HM2301)', DHT21);
	setup(SPECIAL_SENSORS, 'humidity & temperature DHT22 (RHT03, AM2302)', DHT22);
	setup(SPECIAL_SENSORS, 'gesture', GESTURE);
	setup(MODE_SOUNDS, 'internal speaker', Cheese.MODE_SOUND_INTERNAL);
	setup(MODE_SOUNDS, 'port Sa', Cheese.MODE_SOUND_SA);
	setup(MODE_SOUNDS, 'port Sb', Cheese.MODE_SOUND_SB);
	setup(MODE_SOUNDS, 'port Sc', Cheese.MODE_SOUND_SC);
	setup(ACTION_STEP_MOTORS, 'stop', STOP);
	setup(ACTION_STEP_MOTORS, 'turn off', TURN_OFF);
	setup(MODE_STEP_MOTORS, 'normal', Cheese.MODE_M_STEP_WAVE);
	setup(MODE_STEP_MOTORS, 'power', Cheese.MODE_M_STEP_FULL);
	PORT_PWM_OUTPUTS['Sa'] = 'Sa';
	PORT_PWM_OUTPUTS['Sb'] = 'Sb';
	PORT_PWM_OUTPUTS['Sc'] = 'Sc';
	PORT_PWM_OUTPUTS['La'] = 'La';
	PORT_PWM_OUTPUTS['Lb'] = 'Lb';
	PORT_PWM_OUTPUTS['Lc'] = 'Lc';
	setup(PORT_PWM_OUTPUTS, 'high current Ma', 'Ma');
	setup(PORT_PWM_OUTPUTS, 'high current Mb', 'Mb');
	setup(SHAPES, 'square');
	setup(SHAPES, 'triangle');
	setup(SHAPES, 'diamond');
	setup(SHAPES, 'circle');
	SHAPES['X'] = 'X';
	setup(SHAPES, 'like');
	setup(SHAPES, 'dislike');
	setup(SHAPES, 'angry');
	setup(SHAPES, 'open mouth');
	setup(SHAPES, 'close mouth');
	setup(SHAPES, 'walk 1');
	setup(SHAPES, 'walk 2');
	setup(SHAPES, 'heart');
	setup(SHAPES, 'star');
	setup(SHAPES, 'airplane');
	setup(SHAPES, 'puppy');
	setup(SHAPES, 'butterfly');
	setup(SHAPES, 'quarter note');
	setup(SHAPES, 'eighth note');
	setup(SHAPES, 'left arrow');
	setup(SHAPES, 'right arrow');
	setup(SHAPES, 'up arrow');
	setup(SHAPES, 'down arrow');
	setup(DIRECTIONS, 'left', LEFT);
	setup(DIRECTIONS, 'right', RIGHT);
	setup(DIRECTIONS, 'top', TOP);
	setup(DIRECTIONS, 'bottom', BOTTOM);
	setup(SERIAL_MODES, 'string', SERIAL_STRING);
	setup(SERIAL_MODES, 'string line', SERIAL_STRING_LINE);
	setup(SERIAL_DELIMITERS, 'new line', SERIAL_NEW_LINE);
	NEOPIXEL_TYPES['GRB'] = NEOPIXEL_GRB;
	NEOPIXEL_TYPES['GRBW'] = NEOPIXEL_GRBW;
	SERIAL_DELIMITERS[','] = SERIAL_COMMA;
	SERIAL_DELIMITERS[':'] = SERIAL_COLON;
	SERIAL_DELIMITERS['$'] = SERIAL_DOLLAR;
	SERIAL_DELIMITERS['#'] = SERIAL_SHARP;
	
	setup(VALUES, 'background', BACKGROUND);
	setup(VALUES, 'all', ALL);
	setup(VALUES, 'on', ON);
	setup(VALUES, 'off', OFF);
	setup(VALUES, 'inverse', INVERSE);
	setup(VALUES, 'clear', CLEAR);
	setup(VALUES, 'show', SHOW);
	setup(VALUES, 'hide', HIDE);
	setup(VALUES, 'x position', POS_X);
	setup(VALUES, 'y position', POS_Y);
	setup(VALUES, 'clockwise', CLOCKWISE);
	setup(VALUES, 'counterclockwise', COUNTERCLOCKWISE);
	setup(VALUES, 'left-right', LEFT_RIGHT);
	setup(VALUES, 'up-down', UP_DOWN);
	setup(VALUES, 'landscape', LANDSCAPE);
	setup(VALUES, 'portrait', PORTRAIT);
	setup(VALUES, 'auto', AUTO);
	setup(VALUES, 'manual', MANUAL);
}

Component.prototype.updateLang = function(translator) { // for sensory
	var lang = translator.getLanguage();
	
	var get = function(key) {
		var value = undefined;
		var data = translations[key];
		if(data) value = data[lang];
		return value || key;
	};
	
	BOOLEAN_TAP = '/' + get('tap') + ' ';
	BOOLEAN_FREE_FALL = '/' + get('free%20fall') + ' ';
	BOOLEAN_BATTERY_NORMAL = '/' + get('battery%20normal') + ' ';
	BOOLEAN_BATTERY_LOW = '/' + get('battery%20low') + ' ';
	BOOLEAN_BATTERY_EMPTY = '/' + get('battery%20empty') + ' ';
	SENSOR_ACCELERATION_X = '/' + get('x%20acceleration') + ' ';
	SENSOR_ACCELERATION_Y = '/' + get('y%20acceleration') + ' ';
	SENSOR_ACCELERATION_Z = '/' + get('z%20acceleration') + ' ';
	SENSOR_SIGNAL_STRENGTH = '/' + get('signal%20strength%20(dBm)') + ' ';
	SENSOR_INPUT = '/' + get('input%20');
	SENSOR_ULTRASONIC = '/' + get('ultrasonic%20(cm)') + ' ';
	SENSOR_ENCODER = '/' + get('encoder') + ' ';
	SENSOR_TEMPERATURE = '/' + get('temperature%20(℃)') + ' ';
	SENSOR_HUMIDITY = '/' + get('humidity%20(%25RH)') + ' ';
	SENSOR_GESTURE = '/' + get('gesture') + ' ';
	TOUCH_BACKGROUND = '/' + get('background') + ' ';
	TOUCH_OTHER_SPRITE = '/' + get('other%20sprite') + ' ';
	TOUCH_LEFT_WALL = '/' + get('left%20wall') + ' ';
	TOUCH_RIGHT_WALL = '/' + get('right%20wall') + ' ';
	TOUCH_TOP_WALL = '/' + get('top%20wall') + ' ';
	TOUCH_BOTTOM_WALL = '/' + get('bottom%20wall') + ' ';
	TOUCH_ANY_WALL = '/' + get('any%20wall') + ' ';
	X_POSITION = '/' + get('x%20position') + ' ';
	Y_POSITION = '/' + get('y%20position') + ' ';
};

Component.prototype.getModule = function() {
	return MODULE;
};

Component.prototype.createRobot = function(index) {
	return new Robot(index);
};

module.exports = new Component();