'use strict';
var context = require('context');
var router = require('router');
var translator = require('translator');
var component = require('./component');
const MODULE = 'cheese';

function Extension() {
	this.robots = {};
}

Extension.prototype.getRobotByIndex = function(index) {
	return this.robots[index];
};

Extension.prototype.getRobotByRoute = function(route) {
	var index = route.index;
	if(index >= 0) {
		var robot = this.robots[index];
		if(!robot) {
			var component = context.getComponentByModule(PROGRAM, route.module);
			robot = component.createRobot(index);
			if(robot) {
				this.robots[index] = robot;
			}
		}
		return robot;
	}
};

Extension.prototype.start = function(express) {
	var self = this;
	var robots = self.robots;
	var problemMessage = '';
	
	self.running = true;
	for(var i in robots) {
		robots[i].clearWaits();
	}
	var setup = function() {
		problemMessage = '_problem ' + translator.translate('First robot is not connected.') + '\n';
		component.updateLang(translator);
	};
	translator.on('cheese-scratch-multi', function() {
		setup();
	});
	setup();
	
	if(express) {
		const HEADER = { 'Content-Type': 'text/plain' };
		
		express.get('/reset_all', function(request, response) {
			response.writeHead(200, HEADER);
			var robot;
			for(var i in robots) {
				robot = robots[i];
				robot.clearWaits();
				robot.clearTimeouts();
				robot.reset();
			}
			router.reset();
			response.end();
		});
		
		express.get('/poll', function(request, response) {
			if(!self.running) {
				response.set('Connection', 'close');
			}
			response.writeHead(200, HEADER);
			var waits, busy = '';
			for(var i in robots) {
				waits = robots[i].getWaits();
				if(waits.length > 0) {
					busy += ' ' + waits.join(' ');
				}
			}
			if(busy.length > 0) {
				response.write('_busy' + busy + '\n');
			}
			for(var i in robots) {
				robots[i].writeMultiSensory(response);
			}
			if(self.state != 'connected') {
				response.write(problemMessage);
			}
			response.end();
		});
		
		express.get('/cheeseSetInputPortTo/:index/:port/:value', function(request, response) {
			response.writeHead(200, HEADER);
			var robot = self.getRobotByIndex(parseInt(request.params.index));
			if(robot) {
				robot.setInputPortTo(request);
			}
			response.end();
		});
		
		express.get('/cheeseSetInputRangeTo/:index/:port/:lower1/:upper1/:lower2/:upper2', function(request, response) {
			response.writeHead(200, HEADER);
			var robot = self.getRobotByIndex(parseInt(request.params.index));
			if(robot) {
				robot.setInputRangeTo(request);
			}
			response.end();
		});
		
		express.get('/cheeseSetInputZeroAndRangeTo/:index/:port/:zero/:lower1/:upper1/:lower2/:upper2', function(request, response) {
			response.writeHead(200, HEADER);
			var robot = self.getRobotByIndex(parseInt(request.params.index));
			if(robot) {
				robot.setInputZeroAndRangeTo(request);
			}
			response.end();
		});
		
		express.get('/cheeseStartSensor/:index/:sensor', function(request, response) {
			response.writeHead(200, HEADER);
			var robot = self.getRobotByIndex(parseInt(request.params.index));
			if(robot) {
				robot.startSensor(request);
			}
			response.end();
		});
		
		express.get('/cheeseResetEncoder/:index', function(request, response) {
			response.writeHead(200, HEADER);
			var robot = self.getRobotByIndex(parseInt(request.params.index));
			if(robot) {
				robot.resetEncoder(request);
			}
			response.end();
		});
		
		express.get('/cheeseSetSoundPortTo/:index/:port', function(request, response) {
			response.writeHead(200, HEADER);
			var robot = self.getRobotByIndex(parseInt(request.params.index));
			if(robot) {
				robot.setSoundPortTo(request);
			}
			response.end();
		});
		
		express.get('/cheesePlaySound/:index/:sound', function(request, response) {
			response.writeHead(200, HEADER);
			var robot = self.getRobotByIndex(parseInt(request.params.index));
			if(robot) {
				robot.playSound(request);
			}
			response.end();
		});
		
		express.get('/cheesePlaySoundTimes/:index/:sound/:count', function(request, response) {
			response.writeHead(200, HEADER);
			var robot = self.getRobotByIndex(parseInt(request.params.index));
			if(robot) {
				robot.playSoundTimes(request);
			}
			response.end();
		});
		
		express.get('/cheesePlaySoundTimesUntilDone/:id/:index/:sound/:count', function(request, response) {
			response.writeHead(200, HEADER);
			var robot = self.getRobotByIndex(parseInt(request.params.index));
			if(robot) {
				robot.playSoundTimesUntilDone(request);
			}
			response.end();
		});
		
		express.get('/cheeseChangeBuzzerBy/:index/:hz', function(request, response) {
			response.writeHead(200, HEADER);
			var robot = self.getRobotByIndex(parseInt(request.params.index));
			if(robot) {
				robot.changeBuzzerBy(request);
			}
			response.end();
		});
		
		express.get('/cheeseSetBuzzerTo/:index/:hz', function(request, response) {
			response.writeHead(200, HEADER);
			var robot = self.getRobotByIndex(parseInt(request.params.index));
			if(robot) {
				robot.setBuzzerTo(request);
			}
			response.end();
		});
		
		express.get('/cheeseClearSound/:index', function(request, response) {
			response.writeHead(200, HEADER);
			var robot = self.getRobotByIndex(parseInt(request.params.index));
			if(robot) {
				robot.clearSound(request);
			}
			response.end();
		});
		
		express.get('/cheesePlayNote/:index/:note/:octave', function(request, response) {
			response.writeHead(200, HEADER);
			var robot = self.getRobotByIndex(parseInt(request.params.index));
			if(robot) {
				robot.playNote(request);
			}
			response.end();
		});
		
		express.get('/cheesePlayNoteForBeats/:id/:index/:note/:octave/:beat', function(request, response) {
			response.writeHead(200, HEADER);
			var robot = self.getRobotByIndex(parseInt(request.params.index));
			if(robot) {
				robot.playNoteForBeats(request);
			}
			response.end();
		});
		
		express.get('/cheeseRestForBeats/:id/:index/:beat', function(request, response) {
			response.writeHead(200, HEADER);
			var robot = self.getRobotByIndex(parseInt(request.params.index));
			if(robot) {
				robot.restForBeats(request);
			}
			response.end();
		});
		
		express.get('/cheeseChangeTempoBy/:index/:bpm', function(request, response) {
			response.writeHead(200, HEADER);
			var robot = self.getRobotByIndex(parseInt(request.params.index));
			if(robot) {
				robot.changeTempoBy(request);
			}
			response.end();
		});
		
		express.get('/cheeseSetTempoTo/:index/:bpm', function(request, response) {
			response.writeHead(200, HEADER);
			var robot = self.getRobotByIndex(parseInt(request.params.index));
			if(robot) {
				robot.setTempoTo(request);
			}
			response.end();
		});
		
		express.get('/cheeseChangeServoMotorAngleBy/:index/:port/:value', function(request, response) {
			response.writeHead(200, HEADER);
			var robot = self.getRobotByIndex(parseInt(request.params.index));
			if(robot) {
				robot.changeServoMotorAngleBy(request);
			}
			response.end();
		});
		
		express.get('/cheeseSetServoMotorAngleTo/:index/:port/:value', function(request, response) {
			response.writeHead(200, HEADER);
			var robot = self.getRobotByIndex(parseInt(request.params.index));
			if(robot) {
				robot.setServoMotorAngleTo(request);
			}
			response.end();
		});
		
		express.get('/cheeseTurnOffServoMotor/:index/:port', function(request, response) {
			response.writeHead(200, HEADER);
			var robot = self.getRobotByIndex(parseInt(request.params.index));
			if(robot) {
				robot.turnOffServoMotor(request);
			}
			response.end();
		});
		
		express.get('/cheeseChangeDcMotorVelocityBy/:index/:port/:value', function(request, response) {
			response.writeHead(200, HEADER);
			var robot = self.getRobotByIndex(parseInt(request.params.index));
			if(robot) {
				robot.changeDcMotorVelocityBy(request);
			}
			response.end();
		});
		
		express.get('/cheeseSetDcMotorVelocityTo/:index/:port/:value', function(request, response) {
			response.writeHead(200, HEADER);
			var robot = self.getRobotByIndex(parseInt(request.params.index));
			if(robot) {
				robot.setDcMotorVelocityTo(request);
			}
			response.end();
		});
		
		express.get('/cheeseStopDcMotor/:index/:port', function(request, response) {
			response.writeHead(200, HEADER);
			var robot = self.getRobotByIndex(parseInt(request.params.index));
			if(robot) {
				robot.stopDcMotor(request);
			}
			response.end();
		});
		
		express.get('/cheeseRotateStepMotor/:id/:index/:value/:velocity', function(request, response) {
			response.writeHead(200, HEADER);
			var robot = self.getRobotByIndex(parseInt(request.params.index));
			if(robot) {
				robot.rotateStepMotor(request);
			}
			response.end();
		});
		
		express.get('/cheeseChangeStepMotorVelocityBy/:index/:value', function(request, response) {
			response.writeHead(200, HEADER);
			var robot = self.getRobotByIndex(parseInt(request.params.index));
			if(robot) {
				robot.changeStepMotorVelocityBy(request);
			}
			response.end();
		});
		
		express.get('/cheeseSetStepMotorVelocityTo/:index/:value', function(request, response) {
			response.writeHead(200, HEADER);
			var robot = self.getRobotByIndex(parseInt(request.params.index));
			if(robot) {
				robot.setStepMotorVelocityTo(request);
			}
			response.end();
		});
		
		express.get('/cheeseStopOffStepMotor/:index/:action', function(request, response) {
			response.writeHead(200, HEADER);
			var robot = self.getRobotByIndex(parseInt(request.params.index));
			if(robot) {
				robot.stopOffStepMotor(request);
			}
			response.end();
		});
		
		express.get('/cheeseSetStepMotorModeTo/:index/:value', function(request, response) {
			response.writeHead(200, HEADER);
			var robot = self.getRobotByIndex(parseInt(request.params.index));
			if(robot) {
				robot.setStepMotorModeTo(request);
			}
			response.end();
		});
		
		express.get('/cheeseSetDigitalOutputTo/:index/:port/:value', function(request, response) {
			response.writeHead(200, HEADER);
			var robot = self.getRobotByIndex(parseInt(request.params.index));
			if(robot) {
				robot.setDigitalOutputTo(request);
			}
			response.end();
		});
		
		express.get('/cheeseChangePwmOutputBy/:index/:port/:value', function(request, response) {
			response.writeHead(200, HEADER);
			var robot = self.getRobotByIndex(parseInt(request.params.index));
			if(robot) {
				robot.changePwmOutputBy(request);
			}
			response.end();
		});
		
		express.get('/cheeseSetPwmOutputTo/:index/:port/:value', function(request, response) {
			response.writeHead(200, HEADER);
			var robot = self.getRobotByIndex(parseInt(request.params.index));
			if(robot) {
				robot.setPwmOutputTo(request);
			}
			response.end();
		});
		
		express.get('/cheeseSetLedTypeTo/:index/:led/:port/:type', function(request, response) {
			response.writeHead(200, HEADER);
			var robot = self.getRobotByIndex(parseInt(request.params.index));
			if(robot) {
				robot.setLedTypeTo(request);
			}
			response.end();
		});
		
		express.get('/cheeseSetLedToColor/:index/:led/:color', function(request, response) {
			response.writeHead(200, HEADER);
			var robot = self.getRobotByIndex(parseInt(request.params.index));
			if(robot) {
				robot.setLedToColor(request);
			}
			response.end();
		});
		
		express.get('/cheeseChangeLedByRGB/:index/:led/:r/:g/:b', function(request, response) {
			response.writeHead(200, HEADER);
			var robot = self.getRobotByIndex(parseInt(request.params.index));
			if(robot) {
				robot.changeLedByRGB(request);
			}
			response.end();
		});
		
		express.get('/cheeseSetLedToRGB/:index/:led/:r/:g/:b', function(request, response) {
			response.writeHead(200, HEADER);
			var robot = self.getRobotByIndex(parseInt(request.params.index));
			if(robot) {
				robot.setLedToRGB(request);
			}
			response.end();
		});
		
		express.get('/cheeseClearLed/:index/:led', function(request, response) {
			response.writeHead(200, HEADER);
			var robot = self.getRobotByIndex(parseInt(request.params.index));
			if(robot) {
				robot.clearLed(request);
			}
			response.end();
		});
		
		express.get('/cheeseClearLedMatrixBackground/:id/:index', function(request, response) {
			response.writeHead(200, HEADER);
			var robot = self.getRobotByIndex(parseInt(request.params.index));
			if(robot) {
				robot.clearLedMatrixBackground(request);
			}
			response.end();
		});
		
		express.get('/cheeseClearLedMatrix/:id/:index/:target', function(request, response) {
			response.writeHead(200, HEADER);
			var robot = self.getRobotByIndex(parseInt(request.params.index));
			if(robot) {
				robot.clearLedMatrix(request);
			}
			response.end();
		});
		
		express.get('/cheeseTurnLedMatrixOnOff/:id/:index/:x/:y/:action', function(request, response) {
			response.writeHead(200, HEADER);
			var robot = self.getRobotByIndex(parseInt(request.params.index));
			if(robot) {
				robot.turnLedMatrixOnOff(request);
			}
			response.end();
		});
		
		express.get('/cheeseDrawLedMatrixShape/:id/:index/:shape', function(request, response) {
			response.writeHead(200, HEADER);
			var robot = self.getRobotByIndex(parseInt(request.params.index));
			if(robot) {
				robot.drawLedMatrixShape(request);
			}
			response.end();
		});
		
		express.get('/cheeseDrawLedMatrixShapeAt/:id/:index/:shape/:x/:y', function(request, response) {
			response.writeHead(200, HEADER);
			var robot = self.getRobotByIndex(parseInt(request.params.index));
			if(robot) {
				robot.drawLedMatrixShapeAt(request);
			}
			response.end();
		});
		
		express.get('/cheeseDrawLedMatrixStringAt/:id/:index/:text/:x/:y', function(request, response) {
			response.writeHead(200, HEADER);
			var robot = self.getRobotByIndex(parseInt(request.params.index));
			if(robot) {
				robot.drawLedMatrixStringAt(request);
			}
			response.end();
		});
		
		express.get('/cheeseDrawLedMatrixBinaryAt/:id/:index/:value/:x/:y', function(request, response) {
			response.writeHead(200, HEADER);
			var robot = self.getRobotByIndex(parseInt(request.params.index));
			if(robot) {
				robot.drawLedMatrixBinaryAt(request);
			}
			response.end();
		});
		
		express.get('/cheeseMoveLedMatrixTo/:id/:index/:direction', function(request, response) {
			response.writeHead(200, HEADER);
			var robot = self.getRobotByIndex(parseInt(request.params.index));
			if(robot) {
				robot.moveLedMatrixTo(request);
			}
			response.end();
		});
		
		express.get('/cheeseMoveLedMatrixBackgroundBy/:id/:index/:x/:y', function(request, response) {
			response.writeHead(200, HEADER);
			var robot = self.getRobotByIndex(parseInt(request.params.index));
			if(robot) {
				robot.moveLedMatrixBackgroundBy(request);
			}
			response.end();
		});
		
		express.get('/cheeseMoveLedMatrixBy/:id/:index/:target/:x/:y', function(request, response) {
			response.writeHead(200, HEADER);
			var robot = self.getRobotByIndex(parseInt(request.params.index));
			if(robot) {
				robot.moveLedMatrixBy(request);
			}
			response.end();
		});
		
		express.get('/cheeseClearShowHideLedMatrixSprite/:id/:index/:sprite/:action', function(request, response) {
			response.writeHead(200, HEADER);
			var robot = self.getRobotByIndex(parseInt(request.params.index));
			if(robot) {
				robot.clearShowHideLedMatrixSprite(request);
			}
			response.end();
		});
		
		express.get('/cheeseSetLedMatrixSpriteShape/:id/:index/:sprite/:shape', function(request, response) {
			response.writeHead(200, HEADER);
			var robot = self.getRobotByIndex(parseInt(request.params.index));
			if(robot) {
				robot.setLedMatrixSpriteShape(request);
			}
			response.end();
		});
		
		express.get('/cheeseSetLedMatrixSpriteString/:id/:index/:sprite/:text', function(request, response) {
			response.writeHead(200, HEADER);
			var robot = self.getRobotByIndex(parseInt(request.params.index));
			if(robot) {
				robot.setLedMatrixSpriteString(request);
			}
			response.end();
		});
		
		express.get('/cheeseSetLedMatrixSpriteBinary/:id/:index/:sprite/:value', function(request, response) {
			response.writeHead(200, HEADER);
			var robot = self.getRobotByIndex(parseInt(request.params.index));
			if(robot) {
				robot.setLedMatrixSpriteBinary(request);
			}
			response.end();
		});
		
		express.get('/cheeseChangeLedMatrixSpritePositionsBy/:id/:index/:sprite/:x/:y', function(request, response) {
			response.writeHead(200, HEADER);
			var robot = self.getRobotByIndex(parseInt(request.params.index));
			if(robot) {
				robot.changeLedMatrixSpritePositionsBy(request);
			}
			response.end();
		});
		
		express.get('/cheeseSetLedMatrixSpritePositionsTo/:id/:index/:sprite/:x/:y', function(request, response) {
			response.writeHead(200, HEADER);
			var robot = self.getRobotByIndex(parseInt(request.params.index));
			if(robot) {
				robot.setLedMatrixSpritePositionsTo(request);
			}
			response.end();
		});
		
		express.get('/cheeseSetLedMatrixSpritePositionTo/:id/:index/:sprite/:position/:value', function(request, response) {
			response.writeHead(200, HEADER);
			var robot = self.getRobotByIndex(parseInt(request.params.index));
			if(robot) {
				robot.setLedMatrixSpritePositionTo(request);
			}
			response.end();
		});
		
		express.get('/cheeseRotateLedMatrixSprite/:id/:index/:sprite/:direction', function(request, response) {
			response.writeHead(200, HEADER);
			var robot = self.getRobotByIndex(parseInt(request.params.index));
			if(robot) {
				robot.rotateLedMatrixSprite(request);
			}
			response.end();
		});
		
		express.get('/cheeseFlipLedMatrixSprite/:id/:index/:sprite/:direction', function(request, response) {
			response.writeHead(200, HEADER);
			var robot = self.getRobotByIndex(parseInt(request.params.index));
			if(robot) {
				robot.flipLedMatrixSprite(request);
			}
			response.end();
		});
		
		express.get('/cheeseStampLedMatrixSprite/:id/:index/:sprite', function(request, response) {
			response.writeHead(200, HEADER);
			var robot = self.getRobotByIndex(parseInt(request.params.index));
			if(robot) {
				robot.stampLedMatrixSprite(request);
			}
			response.end();
		});
		
		express.get('/cheeseChangeLedMatrixBrightnessBy/:id/:index/:value', function(request, response) {
			response.writeHead(200, HEADER);
			var robot = self.getRobotByIndex(parseInt(request.params.index));
			if(robot) {
				robot.changeLedMatrixBrightnessBy(request);
			}
			response.end();
		});
		
		express.get('/cheeseSetLedMatrixBrightnessTo/:id/:index/:value', function(request, response) {
			response.writeHead(200, HEADER);
			var robot = self.getRobotByIndex(parseInt(request.params.index));
			if(robot) {
				robot.setLedMatrixBrightnessTo(request);
			}
			response.end();
		});
		
		express.get('/cheeseSetLedMatrixDirectionTo/:id/:index/:direction', function(request, response) {
			response.writeHead(200, HEADER);
			var robot = self.getRobotByIndex(parseInt(request.params.index));
			if(robot) {
				robot.setLedMatrixDirectionTo(request);
			}
			response.end();
		});
		
		express.get('/cheeseSetLedMatrixModeTo/:id/:index/:auto', function(request, response) {
			response.writeHead(200, HEADER);
			var robot = self.getRobotByIndex(parseInt(request.params.index));
			if(robot) {
				robot.setLedMatrixModeTo(request);
			}
			response.end();
		});
		
		express.get('/cheeseUpdateLedMatrix/:id/:index', function(request, response) {
			response.writeHead(200, HEADER);
			var robot = self.getRobotByIndex(parseInt(request.params.index));
			if(robot) {
				robot.updateLedMatrix(request);
			}
			response.end();
		});
		
		express.get('/cheeseNeopixelSetNumberAndTypeTo/:id/:index/:num/:type', function(request, response) {
			response.writeHead(200, HEADER);
			var robot = self.getRobotByIndex(parseInt(request.params.index));
			if(robot) {
				robot.neopixelSetNumberAndTypeTo(request);
			}
			response.end();
		});
		
		express.get('/cheeseNeopixelSetAllLedsToColor/:id/:index/:color', function(request, response) {
			response.writeHead(200, HEADER);
			var robot = self.getRobotByIndex(parseInt(request.params.index));
			if(robot) {
				robot.neopixelSetAllLedsToColor(request);
			}
			response.end();
		});
		
		express.get('/cheeseNeopixelChangeAllLedsByRGB/:id/:index/:r/:g/:b', function(request, response) {
			response.writeHead(200, HEADER);
			var robot = self.getRobotByIndex(parseInt(request.params.index));
			if(robot) {
				robot.neopixelChangeAllLedsByRGB(request);
			}
			response.end();
		});
		
		express.get('/cheeseNeopixelSetAllLedsToRGB/:id/:index/:r/:g/:b', function(request, response) {
			response.writeHead(200, HEADER);
			var robot = self.getRobotByIndex(parseInt(request.params.index));
			if(robot) {
				robot.neopixelSetAllLedsToRGB(request);
			}
			response.end();
		});
		
		express.get('/cheeseNeopixelClearAllLeds/:id/:index', function(request, response) {
			response.writeHead(200, HEADER);
			var robot = self.getRobotByIndex(parseInt(request.params.index));
			if(robot) {
				robot.neopixelClearAllLeds(request);
			}
			response.end();
		});
		
		express.get('/cheeseNeopixelSetLedAtToColor/:id/:index/:pos/:color', function(request, response) {
			response.writeHead(200, HEADER);
			var robot = self.getRobotByIndex(parseInt(request.params.index));
			if(robot) {
				robot.neopixelSetLedAtToColor(request);
			}
			response.end();
		});
		
		express.get('/cheeseNeopixelChangeLedAtByRGB/:id/:index/:pos/:r/:g/:b', function(request, response) {
			response.writeHead(200, HEADER);
			var robot = self.getRobotByIndex(parseInt(request.params.index));
			if(robot) {
				robot.neopixelChangeLedAtByRGB(request);
			}
			response.end();
		});
		
		express.get('/cheeseNeopixelSetLedAtToRGB/:id/:index/:pos/:r/:g/:b', function(request, response) {
			response.writeHead(200, HEADER);
			var robot = self.getRobotByIndex(parseInt(request.params.index));
			if(robot) {
				robot.neopixelSetLedAtToRGB(request);
			}
			response.end();
		});
		
		express.get('/cheeseNeopixelClearLedAt/:id/:index/:pos', function(request, response) {
			response.writeHead(200, HEADER);
			var robot = self.getRobotByIndex(parseInt(request.params.index));
			if(robot) {
				robot.neopixelClearLedAt(request);
			}
			response.end();
		});
		
		express.get('/cheeseNeopixelSetLedFromToColor/:id/:index/:from/:to/:interval/:color', function(request, response) {
			response.writeHead(200, HEADER);
			var robot = self.getRobotByIndex(parseInt(request.params.index));
			if(robot) {
				robot.neopixelSetLedFromToColor(request);
			}
			response.end();
		});
		
		express.get('/cheeseNeopixelChangeLedFromByRGB/:id/:index/:from/:to/:interval/:r/:g/:b', function(request, response) {
			response.writeHead(200, HEADER);
			var robot = self.getRobotByIndex(parseInt(request.params.index));
			if(robot) {
				robot.neopixelChangeLedFromByRGB(request);
			}
			response.end();
		});
		
		express.get('/cheeseNeopixelSetLedFromToRGB/:id/:index/:from/:to/:interval/:r/:g/:b', function(request, response) {
			response.writeHead(200, HEADER);
			var robot = self.getRobotByIndex(parseInt(request.params.index));
			if(robot) {
				robot.neopixelSetLedFromToRGB(request);
			}
			response.end();
		});
		
		express.get('/cheeseNeopixelClearLedFromTo/:id/:index/:from/:to/:interval', function(request, response) {
			response.writeHead(200, HEADER);
			var robot = self.getRobotByIndex(parseInt(request.params.index));
			if(robot) {
				robot.neopixelClearLedFromTo(request);
			}
			response.end();
		});
		
		express.get('/cheeseNeopixelChangeBrightnessBy/:id/:index/:value', function(request, response) {
			response.writeHead(200, HEADER);
			var robot = self.getRobotByIndex(parseInt(request.params.index));
			if(robot) {
				robot.neopixelChangeBrightnessBy(request);
			}
			response.end();
		});
		
		express.get('/cheeseNeopixelSetBrightnessTo/:id/:index/:value', function(request, response) {
			response.writeHead(200, HEADER);
			var robot = self.getRobotByIndex(parseInt(request.params.index));
			if(robot) {
				robot.neopixelSetBrightnessTo(request);
			}
			response.end();
		});
		
		express.get('/cheeseNeopixelSetModeTo/:id/:index/:auto', function(request, response) {
			response.writeHead(200, HEADER);
			var robot = self.getRobotByIndex(parseInt(request.params.index));
			if(robot) {
				robot.neopixelSetModeTo(request);
			}
			response.end();
		});
		
		express.get('/cheeseNeopixelUpdate/:id/:index', function(request, response) {
			response.writeHead(200, HEADER);
			var robot = self.getRobotByIndex(parseInt(request.params.index));
			if(robot) {
				robot.neopixelUpdate(request);
			}
			response.end();
		});
		
		express.get('/cheeseWriteSerial/:id/:index/:mode/:value', function(request, response) {
			response.writeHead(200, HEADER);
			var robot = self.getRobotByIndex(parseInt(request.params.index));
			if(robot) {
				robot.writeSerial(request);
			}
			response.end();
		});
		
		express.get('/cheeseReadSerial/:id/:index', function(request, response) {
			response.writeHead(200, HEADER);
			var robot = self.getRobotByIndex(parseInt(request.params.index));
			if(robot) {
				robot.readSerial(request);
			}
			response.end();
		});
		
		express.get('/cheeseReadSerialUntil/:id/:index/:delimiter', function(request, response) {
			response.writeHead(200, HEADER);
			var robot = self.getRobotByIndex(parseInt(request.params.index));
			if(robot) {
				robot.readSerialUntil(request);
			}
			response.end();
		});
	}
};

Extension.prototype.stop = function() {
	this.running = false;
	translator.remove('cheese-scratch-multi');
	router.reset();
	this.robots = {};
};

Extension.prototype.setState = function(route) {
	if(route == router.firstRoute && route.module == MODULE) {
		this.state = route.state;
	}
};

Extension.prototype.setOption = function(option) {
};

Extension.prototype.handleSensory = function(data, route) {
	if(route.module == MODULE) {
		var robot = this.getRobotByRoute(route);
		if(robot) {
			robot.handleSensory(data);
		}
	}
};

Extension.prototype.handleMotoring = function(data, route) {
	if(route.module == MODULE) {
		var robot = this.getRobotByRoute(route);
		if(robot) {
			var roboid = robot.roboid;
			roboid.handleMotoring(data);
			roboid.updateMotoring();
		}
	}
};

Extension.prototype.requestMotoring = function(route) {
	if(route.module == MODULE) {
		var robot = this.getRobotByRoute(route);
		if(robot) {
			return robot.roboid.requestMotoring();
		}
	}
};

Extension.prototype.reset = function(route) {
	if(route.module == MODULE) {
		var robot = this.getRobotByRoute(route);
		if(robot) {
			robot.roboid.reset();
			robot.motoring.reset();
		}
	}
};

module.exports.create = function() {
	return new Extension();
};