'use strict';
var router = require('router');

const BATTERY_LEVEL_LOW = 3.6;
const BATTERY_LEVEL_EMPTY = 3.0;
const BATTERY_LEVEL_NO = 2.35;
const HEX_LEDS = [
	'#ff0000', '#ffff00', '#00ff00', '#00ffff', '#0000ff', '#ff00ff', '#ffffff'
];
const COLOR_LEDS = [
	4, 6, 2, 3, 1, 5, 7
];
const COLOR2RGB = {
	0: [0, 0, 0],
	1: [0, 0, 255],
	2: [0, 255, 0],
	3: [0, 255, 255],
	4: [255, 0, 0],
	5: [255, 0, 255],
	6: [255, 255, 0],
	7: [255, 255, 255]
};

function Controller() {
	this.sensory = {
		map: 0,
		signalStrength: 0,
		leftProximity: 0,
		rightProximity: 0,
		leftFloor: 0,
		rightFloor: 0,
		acceleration: { x: 0, y: 0, z: 0 },
		light: 0,
		temperature: 0,
		inputA: 0,
		inputB: 0,
		lineTracerState: 0,
		tilt: 0,
		batteryState: 2
	};
	this.motoring = {
		map: 0,
		leftWheel: 0,
		rightWheel: 0,
		buzzer: 0,
		outputA: 0,
		outputB: 0,
		topology: 0,
		leftLed: 0,
		rightLed: 0,
		note: 0,
		lineTracerMode: 0,
		lineTracerSpeed: 5,
		ioModeA: 0,
		ioModeB: 0,
		configProximity: 2,
		configGravity: 0,
		configBandWidth: 3
	};
	this.lineTracer = {
		flag: 0,
		event: 0,
		state: 0
	};
	this.event = {
		tilt: -4,
		batteryState: -1
	};
	this.alignment = {
		active: false,
		move: false,
		state: 0,
		speed: 30,
		balance: 0
	};
	this.battery = {
		data: new Array(10),
		sum: 0.0,
		index: 0,
		count: 0
	};
	this.sensorMonitor = {
		count: 0,
		index: 0,
		duration: 0,
		averageIndex: 0,
		signalStrength: { current: 0, sum: 0.0, values: new Array(100) },
		leftProximity: { current: 0, sum: 0.0, values: new Array(100) },
		rightProximity: { current: 0, sum: 0.0, values: new Array(100) },
		leftFloor: { current: 0, sum: 0.0, values: new Array(100) },
		rightFloor: { current: 0, sum: 0.0, values: new Array(100) },
		accelerationX: { current: 0, sum: 0.0, values: new Array(100) },
		accelerationY: { current: 0, sum: 0.0, values: new Array(100) },
		accelerationZ: { current: 0, sum: 0.0, values: new Array(100) },
		light: { current: 0, sum: 0.0, values: new Array(100) },
		temperature: { current: 0, sum: 0.0, values: new Array(100) },
		inputA: { current: 0, sum: 0.0, values: new Array(100) },
		inputB: { current: 0, sum: 0.0, values: new Array(100) },
		reset: function() {
			this.signalStrength.current = 0;
			this.signalStrength.sum = 0.0;
			this.leftProximity.current = 0;
			this.leftProximity.sum = 0.0;
			this.rightProximity.current = 0;
			this.rightProximity.sum = 0.0;
			this.leftFloor.current = 0;
			this.leftFloor.sum = 0.0;
			this.rightFloor.current = 0;
			this.rightFloor.sum = 0.0;
			this.accelerationX.current = 0;
			this.accelerationX.sum = 0.0;
			this.accelerationY.current = 0;
			this.accelerationY.sum = 0.0;
			this.accelerationZ.current = 0;
			this.accelerationZ.sum = 0.0;
			this.light.current = 0;
			this.light.sum = 0.0;
			this.temperature.current = 0;
			this.temperature.sum = 0.0;
			this.inputA.current = 0;
			this.inputA.sum = 0.0;
			this.inputB.current = 0;
			this.inputB.sum = 0.0;
			this.count = 0;
			this.index = 0;
			this.averageIndex = 0;
		},
		put: function(device, value) {
			if(this.duration > 0) {
				device.sum += value;
				if(this.averageIndex >= this.duration - 1) {
					device.current = parseInt(device.sum * 100 / this.duration) / 100;
					device.sum = 0.0;
					device.values[this.index] = device.current;
				}
			} else {
				device.current = value;
				device.values[this.index] = value;
			}
		},
		update: function() {
			if(this.duration > 0) {
				++ this.averageIndex;
				if(this.averageIndex >= this.duration) {
					this.averageIndex = 0;
					if(this.count < 100) ++ this.count;
					++ this.index;
					if(this.index >= 100) this.index = 0;
				}
			} else {
				if(this.count < 100) ++ this.count;
				++ this.index;
				if(this.index >= 100) this.index = 0;
			}
		}
	};
	this.sensorNames = [
		'proximity sensor and floor sensor', 'acceleration sensor', 'light sensor', 'signal strength and temperature sensor', 'input A/B'
	];
	this.updateIdentity();
}

Controller.prototype.updateIdentity = function() {
	this.identity = Math.floor(Math.random() * 49); // 0 ~ 48
};

Controller.prototype.getIdentitySize = function() {
	return 2;
};

Controller.prototype.getIdentityHex = function(index) {
	if(index == 0) return HEX_LEDS[Math.floor(this.identity / 7)];
	else return HEX_LEDS[this.identity % 7];
};

Controller.prototype.getIdentityColor = function(index) {
	if(index == 0) return COLOR_LEDS[Math.floor(this.identity / 7)];
	else return COLOR_LEDS[this.identity % 7];
};

Controller.prototype.getIdentityX = function(index) {
	return (index == 0) ? '45px' : '19px'
};

Controller.prototype.getIdentityY = function(index) {
	return (index == 0) ? '75px' : '61px'
};

Controller.prototype.toHex = function(number) {
	var value = parseInt(number);
	if(value < 0) value += 0x100;

	value = value.toString(16).toUpperCase();
	if(value.length > 1) return value;
	else return '0' + value;
};

Controller.prototype.toHex2 = function(number) {
	var value = parseInt(number);
	if(value < 0) value += 0x10000;
	
	value = value.toString(16).toUpperCase();
	var result = '';
	for(var i = value.length; i < 4; ++i) {
		result += '0';
	}
	return result + value;
};

Controller.prototype.toHex3 = function(number) {
	var value = parseInt(number);
	if(value < 0) value += 0x1000000;
	
	value = value.toString(16).toUpperCase();
	var result = '';
	for(var i = value.length; i < 6; ++i) {
		result += '0';
	}
	return result + value;
};

Controller.prototype.load = function(connector, route, callback) {
	callback();
};

Controller.prototype.handleLocalData = function(data, route) { // data: string
	if(data.length != 53) return;

	if(route.hwModule == 'hamster') {
		var str = data.slice(4, 5);
		var value = parseInt(str, 16);
		if(value != 1) return; // invalid data
		
		var sensory = this.sensory;
		sensory.map = 0x80007fe0;
		
		var sensorMonitor = this.sensorMonitor;
		// signal strength
		str = data.slice(6, 8);
		value = parseInt(str, 16);
		value -= 0x100;
		sensory.signalStrength = value;
		sensorMonitor.put(sensorMonitor.signalStrength, value);
		// left proximity
		str = data.slice(8, 10);
		value = parseInt(str, 16);
		sensory.leftProximity = value;
		sensorMonitor.put(sensorMonitor.leftProximity, value);
		// right proximity
		str = data.slice(10, 12);
		value = parseInt(str, 16);
		sensory.rightProximity = value;
		sensorMonitor.put(sensorMonitor.rightProximity, value);
		// left floor
		str = data.slice(12, 14);
		value = parseInt(str, 16);
		sensory.leftFloor = value;
		sensorMonitor.put(sensorMonitor.leftFloor, value);
		// right floor
		str = data.slice(14, 16);
		value = parseInt(str, 16);
		sensory.rightFloor = value;
		sensorMonitor.put(sensorMonitor.rightFloor, value);
		// acceleration x
		var acceleration = sensory.acceleration;
		str = data.slice(16, 20);
		value = parseInt(str, 16);
		if(value > 0x7fff) value -= 0x10000;
		acceleration.x = value;
		sensorMonitor.put(sensorMonitor.accelerationX, value);
		// acceleration y
		str = data.slice(20, 24);
		value = parseInt(str, 16);
		if(value > 0x7fff) value -= 0x10000;
		acceleration.y = value;
		sensorMonitor.put(sensorMonitor.accelerationY, value);
		// acceleration z
		str = data.slice(24, 28);
		value = parseInt(str, 16);
		if(value > 0x7fff) value -= 0x10000;
		acceleration.z = value;
		sensorMonitor.put(sensorMonitor.accelerationZ, value);
		// tilt
		if(acceleration.z < 8192 && acceleration.x > 8192 && acceleration.y > -4096 && acceleration.y < 4096) value = 1;
		else if(acceleration.z < 8192 && acceleration.x < -8192 && acceleration.y > -4096 && acceleration.y < 4096) value = -1;
		else if(acceleration.z < 8192 && acceleration.y > 8192 && acceleration.x > -4096 && acceleration.x < 4096) value = 2;
		else if(acceleration.z < 8192 && acceleration.y < -8192 && acceleration.x > -4096 && acceleration.x < 4096) value = -2;
		else if(acceleration.z > 12288 && acceleration.x > -8192 && acceleration.x < 8192 && acceleration.y > -8192 && acceleration.y < 8192) value = 3;
		else if(acceleration.z < -12288 && acceleration.x > -4096 && acceleration.x < 4096 && acceleration.y > -4096 && acceleration.y < 4096) value = -3;
		else value = 0;
		var event = this.event;
		if(value != event.tilt) {
			if(event.tilt != -4) {
				sensory.map |= 0x00000008;
			}
			event.tilt = value;
			sensory.tilt = value;
		}
		// flag
		str = data.slice(28, 30);
		var flag = parseInt(str, 16);
		if(flag == 0) {
			// light
			str = data.slice(30, 34);
			value = parseInt(str, 16);
			sensory.light = value;
		} else {
			// temperature
			str = data.slice(30, 32);
			value = parseInt(str, 16);
			if(value > 0x7f) value -= 0x100;
			value = value / 2.0 + 24;
			sensory.temperature = parseInt(value);
			// battery
			str = data.slice(32, 34);
			value = parseInt(str, 16);
			value = value / 100.0 + 2;
			// battery state
			var batt = this.battery;
			if(batt.count < 10) {
				++ batt.count;
			} else {
				batt.index %= 10;
				batt.sum -= batt.data[batt.index];
			}
			batt.sum += value;
			batt.data[batt.index] = value;
			++ batt.index;
			value = batt.sum / batt.count;
			var state = 2;
			if(value < BATTERY_LEVEL_EMPTY) state = 0;
			else if(value < BATTERY_LEVEL_LOW) state = 1;
			if(state != event.batteryState) {
				if(event.batteryState != -1) {
					sensory.map |= 0x00000004;
				}
				event.batteryState = state;
				sensory.batteryState = state;
			}
		}
		sensorMonitor.put(sensorMonitor.light, sensory.light);
		sensorMonitor.put(sensorMonitor.temperature, sensory.temperature);
		
		// input a
		str = data.slice(34, 36);
		value = parseInt(str, 16);
		sensory.inputA = value;
		sensorMonitor.put(sensorMonitor.inputA, value);
		// input b
		str = data.slice(36, 38);
		value = parseInt(str, 16);
		sensory.inputB = value;
		sensorMonitor.put(sensorMonitor.inputB, value);
		// line tracer state
		str = data.slice(38, 40);
		value = parseInt(str, 16);
		if((value & 0x40) != 0) {
			var lineTracer = this.lineTracer;
			if(lineTracer.event == 1) {
				if(value != 0x40) {
					lineTracer.event = 2;
				}
			}
			if(lineTracer.event == 2) {
				if(value != lineTracer.state) {
					lineTracer.state = value;
					sensory.lineTracerState = value;
					sensory.map |= 0x00000010;
					if(value == 0x40) {
						lineTracer.event = 0;
					}
				}
			}
		}
		sensorMonitor.update();
		return sensory;
	} else if(route.hwModule == 'hamsterS') {
		var sensory = this.sensory;
		sensory.map = 0x80007fe0;
		var str = data.slice(0, 1);
		var value = parseInt(str, 16);
		if(value == 1) { // normal
			var sensorMonitor = this.sensorMonitor;
			var event = this.event;
			// left proximity
			str = data.slice(6, 8);
			value = parseInt(str, 16);
			sensory.leftProximity = value;
			sensorMonitor.put(sensorMonitor.leftProximity, value);
			// right proximity
			str = data.slice(8, 10);
			value = parseInt(str, 16);
			sensory.rightProximity = value;
			sensorMonitor.put(sensorMonitor.rightProximity, value);
			// flag
			str = data.slice(38, 40);
			var value2 = parseInt(str, 16);
			if((value2 & 0x01) == 0) {
				// light
				str = data.slice(10, 14);
				value = parseInt(str, 16);
				sensory.light = value;
			} else {
				// temperature
				str = data.slice(10, 12);
				value = parseInt(str, 16);
				if(value > 0x7f) value -= 0x100;
				value = value / 2.0 + 23;
				sensory.temperature = parseInt(value);
			}
			sensorMonitor.put(sensorMonitor.light, sensory.light);
			sensorMonitor.put(sensorMonitor.temperature, sensory.temperature);
			// left floor
			str = data.slice(14, 16);
			value = parseInt(str, 16);
			sensory.leftFloor = value;
			sensorMonitor.put(sensorMonitor.leftFloor, value);
			// right floor
			str = data.slice(16, 18);
			value = parseInt(str, 16);
			sensory.rightFloor = value;
			sensorMonitor.put(sensorMonitor.rightFloor, value);
			// acceleration x
			var acceleration = sensory.acceleration;
			str = data.slice(18, 22);
			value = parseInt(str, 16);
			if(value > 0x7fff) value -= 0x10000;
			acceleration.x = value;
			sensorMonitor.put(sensorMonitor.accelerationX, value);
			// acceleration y
			str = data.slice(22, 26);
			value = parseInt(str, 16);
			if(value > 0x7fff) value -= 0x10000;
			acceleration.y = value;
			sensorMonitor.put(sensorMonitor.accelerationY, value);
			// acceleration z
			str = data.slice(26, 30);
			value = parseInt(str, 16);
			if(value > 0x7fff) value -= 0x10000;
			acceleration.z = value;
			sensorMonitor.put(sensorMonitor.accelerationZ, value);
			// tilt
			if(acceleration.z < 8192 && acceleration.x > 8192 && acceleration.y > -4096 && acceleration.y < 4096) value = 1;
			else if(acceleration.z < 8192 && acceleration.x < -8192 && acceleration.y > -4096 && acceleration.y < 4096) value = -1;
			else if(acceleration.z < 8192 && acceleration.y > 8192 && acceleration.x > -4096 && acceleration.x < 4096) value = 2;
			else if(acceleration.z < 8192 && acceleration.y < -8192 && acceleration.x > -4096 && acceleration.x < 4096) value = -2;
			else if(acceleration.z > 12288 && acceleration.x > -8192 && acceleration.x < 8192 && acceleration.y > -8192 && acceleration.y < 8192) value = 3;
			else if(acceleration.z < -12288 && acceleration.x > -4096 && acceleration.x < 4096 && acceleration.y > -4096 && acceleration.y < 4096) value = -3;
			else value = 0;
			if(value != event.tilt) {
				if(event.tilt != -4) {
					sensory.map |= 0x00000008;
				}
				event.tilt = value;
				sensory.tilt = value;
			}
			// input a
			str = data.slice(30, 32);
			value = parseInt(str, 16);
			sensory.inputA = value;
			sensorMonitor.put(sensorMonitor.inputA, value);
			// input b
			str = data.slice(32, 34);
			value = parseInt(str, 16);
			sensory.inputB = value;
			sensorMonitor.put(sensorMonitor.inputB, value);
			// signal strength
			str = data.slice(36, 38);
			value = parseInt(str, 16);
			value -= 0x100;
			sensory.signalStrength = value;
			sensorMonitor.put(sensorMonitor.signalStrength, value);
			// linetracer state
			state = (value2 >> 6) & 0x03;
			if((state & 0x02) != 0) {
				var lineTracer = this.lineTracer;
				if(lineTracer.event == 1) {
					if(state == 0x02) {
						if(++lineTracer.count > 5) lineTracer.event = 2;
					} else {
						lineTracer.event = 2;
					}
				}
				if(lineTracer.event == 2) {
					if(state != lineTracer.state || lineTracer.count > 5) {
						lineTracer.state = state;
						sensory.lineTracerState = (state << 5);
						sensory.map |= 0x00000010;
						if(state == 0x02) {
							lineTracer.event = 0;
							lineTracer.count = 0;
						}
					}
				}
			}
			// battery state
			state = (value2 >> 1) & 0x03;
			if(state == 0) state = 2; // normal
			else if(state >= 2) state = 0; // empty
			if(state != event.batteryState) {
				if(event.batteryState != -1) {
					sensory.map |= 0x00000004;
				}
				event.batteryState = state;
				sensory.batteryState = state;
			}
			sensorMonitor.update();
			return sensory;
		}
	}
};

Controller.prototype.write = function(data) {
	if(!data) return this.motoring;
	
	var map = data.map;
	var writeData = data.writeData;
	var motoring = this.motoring;
	// left wheel
	var t = motoring.leftWheel + writeData[0][0];
	if(t < -100) t = -100;
	else if(t > 100) t = 100;
	motoring.leftWheel = t;
	// right wheel
	t = motoring.rightWheel + writeData[1][0];
	if(t < -100) t = -100;
	else if(t > 100) t = 100;
	motoring.rightWheel = t;
	// buzzer;
	t = writeData[2][0];
	if(t == 0) {
		motoring.buzzer = 0;
	} else {
		t = motoring.buzzer + t;
		if(t < 0.01) t = 0.01;
		else if(t > 167772.15) t = 167772.15;
		motoring.buzzer = t;
	}
	// output a
	t = motoring.outputA + writeData[3][0];
	if(t < 0) t = 0;
	else if(t > 255) t = 255;
	motoring.outputA = t;
	// output b
	t = motoring.outputB + writeData[4][0];
	if(t < 0) t = 0;
	else if(t > 255) t = 255;
	motoring.outputB = t;
	// commands
	if(map & 0x02000000) motoring.topology = writeData[5][0];
	if(map & 0x01000000) motoring.leftLed = writeData[6][0];
	if(map & 0x00800000) motoring.rightLed = writeData[7][0];
	if(map & 0x00400000) motoring.note = writeData[8][0];
	if(map & 0x00200000) motoring.lineTracerMode = writeData[9][0];
	if(map & 0x00100000) motoring.lineTracerSpeed = writeData[10][0];
	if(map & 0x00080000) motoring.ioModeA = writeData[11][0];
	if(map & 0x00040000) motoring.ioModeB = writeData[12][0];
	if(map & 0x00020000) motoring.configProximity = writeData[13][0];
	if(map & 0x00010000) motoring.configGravity = writeData[14][0];
	if(map & 0x00008000) motoring.configBandWidth = writeData[15][0];
	
	motoring.map |= map;
	return motoring;
};

Controller.prototype.speed2gain = function(speed) {
	switch(speed) {
		case 1:
		case 2: return 6;
		case 3:
		case 4: return 5;
		case 5:
		case 6: return 4;
		case 7:
		case 8: return 3;
	}
	if(speed < 1) return 6;
	return 3;
};

Controller.prototype.requestLocalData = function(address, route) {
	var motoring = this.motoring;
	
	if(route.hwModule == 'hamster') {
		var str = this.toHex(motoring.topology & 0x0f); // version, topology
		str += '00'; // network
		var alignment = this.alignment;
		if(alignment.active) {
			if(alignment.state > 0) {
				alignment.state --;
				if(alignment.state == 2) {
					return null;
				} else if(alignment.state == 1) {
					str += 'E0'; // command, security
				} else if(alignment.state == 0) {
					alignment.balance = 0;
					return null;
				} else {
					str += '10'; // command, security
				}
			} else {
				str += '10'; // command, security
			}
			if(alignment.move) {
				var speed = this.toHex(alignment.speed);
				str += speed;
				str += speed;
			} else {
				str += '0000';
			}
		} else {
			str += '10'; // command, security
			str += this.toHex(motoring.leftWheel);
			str += this.toHex(motoring.rightWheel);
		}
		if(router.isExtraButtonActive()) {
			str += this.toHex(this.getIdentityColor(0));
			str += this.toHex(this.getIdentityColor(1));
		} else {
			str += this.toHex(motoring.leftLed);
			str += this.toHex(motoring.rightLed);
		}
		str += this.toHex3(motoring.buzzer * 100);
		str += this.toHex(motoring.note);
		var lineTracer = this.lineTracer;
		if(motoring.map & 0x00200000) {
			if(motoring.lineTracerMode > 0) {
				lineTracer.flag ^= 0x80;
				lineTracer.event = 1;
			}
		}
		var tmp = (motoring.lineTracerMode & 0x0f) << 3;
		tmp |= ((motoring.lineTracerSpeed - 1) & 0x07);
		tmp |= lineTracer.flag & 0x80;
		str += this.toHex(tmp);
		str += this.toHex(motoring.configProximity);
		tmp = (motoring.configGravity & 0x0f) << 4;
		tmp |= (motoring.configBandWidth & 0x0f);
		str += this.toHex(tmp);
		tmp = (motoring.ioModeA & 0x0f) << 4;
		tmp |= (motoring.ioModeB & 0x0f);
		str += this.toHex(tmp);
		str += this.toHex(motoring.outputA);
		str += this.toHex(motoring.outputB);
		str += this.toHex(this.alignment.balance);
		str += '0000-';
		str += address;
		str += '\r';
		return str;
	} else if(route.hwModule == 'hamsterS') {
		var str = '10';
		if(motoring.leftWheel < 0) str += this.toHex(motoring.leftWheel * 1.27 - 0.5);
		else str += this.toHex(motoring.leftWheel * 1.27 + 0.5);
		if(motoring.rightWheel < 0) str += this.toHex(motoring.rightWheel * 1.27 - 0.5);
		else str += this.toHex(motoring.rightWheel * 1.27 + 0.5);
		if(router.isExtraButtonActive()) {
			var rgb = COLOR2RGB[this.getIdentityColor(0)] || COLOR2RGB[0];
			str += this.toHex(rgb[0]);
			str += this.toHex(rgb[1]);
			str += this.toHex(rgb[2]);
			rgb = COLOR2RGB[this.getIdentityColor(1)] || COLOR2RGB[0];
			str += this.toHex(rgb[0]);
			str += this.toHex(rgb[1]);
			str += this.toHex(rgb[2]);
		} else {
			var rgb = COLOR2RGB[motoring.leftLed] || COLOR2RGB[0];
			str += this.toHex(rgb[0]);
			str += this.toHex(rgb[1]);
			str += this.toHex(rgb[2]);
			rgb = COLOR2RGB[motoring.rightLed] || COLOR2RGB[0];
			str += this.toHex(rgb[0]);
			str += this.toHex(rgb[1]);
			str += this.toHex(rgb[2]);
		}
		str += this.toHex(0);
		str += this.toHex2(0);
		var lineTracer = this.lineTracer;
		var tmp = motoring.lineTracerMode & 0x0f;
		if(tmp > 7) tmp += 1; // for white line
		if(motoring.map & 0x00200000) {
			lineTracer.count = 0;
			if(tmp > 0) {
				lineTracer.id = (lineTracer.id % 15) + 1;
				lineTracer.event = 1;
			} else {
				lineTracer.event = 0;
			}
		}
		tmp |= (lineTracer.id & 0x0f) << 4;
		str += this.toHex(tmp);
		tmp = (motoring.lineTracerSpeed & 0x0f) << 4;
		tmp |= this.speed2gain(motoring.lineTracerSpeed) & 0x0f;
		str += this.toHex(tmp);
		tmp = (motoring.configProximity & 0x07) << 5;
		tmp |= (motoring.configBandWidth & 0x07) << 2;
		tmp |= (motoring.configGravity & 0x03);
		str += this.toHex(tmp);
		tmp = (motoring.ioModeA & 0x0f) << 4;
		tmp |= (motoring.ioModeB & 0x0f);
		str += this.toHex(tmp);
		str += this.toHex(motoring.outputA);
		str += this.toHex(motoring.outputB);
		
		if(motoring.note > 0) { // note
			str += '01';
			str += this.toHex(motoring.note);
		} else { // buzzer
			tmp = parseInt(motoring.buzzer * 10) + 512;
			str += this.toHex2(tmp);
		}
		str += '-';
		str += address;
		str += '\r';
		return str;
	}
};

Controller.prototype.clear = function() {
	var motoring = this.motoring;
	motoring.map = 0;
	motoring.leftWheel = 0;
	motoring.rightWheel = 0;
	motoring.buzzer = 0.0;
	motoring.outputA = 0;
	motoring.outputB = 0;
	var sensory = this.sensory;
	sensory.map = 0;
	sensory.lineTracerState = 0;
	sensory.tilt = 0;
	sensory.batteryState = 2;
};

Controller.prototype.reset = function() {
	this.clear();
	var motoring = this.motoring;
	motoring.topology = 0;
	motoring.leftLed = 0;
	motoring.rightLed = 0;
	motoring.note = 0;
	motoring.lineTracerMode = 0;
	motoring.lineTracerSpeed = 5;
	motoring.ioModeA = 0;
	motoring.ioModeB = 0;
	motoring.configProximity = 2;
	motoring.configGravity = 0;
	motoring.configBandWidth = 3;
	var lineTracer = this.lineTracer;
	lineTracer.event = 0;
	lineTracer.state = 0;
	var event = this.event;
	event.tilt = -4;
	event.batteryState = -1;
	var batt = this.battery;
	batt.sum = 0.0;
	batt.index = 0;
	batt.count = 0;
};

Controller.prototype.activateAlignment = function() {
	this.alignment.active = true;
};

Controller.prototype.deactivateAlignment = function() {
	this.alignment.active = false;
};

Controller.prototype.startAlignmentMovement = function() {
	this.alignment.move = true;
};

Controller.prototype.stopAlignmentMovement = function() {
	this.alignment.move = false;
};

Controller.prototype.setAlignmentSpeed = function(speed) {
	this.alignment.speed = speed;
};

Controller.prototype.setAlignmentBalance = function(balance) {
	this.alignment.balance = balance;
};

Controller.prototype.resetAlignment = function() {
	this.alignment.move = false;
	this.alignment.state = 0;
	this.alignment.speed = 30;
	this.alignment.balance = 0;
};

Controller.prototype.saveAlignment = function(balance) {
	this.alignment.balance = balance;
	this.alignment.state = 3;
};

Controller.prototype.getSensorMonitor = function() {
	return this.sensorMonitor;
};

Controller.prototype.getSensorNames = function() {
	return this.sensorNames;
};

Controller.prototype.setDuration = function(duration) {
	var sensorMonitor = this.sensorMonitor;
	sensorMonitor.reset();
	sensorMonitor.duration = duration;
};

module.exports.create = function() {
	return new Controller();
};