'use strict';
var context = require('context');
var translations = require('./translations');
var Hamster = require('../hamster');
var factory = require('../hamster-roboid');
const MODULE = 'hamster';

var PARTS = {};
var DIRECTIONS = {};
var TOWARDS = {};
var UNITS = {};
var COLORS = {};
var NOTES = {};
var BEATS = { '¼': 0.25, '½': 0.5, '¾': 0.75, '1¼': 1.25, '1½': 1.5, '1¾': 1.75 };
var SOUNDS = {};
var IO_MODES = {};
var GRIPPERS = {};
var VALUES = {};

var BOOLEAN_TILT_FORWARD = '/tilt%20forward ';
var BOOLEAN_TILT_BACKWARD = '/tilt%20backward ';
var BOOLEAN_TILT_LEFT = '/tilt%20left ';
var BOOLEAN_TILT_RIGHT = '/tilt%20right ';
var BOOLEAN_TILT_FLIP = '/tilt%20flip ';
var BOOLEAN_TILT_NONE = '/not%20tilt ';
var BOOLEAN_BATTERY_NORMAL = '/normal ';
var BOOLEAN_BATTERY_LOW = '/low ';
var BOOLEAN_BATTERY_EMPTY = '/empty ';

const LEFT = 1;
const RIGHT = 2;
const BOTH = 3;
const FRONT = 4;
const REAR = 5;
const LEFT_PEN = 1;
const RIGHT_PEN = 2;
const LEFT_WHEEL = 3;
const RIGHT_WHEEL = 4;
const FORWARD = 1;
const SECONDS = 1;
const OPEN = 1;
const CLOSE = 2;
const WHITE = 1;

function Robot(index) {
	var roboid = this.roboid = factory.create();
	this.index = index;
	this.waits = [];
	this.timeouts = [];
	this.sensory = {
		signalStrength: 0,
		leftProximity: 0,
		rightProximity: 0,
		leftFloor: 0,
		rightFloor: 0,
		accelerationX: 0,
		accelerationY: 0,
		accelerationZ: 0,
		light: 0,
		temperature: 0,
		inputA: 0,
		inputB: 0,
		tilt: 0,
		batteryState: 2
	};
	this.motoring = {
		topology: roboid.findDeviceById(Hamster.TOPOLOGY),
		leftLed: roboid.findDeviceById(Hamster.LEFT_LED),
		rightLed: roboid.findDeviceById(Hamster.RIGHT_LED),
		note: roboid.findDeviceById(Hamster.NOTE),
		lineTracerMode: roboid.findDeviceById(Hamster.LINE_TRACER_MODE),
		lineTracerSpeed: roboid.findDeviceById(Hamster.LINE_TRACER_SPEED),
		ioModeA: roboid.findDeviceById(Hamster.IO_MODE_A),
		ioModeB: roboid.findDeviceById(Hamster.IO_MODE_B),
		configProximity: roboid.findDeviceById(Hamster.CONFIG_PROXIMITY),
		configGravity: roboid.findDeviceById(Hamster.CONFIG_GRAVITY),
		configBandWidth: roboid.findDeviceById(Hamster.CONFIG_BAND_WIDTH),
		reset: function() {
			this.leftWheel.data = 0;
			this.rightWheel.data = 0;
			this.buzzer.data = 0;
			this.outputA.data = 0;
			this.outputB.data = 0;
		}
	};
	this.register('leftWheel', Hamster.LEFT_WHEEL);
	this.register('rightWheel', Hamster.RIGHT_WHEEL);
	this.register('buzzer', Hamster.BUZZER);
	this.register('outputA', Hamster.OUTPUT_A);
	this.register('outputB', Hamster.OUTPUT_B);
	this.tempo = 60;
	this.wheelId = -1;
	this.wheelTimer = undefined;
	this.lineTracerId = -1;
	this.boardId = -1;
	this.boardCommand = 0;
	this.boardState = 0;
	this.boardCount = 0;
	this.noteId = -1;
	this.noteTimer1 = undefined;
	this.noteTimer2 = undefined;
	this.ioId = -1;
	this.ioTimer = undefined;
	this.active = true;
	this.received = false;
}

Robot.prototype.reset = function() {
	this.tempo = 60;
	this.wheelId = -1;
	this.wheelTimer = undefined;
	this.lineTracerId = -1;
	this.boardId = -1;
	this.boardCommand = 0;
	this.boardState = 0;
	this.boardCount = 0;
	this.noteId = -1;
	this.noteTimer1 = undefined;
	this.noteTimer2 = undefined;
	this.ioId = -1;
	this.ioTimer = undefined;
	this.roboid.reset();
	this.motoring.reset();
};

Robot.prototype.register = function(name, deviceId) {
	var self = this;
	self.motoring[name] = {
		data: 0,
		device: self.roboid.findDeviceById(deviceId),
		write: function(value) {
			this.data = value;
			this.device.write(value);
		},
		read: function() {
			return this.data;
		}
	};
};

Robot.prototype.getWaits = function() {
	return this.waits;
};

Robot.prototype.removeWait = function(id) {
	var waits = this.waits;
	var index = waits.indexOf(id);
	if(index >= 0) {
		waits.splice(index, 1);
	}
};

Robot.prototype.clearWaits = function() {
	this.waits = [];
};

Robot.prototype.removeTimeout = function(id) {
	clearTimeout(id);
	var timeouts = this.timeouts;
	var index = timeouts.indexOf(id);
	if(index >= 0) {
		timeouts.splice(index, 1);
	}
};

Robot.prototype.clearTimeouts = function() {
	var timeouts = this.timeouts;
	for(var i in timeouts) {
		clearTimeout(timeouts[i]);
	}
	this.timeouts = [];
};

Robot.prototype.cancelWheel = function() {
	this.wheelId = -1;
	if(this.wheelTimer !== undefined) {
		this.removeTimeout(this.wheelTimer);
	}
	this.wheelTimer = undefined;
};

Robot.prototype.cancelLineTracer = function() {
	this.lineTracerId = -1;
};

Robot.prototype.cancelBoard = function() {
	this.boardId = -1;
	this.boardCommand = 0;
	this.boardState = 0;
	this.boardCount = 0;
};

Robot.prototype.cancelNote = function() {
	this.noteId = -1;
	if(this.noteTimer1 !== undefined) {
		this.removeTimeout(this.noteTimer1);
	}
	if(this.noteTimer2 !== undefined) {
		this.removeTimeout(this.noteTimer2);
	}
	this.noteTimer1 = undefined;
	this.noteTimer2 = undefined;
};

Robot.prototype.cancelIo = function() {
	this.ioId = -1;
	if(this.ioTimer !== undefined) {
		this.removeTimeout(this.ioTimer);
	}
	this.ioTimer = undefined;
};

Robot.prototype.writeSingleSensory = function(response) {
	var sensory = this.sensory;
	response.write('signalStrength ' + sensory.signalStrength + '\n');
	response.write('leftProximity ' + sensory.leftProximity + '\n');
	response.write('rightProximity ' + sensory.rightProximity + '\n');
	response.write('leftFloor ' + sensory.leftFloor + '\n');
	response.write('rightFloor ' + sensory.rightFloor + '\n');
	response.write('accelerationX ' + sensory.accelerationX + '\n');
	response.write('accelerationY ' + sensory.accelerationY + '\n');
	response.write('accelerationZ ' + sensory.accelerationZ + '\n');
	response.write('light ' + sensory.light + '\n');
	response.write('temperature ' + sensory.temperature + '\n');
	response.write('inputA ' + sensory.inputA + '\n');
	response.write('inputB ' + sensory.inputB + '\n');
	var threshold = context.getHandFoundThreshold(MODULE);
	var handFound = (sensory.leftProximity > threshold || sensory.rightProximity > threshold) ? true : false;
	response.write('handFound ' + handFound + '\n');
	response.write('tilt' + BOOLEAN_TILT_FORWARD + (sensory.tilt == 1) + '\n');
	response.write('tilt' + BOOLEAN_TILT_BACKWARD + (sensory.tilt == -1) + '\n');
	response.write('tilt' + BOOLEAN_TILT_LEFT + (sensory.tilt == 2) + '\n');
	response.write('tilt' + BOOLEAN_TILT_RIGHT + (sensory.tilt == -2) + '\n');
	response.write('tilt' + BOOLEAN_TILT_FLIP + (sensory.tilt == 3) + '\n');
	response.write('tilt' + BOOLEAN_TILT_NONE + (sensory.tilt == -3) + '\n');
	response.write('battery' + BOOLEAN_BATTERY_NORMAL + (sensory.batteryState == 2) + '\n');
	response.write('battery' + BOOLEAN_BATTERY_LOW + (sensory.batteryState == 1) + '\n');
	response.write('battery' + BOOLEAN_BATTERY_EMPTY + (sensory.batteryState == 0) + '\n');
	
	response.write('sSignalStrength ' + sensory.signalStrength + '\n');
	response.write('sLeftProximity ' + sensory.leftProximity + '\n');
	response.write('sRightProximity ' + sensory.rightProximity + '\n');
	response.write('sLeftFloor ' + sensory.leftFloor + '\n');
	response.write('sRightFloor ' + sensory.rightFloor + '\n');
	response.write('sAccelerationX ' + sensory.accelerationX + '\n');
	response.write('sAccelerationY ' + sensory.accelerationY + '\n');
	response.write('sAccelerationZ ' + sensory.accelerationZ + '\n');
	response.write('sLight ' + sensory.light + '\n');
	response.write('sTemperature ' + sensory.temperature + '\n');
	response.write('sInputA ' + sensory.inputA + '\n');
	response.write('sInputB ' + sensory.inputB + '\n');
	response.write('sHandFound ' + handFound + '\n');
	response.write('sTilt' + BOOLEAN_TILT_FORWARD + (sensory.tilt == 1) + '\n');
	response.write('sTilt' + BOOLEAN_TILT_BACKWARD + (sensory.tilt == -1) + '\n');
	response.write('sTilt' + BOOLEAN_TILT_LEFT + (sensory.tilt == 2) + '\n');
	response.write('sTilt' + BOOLEAN_TILT_RIGHT + (sensory.tilt == -2) + '\n');
	response.write('sTilt' + BOOLEAN_TILT_FLIP + (sensory.tilt == 3) + '\n');
	response.write('sTilt' + BOOLEAN_TILT_NONE + (sensory.tilt == -3) + '\n');
	response.write('sBattery' + BOOLEAN_BATTERY_NORMAL + (sensory.batteryState == 2) + '\n');
	response.write('sBattery' + BOOLEAN_BATTERY_LOW + (sensory.batteryState == 1) + '\n');
	response.write('sBattery' + BOOLEAN_BATTERY_EMPTY + (sensory.batteryState == 0) + '\n');
};

Robot.prototype.writeMultiSensory = function(response) {
	if(this.active) {
		var sensory = this.sensory;
		var index = this.index;
		var tag = index + ' ';
		response.write('signalStrength/' + tag + sensory.signalStrength + '\n');
		response.write('leftProximity/' + tag + sensory.leftProximity + '\n');
		response.write('rightProximity/' + tag + sensory.rightProximity + '\n');
		response.write('leftFloor/' + tag + sensory.leftFloor + '\n');
		response.write('rightFloor/' + tag + sensory.rightFloor + '\n');
		response.write('accelerationX/' + tag + sensory.accelerationX + '\n');
		response.write('accelerationY/' + tag + sensory.accelerationY + '\n');
		response.write('accelerationZ/' + tag + sensory.accelerationZ + '\n');
		response.write('light/' + tag + sensory.light + '\n');
		response.write('temperature/' + tag + sensory.temperature + '\n');
		response.write('inputA/' + tag + sensory.inputA + '\n');
		response.write('inputB/' + tag + sensory.inputB + '\n');
		var threshold = context.getHandFoundThreshold(MODULE);
		var handFound = (sensory.leftProximity > threshold || sensory.rightProximity > threshold) ? true : false;
		response.write('handFound/' + tag + handFound + '\n');
		response.write('tilt/' + index + BOOLEAN_TILT_FORWARD + (sensory.tilt == 1) + '\n');
		response.write('tilt/' + index + BOOLEAN_TILT_BACKWARD + (sensory.tilt == -1) + '\n');
		response.write('tilt/' + index + BOOLEAN_TILT_LEFT + (sensory.tilt == 2) + '\n');
		response.write('tilt/' + index + BOOLEAN_TILT_RIGHT + (sensory.tilt == -2) + '\n');
		response.write('tilt/' + index + BOOLEAN_TILT_FLIP + (sensory.tilt == 3) + '\n');
		response.write('tilt/' + index + BOOLEAN_TILT_NONE + (sensory.tilt == -3) + '\n');
		response.write('battery/' + index + BOOLEAN_BATTERY_NORMAL + (sensory.batteryState == 2) + '\n');
		response.write('battery/' + index + BOOLEAN_BATTERY_LOW + (sensory.batteryState == 1) + '\n');
		response.write('battery/' + index + BOOLEAN_BATTERY_EMPTY + (sensory.batteryState == 0) + '\n');
		
		response.write('signalStrength' + tag + sensory.signalStrength + '\n');
		response.write('leftProximity' + tag + sensory.leftProximity + '\n');
		response.write('rightProximity' + tag + sensory.rightProximity + '\n');
		response.write('leftFloor' + tag + sensory.leftFloor + '\n');
		response.write('rightFloor' + tag + sensory.rightFloor + '\n');
		response.write('accelerationX' + tag + sensory.accelerationX + '\n');
		response.write('accelerationY' + tag + sensory.accelerationY + '\n');
		response.write('accelerationZ' + tag + sensory.accelerationZ + '\n');
		response.write('light' + tag + sensory.light + '\n');
		response.write('temperature' + tag + sensory.temperature + '\n');
		response.write('inputA' + tag + sensory.inputA + '\n');
		response.write('inputB' + tag + sensory.inputB + '\n');
		response.write('handFound' + tag + handFound + '\n');
		
		response.write('sSignalStrength/' + tag + sensory.signalStrength + '\n');
		response.write('sLeftProximity/' + tag + sensory.leftProximity + '\n');
		response.write('sRightProximity/' + tag + sensory.rightProximity + '\n');
		response.write('sLeftFloor/' + tag + sensory.leftFloor + '\n');
		response.write('sRightFloor/' + tag + sensory.rightFloor + '\n');
		response.write('sAccelerationX/' + tag + sensory.accelerationX + '\n');
		response.write('sAccelerationY/' + tag + sensory.accelerationY + '\n');
		response.write('sAccelerationZ/' + tag + sensory.accelerationZ + '\n');
		response.write('sLight/' + tag + sensory.light + '\n');
		response.write('sTemperature/' + tag + sensory.temperature + '\n');
		response.write('sInputA/' + tag + sensory.inputA + '\n');
		response.write('sInputB/' + tag + sensory.inputB + '\n');
		response.write('sHandFound/' + tag + handFound + '\n');
		response.write('sTilt/' + index + BOOLEAN_TILT_FORWARD + (sensory.tilt == 1) + '\n');
		response.write('sTilt/' + index + BOOLEAN_TILT_BACKWARD + (sensory.tilt == -1) + '\n');
		response.write('sTilt/' + index + BOOLEAN_TILT_LEFT + (sensory.tilt == 2) + '\n');
		response.write('sTilt/' + index + BOOLEAN_TILT_RIGHT + (sensory.tilt == -2) + '\n');
		response.write('sTilt/' + index + BOOLEAN_TILT_FLIP + (sensory.tilt == 3) + '\n');
		response.write('sTilt/' + index + BOOLEAN_TILT_NONE + (sensory.tilt == -3) + '\n');
		response.write('sBattery/' + index + BOOLEAN_BATTERY_NORMAL + (sensory.batteryState == 2) + '\n');
		response.write('sBattery/' + index + BOOLEAN_BATTERY_LOW + (sensory.batteryState == 1) + '\n');
		response.write('sBattery/' + index + BOOLEAN_BATTERY_EMPTY + (sensory.batteryState == 0) + '\n');
	}
	if(this.received == false) this.active = false;
};

Robot.prototype.boardMoveForward = function(request) {
	var motoring = this.motoring;
	this.cancelWheel();
	this.cancelLineTracer();
	
	motoring.leftWheel.write(45);
	motoring.rightWheel.write(45);
	this.boardCommand = 1;
	this.boardCount = 0;
	this.boardState = 1;
	this.boardId = parseInt(request.params.id);
	this.waits.push(this.boardId);
	motoring.lineTracerMode.write(0);
};

Robot.prototype.boardTurn = function(request) {
	var motoring = this.motoring;
	this.cancelWheel();
	this.cancelLineTracer();
	
	if(DIRECTIONS[request.params.direction] === LEFT) {
		motoring.leftWheel.write(-45);
		motoring.rightWheel.write(45);
		this.boardCommand = 2;
	} else {
		motoring.leftWheel.write(45);
		motoring.rightWheel.write(-45);
		this.boardCommand = 3;
	}
	this.boardCount = 0;
	this.boardState = 1;
	this.boardId = parseInt(request.params.id);
	this.waits.push(this.boardId);
	motoring.lineTracerMode.write(0);
};

Robot.prototype.moveForward = function(request) {
	var self = this;
	var motoring = self.motoring;
	self.cancelBoard();
	self.cancelWheel();
	self.cancelLineTracer();

	var speed = context.getMovementSpeed(MODULE);
	var id = self.wheelId = parseInt(request.params.id);
	self.waits.push(id);
	motoring.leftWheel.write(speed);
	motoring.rightWheel.write(speed);
	motoring.lineTracerMode.write(0);
	self.wheelTimer = setTimeout(function() {
		if(self.wheelId == id) {
			motoring.leftWheel.write(0);
			motoring.rightWheel.write(0);
			self.removeWait(self.wheelId);
			self.cancelWheel();
		}
	}, 1000);
	self.timeouts.push(self.wheelTimer);
};

Robot.prototype.moveBackward = function(request) {
	var self = this;
	var motoring = self.motoring;
	self.cancelBoard();
	self.cancelWheel();
	self.cancelLineTracer();
	
	var speed = context.getMovementSpeed(MODULE);
	var id = self.wheelId = parseInt(request.params.id);
	self.waits.push(id);
	motoring.leftWheel.write(-speed);
	motoring.rightWheel.write(-speed);
	motoring.lineTracerMode.write(0);
	self.wheelTimer = setTimeout(function() {
		if(self.wheelId == id) {
			motoring.leftWheel.write(0);
			motoring.rightWheel.write(0);
			self.removeWait(self.wheelId);
			self.cancelWheel();
		}
	}, 1000);
	self.timeouts.push(self.wheelTimer);
};

Robot.prototype.turn = function(request) {
	var self = this;
	var motoring = self.motoring;
	self.cancelBoard();
	self.cancelWheel();
	self.cancelLineTracer();
	
	var speed = context.getTurningSpeed(MODULE);
	var id = self.wheelId = parseInt(request.params.id);
	self.waits.push(id);
	if(DIRECTIONS[request.params.direction] === LEFT) {
		motoring.leftWheel.write(-speed);
		motoring.rightWheel.write(speed);
	} else {
		motoring.leftWheel.write(speed);
		motoring.rightWheel.write(-speed);
	}
	motoring.lineTracerMode.write(0);
	self.wheelTimer = setTimeout(function() {
		if(self.wheelId == id) {
			motoring.leftWheel.write(0);
			motoring.rightWheel.write(0);
			self.removeWait(self.wheelId);
			self.cancelWheel();
		}
	}, 1000);
	self.timeouts.push(self.wheelTimer);
};

Robot.prototype.moveForwardForSecs = function(request) {
	var self = this;
	var motoring = self.motoring;
	self.cancelBoard();
	self.cancelWheel();
	self.cancelLineTracer();
	
	var speed = context.getMovementSpeed(MODULE);
	var sec = parseFloat(request.params.sec);
	if(sec) {
		var id = self.wheelId = parseInt(request.params.id);
		self.waits.push(id);
		if(sec < 0) {
			sec = -sec;
			speed = -speed;
		}
		motoring.leftWheel.write(speed);
		motoring.rightWheel.write(speed);
		motoring.lineTracerMode.write(0);
		self.wheelTimer = setTimeout(function() {
			if(self.wheelId == id) {
				motoring.leftWheel.write(0);
				motoring.rightWheel.write(0);
				self.removeWait(self.wheelId);
				self.cancelWheel();
			}
		}, sec * 1000);
		self.timeouts.push(self.wheelTimer);
	} else {
		motoring.leftWheel.write(0);
		motoring.rightWheel.write(0);
		motoring.lineTracerMode.write(0);
	}
};

Robot.prototype.moveBackwardForSecs = function(request) {
	var self = this;
	var motoring = self.motoring;
	self.cancelBoard();
	self.cancelWheel();
	self.cancelLineTracer();

	var speed = context.getMovementSpeed(MODULE);
	var sec = parseFloat(request.params.sec);
	if(sec) {
		var id = self.wheelId = parseInt(request.params.id);
		self.waits.push(id);
		if(sec < 0) {
			sec = -sec;
			speed = -speed;
		}
		motoring.leftWheel.write(-speed);
		motoring.rightWheel.write(-speed);
		motoring.lineTracerMode.write(0);
		self.wheelTimer = setTimeout(function() {
			if(self.wheelId == id) {
				motoring.leftWheel.write(0);
				motoring.rightWheel.write(0);
				self.removeWait(self.wheelId);
				self.cancelWheel();
			}
		}, sec * 1000);
		self.timeouts.push(self.wheelTimer);
	} else {
		motoring.leftWheel.write(0);
		motoring.rightWheel.write(0);
		motoring.lineTracerMode.write(0);
	}
};

Robot.prototype.turnForSecs = function(request) {
	var self = this;
	var motoring = self.motoring;
	self.cancelBoard();
	self.cancelWheel();
	self.cancelLineTracer();
	
	var speed = context.getTurningSpeed(MODULE);
	var sec = parseFloat(request.params.sec);
	if(sec) {
		var id = self.wheelId = parseInt(request.params.id);
		self.waits.push(id);
		if(sec < 0) {
			sec = -sec;
			speed = -speed;
		}
		if(DIRECTIONS[request.params.direction] === LEFT) {
			motoring.leftWheel.write(-speed);
			motoring.rightWheel.write(speed);
		} else {
			motoring.leftWheel.write(speed);
			motoring.rightWheel.write(-speed);
		}
		motoring.lineTracerMode.write(0);
		self.wheelTimer = setTimeout(function() {
			if(self.wheelId == id) {
				motoring.leftWheel.write(0);
				motoring.rightWheel.write(0);
				self.removeWait(self.wheelId);
				self.cancelWheel();
			}
		}, sec * 1000);
		self.timeouts.push(self.wheelTimer);
	} else {
		motoring.leftWheel.write(0);
		motoring.rightWheel.write(0);
		motoring.lineTracerMode.write(0);
	}
};

Robot.prototype.moveForwardUnit = function(request) {
	var self = this;
	var motoring = self.motoring;
	self.cancelBoard();
	self.cancelWheel();
	self.cancelLineTracer();
	
	if(UNITS[request.params.unit] === SECONDS) {
		var speed = context.getMovementSpeed(MODULE);
		var sec = parseFloat(request.params.value);
		if(sec) {
			var id = self.wheelId = parseInt(request.params.id);
			self.waits.push(id);
			if(sec < 0) {
				sec = -sec;
				speed = -speed;
			}
			motoring.leftWheel.write(speed);
			motoring.rightWheel.write(speed);
			motoring.lineTracerMode.write(0);
			self.wheelTimer = setTimeout(function() {
				if(self.wheelId == id) {
					motoring.leftWheel.write(0);
					motoring.rightWheel.write(0);
					self.removeWait(self.wheelId);
					self.cancelWheel();
				}
			}, sec * 1000);
			self.timeouts.push(self.wheelTimer);
		} else {
			motoring.leftWheel.write(0);
			motoring.rightWheel.write(0);
			motoring.lineTracerMode.write(0);
		}
	} else {
		self.wheelId = parseInt(request.params.id);
		self.waits.push(self.wheelId);
		motoring.leftWheel.write(0);
		motoring.rightWheel.write(0);
		motoring.lineTracerMode.write(0);
	}
};

Robot.prototype.moveBackwardUnit = function(request) {
	var self = this;
	var motoring = self.motoring;
	self.cancelBoard();
	self.cancelWheel();
	self.cancelLineTracer();

	if(UNITS[request.params.unit] === SECONDS) {
		var speed = context.getMovementSpeed(MODULE);
		var sec = parseFloat(request.params.value);
		if(sec) {
			var id = self.wheelId = parseInt(request.params.id);
			self.waits.push(id);
			if(sec < 0) {
				sec = -sec;
				speed = -speed;
			}
			motoring.leftWheel.write(-speed);
			motoring.rightWheel.write(-speed);
			motoring.lineTracerMode.write(0);
			self.wheelTimer = setTimeout(function() {
				if(self.wheelId == id) {
					motoring.leftWheel.write(0);
					motoring.rightWheel.write(0);
					self.removeWait(self.wheelId);
					self.cancelWheel();
				}
			}, sec * 1000);
			self.timeouts.push(self.wheelTimer);
		} else {
			motoring.leftWheel.write(0);
			motoring.rightWheel.write(0);
			motoring.lineTracerMode.write(0);
		}
	} else {
		self.wheelId = parseInt(request.params.id);
		self.waits.push(self.wheelId);
		motoring.leftWheel.write(0);
		motoring.rightWheel.write(0);
		motoring.lineTracerMode.write(0);
	}
};

Robot.prototype.turnUnitInPlace = function(request) {
	var self = this;
	var motoring = self.motoring;
	self.cancelBoard();
	self.cancelWheel();
	self.cancelLineTracer();
	
	if(UNITS[request.params.unit] === SECONDS) {
		var speed = context.getTurningSpeed(MODULE);
		var sec = parseFloat(request.params.value);
		if(sec) {
			var id = self.wheelId = parseInt(request.params.id);
			self.waits.push(id);
			if(sec < 0) {
				sec = -sec;
				speed = -speed;
			}
			if(DIRECTIONS[request.params.direction] === LEFT) {
				motoring.leftWheel.write(-speed);
				motoring.rightWheel.write(speed);
			} else {
				motoring.leftWheel.write(speed);
				motoring.rightWheel.write(-speed);
			}
			motoring.lineTracerMode.write(0);
			self.wheelTimer = setTimeout(function() {
				if(self.wheelId == id) {
					motoring.leftWheel.write(0);
					motoring.rightWheel.write(0);
					self.removeWait(self.wheelId);
					self.cancelWheel();
				}
			}, sec * 1000);
			self.timeouts.push(self.wheelTimer);
		} else {
			motoring.leftWheel.write(0);
			motoring.rightWheel.write(0);
			motoring.lineTracerMode.write(0);
		}
	} else {
		self.wheelId = parseInt(request.params.id);
		self.waits.push(self.wheelId);
		motoring.leftWheel.write(0);
		motoring.rightWheel.write(0);
		motoring.lineTracerMode.write(0);
	}
};

Robot.prototype.pivotAroundWheelUnitInDirection = function(request) {
	var self = this;
	var motoring = self.motoring;
	self.cancelBoard();
	self.cancelWheel();
	self.cancelLineTracer();
	
	if(UNITS[request.params.unit] === SECONDS) {
		var speed = context.getTurningSpeed(MODULE);
		var sec = parseFloat(request.params.value);
		if(sec) {
			var id = self.wheelId = parseInt(request.params.id);
			self.waits.push(id);
			if(sec < 0) {
				sec = -sec;
				speed = -speed;
			}
			if(PARTS[request.params.wheel] === LEFT) {
				motoring.leftWheel.write(0);
				if(TOWARDS[request.params.forward] === FORWARD) {
					motoring.rightWheel.write(speed);
				} else {
					motoring.rightWheel.write(-speed);
				}
			} else {
				if(TOWARDS[request.params.forward] === FORWARD) {
					motoring.leftWheel.write(speed);
				} else {
					motoring.leftWheel.write(-speed);
				}
				motoring.rightWheel.write(0);
			}
			motoring.lineTracerMode.write(0);
			self.wheelTimer = setTimeout(function() {
				if(self.wheelId == id) {
					motoring.leftWheel.write(0);
					motoring.rightWheel.write(0);
					self.removeWait(self.wheelId);
					self.cancelWheel();
				}
			}, sec * 1000);
			self.timeouts.push(self.wheelTimer);
		} else {
			motoring.leftWheel.write(0);
			motoring.rightWheel.write(0);
			motoring.lineTracerMode.write(0);
		}
	} else {
		self.wheelId = parseInt(request.params.id);
		self.waits.push(self.wheelId);
		motoring.leftWheel.write(0);
		motoring.rightWheel.write(0);
		motoring.lineTracerMode.write(0);
	}
};

Robot.prototype.calculateInnerSpeed = function(speed, radius) {
	return speed * (radius - Hamster.WHEEL_CENTER_DISTANCE) / (radius + Hamster.WHEEL_CENTER_DISTANCE);
};

Robot.prototype.turnBodyUnitWithRadiusInDirection = function(request) {
	var self = this;
	var motoring = self.motoring;
	self.cancelBoard();
	self.cancelWheel();
	self.cancelLineTracer();
	
	if(UNITS[request.params.unit] === SECONDS) {
		var speed = context.getTurningSpeed(MODULE);
		var sec = parseFloat(request.params.value);
		var radius = parseFloat(request.params.radius);
		if(sec && (typeof radius == 'number') && radius >= 0) {
			var id = self.wheelId = parseInt(request.params.id);
			self.waits.push(id);
			if(sec < 0) {
				sec = -sec;
				speed = -speed;
			}
			if(DIRECTIONS[request.params.direction] === LEFT) {
				if(TOWARDS[request.params.forward] === FORWARD) {
					motoring.leftWheel.write(self.calculateInnerSpeed(speed, radius));
					motoring.rightWheel.write(speed);
				} else {
					motoring.leftWheel.write(-self.calculateInnerSpeed(speed, radius));
					motoring.rightWheel.write(-speed);
				}
			} else {
				if(TOWARDS[request.params.forward] === FORWARD) {
					motoring.leftWheel.write(speed);
					motoring.rightWheel.write(self.calculateInnerSpeed(speed, radius));
				} else {
					motoring.leftWheel.write(-speed);
					motoring.rightWheel.write(-self.calculateInnerSpeed(speed, radius));
				}
			}
			motoring.lineTracerMode.write(0);
			self.wheelTimer = setTimeout(function() {
				if(self.wheelId == id) {
					motoring.leftWheel.write(0);
					motoring.rightWheel.write(0);
					self.removeWait(self.wheelId);
					self.cancelWheel();
				}
			}, sec * 1000);
			self.timeouts.push(self.wheelTimer);
		} else {
			motoring.leftWheel.write(0);
			motoring.rightWheel.write(0);
			motoring.lineTracerMode.write(0);
		}
	} else {
		self.wheelId = parseInt(request.params.id);
		self.waits.push(self.wheelId);
		motoring.leftWheel.write(0);
		motoring.rightWheel.write(0);
		motoring.lineTracerMode.write(0);
	}
};

Robot.prototype.pivotAroundUnitInDirection = function(request) {
	var self = this;
	var motoring = self.motoring;
	self.cancelBoard();
	self.cancelWheel();
	self.cancelLineTracer();
	
	if(UNITS[request.params.unit] === SECONDS) {
		var speed = context.getTurningSpeed(MODULE);
		var sec = parseFloat(request.params.value);
		if(sec) {
			var id = self.wheelId = parseInt(request.params.id);
			self.waits.push(id);
			if(sec < 0) {
				sec = -sec;
				speed = -speed;
			}
			var part = PARTS[request.params.part];
			if(part === LEFT_PEN) {
				if(TOWARDS[request.params.forward] === FORWARD) {
					motoring.leftWheel.write(self.calculateInnerSpeed(speed, Hamster.PEN_CENTER_DISTANCE));
					motoring.rightWheel.write(speed);
				} else {
					motoring.leftWheel.write(-self.calculateInnerSpeed(speed, Hamster.PEN_CENTER_DISTANCE));
					motoring.rightWheel.write(-speed);
				}
			} else if(part === RIGHT_PEN) {
				if(TOWARDS[request.params.forward] === FORWARD) {
					motoring.leftWheel.write(speed);
					motoring.rightWheel.write(self.calculateInnerSpeed(speed, Hamster.PEN_CENTER_DISTANCE));
				} else {
					motoring.leftWheel.write(-speed);
					motoring.rightWheel.write(-self.calculateInnerSpeed(speed, Hamster.PEN_CENTER_DISTANCE));
				}
			} else if(part === LEFT_WHEEL) {
				motoring.leftWheel.write(0);
				if(TOWARDS[request.params.forward] === FORWARD) {
					motoring.rightWheel.write(speed);
				} else {
					motoring.rightWheel.write(-speed);
				}
			} else {
				if(TOWARDS[request.params.forward] === FORWARD) {
					motoring.leftWheel.write(speed);
				} else {
					motoring.leftWheel.write(-speed);
				}
				motoring.rightWheel.write(0);
			}
			motoring.lineTracerMode.write(0);
			self.wheelTimer = setTimeout(function() {
				if(self.wheelId == id) {
					motoring.leftWheel.write(0);
					motoring.rightWheel.write(0);
					self.removeWait(self.wheelId);
					self.cancelWheel();
				}
			}, sec * 1000);
			self.timeouts.push(self.wheelTimer);
		} else {
			motoring.leftWheel.write(0);
			motoring.rightWheel.write(0);
			motoring.lineTracerMode.write(0);
		}
	} else {
		self.wheelId = parseInt(request.params.id);
		self.waits.push(self.wheelId);
		motoring.leftWheel.write(0);
		motoring.rightWheel.write(0);
		motoring.lineTracerMode.write(0);
	}
};

Robot.prototype.turnUnitWithRadiusInDirection = function(request) {
	var self = this;
	var motoring = self.motoring;
	self.cancelBoard();
	self.cancelWheel();
	self.cancelLineTracer();
	
	if(UNITS[request.params.unit] === SECONDS) {
		var speed = context.getTurningSpeed(MODULE);
		var sec = parseFloat(request.params.value);
		var radius = parseFloat(request.params.radius);
		if(sec && (typeof radius == 'number') && radius >= 0) {
			var id = self.wheelId = parseInt(request.params.id);
			self.waits.push(id);
			if(sec < 0) {
				sec = -sec;
				speed = -speed;
			}
			var part = PARTS[request.params.part];
			if(part === LEFT_PEN) {
				if(DIRECTIONS[request.params.direction] === LEFT) {
					if(TOWARDS[request.params.forward] === FORWARD) {
						motoring.leftWheel.write(self.calculateInnerSpeed(speed, radius + Hamster.PEN_CENTER_DISTANCE));
						motoring.rightWheel.write(speed);
					} else {
						motoring.leftWheel.write(-self.calculateInnerSpeed(speed, radius + Hamster.PEN_CENTER_DISTANCE));
						motoring.rightWheel.write(-speed);
					}
				} else {
					if(TOWARDS[request.params.forward] === FORWARD) {
						if(radius >= Hamster.PEN_CENTER_DISTANCE) {
							motoring.leftWheel.write(speed);
							motoring.rightWheel.write(self.calculateInnerSpeed(speed, radius - Hamster.PEN_CENTER_DISTANCE));
						} else {
							motoring.leftWheel.write(self.calculateInnerSpeed(speed, Hamster.PEN_CENTER_DISTANCE - radius));
							motoring.rightWheel.write(speed);
						}
					} else {
						if(radius >= Hamster.PEN_CENTER_DISTANCE) {
							motoring.leftWheel.write(-speed);
							motoring.rightWheel.write(-self.calculateInnerSpeed(speed, radius - Hamster.PEN_CENTER_DISTANCE));
						} else {
							motoring.leftWheel.write(-self.calculateInnerSpeed(speed, Hamster.PEN_CENTER_DISTANCE - radius));
							motoring.rightWheel.write(-speed);
						}
					}
				}
			} else if(part === RIGHT_PEN) {
				if(DIRECTIONS[request.params.direction] === LEFT) {
					if(TOWARDS[request.params.forward] === FORWARD) {
						if(radius >= Hamster.PEN_CENTER_DISTANCE) {
							motoring.leftWheel.write(self.calculateInnerSpeed(speed, radius - Hamster.PEN_CENTER_DISTANCE));
							motoring.rightWheel.write(speed);
						} else {
							motoring.leftWheel.write(speed);
							motoring.rightWheel.write(self.calculateInnerSpeed(speed, Hamster.PEN_CENTER_DISTANCE - radius));
						}
					} else {
						if(radius >= Hamster.PEN_CENTER_DISTANCE) {
							motoring.leftWheel.write(-self.calculateInnerSpeed(speed, radius - Hamster.PEN_CENTER_DISTANCE));
							motoring.rightWheel.write(-speed);
						} else {
							motoring.leftWheel.write(-speed);
							motoring.rightWheel.write(-self.calculateInnerSpeed(speed, Hamster.PEN_CENTER_DISTANCE - radius));
						}
					}
				} else {
					if(TOWARDS[request.params.forward] === FORWARD) {
						motoring.leftWheel.write(speed);
						motoring.rightWheel.write(self.calculateInnerSpeed(speed, radius + Hamster.PEN_CENTER_DISTANCE));
					} else {
						motoring.leftWheel.write(-speed);
						motoring.rightWheel.write(-self.calculateInnerSpeed(speed, radius + Hamster.PEN_CENTER_DISTANCE));
					}
				}
			} else {
				if(DIRECTIONS[request.params.direction] === LEFT) {
					if(TOWARDS[request.params.forward] === FORWARD) {
						motoring.leftWheel.write(self.calculateInnerSpeed(speed, radius));
						motoring.rightWheel.write(speed);
					} else {
						motoring.leftWheel.write(-self.calculateInnerSpeed(speed, radius));
						motoring.rightWheel.write(-speed);
					}
				} else {
					if(TOWARDS[request.params.forward] === FORWARD) {
						motoring.leftWheel.write(speed);
						motoring.rightWheel.write(self.calculateInnerSpeed(speed, radius));
					} else {
						motoring.leftWheel.write(-speed);
						motoring.rightWheel.write(-self.calculateInnerSpeed(speed, radius));
					}
				}
			}
			motoring.lineTracerMode.write(0);
			self.wheelTimer = setTimeout(function() {
				if(self.wheelId == id) {
					motoring.leftWheel.write(0);
					motoring.rightWheel.write(0);
					self.removeWait(self.wheelId);
					self.cancelWheel();
				}
			}, sec * 1000);
			self.timeouts.push(self.wheelTimer);
		} else {
			motoring.leftWheel.write(0);
			motoring.rightWheel.write(0);
			motoring.lineTracerMode.write(0);
		}
	} else {
		self.wheelId = parseInt(request.params.id);
		self.waits.push(self.wheelId);
		motoring.leftWheel.write(0);
		motoring.rightWheel.write(0);
		motoring.lineTracerMode.write(0);
	}
};

Robot.prototype.changeBothWheelsBy = function(request) {
	var motoring = this.motoring;
	this.cancelBoard();
	this.cancelWheel();
	this.cancelLineTracer();
	
	var left = parseFloat(request.params.left);
	var right = parseFloat(request.params.right);
	if(typeof left == 'number') {
		left += motoring.leftWheel.read();
		motoring.leftWheel.write(left);
	}
	if(typeof right == 'number') {
		right += motoring.rightWheel.read();
		motoring.rightWheel.write(right);
	}
	motoring.lineTracerMode.write(0);
};

Robot.prototype.setBothWheelsTo = function(request) {
	var motoring = this.motoring;
	this.cancelBoard();
	this.cancelWheel();
	this.cancelLineTracer();
	
	var left = parseFloat(request.params.left);
	var right = parseFloat(request.params.right);
	if(typeof left == 'number') {
		motoring.leftWheel.write(left);
	}
	if(typeof right == 'number') {
		motoring.rightWheel.write(right);
	}
	motoring.lineTracerMode.write(0);
};

Robot.prototype.changeWheelBy = function(request) {
	var motoring = this.motoring;
	this.cancelBoard();
	this.cancelWheel();
	this.cancelLineTracer();
	
	var speed = parseFloat(request.params.speed);
	if(typeof speed == 'number') {
		var which = PARTS[request.params.which];
		if(which === LEFT) {
			speed += motoring.leftWheel.read();
			motoring.leftWheel.write(speed);
		}
		else if(which === RIGHT) {
			speed += motoring.rightWheel.read();
			motoring.rightWheel.write(speed);
		}
		else {
			var left = speed + motoring.leftWheel.read();
			var right = speed + motoring.rightWheel.read();
			motoring.leftWheel.write(left);
			motoring.rightWheel.write(right);
		}
	}
	motoring.lineTracerMode.write(0);
};

Robot.prototype.setWheelTo = function(request) {
	var motoring = this.motoring;
	this.cancelBoard();
	this.cancelWheel();
	this.cancelLineTracer();
	
	var speed = parseFloat(request.params.speed);
	if(typeof speed == 'number') {
		var which = PARTS[request.params.which];
		if(which === LEFT) {
			motoring.leftWheel.write(speed);
		} else if(which === RIGHT) {
			motoring.rightWheel.write(speed);
		} else {
			motoring.leftWheel.write(speed);
			motoring.rightWheel.write(speed);
		}
	}
	motoring.lineTracerMode.write(0);
};

Robot.prototype.followLineUsingFloorSensor = function(request) {
	var motoring = this.motoring;
	this.cancelBoard();
	this.cancelWheel();
	this.cancelLineTracer();
	
	var mode = 1;
	var which = PARTS[request.params.which];
	if(which === RIGHT)
		mode = 2;
	else if(which === BOTH)
		mode = 3;
	if(VALUES[request.params.color] === WHITE)
		mode += 7;
	
	motoring.leftWheel.write(0);
	motoring.rightWheel.write(0);
	motoring.lineTracerMode.write(mode);
};

Robot.prototype.followLineUntilIntersection = function(request) {
	var motoring = this.motoring;
	this.cancelBoard();
	this.cancelWheel();
	
	var mode = 4;
	var which = DIRECTIONS[request.params.which];
	if(which === RIGHT)
		mode = 5;
	else if(which === FRONT)
		mode = 6;
	else if(which === REAR)
		mode = 7;
	if(VALUES[request.params.color] === WHITE)
		mode += 7;
	
	motoring.leftWheel.write(0);
	motoring.rightWheel.write(0);
	this.lineTracerId = parseInt(request.params.id);
	this.waits.push(this.lineTracerId);
	motoring.lineTracerMode.write(mode);
};

Robot.prototype.setFollowingSpeedTo = function(request) {
	var motoring = this.motoring;
	var speed = parseInt(request.params.speed);
	if(typeof speed == 'number') {
		motoring.lineTracerSpeed.write(speed);
	}
};

Robot.prototype.setFollowingGainTo = function(request) {
};

Robot.prototype.stop = function(request) {
	var motoring = this.motoring;
	this.cancelBoard();
	this.cancelWheel();
	this.cancelLineTracer();
	
	motoring.leftWheel.write(0);
	motoring.rightWheel.write(0);
	motoring.lineTracerMode.write(0);
};

Robot.prototype.setLedTo = function(request) {
	var motoring = this.motoring;
	var which = PARTS[request.params.which];
	var color = COLORS[request.params.color];
	if(color && color > 0) {
		if(which === LEFT) {
			motoring.leftLed.write(color);
		} else if(which === RIGHT) {
			motoring.rightLed.write(color);
		} else {
			motoring.leftLed.write(color);
			motoring.rightLed.write(color);
		}
	}
};

Robot.prototype.changeLedByRGB = function(request) {
};

Robot.prototype.setLedToRGB = function(request) {
};

Robot.prototype.clearLed = function(request) {
	var motoring = this.motoring;
	var which = PARTS[request.params.which];
	if(which === LEFT) {
		motoring.leftLed.write(0);
	} else if(which === RIGHT) {
		motoring.rightLed.write(0);
	} else {
		motoring.leftLed.write(0);
		motoring.rightLed.write(0);
	}
};

Robot.prototype.beep = function(request) {
	this.cancelNote();
	
	var id = this.noteId = parseInt(request.params.id);
	this.waits.push(id);
	this.runBeep(1, id);
};

Robot.prototype.runBeep = function(count, id) {
	if(count) {
		var self = this;
		var motoring = self.motoring;
		motoring.buzzer.write(440);
		motoring.note.write(0);
		self.noteTimer1 = setTimeout(function() {
			if(!id || self.noteId == id) {
				motoring.buzzer.write(0);
				if(self.noteTimer1 !== undefined) self.removeTimeout(self.noteTimer1);
				self.noteTimer1 = undefined;
			}
		}, 100);
		self.timeouts.push(self.noteTimer1);
		self.noteTimer2 = setTimeout(function() {
			if(!id || self.noteId == id) {
				motoring.buzzer.write(0);
				if(self.noteTimer2 !== undefined) self.removeTimeout(self.noteTimer2);
				self.noteTimer2 = undefined;
				if(count < 0) {
					self.runBeep(-1, id);
				} else if(count == 1) {
					if(id) self.removeWait(self.noteId);
					self.cancelNote();
				} else {
					self.runBeep(count - 1, id);
				}
			}
		}, 200);
		self.timeouts.push(self.noteTimer2);
	}
};

Robot.prototype.playSound = function(request) {
	var motoring = this.motoring;
	this.cancelNote();
	
	var sound = SOUNDS[request.params.sound];
	motoring.buzzer.write(0);
	motoring.note.write(0);
	if(sound) {
		this.runBeep(1);
	}
};

Robot.prototype.playSoundTimes = function(request) {
	var motoring = this.motoring;
	this.cancelNote();
	
	var sound = SOUNDS[request.params.sound];
	var count = parseInt(request.params.count);
	motoring.buzzer.write(0);
	motoring.note.write(0);
	if(sound && count) {
		this.runBeep(count);
	}
};

Robot.prototype.playSoundTimesUntilDone = function(request) {
	var motoring = this.motoring;
	this.cancelNote();
	
	var sound = SOUNDS[request.params.sound];
	var count = parseInt(request.params.count);
	motoring.buzzer.write(0);
	motoring.note.write(0);
	if(count) {
		var id = this.noteId = parseInt(request.params.id);
		this.waits.push(id);
		if(sound) {
			this.runBeep(count, id);
		}
	}
};

Robot.prototype.changeBuzzerBy = function(request) {
	var motoring = this.motoring;
	this.cancelNote();
	
	var buzzer = parseFloat(request.params.value);
	if(typeof buzzer == 'number') {
		buzzer += motoring.buzzer.read();
		motoring.buzzer.write(buzzer);
	}
	motoring.note.write(0);
};

Robot.prototype.setBuzzerTo = function(request) {
	var motoring = this.motoring;
	this.cancelNote();
	
	var buzzer = parseFloat(request.params.value);
	if(typeof buzzer == 'number') {
		motoring.buzzer.write(buzzer);
	}
	motoring.note.write(0);
};

Robot.prototype.clearBuzzer = function(request) {
	var motoring = this.motoring;
	this.cancelNote();
	
	motoring.buzzer.write(0);
	motoring.note.write(0);
};

Robot.prototype.clearSound = function(request) {
	var motoring = this.motoring;
	this.cancelNote();
	
	motoring.buzzer.write(0);
	motoring.note.write(0);
};

Robot.prototype.playNote = function(request) {
	var motoring = this.motoring;
	this.cancelNote();
	
	var note = NOTES[request.params.note];
	var octave = parseInt(request.params.octave);
	motoring.buzzer.write(0);
	if(note && octave && octave > 0 && octave < 8) {
		note += (octave - 1) * 12;
		motoring.note.write(note);
	} else {
		motoring.note.write(0);
	}
};

Robot.prototype.playNoteFor = function(request) {
	var self = this;
	var motoring = self.motoring;
	self.cancelNote();
	
	var note = NOTES[request.params.note];
	var octave = parseInt(request.params.octave);
	var beat = BEATS[request.params.beat];
	if(!beat) {
		beat = parseFloat(request.params.beat);
	}
	motoring.buzzer.write(0);
	if(note && octave && octave > 0 && octave < 8 && beat && beat > 0 && self.tempo > 0) {
		var id = self.noteId = parseInt(request.params.id);
		self.waits.push(id);
		note += (octave - 1) * 12;
		motoring.note.write(note);
		var timeout = beat * 60 * 1000 / self.tempo;
		var tail = 0;
		if(timeout > 100) {
			tail = 100;
		}
		if(tail > 0) {
			self.noteTimer1 = setTimeout(function() {
				if(self.noteId == id) {
					motoring.note.write(0);
					if(self.noteTimer1 !== undefined) self.removeTimeout(self.noteTimer1);
					self.noteTimer1 = undefined;
				}
			}, timeout - tail);
			self.timeouts.push(self.noteTimer1);
		}
		self.noteTimer2 = setTimeout(function() {
			if(self.noteId == id) {
				motoring.note.write(0);
				self.removeWait(self.noteId);
				self.cancelNote();
			}
		}, timeout);
		self.timeouts.push(self.noteTimer2);
	} else {
		motoring.note.write(0);
	}
};

Robot.prototype.restFor = function(request) {
	var self = this;
	var motoring = self.motoring;
	self.cancelNote();
	
	var beat = BEATS[request.params.beat];
	if(!beat) {
		beat = parseFloat(request.params.beat);
	}
	motoring.buzzer.write(0);
	motoring.note.write(0);
	if(beat && beat > 0 && self.tempo > 0) {
		var id = self.noteId = parseInt(request.params.id);
		self.waits.push(id);
		self.noteTimer1 = setTimeout(function() {
			if(self.noteId == id) {
				self.removeWait(self.noteId);
				self.cancelNote();
			}
		}, beat * 60 * 1000 / self.tempo);
		self.timeouts.push(self.noteTimer1);
	}
};

Robot.prototype.changeTempoBy = function(request) {
	var value = parseFloat(request.params.value);
	if(typeof value == 'number') {
		this.tempo += value;
		if(this.tempo < 1) this.tempo = 1;
	}
};

Robot.prototype.setTempoTo = function(request) {
	var value = parseFloat(request.params.value);
	if(typeof value == 'number') {
		this.tempo = value;
		if(this.tempo < 1) this.tempo = 1;
	}
};

Robot.prototype.setPortTo = function(request) {
	var motoring = this.motoring;
	this.cancelIo();
	var mode = IO_MODES[request.params.value];
	if(typeof mode == 'number') {
		if(request.params.which == 'A') {
			motoring.ioModeA.write(mode);
		} else if(request.params.which == 'B') {
			motoring.ioModeB.write(mode);
		} else {
			motoring.ioModeA.write(mode);
			motoring.ioModeB.write(mode);
		}
	}
};

Robot.prototype.changeOutputBy = function(request) {
	var motoring = this.motoring;
	this.cancelIo();
	var value = parseFloat(request.params.value);
	if(typeof value == 'number') {
		if(request.params.which == 'A') {
			value += motoring.outputA.read();
			motoring.outputA.write(value);
		} else if(request.params.which == 'B') {
			value += motoring.outputB.read();
			motoring.outputB.write(value);
		} else {
			var valueA = value + motoring.outputA.read();
			var valueB = value + motoring.outputB.read();
			motoring.outputA.write(valueA);
			motoring.outputB.write(valueB);
		}
	}
};

Robot.prototype.setOutputTo = function(request) {
	var motoring = this.motoring;
	this.cancelIo();
	var value = parseFloat(request.params.value);
	if(typeof value == 'number') {
		if(request.params.which == 'A') {
			motoring.outputA.write(value);
		} else if(request.params.which == 'B') {
			motoring.outputB.write(value);
		} else {
			motoring.outputA.write(value);
			motoring.outputB.write(value);
		}
	}
};

Robot.prototype.gripper = function(request) {
	var self = this;
	var motoring = self.motoring;
	self.cancelIo();
	
	var action = GRIPPERS[request.params.action];
	var id = self.ioId = parseInt(request.params.id);
	self.waits.push(id);
	
	motoring.ioModeA.write(Hamster.IO_MODE_DIGITAL_OUTPUT);
	motoring.ioModeB.write(Hamster.IO_MODE_DIGITAL_OUTPUT);
	if(action == OPEN) {
		motoring.outputA.write(1);
		motoring.outputB.write(0);
	} else {
		motoring.outputA.write(0);
		motoring.outputB.write(1);
	}
	self.ioTimer = setTimeout(function() {
		if(self.ioId == id) {
			self.removeWait(self.ioId);
			self.cancelIo();
		}
	}, 500);
	self.timeouts.push(self.ioTimer);
};

Robot.prototype.releaseGripper = function(request) {
	var motoring = this.motoring;
	this.cancelIo();
	motoring.ioModeA.write(Hamster.IO_MODE_DIGITAL_OUTPUT);
	motoring.ioModeB.write(Hamster.IO_MODE_DIGITAL_OUTPUT);
	motoring.outputA.write(0);
	motoring.outputB.write(0);
};

Robot.prototype.writeSerial = function(request) {
	this.cancelIo();
	this.ioId = parseInt(request.params.id);
	this.waits.push(this.ioId);
};

Robot.prototype.readSerialUntil = function(request) {
	this.cancelIo();
	this.ioId = parseInt(request.params.id);
	this.waits.push(this.ioId);
};

Robot.prototype.setSerialRateTo = function(request) {
	this.cancelIo();
};

Robot.prototype.handleSensory = function(data) {
	var self = this;
	self.received = true;
	self.active = true;
	var sensory = self.sensory;
	sensory.signalStrength = data.signalStrength;
	sensory.leftProximity = data.leftProximity;
	sensory.rightProximity = data.rightProximity;
	sensory.leftFloor = data.leftFloor;
	sensory.rightFloor = data.rightFloor;
	sensory.accelerationX = data.acceleration.x;
	sensory.accelerationY = data.acceleration.y;
	sensory.accelerationZ = data.acceleration.z;
	sensory.light = data.light;
	sensory.temperature = data.temperature;
	sensory.inputA = data.inputA;
	sensory.inputB = data.inputB;
	
	if(data.map & 0x00000008) {
		sensory.tilt = data.tilt;
	}
	if(data.map & 0x00000004) {
		sensory.batteryState = data.batteryState;
	}
	
	var motoring = self.motoring;
	if(self.lineTracerId >= 0 && (data.map & 0x00000010) != 0) {
		if(data.lineTracerState == 0x40) {
			motoring.lineTracerMode.write(0);
			self.removeWait(self.lineTracerId);
			self.cancelLineTracer();
		}
	}
	if(self.boardId >= 0) {
		if(self.boardCommand == 1) {
			switch(self.boardState) {
				case 1: {
					if(self.boardCount < 2) {
						if(data.leftFloor < 50 && data.rightFloor < 50)
							self.boardCount ++;
						else
							self.boardCount = 0;
						var diff = data.leftFloor - data.rightFloor;
						motoring.leftWheel.write(45 + diff * 0.25);
						motoring.rightWheel.write(45 - diff * 0.25);
					} else {
						self.boardCount = 0;
						self.boardState = 2;
					}
					break;
				}
				case 2: {
					var diff = data.leftFloor - data.rightFloor;
					motoring.leftWheel.write(45 + diff * 0.25);
					motoring.rightWheel.write(45 - diff * 0.25);
					self.boardState = 3;
					self.wheelTimer = setTimeout(function() {
						motoring.leftWheel.write(0);
						motoring.rightWheel.write(0);
						self.boardState = 4;
						if(self.wheelTimer !== undefined) self.removeTimeout(self.wheelTimer);
						self.wheelTimer = undefined;
					}, 250);
					self.timeouts.push(self.wheelTimer);
					break;
				}
				case 3: {
					var diff = data.leftFloor - data.rightFloor;
					motoring.leftWheel.write(45 + diff * 0.25);
					motoring.rightWheel.write(45 - diff * 0.25);
					break;
				}
				case 4: {
					motoring.leftWheel.write(0);
					motoring.rightWheel.write(0);
					self.removeWait(self.boardId);
					self.cancelBoard();
					break;
				}
			}
		} else if(self.boardCommand == 2) {
			switch(self.boardState) {
				case 1: {
					if(self.boardCount < 2) {
						if(data.leftFloor > 50)
							self.boardCount ++;
					} else {
						self.boardCount = 0;
						self.boardState = 2;
					}
					break;
				}
				case 2: {
					if(data.leftFloor < 20) {
						self.boardState = 3;
					}
					break;
				}
				case 3: {
					if(self.boardCount < 2) {
						if(data.leftFloor < 20)
							self.boardCount ++;
					} else {
						self.boardCount = 0;
						self.boardState = 4;
					}
					break;
				}
				case 4: {
					if(data.leftFloor > 50) {
						self.boardState = 5;
					}
					break;
				}
				case 5: {
					var diff = data.leftFloor - data.rightFloor;
					if(diff > -15) {
						motoring.leftWheel.write(0);
						motoring.rightWheel.write(0);
						self.removeWait(self.boardId);
						self.cancelBoard();
					} else {
						motoring.leftWheel.write(diff * 0.5);
						motoring.rightWheel.write(-diff * 0.5);
					}
					break;
				}
			}
		} else if(self.boardCommand == 3) {
			switch(self.boardState) {
				case 1: {
					if(self.boardCount < 2) {
						if(data.rightFloor > 50)
							self.boardCount ++;
					} else {
						self.boardCount = 0;
						self.boardState = 2;
					}
					break;
				}
				case 2: {
					if(data.rightFloor < 20) {
						self.boardState = 3;
					}
					break;
				}
				case 3: {
					if(self.boardCount < 2) {
						if(data.rightFloor < 20)
							self.boardCount ++;
					} else {
						self.boardCount = 0;
						self.boardState = 4;
					}
					break;
				}
				case 4: {
					if(data.rightFloor > 50) {
						self.boardState = 5;
					}
					break;
				}
				case 5: {
					var diff = data.rightFloor - data.leftFloor;
					if(diff > -15) {
						motoring.leftWheel.write(0);
						motoring.rightWheel.write(0);
						self.removeWait(self.boardId);
						self.cancelBoard();
					} else {
						motoring.leftWheel.write(-diff * 0.5);
						motoring.rightWheel.write(diff * 0.5);
					}
					break;
				}
			}
		}
	}
};

function Component() {
	var setup = function(map, key, value) {
		var data = translations[key];
		if(data) {
			if(value === undefined) value = key;
			for(var i in data) {
				map[data[i]] = value;
			}
			map[key] = value;
		}
	};
	setup(PARTS, 'left', LEFT);
	setup(PARTS, 'right', RIGHT);
	setup(PARTS, 'both', BOTH);
	
	setup(PARTS, 'left pen', LEFT_PEN);
	setup(PARTS, 'right pen', RIGHT_PEN);
	setup(PARTS, 'left wheel', LEFT_WHEEL);
	setup(PARTS, 'right wheel', RIGHT_WHEEL);
	
	setup(DIRECTIONS, 'left', LEFT);
	setup(DIRECTIONS, 'right', RIGHT);
	setup(DIRECTIONS, 'front', FRONT);
	setup(DIRECTIONS, 'rear', REAR);
	
	setup(TOWARDS, 'forward', FORWARD);
	setup(UNITS, 'seconds', SECONDS);
	
	setup(COLORS, 'blue', Hamster.LED_BLUE);
	setup(COLORS, 'green', Hamster.LED_GREEN);
	setup(COLORS, 'cyan', Hamster.LED_CYAN);
	setup(COLORS, 'sky blue', Hamster.LED_CYAN);
	setup(COLORS, 'red', Hamster.LED_RED);
	setup(COLORS, 'magenta', Hamster.LED_MAGENTA);
	setup(COLORS, 'purple', Hamster.LED_MAGENTA);
	setup(COLORS, 'yellow', Hamster.LED_YELLOW);
	setup(COLORS, 'white', Hamster.LED_WHITE);
	setup(COLORS, 'black', Hamster.LED_OFF);
	setup(COLORS, 'orange', Hamster.LED_RED);
	setup(COLORS, 'violet', Hamster.LED_MAGENTA);
	
	setup(NOTES, 'C', Hamster.NOTE_C_1);
	setup(NOTES, 'C#', Hamster.NOTE_C_SHARP_1);
	setup(NOTES, 'C♯ (D♭)', Hamster.NOTE_C_SHARP_1);
	setup(NOTES, 'D', Hamster.NOTE_D_1);
	setup(NOTES, 'D♯ (E♭)', Hamster.NOTE_E_FLAT_1);
	setup(NOTES, 'Eb', Hamster.NOTE_E_FLAT_1);
	setup(NOTES, 'E', Hamster.NOTE_E_1);
	setup(NOTES, 'F', Hamster.NOTE_F_1);
	setup(NOTES, 'F#', Hamster.NOTE_F_SHARP_1);
	setup(NOTES, 'F♯ (G♭)', Hamster.NOTE_F_SHARP_1);
	setup(NOTES, 'G', Hamster.NOTE_G_1);
	setup(NOTES, 'G#', Hamster.NOTE_G_SHARP_1);
	setup(NOTES, 'G♯ (A♭)', Hamster.NOTE_G_SHARP_1);
	setup(NOTES, 'A', Hamster.NOTE_A_1);
	setup(NOTES, 'A♯ (B♭)', Hamster.NOTE_B_FLAT_1);
	setup(NOTES, 'Bb', Hamster.NOTE_B_FLAT_1);
	setup(NOTES, 'B', Hamster.NOTE_B_1);
	
	setup(SOUNDS, 'beep', 1);
	
	setup(IO_MODES, 'analog input', Hamster.IO_MODE_ANALOG_INPUT);
	setup(IO_MODES, 'digital input', Hamster.IO_MODE_DIGITAL_INPUT);
	setup(IO_MODES, 'servo output', Hamster.IO_MODE_SERVO_OUTPUT);
	setup(IO_MODES, 'pwm output', Hamster.IO_MODE_PWM_OUTPUT);
	setup(IO_MODES, 'digital output', Hamster.IO_MODE_DIGITAL_OUTPUT);
	
	setup(GRIPPERS, 'open', OPEN);
	setup(GRIPPERS, 'close', CLOSE);
	
	setup(VALUES, 'white', WHITE);
}

Component.prototype.updateLang = function(translator) {
	var lang = translator.getLanguage();
	
	var get = function(key) {
		var value = undefined;
		var data = translations[key];
		if(data) value = data[lang];
		return value || key;
	};
	
	BOOLEAN_TILT_FORWARD = '/' + get('tilt%20forward') + ' ';
	BOOLEAN_TILT_BACKWARD = '/' + get('tilt%20backward') + ' ';
	BOOLEAN_TILT_LEFT = '/' + get('tilt%20left') + ' ';
	BOOLEAN_TILT_RIGHT = '/' + get('tilt%20right') + ' ';
	BOOLEAN_TILT_FLIP = '/' + get('tilt%20flip') + ' ';
	BOOLEAN_TILT_NONE = '/' + get('not%20tilt') + ' ';
	BOOLEAN_BATTERY_NORMAL = '/' + get('normal') + ' ';
	BOOLEAN_BATTERY_LOW = '/' + get('low') + ' ';
	BOOLEAN_BATTERY_EMPTY = '/' + get('empty') + ' ';
};

Component.prototype.getModule = function() {
	return MODULE;
};

Component.prototype.createRobot = function(index) {
	return new Robot(index);
};

module.exports = new Component();