'use strict';
var router = require('router');
var translator = require('translator');
var component = require('./component');
const MODULE = 'hamster';

function Extension() {
	this.robots = {};
}

Extension.prototype.getRobotByIndex = function(index) {
	if(index >= 0) {
		var robot = this.robots[index];
		if(!robot) {
			robot = component.createRobot(index);
			if(robot) {
				this.robots[index] = robot;
			}
		}
		return robot;
	}
};

Extension.prototype.getRobotByRoute = function(route) {
	return this.getRobotByIndex(route.index);
};

Extension.prototype.start = function(express) {
	var self = this;
	var robots = self.robots;
	var problemMessage = '';
	
	self.running = true;
	for(var i in robots) {
		robots[i].clearWaits();
	}
	var setup = function() {
		problemMessage = '_problem ' + translator.translate('First robot is not connected.') + '\n';
		component.updateLang(translator);
	};
	translator.on('hamster-scratch-multi', function() {
		setup();
	});
	setup();
	
	if(express) {
		const HEADER = { 'Content-Type': 'text/plain' };
		
		express.get('/reset_all', function(request, response) {
			response.writeHead(200, HEADER);
			var robot;
			for(var i in robots) {
				robot = robots[i];
				robot.clearWaits();
				robot.clearTimeouts();
				robot.reset();
			}
			router.reset();
			response.end();
		});
		
		express.get('/poll', function(request, response) {
			if(!self.running) {
				response.set('Connection', 'close');
			}
			response.writeHead(200, HEADER);
			var waits, busy = '';
			for(var i in robots) {
				waits = robots[i].getWaits();
				if(waits.length > 0) {
					busy += ' ' + waits.join(' ');
				}
			}
			if(busy.length > 0) {
				response.write('_busy' + busy + '\n');
			}
			for(var i in robots) {
				robots[i].writeMultiSensory(response);
			}
			if(self.state != 'connected') {
				response.write(problemMessage);
			}
			response.end();
		});
		
		express.get('/boardMoveForward/:id/:index', function(request, response) {
			response.writeHead(200, HEADER);
			var robot = self.getRobotByIndex(parseInt(request.params.index));
			if(robot) {
				robot.boardMoveForward(request);
			}
			response.end();
		});
		
		express.get('/boardTurn/:id/:index/:direction', function(request, response) {
			response.writeHead(200, HEADER);
			var robot = self.getRobotByIndex(parseInt(request.params.index));
			if(robot) {
				robot.boardTurn(request);
			}
			response.end();
		});
		
		express.get('/moveForward/:id/:index', function(request, response) {
			response.writeHead(200, HEADER);
			var robot = self.getRobotByIndex(parseInt(request.params.index));
			if(robot) {
				robot.moveForward(request);
			}
			response.end();
		});
		
		express.get('/moveBackward/:id/:index', function(request, response) {
			response.writeHead(200, HEADER);
			var robot = self.getRobotByIndex(parseInt(request.params.index));
			if(robot) {
				robot.moveBackward(request);
			}
			response.end();
		});
		
		express.get('/turn/:id/:index/:direction', function(request, response) {
			response.writeHead(200, HEADER);
			var robot = self.getRobotByIndex(parseInt(request.params.index));
			if(robot) {
				robot.turn(request);
			}
			response.end();
		});
		
		express.get('/moveForwardForSecs/:id/:index/:sec', function(request, response) {
			response.writeHead(200, HEADER);
			var robot = self.getRobotByIndex(parseInt(request.params.index));
			if(robot) {
				robot.moveForwardForSecs(request);
			}
			response.end();
		});
		
		express.get('/moveBackwardForSecs/:id/:index/:sec', function(request, response) {
			response.writeHead(200, HEADER);
			var robot = self.getRobotByIndex(parseInt(request.params.index));
			if(robot) {
				robot.moveBackwardForSecs(request);
			}
			response.end();
		});
		
		express.get('/turnForSecs/:id/:index/:direction/:sec', function(request, response) {
			response.writeHead(200, HEADER);
			var robot = self.getRobotByIndex(parseInt(request.params.index));
			if(robot) {
				robot.turnForSecs(request);
			}
			response.end();
		});
		
		express.get('/changeBothWheelsBy/:index/:left/:right', function(request, response) {
			response.writeHead(200, HEADER);
			var robot = self.getRobotByIndex(parseInt(request.params.index));
			if(robot) {
				robot.changeBothWheelsBy(request);
			}
			response.end();
		});
		
		express.get('/setBothWheelsTo/:index/:left/:right', function(request, response) {
			response.writeHead(200, HEADER);
			var robot = self.getRobotByIndex(parseInt(request.params.index));
			if(robot) {
				robot.setBothWheelsTo(request);
			}
			response.end();
		});
		
		express.get('/changeWheelBy/:index/:which/:speed', function(request, response) {
			response.writeHead(200, HEADER);
			var robot = self.getRobotByIndex(parseInt(request.params.index));
			if(robot) {
				robot.changeWheelBy(request);
			}
			response.end();
		});
		
		express.get('/setWheelTo/:index/:which/:speed', function(request, response) {
			response.writeHead(200, HEADER);
			var robot = self.getRobotByIndex(parseInt(request.params.index));
			if(robot) {
				robot.setWheelTo(request);
			}
			response.end();
		});
		
		express.get('/followLineUsingFloorSensor/:index/:color/:which', function(request, response) {
			response.writeHead(200, HEADER);
			var robot = self.getRobotByIndex(parseInt(request.params.index));
			if(robot) {
				robot.followLineUsingFloorSensor(request);
			}
			response.end();
		});
		
		express.get('/followLineUntilIntersection/:id/:index/:color/:which', function(request, response) {
			response.writeHead(200, HEADER);
			var robot = self.getRobotByIndex(parseInt(request.params.index));
			if(robot) {
				robot.followLineUntilIntersection(request);
			}
			response.end();
		});
		
		express.get('/setFollowingSpeedTo/:index/:speed', function(request, response) {
			response.writeHead(200, HEADER);
			var robot = self.getRobotByIndex(parseInt(request.params.index));
			if(robot) {
				robot.setFollowingSpeedTo(request);
			}
			response.end();
		});
		
		express.get('/stop/:index', function(request, response) {
			response.writeHead(200, HEADER);
			var robot = self.getRobotByIndex(parseInt(request.params.index));
			if(robot) {
				robot.stop(request);
			}
			response.end();
		});
		
		express.get('/setLedTo/:index/:which/:color', function(request, response) {
			response.writeHead(200, HEADER);
			var robot = self.getRobotByIndex(parseInt(request.params.index));
			if(robot) {
				robot.setLedTo(request);
			}
			response.end();
		});
		
		express.get('/clearLed/:index/:which', function(request, response) {
			response.writeHead(200, HEADER);
			var robot = self.getRobotByIndex(parseInt(request.params.index));
			if(robot) {
				robot.clearLed(request);
			}
			response.end();
		});
		
		express.get('/beep/:id/:index', function(request, response) {
			response.writeHead(200, HEADER);
			var robot = self.getRobotByIndex(parseInt(request.params.index));
			if(robot) {
				robot.beep(request);
			}
			response.end();
		});
		
		express.get('/changeBuzzerBy/:index/:value', function(request, response) {
			response.writeHead(200, HEADER);
			var robot = self.getRobotByIndex(parseInt(request.params.index));
			if(robot) {
				robot.changeBuzzerBy(request);
			}
			response.end();
		});
		
		express.get('/setBuzzerTo/:index/:value', function(request, response) {
			response.writeHead(200, HEADER);
			var robot = self.getRobotByIndex(parseInt(request.params.index));
			if(robot) {
				robot.setBuzzerTo(request);
			}
			response.end();
		});
		
		express.get('/clearBuzzer/:index', function(request, response) {
			response.writeHead(200, HEADER);
			var robot = self.getRobotByIndex(parseInt(request.params.index));
			if(robot) {
				robot.clearBuzzer(request);
			}
			response.end();
		});
		
		express.get('/playNote/:index/:note/:octave', function(request, response) {
			response.writeHead(200, HEADER);
			var robot = self.getRobotByIndex(parseInt(request.params.index));
			if(robot) {
				robot.playNote(request);
			}
			response.end();
		});
		
		express.get('/playNoteFor/:id/:index/:note/:octave/:beat', function(request, response) {
			response.writeHead(200, HEADER);
			var robot = self.getRobotByIndex(parseInt(request.params.index));
			if(robot) {
				robot.playNoteFor(request);
			}
			response.end();
		});
		
		express.get('/restFor/:id/:index/:beat', function(request, response) {
			response.writeHead(200, HEADER);
			var robot = self.getRobotByIndex(parseInt(request.params.index));
			if(robot) {
				robot.restFor(request);
			}
			response.end();
		});
		
		express.get('/changeTempoBy/:index/:value', function(request, response) {
			response.writeHead(200, HEADER);
			var robot = self.getRobotByIndex(parseInt(request.params.index));
			if(robot) {
				robot.changeTempoBy(request);
			}
			response.end();
		});
		
		express.get('/setTempoTo/:index/:value', function(request, response) {
			response.writeHead(200, HEADER);
			var robot = self.getRobotByIndex(parseInt(request.params.index));
			if(robot) {
				robot.setTempoTo(request);
			}
			response.end();
		});
		
		express.get('/setPortTo/:index/:which/:value', function(request, response) {
			response.writeHead(200, HEADER);
			var robot = self.getRobotByIndex(parseInt(request.params.index));
			if(robot) {
				robot.setPortTo(request);
			}
			response.end();
		});
		
		express.get('/changeOutputBy/:index/:which/:value', function(request, response) {
			response.writeHead(200, HEADER);
			var robot = self.getRobotByIndex(parseInt(request.params.index));
			if(robot) {
				robot.changeOutputBy(request);
			}
			response.end();
		});
		
		express.get('/setOutputTo/:index/:which/:value', function(request, response) {
			response.writeHead(200, HEADER);
			var robot = self.getRobotByIndex(parseInt(request.params.index));
			if(robot) {
				robot.setOutputTo(request);
			}
			response.end();
		});
		
		express.get('/gripper/:id/:index/:action', function(request, response) {
			response.writeHead(200, HEADER);
			var robot = self.getRobotByIndex(parseInt(request.params.index));
			if(robot) {
				robot.gripper(request);
			}
			response.end();
		});
		
		express.get('/releaseGripper/:index', function(request, response) {
			response.writeHead(200, HEADER);
			var robot = self.getRobotByIndex(parseInt(request.params.index));
			if(robot) {
				robot.releaseGripper(request);
			}
			response.end();
		});
		
		express.get('/sBoardMoveForward/:id/:index', function(request, response) {
			response.writeHead(200, HEADER);
			var robot = self.getRobotByIndex(parseInt(request.params.index));
			if(robot) {
				robot.boardMoveForward(request);
			}
			response.end();
		});
		
		express.get('/sBoardTurn/:id/:index/:direction', function(request, response) {
			response.writeHead(200, HEADER);
			var robot = self.getRobotByIndex(parseInt(request.params.index));
			if(robot) {
				robot.boardTurn(request);
			}
			response.end();
		});
		
		express.get('/sMoveForward/:id/:index', function(request, response) {
			response.writeHead(200, HEADER);
			var robot = self.getRobotByIndex(parseInt(request.params.index));
			if(robot) {
				robot.moveForward(request);
			}
			response.end();
		});
		
		express.get('/sMoveBackward/:id/:index', function(request, response) {
			response.writeHead(200, HEADER);
			var robot = self.getRobotByIndex(parseInt(request.params.index));
			if(robot) {
				robot.moveBackward(request);
			}
			response.end();
		});
		
		express.get('/sTurn/:id/:index/:direction', function(request, response) {
			response.writeHead(200, HEADER);
			var robot = self.getRobotByIndex(parseInt(request.params.index));
			if(robot) {
				robot.turn(request);
			}
			response.end();
		});
		
		express.get('/sMoveForwardUnit/:id/:index/:value/:unit', function(request, response) {
			response.writeHead(200, HEADER);
			var robot = self.getRobotByIndex(parseInt(request.params.index));
			if(robot) {
				robot.moveForwardUnit(request);
			}
			response.end();
		});

		express.get('/sMoveBackwardUnit/:id/:index/:value/:unit', function(request, response) {
			response.writeHead(200, HEADER);
			var robot = self.getRobotByIndex(parseInt(request.params.index));
			if(robot) {
				robot.moveBackwardUnit(request);
			}
			response.end();
		});

		express.get('/sTurnUnitInPlace/:id/:index/:direction/:value/:unit', function(request, response) {
			response.writeHead(200, HEADER);
			var robot = self.getRobotByIndex(parseInt(request.params.index));
			if(robot) {
				robot.turnUnitInPlace(request);
			}
			response.end();
		});
		
		express.get('/sPivotAroundWheelUnitInDirection/:id/:index/:wheel/:value/:unit/:forward', function(request, response) {
			response.writeHead(200, HEADER);
			var robot = self.getRobotByIndex(parseInt(request.params.index));
			if(robot) {
				robot.pivotAroundWheelUnitInDirection(request);
			}
			response.end();
		});
		
		express.get('/sTurnBodyUnitWithRadiusInDirection/:id/:index/:direction/:value/:unit/:radius/:forward', function(request, response) {
			response.writeHead(200, HEADER);
			var robot = self.getRobotByIndex(parseInt(request.params.index));
			if(robot) {
				robot.turnBodyUnitWithRadiusInDirection(request);
			}
			response.end();
		});
		
		express.get('/sPivotAroundUnitInDirection/:id/:index/:part/:value/:unit/:forward', function(request, response) {
			response.writeHead(200, HEADER);
			var robot = self.getRobotByIndex(parseInt(request.params.index));
			if(robot) {
				robot.pivotAroundUnitInDirection(request);
			}
			response.end();
		});
		
		express.get('/sTurnUnitWithRadiusInDirection/:id/:index/:part/:direction/:value/:unit/:radius/:forward', function(request, response) {
			response.writeHead(200, HEADER);
			var robot = self.getRobotByIndex(parseInt(request.params.index));
			if(robot) {
				robot.turnUnitWithRadiusInDirection(request);
			}
			response.end();
		});
		
		express.get('/sChangeBothWheelsBy/:index/:left/:right', function(request, response) {
			response.writeHead(200, HEADER);
			var robot = self.getRobotByIndex(parseInt(request.params.index));
			if(robot) {
				robot.changeBothWheelsBy(request);
			}
			response.end();
		});
		
		express.get('/sSetBothWheelsTo/:index/:left/:right', function(request, response) {
			response.writeHead(200, HEADER);
			var robot = self.getRobotByIndex(parseInt(request.params.index));
			if(robot) {
				robot.setBothWheelsTo(request);
			}
			response.end();
		});
		
		express.get('/sChangeWheelBy/:index/:which/:speed', function(request, response) {
			response.writeHead(200, HEADER);
			var robot = self.getRobotByIndex(parseInt(request.params.index));
			if(robot) {
				robot.changeWheelBy(request);
			}
			response.end();
		});
		
		express.get('/sSetWheelTo/:index/:which/:speed', function(request, response) {
			response.writeHead(200, HEADER);
			var robot = self.getRobotByIndex(parseInt(request.params.index));
			if(robot) {
				robot.setWheelTo(request);
			}
			response.end();
		});
		
		express.get('/sFollowLineUsingFloorSensor/:index/:color/:which', function(request, response) {
			response.writeHead(200, HEADER);
			var robot = self.getRobotByIndex(parseInt(request.params.index));
			if(robot) {
				robot.followLineUsingFloorSensor(request);
			}
			response.end();
		});
		
		express.get('/sFollowLineUntilIntersection/:id/:index/:color/:which', function(request, response) {
			response.writeHead(200, HEADER);
			var robot = self.getRobotByIndex(parseInt(request.params.index));
			if(robot) {
				robot.followLineUntilIntersection(request);
			}
			response.end();
		});
		
		express.get('/sSetFollowingSpeedTo/:index/:speed', function(request, response) {
			response.writeHead(200, HEADER);
			var robot = self.getRobotByIndex(parseInt(request.params.index));
			if(robot) {
				robot.setFollowingSpeedTo(request);
			}
			response.end();
		});
		
		express.get('/sSetFollowingGainTo/:index/:gain', function(request, response) {
			response.writeHead(200, HEADER);
			var robot = self.getRobotByIndex(parseInt(request.params.index));
			if(robot) {
				robot.setFollowingGainTo(request);
			}
			response.end();
		});
		
		express.get('/sStop/:index', function(request, response) {
			response.writeHead(200, HEADER);
			var robot = self.getRobotByIndex(parseInt(request.params.index));
			if(robot) {
				robot.stop(request);
			}
			response.end();
		});
		
		express.get('/sSetLedTo/:index/:which/:color', function(request, response) {
			response.writeHead(200, HEADER);
			var robot = self.getRobotByIndex(parseInt(request.params.index));
			if(robot) {
				robot.setLedTo(request);
			}
			response.end();
		});
		
		express.get('/sChangeLedByRGB/:index/:which/:red/:green/:blue', function(request, response) {
			response.writeHead(200, HEADER);
			var robot = self.getRobotByIndex(parseInt(request.params.index));
			if(robot) {
				robot.changeLedByRGB(request);
			}
			response.end();
		});
		
		express.get('/sSetLedToRGB/:index/:which/:red/:green/:blue', function(request, response) {
			response.writeHead(200, HEADER);
			var robot = self.getRobotByIndex(parseInt(request.params.index));
			if(robot) {
				robot.setLedToRGB(request);
			}
			response.end();
		});
		
		express.get('/sClearLed/:index/:which', function(request, response) {
			response.writeHead(200, HEADER);
			var robot = self.getRobotByIndex(parseInt(request.params.index));
			if(robot) {
				robot.clearLed(request);
			}
			response.end();
		});
		
		express.get('/sPlaySound/:index/:sound', function(request, response) {
			response.writeHead(200, HEADER);
			var robot = self.getRobotByIndex(parseInt(request.params.index));
			if(robot) {
				robot.playSound(request);
			}
			response.end();
		});
		
		express.get('/sPlaySoundTimes/:index/:sound/:count', function(request, response) {
			response.writeHead(200, HEADER);
			var robot = self.getRobotByIndex(parseInt(request.params.index));
			if(robot) {
				robot.playSoundTimes(request);
			}
			response.end();
		});
		
		express.get('/sPlaySoundTimesUntilDone/:id/:index/:sound/:count', function(request, response) {
			response.writeHead(200, HEADER);
			var robot = self.getRobotByIndex(parseInt(request.params.index));
			if(robot) {
				robot.playSoundTimesUntilDone(request);
			}
			response.end();
		});
		
		express.get('/sChangeBuzzerBy/:index/:value', function(request, response) {
			response.writeHead(200, HEADER);
			var robot = self.getRobotByIndex(parseInt(request.params.index));
			if(robot) {
				robot.changeBuzzerBy(request);
			}
			response.end();
		});
		
		express.get('/sSetBuzzerTo/:index/:value', function(request, response) {
			response.writeHead(200, HEADER);
			var robot = self.getRobotByIndex(parseInt(request.params.index));
			if(robot) {
				robot.setBuzzerTo(request);
			}
			response.end();
		});
		
		express.get('/sClearSound/:index', function(request, response) {
			response.writeHead(200, HEADER);
			var robot = self.getRobotByIndex(parseInt(request.params.index));
			if(robot) {
				robot.clearSound(request);
			}
			response.end();
		});
		
		express.get('/sPlayNote/:index/:note/:octave', function(request, response) {
			response.writeHead(200, HEADER);
			var robot = self.getRobotByIndex(parseInt(request.params.index));
			if(robot) {
				robot.playNote(request);
			}
			response.end();
		});
		
		express.get('/sPlayNoteFor/:id/:index/:note/:octave/:beat', function(request, response) {
			response.writeHead(200, HEADER);
			var robot = self.getRobotByIndex(parseInt(request.params.index));
			if(robot) {
				robot.playNoteFor(request);
			}
			response.end();
		});
		
		express.get('/sRestFor/:id/:index/:beat', function(request, response) {
			response.writeHead(200, HEADER);
			var robot = self.getRobotByIndex(parseInt(request.params.index));
			if(robot) {
				robot.restFor(request);
			}
			response.end();
		});
		
		express.get('/sChangeTempoBy/:index/:value', function(request, response) {
			response.writeHead(200, HEADER);
			var robot = self.getRobotByIndex(parseInt(request.params.index));
			if(robot) {
				robot.changeTempoBy(request);
			}
			response.end();
		});
		
		express.get('/sSetTempoTo/:index/:value', function(request, response) {
			response.writeHead(200, HEADER);
			var robot = self.getRobotByIndex(parseInt(request.params.index));
			if(robot) {
				robot.setTempoTo(request);
			}
			response.end();
		});
		
		express.get('/sSetPortTo/:index/:which/:value', function(request, response) {
			response.writeHead(200, HEADER);
			var robot = self.getRobotByIndex(parseInt(request.params.index));
			if(robot) {
				robot.setPortTo(request);
			}
			response.end();
		});
		
		express.get('/sChangeOutputBy/:index/:which/:value', function(request, response) {
			response.writeHead(200, HEADER);
			var robot = self.getRobotByIndex(parseInt(request.params.index));
			if(robot) {
				robot.changeOutputBy(request);
			}
			response.end();
		});
		
		express.get('/sSetOutputTo/:index/:which/:value', function(request, response) {
			response.writeHead(200, HEADER);
			var robot = self.getRobotByIndex(parseInt(request.params.index));
			if(robot) {
				robot.setOutputTo(request);
			}
			response.end();
		});
		
		express.get('/sGripper/:id/:index/:action', function(request, response) {
			response.writeHead(200, HEADER);
			var robot = self.getRobotByIndex(parseInt(request.params.index));
			if(robot) {
				robot.gripper(request);
			}
			response.end();
		});
		
		express.get('/sReleaseGripper/:index', function(request, response) {
			response.writeHead(200, HEADER);
			var robot = self.getRobotByIndex(parseInt(request.params.index));
			if(robot) {
				robot.releaseGripper(request);
			}
			response.end();
		});
		
		express.get('/sWriteSerial/:id/:index/:mode/:value', function(request, response) {
			response.writeHead(200, HEADER);
			var robot = self.getRobotByIndex(parseInt(request.params.index));
			if(robot) {
				robot.writeSerial(request);
			}
			response.end();
		});
		
		express.get('/sReadSerialUntil/:id/:index/:delimiter', function(request, response) {
			response.writeHead(200, HEADER);
			var robot = self.getRobotByIndex(parseInt(request.params.index));
			if(robot) {
				robot.readSerialUntil(request);
			}
			response.end();
		});
		
		express.get('/sSetSerialRateTo/:index/:baud', function(request, response) {
			response.writeHead(200, HEADER);
			var robot = self.getRobotByIndex(parseInt(request.params.index));
			if(robot) {
				robot.setSerialRateTo(request);
			}
			response.end();
		});
	}
};

Extension.prototype.stop = function() {
	this.running = false;
	translator.remove('hamster-scratch-multi');
	router.reset();
	this.robots = {};
};

Extension.prototype.setState = function(route) {
	if(route == router.firstRoute && route.module == MODULE) {
		this.state = route.state;
	}
};

Extension.prototype.setOption = function(option) {
};

Extension.prototype.handleSensory = function(data, route) {
	if(route.module == MODULE) {
		var robot = this.getRobotByRoute(route);
		if(robot) {
			robot.handleSensory(data);
		}
	}
};

Extension.prototype.handleMotoring = function(data, route) {
	if(route.module == MODULE) {
		var robot = this.getRobotByRoute(route);
		if(robot) {
			var roboid = robot.roboid;
			roboid.handleMotoring(data);
			roboid.updateMotoring();
		}
	}
};

Extension.prototype.requestMotoring = function(route) {
	if(route.module == MODULE) {
		var robot = this.getRobotByRoute(route);
		if(robot) {
			return robot.roboid.requestMotoring();
		}
	}
};

Extension.prototype.reset = function(route) {
	if(route.module == MODULE) {
		var robot = this.getRobotByRoute(route);
		if(robot) {
			robot.roboid.reset();
			robot.motoring.reset();
		}
	}
};

module.exports.create = function() {
	return new Extension();
};