'use strict';
var context = require('context');
var translations = require('./translations');
var HamsterS = require('../hamsterS');
var factory = require('../hamsterS-roboid');
var SerialQueue = require('../serial-queue');
const MODULE = 'hamsterS';

var SPEED2GAIN = { 1: 6, 2: 6, 3: 5, 4: 5, 5: 4, 6: 4, 7: 3, 8: 3 };
var PARTS = {};
var DIRECTIONS = {};
var TOWARDS = {};
var UNITS = {};
var RGB_COLORS = {};
var NOTES = {};
var BEATS = { '¼': 0.25, '½': 0.5, '¾': 0.75, '1¼': 1.25, '1½': 1.5, '1¾': 1.75 };
var SOUNDS = {};
var IO_MODES = {};
var GRIPPERS = {};
var SERIAL_MODES = {};
var SERIAL_DELIMITERS = {};
var SERIAL_BAUDS = {};
var VALUES = {};

var BOOLEAN_TILT_FORWARD = '/tilt%20forward ';
var BOOLEAN_TILT_BACKWARD = '/tilt%20backward ';
var BOOLEAN_TILT_LEFT = '/tilt%20left ';
var BOOLEAN_TILT_RIGHT = '/tilt%20right ';
var BOOLEAN_TILT_FLIP = '/tilt%20flip ';
var BOOLEAN_TILT_NONE = '/not%20tilt ';
var BOOLEAN_FREE_FALL = '/free%20fall ';
var BOOLEAN_TAP = '/tap ';
var BOOLEAN_BATTERY_NORMAL = '/normal ';
var BOOLEAN_BATTERY_LOW = '/low ';
var BOOLEAN_BATTERY_EMPTY = '/empty ';

const LEFT = 1;
const RIGHT = 2;
const BOTH = 3;
const FRONT = 4;
const REAR = 5;
const LEFT_PEN = 1;
const RIGHT_PEN = 2;
const LEFT_WHEEL = 3;
const RIGHT_WHEEL = 4;
const FORWARD = 1;
const BACKWARD = 2;
const SECONDS = 1;
const PULSES = 2;
const DEGREES = 3;
const OPEN = 1;
const CLOSE = 2;
const SERIAL_STRING = 1;
const SERIAL_STRING_LINE = 2;
const SERIAL_ALL = 0;
const SERIAL_NEW_LINE = 0x0D;
const SERIAL_COMMA = 0x2C;
const SERIAL_COLON = 0x3A;
const SERIAL_DOLLAR = 0x24;
const SERIAL_SHARP = 0x23;
const WHITE = 1;

function Robot(index) {
	var roboid = this.roboid = factory.create();
	this.index = index;
	this.waits = [];
	this.timeouts = [];
	this.sensory = {
		signalStrength: 0,
		leftProximity: 0,
		rightProximity: 0,
		leftFloor: 0,
		rightFloor: 0,
		accelerationX: 0,
		accelerationY: 0,
		accelerationZ: 0,
		light: 0,
		temperature: 0,
		inputA: 0,
		inputB: 0,
		tilt: 0,
		serial: '',
		batteryState: 2
	};
	this.motoring = {
		pulse: roboid.findDeviceById(HamsterS.PULSE),
		note: roboid.findDeviceById(HamsterS.NOTE),
		sound: roboid.findDeviceById(HamsterS.SOUND),
		lineTracerMode: roboid.findDeviceById(HamsterS.LINE_TRACER_MODE),
		lineTracerGain: roboid.findDeviceById(HamsterS.LINE_TRACER_GAIN),
		lineTracerSpeed: roboid.findDeviceById(HamsterS.LINE_TRACER_SPEED),
		ioModeA: roboid.findDeviceById(HamsterS.IO_MODE_A),
		ioModeB: roboid.findDeviceById(HamsterS.IO_MODE_B),
		writeSerial: roboid.findDeviceById(HamsterS.WRITE_SERIAL),
		motion: roboid.findDeviceById(HamsterS.MOTION),
		motorMode: roboid.findDeviceById(HamsterS.MOTOR_MODE),
		configProximity: roboid.findDeviceById(HamsterS.CONFIG_PROXIMITY),
		configGravity: roboid.findDeviceById(HamsterS.CONFIG_GRAVITY),
		configBandWidth: roboid.findDeviceById(HamsterS.CONFIG_BAND_WIDTH),
		reset: function() {
			this.leftWheel.data = 0;
			this.rightWheel.data = 0;
			this.leftRgb.data[0] = 0;
			this.leftRgb.data[1] = 0;
			this.leftRgb.data[2] = 0;
			this.rightRgb.data[0] = 0;
			this.rightRgb.data[1] = 0;
			this.rightRgb.data[2] = 0;
			this.buzzer.data = 0;
			this.outputA.data = 0;
			this.outputB.data = 0;
		}
	};
	this.register('leftWheel', HamsterS.LEFT_WHEEL);
	this.register('rightWheel', HamsterS.RIGHT_WHEEL);
	this.register3('leftRgb', HamsterS.LEFT_RGB);
	this.register3('rightRgb', HamsterS.RIGHT_RGB);
	this.register('buzzer', HamsterS.BUZZER);
	this.register('outputA', HamsterS.OUTPUT_A);
	this.register('outputB', HamsterS.OUTPUT_B);
	this.serialRate = HamsterS.IO_MODE_SERIAL_9600;
	this.serialWriteWaits = [];
	this.serialReadWaits = [];
	this.writeQueue = SerialQueue.createWriteQueue(64);
	this.readQueue = SerialQueue.createReadQueue(64);
	this.tempo = 60;
	this.speed = 5;
	this.gain = -1;
	this.motionId = -1;
	this.lineTracerId = -1;
	this.boardId = -1;
	this.boardCommand = 0;
	this.boardState = 0;
	this.boardCount = 0;
	this.soundId = -1;
	this.currentSound = 0;
	this.soundRepeat = 1;
	this.noteId = -1;
	this.noteTimer1 = undefined;
	this.noteTimer2 = undefined;
	this.ioId = -1;
	this.ioTimer = undefined;
	this.serialDelimiter = 0;
	this.freeFall = false;
	this.tap = false;
	this.active = true;
	this.received = false;
}

Robot.prototype.reset = function() {
	this.serialRate = HamsterS.IO_MODE_SERIAL_9600;
	this.tempo = 60;
	this.speed = 5;
	this.gain = -1;
	this.motionId = -1;
	this.lineTracerId = -1;
	this.boardId = -1;
	this.boardCommand = 0;
	this.boardState = 0;
	this.boardCount = 0;
	this.soundId = -1;
	this.currentSound = 0;
	this.soundRepeat = 1;
	this.noteId = -1;
	this.noteTimer1 = undefined;
	this.noteTimer2 = undefined;
	this.ioId = -1;
	this.ioTimer = undefined;
	this.serialDelimiter = 0;
	this.freeFall = false;
	this.tap = false;
	this.roboid.reset();
	this.motoring.reset();
	this.writeQueue.reset();
	this.readQueue.reset();
};

Robot.prototype.register = function(name, deviceId) {
	var self = this;
	self.motoring[name] = {
		data: 0,
		device: self.roboid.findDeviceById(deviceId),
		write: function(value) {
			this.data = value;
			this.device.write(value);
		},
		read: function() {
			return this.data;
		}
	};
};

Robot.prototype.register3 = function(name, deviceId) {
	var self = this;
	self.motoring[name] = {
		data: [0, 0, 0],
		device: self.roboid.findDeviceById(deviceId),
		write: function(index, value) {
			this.data[index] = value;
			this.device.write(index, value);
		},
		read: function(index) {
			return this.data[index];
		}
	};
};

Robot.prototype.getWaits = function() {
	return this.waits;
};

Robot.prototype.addSerialWriteWait = function(id) {
	this.waits.push(id);
	this.serialWriteWaits.push(id);
};

Robot.prototype.addSerialReadWait = function(id) {
	this.waits.push(id);
	this.serialReadWaits.push(id);
};

Robot.prototype.removeWait = function(id) {
	var waits = this.waits;
	var index = waits.indexOf(id);
	if(index >= 0) {
		waits.splice(index, 1);
	}
};

Robot.prototype.removeSerialWriteWaits = function() {
	var serialWaits = this.serialWriteWaits;
	for(var i in serialWaits) {
		this.removeWait(serialWaits[i]);
	}
	this.serialWriteWaits = [];
};

Robot.prototype.removeSerialReadWaits = function() {
	var serialWaits = this.serialReadWaits;
	for(var i in serialWaits) {
		this.removeWait(serialWaits[i]);
	}
	this.serialReadWaits = [];
};

Robot.prototype.clearWaits = function() {
	this.waits = [];
	this.serialWriteWaits = [];
	this.serialReadWaits = [];
};

Robot.prototype.removeTimeout = function(id) {
	clearTimeout(id);
	var timeouts = this.timeouts;
	var index = timeouts.indexOf(id);
	if(index >= 0) {
		timeouts.splice(index, 1);
	}
};

Robot.prototype.clearTimeouts = function() {
	var timeouts = this.timeouts;
	for(var i in timeouts) {
		clearTimeout(timeouts[i]);
	}
	this.timeouts = [];
};

Robot.prototype.cancelMotion = function() {
	this.motionId = -1;
};

Robot.prototype.cancelLineTracer = function() {
	this.lineTracerId = -1;
};

Robot.prototype.cancelBoard = function() {
	this.boardId = -1;
	this.boardCommand = 0;
	this.boardState = 0;
	this.boardCount = 0;
};

Robot.prototype.cancelNote = function() {
	this.noteId = -1;
	if(this.noteTimer1 !== undefined) {
		this.removeTimeout(this.noteTimer1);
	}
	if(this.noteTimer2 !== undefined) {
		this.removeTimeout(this.noteTimer2);
	}
	this.noteTimer1 = undefined;
	this.noteTimer2 = undefined;
};

Robot.prototype.cancelSound = function() {
	this.soundId = -1;
};

Robot.prototype.cancelIo = function() {
	this.ioId = -1;
	if(this.ioTimer !== undefined) {
		this.removeTimeout(this.ioTimer);
	}
	this.ioTimer = undefined;
};

Robot.prototype.writeSingleSensory = function(response) {
	var sensory = this.sensory;
	response.write('sSignalStrength ' + sensory.signalStrength + '\n');
	response.write('sLeftProximity ' + sensory.leftProximity + '\n');
	response.write('sRightProximity ' + sensory.rightProximity + '\n');
	response.write('sLeftFloor ' + sensory.leftFloor + '\n');
	response.write('sRightFloor ' + sensory.rightFloor + '\n');
	response.write('sAccelerationX ' + sensory.accelerationX + '\n');
	response.write('sAccelerationY ' + sensory.accelerationY + '\n');
	response.write('sAccelerationZ ' + sensory.accelerationZ + '\n');
	response.write('sLight ' + sensory.light + '\n');
	response.write('sTemperature ' + sensory.temperature + '\n');
	response.write('sInputA ' + sensory.inputA + '\n');
	response.write('sInputB ' + sensory.inputB + '\n');
	var threshold = context.getHandFoundThreshold(MODULE);
	var handFound = (sensory.leftProximity > threshold || sensory.rightProximity > threshold) ? true : false;
	response.write('sHandFound ' + handFound + '\n');
	
	response.write('sTilt' + BOOLEAN_TILT_FORWARD + (sensory.tilt == 1) + '\n');
	response.write('sTilt' + BOOLEAN_TILT_BACKWARD + (sensory.tilt == -1) + '\n');
	response.write('sTilt' + BOOLEAN_TILT_LEFT + (sensory.tilt == 2) + '\n');
	response.write('sTilt' + BOOLEAN_TILT_RIGHT + (sensory.tilt == -2) + '\n');
	response.write('sTilt' + BOOLEAN_TILT_FLIP + (sensory.tilt == 3) + '\n');
	response.write('sTilt' + BOOLEAN_TILT_NONE + (sensory.tilt == -3) + '\n');
	response.write('sTilt' + BOOLEAN_FREE_FALL + this.freeFall + '\n');
	response.write('sTilt' + BOOLEAN_TAP + this.tap + '\n');
	response.write('sBattery' + BOOLEAN_BATTERY_NORMAL + (sensory.batteryState == 2) + '\n');
	response.write('sBattery' + BOOLEAN_BATTERY_LOW + (sensory.batteryState == 1) + '\n');
	response.write('sBattery' + BOOLEAN_BATTERY_EMPTY + (sensory.batteryState == 0) + '\n');
	response.write('sSerial ' + sensory.serial + '\n');
	
	response.write('signalStrength ' + sensory.signalStrength + '\n');
	response.write('leftProximity ' + sensory.leftProximity + '\n');
	response.write('rightProximity ' + sensory.rightProximity + '\n');
	response.write('leftFloor ' + sensory.leftFloor + '\n');
	response.write('rightFloor ' + sensory.rightFloor + '\n');
	response.write('accelerationX ' + sensory.accelerationX + '\n');
	response.write('accelerationY ' + sensory.accelerationY + '\n');
	response.write('accelerationZ ' + sensory.accelerationZ + '\n');
	response.write('light ' + sensory.light + '\n');
	response.write('temperature ' + sensory.temperature + '\n');
	response.write('inputA ' + sensory.inputA + '\n');
	response.write('inputB ' + sensory.inputB + '\n');
	response.write('handFound ' + handFound + '\n');
	response.write('tilt' + BOOLEAN_TILT_FORWARD + (sensory.tilt == 1) + '\n');
	response.write('tilt' + BOOLEAN_TILT_BACKWARD + (sensory.tilt == -1) + '\n');
	response.write('tilt' + BOOLEAN_TILT_LEFT + (sensory.tilt == 2) + '\n');
	response.write('tilt' + BOOLEAN_TILT_RIGHT + (sensory.tilt == -2) + '\n');
	response.write('tilt' + BOOLEAN_TILT_FLIP + (sensory.tilt == 3) + '\n');
	response.write('tilt' + BOOLEAN_TILT_NONE + (sensory.tilt == -3) + '\n');
	response.write('battery' + BOOLEAN_BATTERY_NORMAL + (sensory.batteryState == 2) + '\n');
	response.write('battery' + BOOLEAN_BATTERY_LOW + (sensory.batteryState == 1) + '\n');
	response.write('battery' + BOOLEAN_BATTERY_EMPTY + (sensory.batteryState == 0) + '\n');
	
	this.freeFall = false;
	this.tap = false;
};

Robot.prototype.writeMultiSensory = function(response) {
	if(this.active) {
		var sensory = this.sensory;
		var index = this.index;
		var tag = index + ' ';
		response.write('sSignalStrength/' + tag + sensory.signalStrength + '\n');
		response.write('sLeftProximity/' + tag + sensory.leftProximity + '\n');
		response.write('sRightProximity/' + tag + sensory.rightProximity + '\n');
		response.write('sLeftFloor/' + tag + sensory.leftFloor + '\n');
		response.write('sRightFloor/' + tag + sensory.rightFloor + '\n');
		response.write('sAccelerationX/' + tag + sensory.accelerationX + '\n');
		response.write('sAccelerationY/' + tag + sensory.accelerationY + '\n');
		response.write('sAccelerationZ/' + tag + sensory.accelerationZ + '\n');
		response.write('sLight/' + tag + sensory.light + '\n');
		response.write('sTemperature/' + tag + sensory.temperature + '\n');
		response.write('sInputA/' + tag + sensory.inputA + '\n');
		response.write('sInputB/' + tag + sensory.inputB + '\n');
		var threshold = context.getHandFoundThreshold(MODULE);
		var handFound = (sensory.leftProximity > threshold || sensory.rightProximity > threshold) ? true : false;
		response.write('sHandFound/' + tag + handFound + '\n');
		response.write('sTilt/' + index + BOOLEAN_TILT_FORWARD + (sensory.tilt == 1) + '\n');
		response.write('sTilt/' + index + BOOLEAN_TILT_BACKWARD + (sensory.tilt == -1) + '\n');
		response.write('sTilt/' + index + BOOLEAN_TILT_LEFT + (sensory.tilt == 2) + '\n');
		response.write('sTilt/' + index + BOOLEAN_TILT_RIGHT + (sensory.tilt == -2) + '\n');
		response.write('sTilt/' + index + BOOLEAN_TILT_FLIP + (sensory.tilt == 3) + '\n');
		response.write('sTilt/' + index + BOOLEAN_TILT_NONE + (sensory.tilt == -3) + '\n');
		response.write('sTilt/' + index + BOOLEAN_FREE_FALL + this.freeFall + '\n');
		response.write('sTilt/' + index + BOOLEAN_TAP + this.tap + '\n');
		response.write('sBattery/' + index + BOOLEAN_BATTERY_NORMAL + (sensory.batteryState == 2) + '\n');
		response.write('sBattery/' + index + BOOLEAN_BATTERY_LOW + (sensory.batteryState == 1) + '\n');
		response.write('sBattery/' + index + BOOLEAN_BATTERY_EMPTY + (sensory.batteryState == 0) + '\n');
		response.write('sSerial/' + index + ' ' + sensory.serial + '\n');
		
		response.write('signalStrength/' + tag + sensory.signalStrength + '\n');
		response.write('leftProximity/' + tag + sensory.leftProximity + '\n');
		response.write('rightProximity/' + tag + sensory.rightProximity + '\n');
		response.write('leftFloor/' + tag + sensory.leftFloor + '\n');
		response.write('rightFloor/' + tag + sensory.rightFloor + '\n');
		response.write('accelerationX/' + tag + sensory.accelerationX + '\n');
		response.write('accelerationY/' + tag + sensory.accelerationY + '\n');
		response.write('accelerationZ/' + tag + sensory.accelerationZ + '\n');
		response.write('light/' + tag + sensory.light + '\n');
		response.write('temperature/' + tag + sensory.temperature + '\n');
		response.write('inputA/' + tag + sensory.inputA + '\n');
		response.write('inputB/' + tag + sensory.inputB + '\n');
		response.write('handFound/' + tag + handFound + '\n');
		response.write('tilt/' + index + BOOLEAN_TILT_FORWARD + (sensory.tilt == 1) + '\n');
		response.write('tilt/' + index + BOOLEAN_TILT_BACKWARD + (sensory.tilt == -1) + '\n');
		response.write('tilt/' + index + BOOLEAN_TILT_LEFT + (sensory.tilt == 2) + '\n');
		response.write('tilt/' + index + BOOLEAN_TILT_RIGHT + (sensory.tilt == -2) + '\n');
		response.write('tilt/' + index + BOOLEAN_TILT_FLIP + (sensory.tilt == 3) + '\n');
		response.write('tilt/' + index + BOOLEAN_TILT_NONE + (sensory.tilt == -3) + '\n');
		response.write('battery/' + index + BOOLEAN_BATTERY_NORMAL + (sensory.batteryState == 2) + '\n');
		response.write('battery/' + index + BOOLEAN_BATTERY_LOW + (sensory.batteryState == 1) + '\n');
		response.write('battery/' + index + BOOLEAN_BATTERY_EMPTY + (sensory.batteryState == 0) + '\n');
		
		response.write('signalStrength' + tag + sensory.signalStrength + '\n');
		response.write('leftProximity' + tag + sensory.leftProximity + '\n');
		response.write('rightProximity' + tag + sensory.rightProximity + '\n');
		response.write('leftFloor' + tag + sensory.leftFloor + '\n');
		response.write('rightFloor' + tag + sensory.rightFloor + '\n');
		response.write('accelerationX' + tag + sensory.accelerationX + '\n');
		response.write('accelerationY' + tag + sensory.accelerationY + '\n');
		response.write('accelerationZ' + tag + sensory.accelerationZ + '\n');
		response.write('light' + tag + sensory.light + '\n');
		response.write('temperature' + tag + sensory.temperature + '\n');
		response.write('inputA' + tag + sensory.inputA + '\n');
		response.write('inputB' + tag + sensory.inputB + '\n');
		response.write('handFound' + tag + handFound + '\n');
	}
	this.freeFall = false;
	this.tap = false;
	if(this.received == false) this.active = false;
};

Robot.prototype.boardMoveForward = function(request) {
	var motoring = this.motoring;
	this.cancelMotion();
	this.cancelLineTracer();
	
	motoring.leftWheel.write(45);
	motoring.rightWheel.write(45);
	this.boardCommand = 1;
	this.boardCount = 0;
	this.boardState = 1;
	this.boardId = parseInt(request.params.id);
	this.waits.push(this.boardId);
	motoring.pulse.write(0);
	motoring.motion.write(0);
	motoring.lineTracerMode.write(0);
};

Robot.prototype.boardTurn = function(request) {
	var motoring = this.motoring;
	this.cancelMotion();
	this.cancelLineTracer();
	
	if(DIRECTIONS[request.params.direction] === LEFT) {
		motoring.leftWheel.write(-45);
		motoring.rightWheel.write(45);
		this.boardCommand = 2;
	} else {
		motoring.leftWheel.write(45);
		motoring.rightWheel.write(-45);
		this.boardCommand = 3;
	}
	this.boardCount = 0;
	this.boardState = 1;
	this.boardId = parseInt(request.params.id);
	this.waits.push(this.boardId);
	motoring.pulse.write(0);
	motoring.motion.write(0);
	motoring.lineTracerMode.write(0);
};

Robot.prototype.moveForward = function(request) {
	var motoring = this.motoring;
	this.cancelBoard();
	this.cancelLineTracer();
	
	motoring.leftWheel.write(0);
	motoring.rightWheel.write(0);
	motoring.pulse.write(0);
	motoring.motion.write(1, HamsterS.UNIT_CM_DEG);
	motoring.motion.write(2, 0);
	motoring.motion.write(3, context.getMovementDistance(MODULE));
	motoring.motion.write(0, HamsterS.MOTION_MOVE_FORWARD);
	this.motionId = parseInt(request.params.id);
	this.waits.push(this.motionId);
	motoring.lineTracerMode.write(0);
};

Robot.prototype.moveBackward = function(request) {
	var motoring = this.motoring;
	this.cancelBoard();
	this.cancelLineTracer();
	
	motoring.leftWheel.write(0);
	motoring.rightWheel.write(0);
	motoring.pulse.write(0);
	motoring.motion.write(1, HamsterS.UNIT_CM_DEG);
	motoring.motion.write(2, 0);
	motoring.motion.write(3, context.getMovementDistance(MODULE));
	motoring.motion.write(0, HamsterS.MOTION_MOVE_BACKWARD);
	this.motionId = parseInt(request.params.id);
	this.waits.push(this.motionId);
	motoring.lineTracerMode.write(0);
};

Robot.prototype.turn = function(request) {
	var motoring = this.motoring;
	this.cancelBoard();
	this.cancelLineTracer();
	
	motoring.leftWheel.write(0);
	motoring.rightWheel.write(0);
	motoring.pulse.write(0);
	motoring.motion.write(1, HamsterS.UNIT_CM_DEG);
	motoring.motion.write(2, 0);
	motoring.motion.write(3, context.getTurningAngle(MODULE));
	if(DIRECTIONS[request.params.direction] === LEFT) {
		motoring.motion.write(0, HamsterS.MOTION_TURN_LEFT);
	} else {
		motoring.motion.write(0, HamsterS.MOTION_TURN_RIGHT);
	}
	this.motionId = parseInt(request.params.id);
	this.waits.push(this.motionId);
	motoring.lineTracerMode.write(0);
};

Robot.prototype.moveForwardForSecs = function(request) {
	var motoring = this.motoring;
	this.cancelBoard();
	this.cancelMotion();
	this.cancelLineTracer();
	
	motoring.leftWheel.write(0);
	motoring.rightWheel.write(0);
	motoring.pulse.write(0);
	var value = parseFloat(request.params.sec);
	if(value) {
		motoring.motion.write(1, HamsterS.UNIT_SEC);
		motoring.motion.write(2, 0);
		if(value < 0) {
			motoring.motion.write(3, -value);
			motoring.motion.write(0, HamsterS.MOTION_MOVE_BACKWARD);
		} else {
			motoring.motion.write(3, value);
			motoring.motion.write(0, HamsterS.MOTION_MOVE_FORWARD);
		}
		this.motionId = parseInt(request.params.id);
		this.waits.push(this.motionId);
		motoring.lineTracerMode.write(0);
	} else {
		motoring.motion.write(0);
		motoring.lineTracerMode.write(0);
	}
};

Robot.prototype.moveBackwardForSecs = function(request) {
	var motoring = this.motoring;
	this.cancelBoard();
	this.cancelMotion();
	this.cancelLineTracer();
	
	motoring.leftWheel.write(0);
	motoring.rightWheel.write(0);
	motoring.pulse.write(0);
	var value = parseFloat(request.params.sec);
	if(value) {
		motoring.motion.write(1, HamsterS.UNIT_SEC);
		motoring.motion.write(2, 0);
		if(value < 0) {
			motoring.motion.write(3, -value);
			motoring.motion.write(0, HamsterS.MOTION_MOVE_FORWARD);
		} else {
			motoring.motion.write(3, value);
			motoring.motion.write(0, HamsterS.MOTION_MOVE_BACKWARD);
		}
		this.motionId = parseInt(request.params.id);
		this.waits.push(this.motionId);
		motoring.lineTracerMode.write(0);
	} else {
		motoring.motion.write(0);
		motoring.lineTracerMode.write(0);
	}
};

Robot.prototype.turnForSecs = function(request) {
	var motoring = this.motoring;
	this.cancelBoard();
	this.cancelMotion();
	this.cancelLineTracer();
	
	motoring.leftWheel.write(0);
	motoring.rightWheel.write(0);
	motoring.pulse.write(0);
	var value = parseFloat(request.params.sec);
	if(value) {
		motoring.motion.write(1, HamsterS.UNIT_SEC);
		motoring.motion.write(2, 0);
		if(value < 0) {
			motoring.motion.write(3, -value);
			if(DIRECTIONS[request.params.direction] === LEFT) {
				motoring.motion.write(0, HamsterS.MOTION_TURN_RIGHT);
			} else {
				motoring.motion.write(0, HamsterS.MOTION_TURN_LEFT);
			}
		} else {
			motoring.motion.write(3, value);
			if(DIRECTIONS[request.params.direction] === LEFT) {
				motoring.motion.write(0, HamsterS.MOTION_TURN_LEFT);
			} else {
				motoring.motion.write(0, HamsterS.MOTION_TURN_RIGHT);
			}
		}
		this.motionId = parseInt(request.params.id);
		this.waits.push(this.motionId);
		motoring.lineTracerMode.write(0);
	} else {
		motoring.motion.write(0);
		motoring.lineTracerMode.write(0);
	}
};

Robot.prototype.moveForwardUnit = function(request) {
	var motoring = this.motoring;
	this.cancelBoard();
	this.cancelMotion();
	this.cancelLineTracer();
	
	motoring.leftWheel.write(0);
	motoring.rightWheel.write(0);
	motoring.pulse.write(0);
	var value = parseFloat(request.params.value);
	if(value) {
		var unit = UNITS[request.params.unit];
		if(unit === SECONDS) motoring.motion.write(1, HamsterS.UNIT_SEC);
		else if(unit === PULSES) motoring.motion.write(1, HamsterS.UNIT_PULSE);
		else motoring.motion.write(1, HamsterS.UNIT_CM_DEG);
		motoring.motion.write(2, 0);
		if(value < 0) {
			motoring.motion.write(3, -value);
			motoring.motion.write(0, HamsterS.MOTION_MOVE_BACKWARD);
		} else {
			motoring.motion.write(3, value);
			motoring.motion.write(0, HamsterS.MOTION_MOVE_FORWARD);
		}
		this.motionId = parseInt(request.params.id);
		this.waits.push(this.motionId);
		motoring.lineTracerMode.write(0);
	} else {
		motoring.motion.write(0);
		motoring.lineTracerMode.write(0);
	}
};

Robot.prototype.moveBackwardUnit = function(request) {
	var motoring = this.motoring;
	this.cancelBoard();
	this.cancelMotion();
	this.cancelLineTracer();
	
	motoring.leftWheel.write(0);
	motoring.rightWheel.write(0);
	motoring.pulse.write(0);
	var value = parseFloat(request.params.value);
	if(value) {
		var unit = UNITS[request.params.unit];
		if(unit === SECONDS) motoring.motion.write(1, HamsterS.UNIT_SEC);
		else if(unit === PULSES) motoring.motion.write(1, HamsterS.UNIT_PULSE);
		else motoring.motion.write(1, HamsterS.UNIT_CM_DEG);
		motoring.motion.write(2, 0);
		if(value < 0) {
			motoring.motion.write(3, -value);
			motoring.motion.write(0, HamsterS.MOTION_MOVE_FORWARD);
		} else {
			motoring.motion.write(3, value);
			motoring.motion.write(0, HamsterS.MOTION_MOVE_BACKWARD);
		}
		this.motionId = parseInt(request.params.id);
		this.waits.push(this.motionId);
		motoring.lineTracerMode.write(0);
	} else {
		motoring.motion.write(0);
		motoring.lineTracerMode.write(0);
	}
};

Robot.prototype.turnUnitInPlace = function(request) {
	var motoring = this.motoring;
	this.cancelBoard();
	this.cancelMotion();
	this.cancelLineTracer();
	
	motoring.leftWheel.write(0);
	motoring.rightWheel.write(0);
	motoring.pulse.write(0);
	var value = parseFloat(request.params.value);
	if(value) {
		var unit = UNITS[request.params.unit];
		if(unit === SECONDS) motoring.motion.write(1, HamsterS.UNIT_SEC);
		else if(unit === PULSES) motoring.motion.write(1, HamsterS.UNIT_PULSE);
		else motoring.motion.write(1, HamsterS.UNIT_CM_DEG);
		motoring.motion.write(2, 0);
		if(value < 0) {
			motoring.motion.write(3, -value);
			if(DIRECTIONS[request.params.direction] === LEFT) {
				motoring.motion.write(0, HamsterS.MOTION_TURN_RIGHT);
			} else {
				motoring.motion.write(0, HamsterS.MOTION_TURN_LEFT);
			}
		} else {
			motoring.motion.write(3, value);
			if(DIRECTIONS[request.params.direction] === LEFT) {
				motoring.motion.write(0, HamsterS.MOTION_TURN_LEFT);
			} else {
				motoring.motion.write(0, HamsterS.MOTION_TURN_RIGHT);
			}
		}
		this.motionId = parseInt(request.params.id);
		this.waits.push(this.motionId);
		motoring.lineTracerMode.write(0);
	} else {
		motoring.motion.write(0);
		motoring.lineTracerMode.write(0);
	}
};

Robot.prototype.pivotAroundWheelUnitInDirection = function(request) {
	var motoring = this.motoring;
	this.cancelBoard();
	this.cancelMotion();
	this.cancelLineTracer();
	
	motoring.leftWheel.write(0);
	motoring.rightWheel.write(0);
	motoring.pulse.write(0);
	var value = parseFloat(request.params.value);
	if(value) {
		var unit = UNITS[request.params.unit];
		if(unit === SECONDS) motoring.motion.write(1, HamsterS.UNIT_SEC);
		else if(unit === PULSES) motoring.motion.write(1, HamsterS.UNIT_PULSE);
		else motoring.motion.write(1, HamsterS.UNIT_CM_DEG);
		motoring.motion.write(2, 0);
		if(value < 0) {
			motoring.motion.write(3, -value);
			if(PARTS[request.params.wheel] === LEFT) {
				if(TOWARDS[request.params.forward] === FORWARD) motoring.motion.write(0, HamsterS.MOTION_PIVOT_LEFT_BACKWARD);
				else motoring.motion.write(0, HamsterS.MOTION_PIVOT_LEFT_FORWARD);
			} else {
				if(TOWARDS[request.params.forward] === FORWARD) motoring.motion.write(0, HamsterS.MOTION_PIVOT_RIGHT_BACKWARD);
				else motoring.motion.write(0, HamsterS.MOTION_PIVOT_RIGHT_FORWARD);
			}
		} else {
			motoring.motion.write(3, value);
			if(PARTS[request.params.wheel] === LEFT) {
				if(TOWARDS[request.params.forward] === FORWARD) motoring.motion.write(0, HamsterS.MOTION_PIVOT_LEFT_FORWARD);
				else motoring.motion.write(0, HamsterS.MOTION_PIVOT_LEFT_BACKWARD);
			} else {
				if(TOWARDS[request.params.forward] === FORWARD) motoring.motion.write(0, HamsterS.MOTION_PIVOT_RIGHT_FORWARD);
				else motoring.motion.write(0, HamsterS.MOTION_PIVOT_RIGHT_BACKWARD);
			}
		}
		this.motionId = parseInt(request.params.id);
		this.waits.push(this.motionId);
		motoring.lineTracerMode.write(0);
	} else {
		motoring.motion.write(0);
		motoring.lineTracerMode.write(0);
	}
};

Robot.prototype.turnBodyUnitWithRadiusInDirection = function(request) {
	var motoring = this.motoring;
	this.cancelBoard();
	this.cancelMotion();
	this.cancelLineTracer();
	
	motoring.leftWheel.write(0);
	motoring.rightWheel.write(0);
	motoring.pulse.write(0);
	var value = parseFloat(request.params.value);
	var radius = parseFloat(request.params.radius);
	if(value && (typeof radius == 'number') && radius >= 0) {
		var unit = UNITS[request.params.unit];
		if(unit === SECONDS) motoring.motion.write(1, HamsterS.UNIT_SEC);
		else if(unit === PULSES) motoring.motion.write(1, HamsterS.UNIT_PULSE);
		else motoring.motion.write(1, HamsterS.UNIT_CM_DEG);
		motoring.motion.write(2, 0);
		motoring.motion.write(4, radius);
		if(value < 0) {
			motoring.motion.write(3, -value);
			if(DIRECTIONS[request.params.direction] === LEFT) {
				if(TOWARDS[request.params.forward] === FORWARD) motoring.motion.write(0, HamsterS.MOTION_SWING_LEFT_BACKWARD);
				else motoring.motion.write(0, HamsterS.MOTION_SWING_LEFT_FORWARD);
			} else {
				if(TOWARDS[request.params.forward] === FORWARD) motoring.motion.write(0, HamsterS.MOTION_SWING_RIGHT_BACKWARD);
				else motoring.motion.write(0, HamsterS.MOTION_SWING_RIGHT_FORWARD);
			}
		} else {
			motoring.motion.write(3, value);
			if(DIRECTIONS[request.params.direction] === LEFT) {
				if(TOWARDS[request.params.forward] === FORWARD) motoring.motion.write(0, HamsterS.MOTION_SWING_LEFT_FORWARD);
				else motoring.motion.write(0, HamsterS.MOTION_SWING_LEFT_BACKWARD);
			} else {
				if(TOWARDS[request.params.forward] === FORWARD) motoring.motion.write(0, HamsterS.MOTION_SWING_RIGHT_FORWARD);
				else motoring.motion.write(0, HamsterS.MOTION_SWING_RIGHT_BACKWARD);
			}
		}
		this.motionId = parseInt(request.params.id);
		this.waits.push(this.motionId);
		motoring.lineTracerMode.write(0);
	} else {
		motoring.motion.write(0);
		motoring.lineTracerMode.write(0);
	}
};

Robot.prototype.pivotAroundUnitInDirection = function(request) {
	var motoring = this.motoring;
	this.cancelBoard();
	this.cancelMotion();
	this.cancelLineTracer();
	
	motoring.leftWheel.write(0);
	motoring.rightWheel.write(0);
	motoring.pulse.write(0);
	var value = parseFloat(request.params.value);
	if(value) {
		var unit = UNITS[request.params.unit];
		if(unit === SECONDS) motoring.motion.write(1, HamsterS.UNIT_SEC);
		else if(unit === PULSES) motoring.motion.write(1, HamsterS.UNIT_PULSE);
		else motoring.motion.write(1, HamsterS.UNIT_CM_DEG);
		motoring.motion.write(2, 0);
		var part = PARTS[request.params.part];
		if(value < 0) {
			motoring.motion.write(3, -value);
			if(part === LEFT_PEN) {
				if(TOWARDS[request.params.forward] === FORWARD) motoring.motion.write(0, HamsterS.MOTION_PIVOT_LEFT_PEN_BACKWARD);
				else motoring.motion.write(0, HamsterS.MOTION_PIVOT_LEFT_PEN_FORWARD);
			} else if(part == RIGHT_PEN) {
				if(TOWARDS[request.params.forward] === FORWARD) motoring.motion.write(0, HamsterS.MOTION_PIVOT_RIGHT_PEN_BACKWARD);
				else motoring.motion.write(0, HamsterS.MOTION_PIVOT_RIGHT_PEN_FORWARD);
			} else if(part == LEFT_WHEEL) {
				if(TOWARDS[request.params.forward] === FORWARD) motoring.motion.write(0, HamsterS.MOTION_PIVOT_LEFT_BACKWARD);
				else motoring.motion.write(0, HamsterS.MOTION_PIVOT_LEFT_FORWARD);
			} else {
				if(TOWARDS[request.params.forward] === FORWARD) motoring.motion.write(0, HamsterS.MOTION_PIVOT_RIGHT_BACKWARD);
				else motoring.motion.write(0, HamsterS.MOTION_PIVOT_RIGHT_FORWARD);
			}
		} else {
			motoring.motion.write(3, value);
			if(part === LEFT_PEN) {
				if(TOWARDS[request.params.forward] === FORWARD) motoring.motion.write(0, HamsterS.MOTION_PIVOT_LEFT_PEN_FORWARD);
				else motoring.motion.write(0, HamsterS.MOTION_PIVOT_LEFT_PEN_BACKWARD);
			} else if(part == RIGHT_PEN) {
				if(TOWARDS[request.params.forward] === FORWARD) motoring.motion.write(0, HamsterS.MOTION_PIVOT_RIGHT_PEN_FORWARD);
				else motoring.motion.write(0, HamsterS.MOTION_PIVOT_RIGHT_PEN_BACKWARD);
			} else if(part == LEFT_WHEEL) {
				if(TOWARDS[request.params.forward] === FORWARD) motoring.motion.write(0, HamsterS.MOTION_PIVOT_LEFT_FORWARD);
				else motoring.motion.write(0, HamsterS.MOTION_PIVOT_LEFT_BACKWARD);
			} else {
				if(TOWARDS[request.params.forward] === FORWARD) motoring.motion.write(0, HamsterS.MOTION_PIVOT_RIGHT_FORWARD);
				else motoring.motion.write(0, HamsterS.MOTION_PIVOT_RIGHT_BACKWARD);
			}
		}
		this.motionId = parseInt(request.params.id);
		this.waits.push(this.motionId);
		motoring.lineTracerMode.write(0);
	} else {
		motoring.motion.write(0);
		motoring.lineTracerMode.write(0);
	}
};

Robot.prototype.turnUnitWithRadiusInDirection = function(request) {
	var motoring = this.motoring;
	this.cancelBoard();
	this.cancelMotion();
	this.cancelLineTracer();
	
	motoring.leftWheel.write(0);
	motoring.rightWheel.write(0);
	motoring.pulse.write(0);
	var value = parseFloat(request.params.value);
	var radius = parseFloat(request.params.radius);
	if(value && (typeof radius == 'number') && radius >= 0) {
		var unit = UNITS[request.params.unit];
		if(unit === SECONDS) motoring.motion.write(1, HamsterS.UNIT_SEC);
		else if(unit === PULSES) motoring.motion.write(1, HamsterS.UNIT_PULSE);
		else motoring.motion.write(1, HamsterS.UNIT_CM_DEG);
		motoring.motion.write(2, 0);
		motoring.motion.write(4, radius);
		var part = PARTS[request.params.part];
		if(value < 0) {
			motoring.motion.write(3, -value);
			if(part === LEFT_PEN) {
				if(DIRECTIONS[request.params.direction] === LEFT) {
					if(TOWARDS[request.params.forward] === FORWARD) motoring.motion.write(0, HamsterS.MOTION_SWING_LEFT_PEN_LEFT_BACKWARD);
					else motoring.motion.write(0, HamsterS.MOTION_SWING_LEFT_PEN_LEFT_FORWARD);
				} else {
					if(TOWARDS[request.params.forward] === FORWARD) motoring.motion.write(0, HamsterS.MOTION_SWING_LEFT_PEN_RIGHT_BACKWARD);
					else motoring.motion.write(0, HamsterS.MOTION_SWING_LEFT_PEN_RIGHT_FORWARD);
				}
			} else if(part == RIGHT_PEN) {
				if(DIRECTIONS[request.params.direction] === LEFT) {
					if(TOWARDS[request.params.forward] === FORWARD) motoring.motion.write(0, HamsterS.MOTION_SWING_RIGHT_PEN_LEFT_BACKWARD);
					else motoring.motion.write(0, HamsterS.MOTION_SWING_RIGHT_PEN_LEFT_FORWARD);
				} else {
					if(TOWARDS[request.params.forward] === FORWARD) motoring.motion.write(0, HamsterS.MOTION_SWING_RIGHT_PEN_RIGHT_BACKWARD);
					else motoring.motion.write(0, HamsterS.MOTION_SWING_RIGHT_PEN_RIGHT_FORWARD);
				}
			} else {
				if(DIRECTIONS[request.params.direction] === LEFT) {
					if(TOWARDS[request.params.forward] === FORWARD) motoring.motion.write(0, HamsterS.MOTION_SWING_LEFT_BACKWARD);
					else motoring.motion.write(0, HamsterS.MOTION_SWING_LEFT_FORWARD);
				} else {
					if(TOWARDS[request.params.forward] === FORWARD) motoring.motion.write(0, HamsterS.MOTION_SWING_RIGHT_BACKWARD);
					else motoring.motion.write(0, HamsterS.MOTION_SWING_RIGHT_FORWARD);
				}
			}
		} else {
			motoring.motion.write(3, value);
			if(part === LEFT_PEN) {
				if(DIRECTIONS[request.params.direction] === LEFT) {
					if(TOWARDS[request.params.forward] === FORWARD) motoring.motion.write(0, HamsterS.MOTION_SWING_LEFT_PEN_LEFT_FORWARD);
					else motoring.motion.write(0, HamsterS.MOTION_SWING_LEFT_PEN_LEFT_BACKWARD);
				} else {
					if(TOWARDS[request.params.forward] === FORWARD) motoring.motion.write(0, HamsterS.MOTION_SWING_LEFT_PEN_RIGHT_FORWARD);
					else motoring.motion.write(0, HamsterS.MOTION_SWING_LEFT_PEN_RIGHT_BACKWARD);
				}
			} else if(part == RIGHT_PEN) {
				if(DIRECTIONS[request.params.direction] === LEFT) {
					if(TOWARDS[request.params.forward] === FORWARD) motoring.motion.write(0, HamsterS.MOTION_SWING_RIGHT_PEN_LEFT_FORWARD);
					else motoring.motion.write(0, HamsterS.MOTION_SWING_RIGHT_PEN_LEFT_BACKWARD);
				} else {
					if(TOWARDS[request.params.forward] === FORWARD) motoring.motion.write(0, HamsterS.MOTION_SWING_RIGHT_PEN_RIGHT_FORWARD);
					else motoring.motion.write(0, HamsterS.MOTION_SWING_RIGHT_PEN_RIGHT_BACKWARD);
				}
			} else {
				if(DIRECTIONS[request.params.direction] === LEFT) {
					if(TOWARDS[request.params.forward] === FORWARD) motoring.motion.write(0, HamsterS.MOTION_SWING_LEFT_FORWARD);
					else motoring.motion.write(0, HamsterS.MOTION_SWING_LEFT_BACKWARD);
				} else {
					if(TOWARDS[request.params.forward] === FORWARD) motoring.motion.write(0, HamsterS.MOTION_SWING_RIGHT_FORWARD);
					else motoring.motion.write(0, HamsterS.MOTION_SWING_RIGHT_BACKWARD);
				}
			}
		}
		this.motionId = parseInt(request.params.id);
		this.waits.push(this.motionId);
		motoring.lineTracerMode.write(0);
	} else {
		motoring.motion.write(0);
		motoring.lineTracerMode.write(0);
	}
};

Robot.prototype.changeBothWheelsBy = function(request) {
	var motoring = this.motoring;
	this.cancelBoard();
	this.cancelMotion();
	this.cancelLineTracer();
	
	var left = parseFloat(request.params.left);
	var right = parseFloat(request.params.right);
	if(typeof left == 'number') {
		left += motoring.leftWheel.read();
		motoring.leftWheel.write(left);
	}
	if(typeof right == 'number') {
		right += motoring.rightWheel.read();
		motoring.rightWheel.write(right);
	}
	motoring.pulse.write(0);
	motoring.motion.write(0);
	motoring.lineTracerMode.write(0);
};

Robot.prototype.setBothWheelsTo = function(request) {
	var motoring = this.motoring;
	this.cancelBoard();
	this.cancelMotion();
	this.cancelLineTracer();
	
	var left = parseFloat(request.params.left);
	var right = parseFloat(request.params.right);
	if(typeof left == 'number') {
		motoring.leftWheel.write(left);
	}
	if(typeof right == 'number') {
		motoring.rightWheel.write(right);
	}
	motoring.pulse.write(0);
	motoring.motion.write(0);
	motoring.lineTracerMode.write(0);
};

Robot.prototype.changeWheelBy = function(request) {
	var motoring = this.motoring;
	this.cancelBoard();
	this.cancelMotion();
	this.cancelLineTracer();
	
	var speed = parseFloat(request.params.speed);
	if(typeof speed == 'number') {
		var which = PARTS[request.params.which];
		if(which === LEFT) {
			speed += motoring.leftWheel.read();
			motoring.leftWheel.write(speed);
		} else if(which === RIGHT) {
			speed += motoring.rightWheel.read();
			motoring.rightWheel.write(speed);
		} else {
			var left = speed + motoring.leftWheel.read();
			var right = speed + motoring.rightWheel.read();
			motoring.leftWheel.write(left);
			motoring.rightWheel.write(right);
		}
	}
	motoring.pulse.write(0);
	motoring.motion.write(0);
	motoring.lineTracerMode.write(0);
};

Robot.prototype.setWheelTo = function(request) {
	var motoring = this.motoring;
	this.cancelBoard();
	this.cancelMotion();
	this.cancelLineTracer();
	
	var speed = parseFloat(request.params.speed);
	if(typeof speed == 'number') {
		var which = PARTS[request.params.which];
		if(which === LEFT) {
			motoring.leftWheel.write(speed);
		} else if(which === RIGHT) {
			motoring.rightWheel.write(speed);
		} else {
			motoring.leftWheel.write(speed);
			motoring.rightWheel.write(speed);
		}
	}
	motoring.pulse.write(0);
	motoring.motion.write(0);
	motoring.lineTracerMode.write(0);
};

Robot.prototype.followLineUsingFloorSensor = function(request) {
	var motoring = this.motoring;
	this.cancelBoard();
	this.cancelMotion();
	this.cancelLineTracer();
	
	var mode = 1;
	var which = PARTS[request.params.which];
	if(which === RIGHT)
		mode = 2;
	else if(which === BOTH)
		mode = 3;
	if(VALUES[request.params.color] === WHITE)
		mode += 7;
	
	motoring.leftWheel.write(0);
	motoring.rightWheel.write(0);
	motoring.pulse.write(0);
	motoring.motion.write(0);
	motoring.lineTracerMode.write(mode);
};

Robot.prototype.followLineUntilIntersection = function(request) {
	var motoring = this.motoring;
	this.cancelBoard();
	this.cancelMotion();
	
	var mode = 4;
	var which = DIRECTIONS[request.params.which];
	if(which === RIGHT)
		mode = 5;
	else if(which === FRONT)
		mode = 6;
	else if(which === REAR)
		mode = 7;
	if(VALUES[request.params.color] === WHITE)
		mode += 7;
	
	motoring.leftWheel.write(0);
	motoring.rightWheel.write(0);
	motoring.pulse.write(0);
	motoring.motion.write(0);
	this.lineTracerId = parseInt(request.params.id);
	this.waits.push(this.lineTracerId);
	motoring.lineTracerMode.write(mode);
};

Robot.prototype.setFollowingSpeedTo = function(request) {
	var speed = parseInt(request.params.speed);
	var gain = this.gain;
	if(gain < 0) gain = SPEED2GAIN[speed];
	if(speed && gain && speed > 0 && gain > 0) {
		this.speed = speed;
		var motoring = this.motoring;
		motoring.lineTracerSpeed.write(speed);
		motoring.lineTracerGain.write(gain);
	}
};

Robot.prototype.setFollowingGainTo = function(request) {
	var gain = parseInt(request.params.gain);
	if(gain && gain > 0) {
		this.gain = gain;
		this.motoring.lineTracerGain.write(gain);
	} else {
		this.gain = -1;
		gain = SPEED2GAIN[this.speed];
		if(gain && gain > 0) {
			this.motoring.lineTracerGain.write(gain);
		}
	}
};

Robot.prototype.stop = function(request) {
	var motoring = this.motoring;
	this.cancelBoard();
	this.cancelMotion();
	this.cancelLineTracer();
	
	motoring.leftWheel.write(0);
	motoring.rightWheel.write(0);
	motoring.pulse.write(0);
	motoring.motion.write(0);
	motoring.lineTracerMode.write(0);
};

Robot.prototype.setLedTo = function(request) {
	var motoring = this.motoring;
	var which = PARTS[request.params.which];
	var rgb = RGB_COLORS[request.params.color];
	if(rgb) {
		if(which === LEFT) {
			motoring.leftRgb.write(rgb);
		} else if(which === RIGHT) {
			motoring.rightRgb.write(rgb);
		} else {
			motoring.leftRgb.write(rgb);
			motoring.rightRgb.write(rgb);
		}
	}
};

Robot.prototype.changeLedByRGB = function(request) {
	var motoring = this.motoring;
	var which = PARTS[request.params.which];
	var red = parseInt(request.params.red);
	var green = parseInt(request.params.green);
	var blue = parseInt(request.params.blue);
	if(which === LEFT) {
		if(typeof red == 'number') {
			motoring.leftRgb.write(0, red + motoring.leftRgb.read(0));
		}
		if(typeof green == 'number') {
			motoring.leftRgb.write(1, green + motoring.leftRgb.read(1));
		}
		if(typeof blue == 'number') {
			motoring.leftRgb.write(2, blue + motoring.leftRgb.read(2));
		}
	} else if(which === RIGHT) {
		if(typeof red == 'number') {
			motoring.rightRgb.write(0, red + motoring.rightRgb.read(0));
		}
		if(typeof green == 'number') {
			motoring.rightRgb.write(1, green + motoring.rightRgb.read(1));
		}
		if(typeof blue == 'number') {
			motoring.rightRgb.write(2, blue + motoring.rightRgb.read(2));
		}
	} else {
		if(typeof red == 'number') {
			motoring.leftRgb.write(0, red + motoring.leftRgb.read(0));
			motoring.rightRgb.write(0, red + motoring.rightRgb.read(0));
		}
		if(typeof green == 'number') {
			motoring.leftRgb.write(1, green + motoring.leftRgb.read(1));
			motoring.rightRgb.write(1, green + motoring.rightRgb.read(1));
		}
		if(typeof blue == 'number') {
			motoring.leftRgb.write(2, blue + motoring.leftRgb.read(2));
			motoring.rightRgb.write(2, blue + motoring.rightRgb.read(2));
		}
	}
};

Robot.prototype.setLedToRGB = function(request) {
	var motoring = this.motoring;
	var which = PARTS[request.params.which];
	var red = parseInt(request.params.red);
	var green = parseInt(request.params.green);
	var blue = parseInt(request.params.blue);
	if(which === LEFT) {
		if(typeof red == 'number') {
			motoring.leftRgb.write(0, red);
		}
		if(typeof green == 'number') {
			motoring.leftRgb.write(1, green);
		}
		if(typeof blue == 'number') {
			motoring.leftRgb.write(2, blue);
		}
	} else if(which === RIGHT) {
		if(typeof red == 'number') {
			motoring.rightRgb.write(0, red);
		}
		if(typeof green == 'number') {
			motoring.rightRgb.write(1, green);
		}
		if(typeof blue == 'number') {
			motoring.rightRgb.write(2, blue);
		}
	} else {
		if(typeof red == 'number') {
			motoring.leftRgb.write(0, red);
			motoring.rightRgb.write(0, red);
		}
		if(typeof green == 'number') {
			motoring.leftRgb.write(1, green);
			motoring.rightRgb.write(1, green);
		}
		if(typeof blue == 'number') {
			motoring.leftRgb.write(2, blue);
			motoring.rightRgb.write(2, blue);
		}
	}
};

Robot.prototype.clearLed = function(request) {
	var motoring = this.motoring;
	var which = PARTS[request.params.which];
	if(which === LEFT) {
		motoring.leftRgb.write(0, 0);
		motoring.leftRgb.write(1, 0);
		motoring.leftRgb.write(2, 0);
	} else if(which === RIGHT) {
		motoring.rightRgb.write(0, 0);
		motoring.rightRgb.write(1, 0);
		motoring.rightRgb.write(2, 0);
	} else {
		motoring.leftRgb.write(0, 0);
		motoring.leftRgb.write(1, 0);
		motoring.leftRgb.write(2, 0);
		motoring.rightRgb.write(0, 0);
		motoring.rightRgb.write(1, 0);
		motoring.rightRgb.write(2, 0);
	}
};

Robot.prototype.beep = function(request) {
	var motoring = this.motoring;
	this.cancelNote();
	this.cancelSound();
	
	motoring.buzzer.write(0);
	motoring.note.write(0);
	this.runSound(HamsterS.SOUND_BEEP, 1);
	this.soundId = parseInt(request.params.id);
	this.waits.push(this.soundId);
};

Robot.prototype.playSound = function(request) {
	var motoring = this.motoring;
	this.cancelNote();
	this.cancelSound();
	
	var sound = SOUNDS[request.params.sound];
	motoring.buzzer.write(0);
	motoring.note.write(0);
	if(sound) {
		this.runSound(sound);
	} else {
		this.runSound(0);
	}
};

Robot.prototype.playSoundTimes = function(request) {
	var motoring = this.motoring;
	this.cancelNote();
	this.cancelSound();
	
	var sound = SOUNDS[request.params.sound];
	var count = parseInt(request.params.count);
	motoring.buzzer.write(0);
	motoring.note.write(0);
	if(sound && count) {
		this.runSound(sound, count);
	} else {
		this.runSound(0);
	}
};

Robot.prototype.playSoundTimesUntilDone = function(request) {
	var motoring = this.motoring;
	this.cancelNote();
	this.cancelSound();
	
	var sound = SOUNDS[request.params.sound];
	var count = parseInt(request.params.count);
	motoring.buzzer.write(0);
	motoring.note.write(0);
	if(count) {
		if(sound) {
			this.runSound(sound, count);
		} else {
			this.runSound(0);
		}
		this.soundId = parseInt(request.params.id);
		this.waits.push(this.soundId);
	} else {
		this.runSound(0);
	}
};

Robot.prototype.changeBuzzerBy = function(request) {
	var motoring = this.motoring;
	this.cancelNote();
	this.cancelSound();
	
	var hz = parseFloat(request.params.value);
	if(typeof hz == 'number') {
		hz += motoring.buzzer.read();
		motoring.buzzer.write(hz);
	}
	motoring.note.write(0);
	this.runSound(0);
};

Robot.prototype.setBuzzerTo = function(request) {
	var motoring = this.motoring;
	this.cancelNote();
	this.cancelSound();
	
	var hz = parseFloat(request.params.value);
	if(typeof hz == 'number') {
		motoring.buzzer.write(hz);
	}
	motoring.note.write(0);
	this.runSound(0);
};

Robot.prototype.clearBuzzer = function(request) {
	var motoring = this.motoring;
	this.cancelNote();
	this.cancelSound();
	
	motoring.buzzer.write(0);
	motoring.note.write(0);
	this.runSound(0);
};

Robot.prototype.clearSound = function(request) {
	var motoring = this.motoring;
	this.cancelNote();
	this.cancelSound();
	
	motoring.buzzer.write(0);
	motoring.note.write(0);
	this.runSound(0);
};

Robot.prototype.playNote = function(request) {
	var motoring = this.motoring;
	this.cancelNote();
	this.cancelSound();
	
	var note = NOTES[request.params.note];
	var octave = parseInt(request.params.octave);
	motoring.buzzer.write(0);
	if(note && octave && octave > 0 && octave < 8) {
		note += (octave - 1) * 12;
		motoring.note.write(note);
	} else {
		motoring.note.write(0);
	}
	this.runSound(0);
};

Robot.prototype.playNoteFor = function(request) {
	var self = this;
	var motoring = self.motoring;
	self.cancelNote();
	self.cancelSound();
	
	var note = NOTES[request.params.note];
	var octave = parseInt(request.params.octave);
	var beat = BEATS[request.params.beat];
	if(!beat) {
		beat = parseFloat(request.params.beat);
	}
	motoring.buzzer.write(0);
	if(note && octave && octave > 0 && octave < 8 && beat && beat > 0 && self.tempo > 0) {
		var id = self.noteId = parseInt(request.params.id);
		self.waits.push(id);
		note += (octave - 1) * 12;
		motoring.note.write(note);
		var timeout = beat * 60 * 1000 / self.tempo;
		var tail = 0;
		if(timeout > 100) {
			tail = 100;
		}
		if(tail > 0) {
			self.noteTimer1 = setTimeout(function() {
				if(self.noteId == id) {
					motoring.note.write(0);
					if(self.noteTimer1 !== undefined) self.removeTimeout(self.noteTimer1);
					self.noteTimer1 = undefined;
				}
			}, timeout - tail);
			self.timeouts.push(self.noteTimer1);
		}
		self.noteTimer2 = setTimeout(function() {
			if(self.noteId == id) {
				motoring.note.write(0);
				self.removeWait(self.noteId);
				self.cancelNote();
			}
		}, timeout);
		self.timeouts.push(self.noteTimer2);
	} else {
		motoring.note.write(0);
	}
	self.runSound(0);
};

Robot.prototype.restFor = function(request) {
	var self = this;
	var motoring = self.motoring;
	self.cancelNote();
	self.cancelSound();
	
	var beat = BEATS[request.params.beat];
	if(!beat) {
		beat = parseFloat(request.params.beat);
	}
	motoring.buzzer.write(0);
	motoring.note.write(0);
	self.runSound(0);
	if(beat && beat > 0 && self.tempo > 0) {
		var id = self.noteId = parseInt(request.params.id);
		self.waits.push(id);
		self.noteTimer1 = setTimeout(function() {
			if(self.noteId == id) {
				self.removeWait(self.noteId);
				self.cancelNote();
			}
		}, beat * 60 * 1000 / self.tempo);
		self.timeouts.push(self.noteTimer1);
	}
};

Robot.prototype.changeTempoBy = function(request) {
	var bpm = parseFloat(request.params.value);
	if(typeof bpm == 'number') {
		this.tempo += bpm;
		if(this.tempo < 1) this.tempo = 1;
	}
};

Robot.prototype.setTempoTo = function(request) {
	var bpm = parseFloat(request.params.value);
	if(typeof bpm == 'number') {
		this.tempo = bpm;
		if(this.tempo < 1) this.tempo = 1;
	}
};

Robot.prototype.runSound = function(sound, count) {
	if(typeof count != 'number') count = 1;
	if(count < 0) count = -1;
	if(count) {
		this.currentSound = sound;
		this.soundRepeat = count;
		this.motoring.sound.write(sound);
	}
};

Robot.prototype.setPortTo = function(request) {
	var motoring = this.motoring;
	this.cancelIo();
	var mode = IO_MODES[request.params.value];
	if(typeof mode == 'number') {
		if(request.params.which == 'A') {
			motoring.ioModeA.write(mode);
		} else if(request.params.which == 'B') {
			motoring.ioModeB.write(mode);
		} else {
			motoring.ioModeA.write(mode);
			motoring.ioModeB.write(mode);
		}
	}
};

Robot.prototype.changeOutputBy = function(request) {
	var motoring = this.motoring;
	this.cancelIo();
	var value = parseFloat(request.params.value);
	if(typeof value == 'number') {
		if(request.params.which == 'A') {
			value += motoring.outputA.read();
			motoring.outputA.write(value);
		} else if(request.params.which == 'B') {
			value += motoring.outputB.read();
			motoring.outputB.write(value);
		} else {
			var valueA = value + motoring.outputA.read();
			var valueB = value + motoring.outputB.read();
			motoring.outputA.write(valueA);
			motoring.outputB.write(valueB);
		}
	}
};

Robot.prototype.setOutputTo = function(request) {
	var motoring = this.motoring;
	this.cancelIo();
	var value = parseFloat(request.params.value);
	if(typeof value == 'number') {
		if(request.params.which == 'A') {
			motoring.outputA.write(value);
		} else if(request.params.which == 'B') {
			motoring.outputB.write(value);
		} else {
			motoring.outputA.write(value);
			motoring.outputB.write(value);
		}
	}
};

Robot.prototype.gripper = function(request) {
	var self = this;
	var motoring = self.motoring;
	self.cancelIo();
	
	var action = GRIPPERS[request.params.action];
	var id = self.ioId = parseInt(request.params.id);
	self.waits.push(id);
	
	motoring.ioModeA.write(HamsterS.IO_MODE_DIGITAL_OUTPUT);
	motoring.ioModeB.write(HamsterS.IO_MODE_DIGITAL_OUTPUT);
	if(action == OPEN) {
		motoring.outputA.write(1);
		motoring.outputB.write(0);
	} else {
		motoring.outputA.write(0);
		motoring.outputB.write(1);
	}
	self.ioTimer = setTimeout(function() {
		if(self.ioId == id) {
			self.removeWait(self.ioId);
			self.cancelIo();
		}
	}, 500);
	self.timeouts.push(self.ioTimer);
};

Robot.prototype.releaseGripper = function(request) {
	var motoring = this.motoring;
	this.cancelIo();
	motoring.ioModeA.write(HamsterS.IO_MODE_DIGITAL_OUTPUT);
	motoring.ioModeB.write(HamsterS.IO_MODE_DIGITAL_OUTPUT);
	motoring.outputA.write(0);
	motoring.outputB.write(0);
};

Robot.prototype.writeSerial = function(request) {
	var motoring = this.motoring;
	this.cancelIo();
	motoring.ioModeA.write(this.serialRate);
	motoring.ioModeB.write(this.serialRate);
	var mode = SERIAL_MODES[request.params.mode];
	if(typeof mode == 'number') {
		var queue = this.writeQueue;
		queue.push(request.params.value, mode == SERIAL_STRING_LINE);
		var data = queue.pop();
		if(data) {
			this.addSerialWriteWait(parseInt(request.params.id));
			motoring.writeSerial.write(data);
		}
	}
};

Robot.prototype.readSerialUntil = function(request) {
	var motoring = this.motoring;
	this.cancelIo();
	motoring.ioModeA.write(this.serialRate);
	motoring.ioModeB.write(this.serialRate);
	var delimiter = SERIAL_DELIMITERS[request.params.delimiter];
	if(typeof delimiter == 'number') {
		this.serialDelimiter = delimiter;
		this.addSerialReadWait(parseInt(request.params.id));
	}
};

Robot.prototype.setSerialRateTo = function(request) {
	var motoring = this.motoring;
	this.cancelIo();
	var baud = SERIAL_BAUDS[request.params.baud];
	if(baud) {
		this.serialRate = baud;
		motoring.ioModeA.write(baud);
		motoring.ioModeB.write(baud);
	}
};

Robot.prototype.handleSensory = function(data) {
	var self = this;
	self.received = true;
	self.active = true;
	var sensory = self.sensory;
	sensory.signalStrength = data.signalStrength;
	sensory.leftProximity = data.leftProximity;
	sensory.rightProximity = data.rightProximity;
	sensory.leftFloor = data.leftFloor;
	sensory.rightFloor = data.rightFloor;
	sensory.accelerationX = data.acceleration.x;
	sensory.accelerationY = data.acceleration.y;
	sensory.accelerationZ = data.acceleration.z;
	sensory.light = data.light;
	sensory.temperature = data.temperature;
	sensory.inputA = data.inputA;
	sensory.inputB = data.inputB;
	
	if(data.map1 & 0x00000008) {
		sensory.tilt = data.tilt;
	}
	if(data.map1 & 0x00000004) {
		sensory.batteryState = data.batteryState;
	}
	if(data.map2 & 0x00008000) {
		self.freeFall = true;
	}
	if(data.map2 & 0x00004000) {
		self.tap = true;
	}
	
	var motoring = self.motoring;
	if(self.lineTracerId >= 0 && (data.map1 & 0x00000010) != 0) {
		if(data.lineTracerState == 0x40) {
			motoring.lineTracerMode.write(0);
			self.removeWait(self.lineTracerId);
			self.cancelLineTracer();
		}
	}
	if(self.boardId >= 0) {
		if(self.boardCommand == 1) {
			switch(self.boardState) {
				case 1: {
					if(self.boardCount < 2) {
						if(data.leftFloor < 50 && data.rightFloor < 50)
							self.boardCount ++;
						else
							self.boardCount = 0;
						var diff = data.leftFloor - data.rightFloor;
						motoring.leftWheel.write(45 + diff * 0.25);
						motoring.rightWheel.write(45 - diff * 0.25);
					} else {
						self.boardCount = 0;
						self.boardState = 2;
					}
					break;
				}
				case 2: {
					var diff = data.leftFloor - data.rightFloor;
					motoring.leftWheel.write(45 + diff * 0.25);
					motoring.rightWheel.write(45 - diff * 0.25);
					self.boardState = 3;
					self.wheelTimer = setTimeout(function() {
						motoring.leftWheel.write(0);
						motoring.rightWheel.write(0);
						self.boardState = 4;
						if(self.wheelTimer !== undefined) self.removeTimeout(self.wheelTimer);
						self.wheelTimer = undefined;
					}, 250);
					self.timeouts.push(self.wheelTimer);
					break;
				}
				case 3: {
					var diff = data.leftFloor - data.rightFloor;
					motoring.leftWheel.write(45 + diff * 0.25);
					motoring.rightWheel.write(45 - diff * 0.25);
					break;
				}
				case 4: {
					motoring.leftWheel.write(0);
					motoring.rightWheel.write(0);
					self.removeWait(self.boardId);
					self.cancelBoard();
					break;
				}
			}
		} else if(self.boardCommand == 2) {
			switch(self.boardState) {
				case 1: {
					if(self.boardCount < 2) {
						if(data.leftFloor > 50)
							self.boardCount ++;
					} else {
						self.boardCount = 0;
						self.boardState = 2;
					}
					break;
				}
				case 2: {
					if(data.leftFloor < 20) {
						self.boardState = 3;
					}
					break;
				}
				case 3: {
					if(self.boardCount < 2) {
						if(data.leftFloor < 20)
							self.boardCount ++;
					} else {
						self.boardCount = 0;
						self.boardState = 4;
					}
					break;
				}
				case 4: {
					if(data.leftFloor > 50) {
						self.boardState = 5;
					}
					break;
				}
				case 5: {
					var diff = data.leftFloor - data.rightFloor;
					if(diff > -15) {
						motoring.leftWheel.write(0);
						motoring.rightWheel.write(0);
						self.removeWait(self.boardId);
						self.cancelBoard();
					} else {
						motoring.leftWheel.write(diff * 0.5);
						motoring.rightWheel.write(-diff * 0.5);
					}
					break;
				}
			}
		} else if(self.boardCommand == 3) {
			switch(self.boardState) {
				case 1: {
					if(self.boardCount < 2) {
						if(data.rightFloor > 50)
							self.boardCount ++;
					} else {
						self.boardCount = 0;
						self.boardState = 2;
					}
					break;
				}
				case 2: {
					if(data.rightFloor < 20) {
						self.boardState = 3;
					}
					break;
				}
				case 3: {
					if(self.boardCount < 2) {
						if(data.rightFloor < 20)
							self.boardCount ++;
					} else {
						self.boardCount = 0;
						self.boardState = 4;
					}
					break;
				}
				case 4: {
					if(data.rightFloor > 50) {
						self.boardState = 5;
					}
					break;
				}
				case 5: {
					var diff = data.rightFloor - data.leftFloor;
					if(diff > -15) {
						motoring.leftWheel.write(0);
						motoring.rightWheel.write(0);
						self.removeWait(self.boardId);
						self.cancelBoard();
					} else {
						motoring.leftWheel.write(-diff * 0.5);
						motoring.rightWheel.write(diff * 0.5);
					}
					break;
				}
			}
		}
	}
	if(self.motionId >= 0 && (data.map2 & 0x00000800) != 0) {
		if(data.wheelState == 2) {
			motoring.leftWheel.write(0);
			motoring.rightWheel.write(0);
			self.removeWait(self.motionId);
			self.cancelMotion();
		}
	}
	if((data.map2 & 0x00000400) != 0) {
		if(data.soundState == 0) {
			if(self.currentSound > 0) {
				if(self.soundRepeat < 0) {
					self.runSound(self.currentSound, -1);
				} else if(self.soundRepeat > 1) {
					self.soundRepeat --;
					self.runSound(self.currentSound, self.soundRepeat);
				} else {
					self.currentSound = 0;
					self.soundRepeat = 1;
					if(self.soundId >= 0) self.removeWait(self.soundId);
					self.cancelSound();
				}
			} else {
				self.currentSound = 0;
				self.soundRepeat = 1;
				if(self.soundId >= 0) self.removeWait(self.soundId);
				self.cancelSound();
			}
		}
	}
	if(data.map2 & 0x00002000) {
		if(data.serial) self.readQueue.push(data.serial);
	}
	if(data.map2 & 0x00000200) {
		var tmp = self.writeQueue.pop();
		if(tmp) {
			motoring.writeSerial.write(tmp);
		} else {
			self.removeSerialWriteWaits();
		}
	}
	if(self.serialReadWaits.length > 0) {
		var tmp = self.readQueue.pop(self.serialDelimiter);
		if(tmp) {
			sensory.serial = tmp;
			self.removeSerialReadWaits();
		}
	}
};

function Component() {
	var setup = function(map, key, value) {
		var data = translations[key];
		if(data) {
			if(value === undefined) value = key;
			for(var i in data) {
				map[data[i]] = value;
			}
			map[key] = value;
		}
	};
	setup(PARTS, 'left', LEFT);
	setup(PARTS, 'right', RIGHT);
	setup(PARTS, 'both', BOTH);
	
	setup(PARTS, 'left pen', LEFT_PEN);
	setup(PARTS, 'right pen', RIGHT_PEN);
	setup(PARTS, 'left wheel', LEFT_WHEEL);
	setup(PARTS, 'right wheel', RIGHT_WHEEL);
	
	setup(DIRECTIONS, 'left', LEFT);
	setup(DIRECTIONS, 'right', RIGHT);
	setup(DIRECTIONS, 'front', FRONT);
	setup(DIRECTIONS, 'rear', REAR);
	
	setup(TOWARDS, 'forward', FORWARD);
	setup(TOWARDS, 'backward', BACKWARD);
	
	setup(UNITS, 'seconds', SECONDS);
	setup(UNITS, 'pulses', PULSES);
	setup(UNITS, 'degrees', DEGREES);
	
	setup(RGB_COLORS, 'red', [255, 0, 0]);
	setup(RGB_COLORS, 'orange', [255, 63, 0]);
	setup(RGB_COLORS, 'yellow', [255, 255, 0]);
	setup(RGB_COLORS, 'green', [0, 255, 0]);
	setup(RGB_COLORS, 'cyan', [0, 255, 255]);
	setup(RGB_COLORS, 'sky blue', [0, 255, 255]);
	setup(RGB_COLORS, 'blue', [0, 0, 255]);
	setup(RGB_COLORS, 'magenta', [255, 0, 255]);
	setup(RGB_COLORS, 'violet', [63, 0, 255]);
	setup(RGB_COLORS, 'purple', [255, 0, 255]);
	setup(RGB_COLORS, 'white', [255, 255, 255]);
	
	setup(NOTES, 'C', HamsterS.NOTE_C_1);
	setup(NOTES, 'C#', HamsterS.NOTE_C_SHARP_1);
	setup(NOTES, 'C♯ (D♭)', HamsterS.NOTE_C_SHARP_1);
	setup(NOTES, 'D', HamsterS.NOTE_D_1);
	setup(NOTES, 'D♯ (E♭)', HamsterS.NOTE_E_FLAT_1);
	setup(NOTES, 'Eb', HamsterS.NOTE_E_FLAT_1);
	setup(NOTES, 'E', HamsterS.NOTE_E_1);
	setup(NOTES, 'F', HamsterS.NOTE_F_1);
	setup(NOTES, 'F#', HamsterS.NOTE_F_SHARP_1);
	setup(NOTES, 'F♯ (G♭)', HamsterS.NOTE_F_SHARP_1);
	setup(NOTES, 'G', HamsterS.NOTE_G_1);
	setup(NOTES, 'G#', HamsterS.NOTE_G_SHARP_1);
	setup(NOTES, 'G♯ (A♭)', HamsterS.NOTE_G_SHARP_1);
	setup(NOTES, 'A', HamsterS.NOTE_A_1);
	setup(NOTES, 'A♯ (B♭)', HamsterS.NOTE_B_FLAT_1);
	setup(NOTES, 'Bb', HamsterS.NOTE_B_FLAT_1);
	setup(NOTES, 'B', HamsterS.NOTE_B_1);
	
	setup(SOUNDS, 'beep', HamsterS.SOUND_BEEP);
	setup(SOUNDS, 'random beep', HamsterS.SOUND_RANDOM_BEEP);
	setup(SOUNDS, 'noise', HamsterS.SOUND_NOISE);
	setup(SOUNDS, 'siren', HamsterS.SOUND_SIREN);
	setup(SOUNDS, 'engine', HamsterS.SOUND_ENGINE);
	setup(SOUNDS, 'chop', HamsterS.SOUND_CHOP);
	setup(SOUNDS, 'robot', HamsterS.SOUND_ROBOT);
	setup(SOUNDS, 'dibidibidip', HamsterS.SOUND_DIBIDIBIDIP);
	setup(SOUNDS, 'good job', HamsterS.SOUND_GOOD_JOB);
	setup(SOUNDS, 'happy', HamsterS.SOUND_HAPPY);
	setup(SOUNDS, 'angry', HamsterS.SOUND_ANGRY);
	setup(SOUNDS, 'sad', HamsterS.SOUND_SAD);
	setup(SOUNDS, 'sleep', HamsterS.SOUND_SLEEP);
	setup(SOUNDS, 'march', HamsterS.SOUND_MARCH);
	setup(SOUNDS, 'birthday', HamsterS.SOUND_BIRTHDAY);
	
	setup(IO_MODES, 'analog input', HamsterS.IO_MODE_ANALOG_INPUT);
	setup(IO_MODES, 'digital input', HamsterS.IO_MODE_DIGITAL_INPUT);
	setup(IO_MODES, 'digital input (pull up)', HamsterS.IO_MODE_DIGITAL_INPUT_PULL_UP);
	setup(IO_MODES, 'digital input (pull down)', HamsterS.IO_MODE_DIGITAL_INPUT_PULL_DOWN);
	setup(IO_MODES, 'voltage input', HamsterS.IO_MODE_VOLTAGE_INPUT);
	setup(IO_MODES, 'servo output', HamsterS.IO_MODE_SERVO_OUTPUT);
	setup(IO_MODES, 'pwm output', HamsterS.IO_MODE_PWM_OUTPUT);
	setup(IO_MODES, 'digital output', HamsterS.IO_MODE_DIGITAL_OUTPUT);
	
	setup(GRIPPERS, 'open', OPEN);
	setup(GRIPPERS, 'close', CLOSE);
	
	setup(SERIAL_MODES, 'string', SERIAL_STRING);
	setup(SERIAL_MODES, 'string line', SERIAL_STRING_LINE);
	
	setup(SERIAL_DELIMITERS, 'all', SERIAL_ALL);
	setup(SERIAL_DELIMITERS, 'until new line', SERIAL_NEW_LINE);
	setup(SERIAL_DELIMITERS, 'until ,(comma)', SERIAL_COMMA);
	setup(SERIAL_DELIMITERS, 'until :(colon)', SERIAL_COLON);
	setup(SERIAL_DELIMITERS, 'until $', SERIAL_DOLLAR);
	setup(SERIAL_DELIMITERS, 'until #', SERIAL_SHARP);
	
	SERIAL_BAUDS["9600"] = HamsterS.IO_MODE_SERIAL_9600;
	SERIAL_BAUDS["14400"] = HamsterS.IO_MODE_SERIAL_14400;
	SERIAL_BAUDS["19200"] = HamsterS.IO_MODE_SERIAL_19200;
	SERIAL_BAUDS["28800"] = HamsterS.IO_MODE_SERIAL_28800;
	SERIAL_BAUDS["38400"] = HamsterS.IO_MODE_SERIAL_38400;
	SERIAL_BAUDS["57600"] = HamsterS.IO_MODE_SERIAL_57600;
	SERIAL_BAUDS["76800"] = HamsterS.IO_MODE_SERIAL_76800;
	SERIAL_BAUDS["115200"] = HamsterS.IO_MODE_SERIAL_115200;
	
	setup(VALUES, 'white', WHITE);
}

Component.prototype.updateLang = function(translator) {
	var lang = translator.getLanguage();
	
	var get = function(key) {
		var value = undefined;
		var data = translations[key];
		if(data) value = data[lang];
		return value || key;
	};
	
	BOOLEAN_TILT_FORWARD = '/' + get('tilt%20forward') + ' ';
	BOOLEAN_TILT_BACKWARD = '/' + get('tilt%20backward') + ' ';
	BOOLEAN_TILT_LEFT = '/' + get('tilt%20left') + ' ';
	BOOLEAN_TILT_RIGHT = '/' + get('tilt%20right') + ' ';
	BOOLEAN_TILT_FLIP = '/' + get('tilt%20flip') + ' ';
	BOOLEAN_TILT_NONE = '/' + get('not%20tilt') + ' ';
	BOOLEAN_FREE_FALL = '/' + get('free%20fall') + ' ';
	BOOLEAN_TAP = '/' + get('tap') + ' ';
	BOOLEAN_BATTERY_NORMAL = '/' + get('normal') + ' ';
	BOOLEAN_BATTERY_LOW = '/' + get('low') + ' ';
	BOOLEAN_BATTERY_EMPTY = '/' + get('empty') + ' ';
};

Component.prototype.getModule = function() {
	return MODULE;
};

Component.prototype.createRobot = function(index) {
	return new Robot(index);
};

module.exports = new Component();