'use strict';
var context = require('context');
var router = require('router');
var Turtle = require('./turtle');
const MODULE = 'turtle';
const ALIGNMENT_TYPE_TURN_RIGHT = 0;
const ALIGNMENT_TYPE_TURN_LEFT = 1;
const ALIGNMENT_TYPE_DISTANCE = 2;

const BATTERY_LEVEL_LOW = 3.65;
const BATTERY_LEVEL_EMPTY = 3.63;
const BATTERY_LEVEL_NO = 3.62;

const HEX_LEDS = [
	['#ff0000', '#ffff00'], ['#ff0000', '#00ff00'], ['#ff0000', '#00ffff'], ['#ff0000', '#0000ff'], ['#ff0000', '#ff00ff'], ['#ff0000', '#ffffff'],
	['#ffff00', '#00ff00'], ['#ffff00', '#00ffff'], ['#ffff00', '#0000ff'], ['#ffff00', '#ff00ff'], ['#ffff00', '#ffffff'],
	['#00ff00', '#00ffff'], ['#00ff00', '#0000ff'], ['#00ff00', '#ff00ff'], ['#00ff00', '#ffffff'],
	['#00ffff', '#0000ff'], ['#00ffff', '#ff00ff'], ['#00ffff', '#ffffff'],
	['#0000ff', '#ff00ff'], ['#0000ff', '#ffffff'],
	['#ff00ff', '#ffffff']
];
const RGB_LEDS = [
	[[255, 0, 0], [255, 255, 0]], [[255, 0, 0], [0, 255, 0]], [[255, 0, 0], [0, 255, 255]], [[255, 0, 0], [0, 0, 255]], [[255, 0, 0], [255, 0, 255]], [[255, 0, 0], [255, 255, 255]],
	[[255, 255, 0], [0, 255, 0]], [[255, 255, 0], [0, 255, 255]], [[255, 255, 0], [0, 0, 255]], [[255, 255, 0], [255, 0, 255]], [[255, 255, 0], [255, 255, 255]],
	[[0, 255, 0], [0, 255, 255]], [[0, 255, 0], [0, 0, 255]], [[0, 255, 0], [255, 0, 255]], [[0, 255, 0], [255, 255, 255]],
	[[0, 255, 255], [0, 0, 255]], [[0, 255, 255], [255, 0, 255]], [[0, 255, 255], [255, 255, 255]],
	[[0, 0, 255], [255, 0, 255]], [[0, 0, 255], [255, 255, 255]],
	[[255, 0, 255], [255, 255, 255]]
];

function Controller() {
	this.sensory = {
		map: 0,
		signalStrength: 0,
		color: { r: 0, g: 0, b: 0, c: 0 },
		floor: 0,
		acceleration: { x: 0, y: 0, z: 0 },
		temperature: 0,
		button: 0,
		colorNumber: -1,
		colorPattern: -1,
		pulseCount: 0,
		wheelState: 0,
		soundState: 0,
		lineTracerState: 0,
		tilt: 0,
		batteryState: 2
	};
	this.motoring = {
		map: 0,
		leftWheel: 0,
		rightWheel: 0,
		led: { r: 0, g: 0, b: 0 },
		buzzer: 0,
		pulse: 0,
		note: 0,
		sound: 0,
		lineTracerMode: 0,
		lineTracerGain: 5,
		lineTracerSpeed: 5,
		lamp: 1,
		lock: 0,
		motion: { type: 0, unit: 0, speed: 0, value: 0, radius: 0 }
	};
	this.motion = {
		type: 0,
		speed: 0
	};
	this.acceleration = {
		x: new Array(10),
		y: new Array(10),
		z: new Array(10),
		sumx: 0.0,
		sumy: 0.0,
		sumz: 0.0,
		index: 0,
		count: 0
	};
	this.button = {
		clickId: -1,
		longPressId: -1
	};
	this.wheel = {
		id: 0,
		pulse: 0,
		pulsePrev: -1,
		event: 0,
		state: 0,
		count: 0,
		next: false,
		nextLeftWheel: 0,
		nextRightWheel: 0,
		nextPulse: 0
	};
	this.sound = {
		flag: 0,
		event: 0,
		state: 0,
		count: 0
	};
	this.lineTracer = {
		flag: 0,
		event: 0,
		state: 0,
		count: 0
	};
	this.event = {
		button: -1,
		colorNumber: -2,
		colorPattern: -1,
		pulseCount: -1,
		tilt: -4,
		batteryState: -1
	};
	this.alignment = {
		active: false,
		moveState: 0,
		type: 0,
		distance: Turtle.WHEEL_DISTANCE,
		saveState: 0,
		saveCount: 0,
		spinRight: 0,
		pivotRight: 0,
		spinLeft: 0,
		pivotLeft: 0,
		offset: 0,
		delta: 0
	};
	this.battery = {
		data: new Array(10),
		sum: 0.0,
		index: 0,
		count: 0
	};
	this.timerId = undefined;
	this.sensorMonitor = {
		count: 0,
		index: 0,
		duration: 0,
		averageIndex: 0,
		signalStrength: { current: 0, sum: 0.0, values: new Array(100) },
		floor: { current: 0, sum: 0.0, values: new Array(100) },
		accelerationX: { current: 0, sum: 0.0, values: new Array(100) },
		accelerationY: { current: 0, sum: 0.0, values: new Array(100) },
		accelerationZ: { current: 0, sum: 0.0, values: new Array(100) },
		temperature: { current: 0, sum: 0.0, values: new Array(100) },
		button: { current: 0, sum: 0.0, values: new Array(100) },
		colorNumber: { current: 0, sum: 0.0, values: new Array(100) },
		reset: function() {
			this.signalStrength.current = 0;
			this.signalStrength.sum = 0.0;
			this.floor.current = 0;
			this.floor.sum = 0.0;
			this.accelerationX.current = 0;
			this.accelerationX.sum = 0.0;
			this.accelerationY.current = 0;
			this.accelerationY.sum = 0.0;
			this.accelerationZ.current = 0;
			this.accelerationZ.sum = 0.0;
			this.temperature.current = 0;
			this.temperature.sum = 0.0;
			this.button.current = 0;
			this.button.sum = 0.0;
			this.colorNumber.current = 0;
			this.colorNumber.sum = 0.0;
			this.count = 0;
			this.index = 0;
			this.averageIndex = 0;
		},
		put: function(device, value) {
			if(this.duration > 0) {
				device.sum += value;
				if(this.averageIndex >= this.duration - 1) {
					device.current = parseInt(device.sum * 100 / this.duration) / 100;
					device.sum = 0.0;
					device.values[this.index] = device.current;
				}
			} else {
				device.current = value;
				device.values[this.index] = value;
			}
		},
		update: function() {
			if(this.duration > 0) {
				++ this.averageIndex;
				if(this.averageIndex >= this.duration) {
					this.averageIndex = 0;
					if(this.count < 100) ++ this.count;
					++ this.index;
					if(this.index >= 100) this.index = 0;
				}
			} else {
				if(this.count < 100) ++ this.count;
				++ this.index;
				if(this.index >= 100) this.index = 0;
			}
		}
	};
	this.sensorNames = [
		'floor sensor', 'button and color number', 'acceleration sensor', 'signal strength and temperature sensor'
	];
	this.identityCount = 0;
	this.updateIdentity();
}

Controller.prototype.updateIdentity = function() {
	this.identity = Math.floor(Math.random() * 21); // 0 ~ 20
};

Controller.prototype.getIdentitySize = function() {
	return 2;
};

Controller.prototype.getIdentityHex = function(index) {
	if(index == 0) return HEX_LEDS[this.identity][0];
	else return HEX_LEDS[this.identity][1];
};

Controller.prototype.getIdentityRgb = function(index) {
	if(index == 0) return RGB_LEDS[this.identity][0];
	else return RGB_LEDS[this.identity][1];
};

Controller.prototype.getIdentityX = function(index) {
	return (index == 0) ? '27px' : '41px'
};

Controller.prototype.getIdentityY = function(index) {
	return (index == 0) ? '71px' : '71px'
};

Controller.prototype.toHex = function(number) {
	var value = parseInt(number);
	if(value < 0) value += 0x100;

	value = value.toString(16).toUpperCase();
	if(value.length > 1) return value;
	else return '0' + value;
};

Controller.prototype.toHex2 = function(number) {
	var value = parseInt(number);
	if(value < 0) value += 0x10000;
	
	value = value.toString(16).toUpperCase();
	var result = '';
	for(var i = value.length; i < 4; ++i) {
		result += '0';
	}
	return result + value;
};
	
Controller.prototype.toHex3 = function(number) {
	var value = parseInt(number);
	if(value < 0) value += 0x1000000;
	
	value = value.toString(16).toUpperCase();
	var result = '';
	for(var i = value.length; i < 6; ++i) {
		result += '0';
	}
	return result + value;
};

Controller.prototype.calculateInnerSpeed = function(speed, radius) {
	return speed * (radius - this.alignment.distance) / (radius + this.alignment.distance);
};

Controller.prototype.calculateSwingPulse = function(deg, radius, offset) {
	return Math.round(deg * (Turtle.DEG_TO_PULSE + offset) * (radius + this.alignment.distance) / 360.0 / this.alignment.distance);
};

Controller.prototype.cancelTimeout = function() {
	if(this.timerId !== undefined) clearTimeout(this.timerId);
	this.timerId = undefined;
};

Controller.prototype.load = function(connector, route, callback) {
	var self = this;
	var loading = true;
	var loaded = false;
	var count = 0;
	connector.load(function(error, data) {
		if(loading) {
			if(error) {
				loading = false;
				callback();
			} else if(loaded) {
				var data1 = data.slice(0, 1);
				var data2 = data.slice(1, 2);
				if(data1 == 'B' && data2 == '0') {
					self.parseAlignment(data);
					connector.send('7000000000000000000000000000000000004400-' + route.address + '\r');
				} else if(data1 == '1') {
					loading = false;
					connector.unload();
					callback();
				}
			} else {
				if(data.slice(0, 2) == 'B0') {
					self.parseAlignment(data);
					connector.send('7000000000000000000000000000000000004400-' + route.address + '\r');
					loaded = true;
				} else {
					if(count % 10 == 0) {
						connector.send('B000000000000000000000000000000000004400-' + route.address + '\r');
					}
					++count;
				}
			}
		}
	});
};

Controller.prototype.parseAlignment = function(data) {
	var alignment = this.alignment;
	var str = data.slice(2, 4);
	var value = parseInt(str, 16);
	if(value > 0x7f) value -= 0x100;
	alignment.spinRight = value + 1;
	str = data.slice(4, 6);
	value = parseInt(str, 16);
	if(value > 0x7f) value -= 0x100;
	alignment.pivotRight = value + 1;
	str = data.slice(6, 8);
	value = parseInt(str, 16);
	if(value > 0x7f) value -= 0x100;
	alignment.spinLeft = value + 1;
	str = data.slice(8, 10);
	value = parseInt(str, 16);
	if(value > 0x7f) value -= 0x100;
	alignment.pivotLeft = value + 1;
	str = data.slice(10, 12);
	value = parseInt(str, 16);
	if(value > 0x7f) value -= 0x100;
	alignment.offset = (value + 1) * 256;
	str = data.slice(12, 14);
	value = parseInt(str, 16);
	if(value > 0x7f) value -= 0x100;
	alignment.offset += value + 1;
	alignment.distance = Turtle.WHEEL_DISTANCE + alignment.offset / 10000.0;
	if(alignment.distance <= 0) alignment.distance = Turtle.WHEEL_DISTANCE;
	alignment.delta = 0;
};

Controller.prototype.handleLocalData = function(data) { // data: string
	if(data.length != 53) return;
	
	var alignment = this.alignment;
	if(alignment.active) {
		if(alignment.saveState == 2) { // wait for 250ms
			if(++alignment.saveCount > 12) {
				alignment.saveCount = 0;
				alignment.saveState = 3; // send B0
			}
		} else if(alignment.saveState == 4) { // parse
			if(data.slice(0, 2) == 'B0') {
				this.parseAlignment(data);
				alignment.saveState = 5; // send 70
			} else {
				if(alignment.saveCount % 10 == 0) alignment.saveState = 3; // send B0
				++alignment.saveCount;
			}
		} else if(alignment.saveState == 6) { // end
			var data1 = data.slice(0, 1);
			var data2 = data.slice(1, 2);
			if(data1 == 'B' && data2 == '0') {
				this.parseAlignment(data);
				alignment.saveState = 5; // send 70
			} else if(data1 == '1') {
				alignment.saveState = 0;
				if(alignment.saveCallback) alignment.saveCallback();
			}
		}
	}
	
	var str = data.slice(0, 1);
	var value = parseInt(str, 16);
	if(value != 1) return; // invalid data
	
	var sensory = this.sensory;
	sensory.map = 0x8003e000;

	var sensorMonitor = this.sensorMonitor;
	// r, g, b
	str = data.slice(2, 6);
	var red = parseInt(str, 16);
	str = data.slice(6, 10);
	var green = parseInt(str, 16);
	str = data.slice(10, 14);
	var blue = parseInt(str, 16);
	str = data.slice(14, 16);
	value = parseInt(str, 16);
	
	var r = parseInt(red * 255 / 1023);
	var g = parseInt(green * 255 / 1023);
	var b = parseInt(blue * 255 / 1023);
	if(r > 255) r = 255;
	else if(r < 0) r = 0;
	if(g > 255) g = 255;
	else if(g < 0) g = 0;
	if(b > 255) b = 255;
	else if(b < 0) b = 0;
	
	sensory.color.r = r;
	sensory.color.g = g;
	sensory.color.b = b;
	sensory.color.c = value;
	
	// color number
	var colorNumber = Turtle.COLOR_NONE;
	if(value < 75) {
		if(red > 600 && green > 600 && blue > 600) {
			colorNumber = Turtle.COLOR_WHITE;
		} else if(green < red * 3/10 && blue < red * 3/10 && red > 350) {
			colorNumber = Turtle.COLOR_RED;
		} else if(blue < red * 5/10 && blue < green * 6/10 && red > 300 && green > 250) {
			if(green < red * 15/20) {
				colorNumber = Turtle.COLOR_ORANGE;
			} else {
				colorNumber = Turtle.COLOR_YELLOW;
			}
		} else if(red < green * 6/10 && blue < green * 6/10 && green > 300) {
			colorNumber = Turtle.COLOR_GREEN;
		} else if(red < blue * 6/10 && green < blue * 7/10 && blue > 250) {
			colorNumber = Turtle.COLOR_BLUE;
		} else if(red < green * 6/10 && red < blue * 6/10 && blue > 250 && green > 250) {
			colorNumber = Turtle.COLOR_CYAN;
		} else if(green < red * 7/10 && green < blue * 7/10 && red > 250) {
			colorNumber = Turtle.COLOR_MAGENTA;
		} else if(red < 300 && green < 300 && blue < 250) {
			colorNumber = Turtle.COLOR_BLACK;
		}
	}
	var event = this.event;
	if(colorNumber != event.colorNumber) {
		if(event.colorNumber != -2) {
			sensory.map |= 0x00000100;
		}
		event.colorNumber = colorNumber;
		sensory.colorNumber = colorNumber;
	}
	sensorMonitor.put(sensorMonitor.colorNumber, colorNumber);
	
	// clicked / double clicked / long pressed
	str = data.slice(16, 18);
	value = parseInt(str, 16);
	var clickId = (value >> 2) & 0x03;
	var longPressId = (value >> 4) & 0x03;
	var button = this.button;
	if(button.clickId < 0) {
		button.clickId = clickId;
	} else if(clickId != button.clickId) {
		button.clickId = clickId;
		value = value & 0x03;
		if(value == 1) {
			sensory.map |= 0x00000800;
		} else if(value == 2) {
			sensory.map |= 0x00000400;
		}
	}
	if(button.longPressId < 0) {
		button.longPressId = longPressId;
	} else if(longPressId != button.longPressId) {
		button.longPressId = longPressId;
		sensory.map |= 0x00000200;
	}
	
	// floor
	str = data.slice(18, 20);
	value = parseInt(str, 16);
	sensory.floor = value;
	sensorMonitor.put(sensorMonitor.floor, value);
	// color pattern / button
	str = data.slice(20, 22);
	value = parseInt(str, 16);
	var pattern = (value >> 1) & 0x7f;
	if(event.colorPattern < 0) {
		event.colorPattern = pattern;
	} else if(pattern != event.colorPattern) {
		event.colorPattern = pattern;
		var pattern1 = (pattern >> 3) & 0x07;
		if(pattern1 > 1) pattern1++;
		var pattern2 = pattern & 0x07;
		if(pattern2 > 1) pattern2++;
		sensory.colorPattern = pattern1 * 10 + pattern2;
		sensory.map |= 0x00000080;
	}
	value = value & 0x01;
	if(value != event.button) {
		if(event.button != -1) {
			sensory.map |= 0x00001000;
		}
		event.button = value;
		sensory.button = value;
	}
	sensorMonitor.put(sensorMonitor.button, value);
	// pulse count
	str = data.slice(22, 26);
	value = parseInt(str, 16);
	if(value != event.pulseCount) {
		if(event.pulseCount != -1) {
			sensory.map |= 0x00000040;
		}
		event.pulseCount = value;
		sensory.pulseCount = value;
	}
	
	var acc = this.acceleration;
	if(acc.count < 10) {
		++ acc.count;
	} else {
		acc.index %= 10;
		acc.sumx -= acc.x[acc.index];
		acc.sumy -= acc.y[acc.index];
		acc.sumz -= acc.z[acc.index];
	}
	// acceleration x
	str = data.slice(26, 28);
	value = parseInt(str, 16);
	if(value > 0x7f) value -= 0x100;
	value *= 256;
	acc.sumx += value;
	acc.x[acc.index] = value;
	// acceleration y
	str = data.slice(28, 30);
	value = parseInt(str, 16);
	if(value > 0x7f) value -= 0x100;
	value *= 256;
	acc.sumy += value;
	acc.y[acc.index] = value;
	// acceleration z
	str = data.slice(30, 32);
	value = parseInt(str, 16);
	if(value > 0x7f) value -= 0x100;
	value *= 256;
	acc.sumz += value;
	acc.z[acc.index] = value;
	++ acc.index;
	var acceleration = sensory.acceleration;
	acceleration.x = Math.round(acc.sumx / acc.count);
	acceleration.y = Math.round(acc.sumy / acc.count);
	acceleration.z = Math.round(acc.sumz / acc.count);
	sensorMonitor.put(sensorMonitor.accelerationX, acceleration.x);
	sensorMonitor.put(sensorMonitor.accelerationY, acceleration.y);
	sensorMonitor.put(sensorMonitor.accelerationZ, acceleration.z);
	// tilt
	if(acceleration.z < 8192 && acceleration.x > 8192 && acceleration.y > -4096 && acceleration.y < 4096) value = 1;
	else if(acceleration.z < 8192 && acceleration.x < -8192 && acceleration.y > -4096 && acceleration.y < 4096) value = -1;
	else if(acceleration.z < 8192 && acceleration.y > 8192 && acceleration.x > -4096 && acceleration.x < 4096) value = 2;
	else if(acceleration.z < 8192 && acceleration.y < -8192 && acceleration.x > -4096 && acceleration.x < 4096) value = -2;
	else if(acceleration.z > 12288 && acceleration.x > -8192 && acceleration.x < 8192 && acceleration.y > -8192 && acceleration.y < 8192) value = 3;
	else if(acceleration.z < -12288 && acceleration.x > -4096 && acceleration.x < 4096 && acceleration.y > -4096 && acceleration.y < 4096) value = -3;
	else value = 0;
	if(value != event.tilt) {
		if(event.tilt != -4) {
			sensory.map |= 0x00000004;
		}
		event.tilt = value;
		sensory.tilt = value;
	}
	
	// temperature
	str = data.slice(32, 34);
	value = parseInt(str, 16);
	if(value > 0x7f) value -= 0x100;
	value = value / 2.0 + 24;
	sensory.temperature = parseInt(value);
	sensorMonitor.put(sensorMonitor.temperature, sensory.temperature);
	// signal strength
	str = data.slice(34, 36);
	value = parseInt(str, 16);
	value -= 0x100;
	sensory.signalStrength = value;
	sensorMonitor.put(sensorMonitor.signalStrength, value);
	// battery
	str = data.slice(36, 38);
	value = parseInt(str, 16);
	value = value / 100.0 + 2;
	// battery state
	var batt = this.battery;
	if(batt.count < 10) {
		++ batt.count;
	} else {
		batt.index %= 10;
		batt.sum -= batt.data[batt.index];
	}
	batt.sum += value;
	batt.data[batt.index] = value;
	++ batt.index;
	value = batt.sum / batt.count;
	var state = 2;
	if(value < BATTERY_LEVEL_EMPTY) state = 0;
	else if(value < BATTERY_LEVEL_LOW) state = 1;
	if(state != event.batteryState) {
		if(event.batteryState != -1) {
			sensory.map |= 0x00000002;
		}
		event.batteryState = state;
		sensory.batteryState = state;
	}
	// wheel state / sound state / linetracer state
	str = data.slice(38, 40);
	value = parseInt(str, 16);
	state = (value >> 4) & 0x01;
	var wheel = this.wheel;
//	console.log('target pulse = ' + wheel.pulse + ' pulse count = ' + sensory.pulseCount + ' state = ' + state);
//	if(wheel.event == 1) {
//		if(state == 0) {
//			if(++wheel.count > 5) wheel.event = 2;
//		} else {
//			wheel.event = 2;
//		}
//	}
	if(wheel.event == 1) {
		//if(state != wheel.state || wheel.count > 5) {
			if(state == 0) {
				if(wheel.pulse != sensory.pulseCount) {
					if(wheel.pulse < 12) {
						if(++wheel.count < 10) state = 1;
					} else {
						state = 1;
					}
				}
			} else if(state == 1 && wheel.pulse == sensory.pulseCount) {
				state = 0;
			}
			if(state != wheel.state) {
				wheel.state = state;
				sensory.wheelState = state;
				sensory.map |= 0x00000020;
			}
			if(state == 0) {
				wheel.event = 0;
				wheel.count = 0;
				this.motion.type = Turtle.MOTION_NONE;
//				console.log('DONE');
				if(alignment.active && alignment.moveState > 0) {
					switch(alignment.moveState) {
						case 2:
						case 4:
						case 6:
						case 8:
						case 10:
						case 12:
						case 14:
						case 16:
							alignment.moveState ++;
							break;
						case 18:
							alignment.moveState = 0;
							if(alignment.moveCallback) alignment.moveCallback();
							break;
						case 102:
						case 104:
						case 106:
							alignment.moveState ++;
							break;
						case 108:
							alignment.moveState = 0;
							if(alignment.moveCallback) alignment.moveCallback();
							break;
					}
				}
			}
		//}
	}
	if(wheel.event == -1) {
		wheel.state = state;
		sensory.wheelState = 0;
		sensory.map |= 0x00000020;
		wheel.event = 0;
		wheel.count = 0;
		this.motion.type = Turtle.MOTION_NONE;
	}
	state = value & 0x03;
	if(state > 1) state = 1;
	var sound = this.sound;
	if(sound.event == 1) {
		if(state == 0) {
			if(++sound.count > 5) sound.event = 2;
		} else {
			sound.event = 2;
		}
	}
	if(sound.event == 2) {
		if(state != sound.state || sound.count > 5) {
			sound.state = state;
			sensory.soundState = state;
			sensory.map |= 0x00000010;
			if(state == 0) {
				sound.event = 0;
				sound.count = 0;
			}
		}
	}
	state = (value >> 2) & 0x03;
	if((state & 0x02) != 0) {
		var lineTracer = this.lineTracer;
		if(lineTracer.event == 1) {
			if(state == 0x02) {
				if(++lineTracer.count > 5) lineTracer.event = 2;
			} else {
				lineTracer.event = 2;
			}
		}
		if(lineTracer.event == 2) {
			if(state != lineTracer.state || lineTracer.count > 5) {
				lineTracer.state = state;
				sensory.lineTracerState = state;
				sensory.map |= 0x00000008;
				if(state == 0x02) {
					lineTracer.event = 0;
					lineTracer.count = 0;
				}
			}
		}
	}
	sensorMonitor.update();
	return sensory;
};

Controller.prototype.write = function(data) {
	if(!data) return this.motoring;
	
	var map = data.map;
	var writeData = data.writeData;
	var motoring = this.motoring;
	// left wheel
	var t = motoring.leftWheel + writeData[0][0];
	if(t < -400) t = -400;
	else if(t > 400) t = 400;
	motoring.leftWheel = t;
	// right wheel
	t = motoring.rightWheel + writeData[1][0];
	if(t < -400) t = -400;
	else if(t > 400) t = 400;
	motoring.rightWheel = t;
	// led
	t = motoring.led.r + writeData[2][0];
	if(t < 0) t = 0;
	else if(t > 255) t = 255;
	motoring.led.r = t;
	t = motoring.led.g + writeData[2][1];
	if(t < 0) t = 0;
	else if(t > 255) t = 255;
	motoring.led.g = t;
	t = motoring.led.b + writeData[2][2];
	if(t < 0) t = 0;
	else if(t > 255) t = 255;
	motoring.led.b = t;
	// buzzer;
	t = writeData[3][0];
	if(t == 0) {
		motoring.buzzer = 0;
	} else {
		t = motoring.buzzer + t;
		if(t < 0.01) t = 0.01;
		else if(t > 167772.15) t = 167772.15;
		motoring.buzzer = t;
	}
	// commands
	if(map & 0x04000000) motoring.pulse = writeData[4][0];
	if(map & 0x02000000) motoring.note = writeData[5][0];
	if(map & 0x01000000) motoring.sound = writeData[6][0];
	if(map & 0x00800000) motoring.lineTracerMode = writeData[7][0];
	if(map & 0x00400000) motoring.lineTracerGain = writeData[8][0];
	if(map & 0x00200000) motoring.lineTracerSpeed = writeData[9][0];
	if(map & 0x00100000) motoring.lamp = writeData[10][0];
	if(map & 0x00080000) motoring.lock = writeData[11][0];
	if(map & 0x00040000) {
		motoring.motion.type = parseInt(writeData[12][0]);
		motoring.motion.unit = parseInt(writeData[12][1]);
		motoring.motion.speed = writeData[12][2];
		motoring.motion.value = writeData[12][3];
		motoring.motion.radius = writeData[12][4];
	}
	motoring.map |= map;
	return motoring;
};

Controller.prototype.requestLocalData = function(address) {
	var self = this;
	var alignment = self.alignment;
	if(alignment.active) {
		self.cancelTimeout();
		if(alignment.saveState > 0) {
			if(alignment.saveState == 1) { // send C0
				alignment.saveState = 2; // wait 250ms
				var spinRight = 0, pivotRight = 0, spinLeft = 0, pivotLeft = 0, offset = 0;
				if(alignment.type == ALIGNMENT_TYPE_TURN_RIGHT) {
					spinRight = alignment.spinRight + alignment.delta;
					pivotRight = alignment.pivotRight + alignment.delta * 2;
					spinLeft = alignment.spinLeft;
					pivotLeft = alignment.pivotLeft;
					offset = alignment.offset;
				} else if(alignment.type == ALIGNMENT_TYPE_TURN_LEFT) {
					spinRight = alignment.spinRight;
					pivotRight = alignment.pivotRight;
					spinLeft = alignment.spinLeft + alignment.delta;
					pivotLeft = alignment.pivotLeft + alignment.delta * 2;
					offset = alignment.offset;
				} else if(alignment.type == ALIGNMENT_TYPE_DISTANCE) {
					spinRight = alignment.spinRight;
					pivotRight = alignment.pivotRight;
					spinLeft = alignment.spinLeft;
					pivotLeft = alignment.pivotLeft;
					offset = alignment.delta;
				}
				var offset1 = parseInt(offset / 256);
				var offset2 = parseInt(offset % 256);
				var str = 'C0';
				str += self.toHex(spinRight - 1);
				str += self.toHex(pivotRight - 1);
				str += self.toHex(spinLeft - 1);
				str += self.toHex(pivotLeft - 1);
				str += self.toHex(offset1 - 1);
				str += self.toHex(offset2 - 1);
				str += 'FFFFFFFFFFFFFFFFFFFFFFFFFF-';
				str += address;
				str += '\r';
				return str;
			} else if(alignment.saveState == 3) { // send B0
				alignment.saveState = 4; // parsing
				return 'B000000000000000000000000000000000004400-' + address + '\r';
			} else if(alignment.saveState == 5) { // send 70
				alignment.saveState = 6; // end
				return '7000000000000000000000000000000000004400-' + address + '\r';
			}
		} else {
			var leftSpeed = 0, rightSpeed = 0;
			var wheel = self.wheel;
			switch(alignment.moveState) {
				case 1:
				case 5:
				case 9:
				case 13:
				case 17:
					alignment.moveState ++;
					wheel.pulse = Math.round(5 * Turtle.CM_TO_PULSE);
					wheel.id = (wheel.id % 255) + 1;
					wheel.count = 0;
					wheel.event = 1;
				case 2:
				case 6:
				case 10:
				case 14:
				case 18:
					leftSpeed = 50;
					rightSpeed = 50;
					break;
				case 3:
				case 7:
				case 11:
				case 15:
					alignment.moveState ++;
					if(alignment.type == ALIGNMENT_TYPE_TURN_RIGHT) {
						wheel.pulse = Math.round(144 * (Turtle.DEG_TO_PULSE + alignment.spinRight + alignment.delta) / 360.0);
					} else if(alignment.type == ALIGNMENT_TYPE_TURN_LEFT) {
						wheel.pulse = Math.round(144 * (Turtle.DEG_TO_PULSE + alignment.spinLeft + alignment.delta) / 360.0);
					}
					wheel.id = (wheel.id % 255) + 1;
					wheel.count = 0;
					wheel.event = 1;
				case 4:
				case 8:
				case 12:
				case 16:
					if(alignment.type == ALIGNMENT_TYPE_TURN_RIGHT) {
						leftSpeed = 50;
						rightSpeed = -50;
					} else if(alignment.type == ALIGNMENT_TYPE_TURN_LEFT) {
						leftSpeed = -50;
						rightSpeed = 50;
					}
					break;
				case 101:
					alignment.moveState ++;
					wheel.pulse = Math.round(Turtle.CM_TO_PULSE);
					wheel.id = (wheel.id % 255) + 1;
					wheel.count = 0;
					wheel.event = 1;
				case 102:
					leftSpeed = -50;
					rightSpeed = -50;
					break;
				case 103:
					alignment.moveState ++;
					wheel.pulse = Math.round(180 * (Turtle.DEG_TO_PULSE * 2 + alignment.pivotRight) / 360.0);
					wheel.id = (wheel.id % 255) + 1;
					wheel.count = 0;
					wheel.event = 1;
				case 104:
					leftSpeed = 50;
					rightSpeed = 0;
					break;
				case 105:
					alignment.moveState ++;
					wheel.pulse = Math.round(180 * (Turtle.DEG_TO_PULSE * 2 + alignment.pivotLeft) / 360.0);
					wheel.id = (wheel.id % 255) + 1;
					wheel.count = 0;
					wheel.event = 1;
				case 106:
					leftSpeed = 0;
					rightSpeed = -50;
					break;
				case 107:
					alignment.moveState ++;
					wheel.pulse = Math.round(Turtle.CM_TO_PULSE);
					wheel.id = (wheel.id % 255) + 1;
					wheel.count = 0;
					wheel.event = 1;
				case 108:
					leftSpeed = 50;
					rightSpeed = 50;
					break;
			}
			var str = '10';
			if(leftSpeed < 0) str += self.toHex2(leftSpeed * 10.68 - 0.5);
			else str += self.toHex2(leftSpeed * 10.68 + 0.5);
			if(rightSpeed < 0) str += self.toHex2(rightSpeed * 10.68 - 0.5);
			else str += self.toHex2(rightSpeed * 10.68 + 0.5);
			str += self.toHex(wheel.id);
			str += self.toHex2(wheel.pulse);
			str += '000000000000000000004400-';
			str += address;
			str += '\r';
			return str;
		}
	} else {
		var motoring = self.motoring;
		var motion = self.motion;
		var wheel = self.wheel;
		var leftWheel = motoring.leftWheel, rightWheel = motoring.rightWheel;
		if(motoring.map & 0x00040000) {
			self.cancelTimeout();
			motion.type = parseInt(motoring.motion.type);
			if(motion.type < Turtle.MOTION_NONE || motion.type > Turtle.MOTION_SWING_RIGHT_TAIL) {
				motion.type = Turtle.MOTION_NONE;
			}
			if(motion.type != Turtle.MOTION_NONE) {
				motion.speed = motoring.motion.speed;
				wheel.pulse = 0;
				var unit = parseInt(motoring.motion.unit);
				var value = motoring.motion.value;
				if(unit == Turtle.UNIT_SEC) {
					if(value == 0) {
						leftWheel = 0;
						rightWheel = 0;
						wheel.count = 0;
						wheel.event = -1;
						motion.type = Turtle.MOTION_NONE;
					} else {
						self.timerId = setTimeout(function() {
							self.cancelTimeout();
							leftWheel = 0;
							rightWheel = 0;
							wheel.count = 0;
							wheel.event = -1;
							motion.type = Turtle.MOTION_NONE;
						}, value * 1000);
					}
				} else {
					if(unit == Turtle.UNIT_PULSE) {
						wheel.pulse = Math.round(value);
					} else if(unit == Turtle.UNIT_CM_DEG) {
						switch(motion.type) {
							case Turtle.MOTION_MOVE_FORWARD:
							case Turtle.MOTION_MOVE_BACKWARD:
								wheel.pulse = Math.round(value * Turtle.CM_TO_PULSE);
								break;
							case Turtle.MOTION_TURN_LEFT:
								wheel.pulse = Math.round(value * (Turtle.DEG_TO_PULSE + alignment.spinLeft) / 360.0);
								break;
							case Turtle.MOTION_TURN_RIGHT:
								wheel.pulse = Math.round(value * (Turtle.DEG_TO_PULSE + alignment.spinRight) / 360.0);
								break;
							case Turtle.MOTION_PIVOT_LEFT_HEAD:
							case Turtle.MOTION_PIVOT_LEFT_TAIL:
								wheel.pulse = Math.round(value * (2 * Turtle.DEG_TO_PULSE + alignment.pivotLeft) / 360.0);
								break;
							case Turtle.MOTION_PIVOT_RIGHT_HEAD:
							case Turtle.MOTION_PIVOT_RIGHT_TAIL:
								wheel.pulse = Math.round(value * (2 * Turtle.DEG_TO_PULSE + alignment.pivotRight) / 360.0);
								break;
							case Turtle.MOTION_SWING_LEFT_HEAD:
							case Turtle.MOTION_SWING_LEFT_TAIL:
								wheel.pulse = self.calculateSwingPulse(value, motoring.motion.radius, alignment.spinLeft);
								break;
							case Turtle.MOTION_SWING_RIGHT_HEAD:
							case Turtle.MOTION_SWING_RIGHT_TAIL:
								wheel.pulse = self.calculateSwingPulse(value, motoring.motion.radius, alignment.spinRight);
								break;
							default:
								motion.type = Turtle.MOTION_NONE;
								break;
						}
					}
					if(wheel.pulse == 0) {
						leftWheel = 0;
						rightWheel = 0;
						wheel.count = 0;
						wheel.event = -1;
						motion.type = Turtle.MOTION_NONE;
					}
				}
			}
		}
		if(motion.type == Turtle.MOTION_NONE) wheel.pulse = motoring.pulse;
		
		var speed = motion.speed;
		switch(motion.type) {
			case Turtle.MOTION_MOVE_FORWARD:
				if(speed == 0) speed = context.getMovementSpeed(MODULE);
				leftWheel = speed;
				rightWheel = speed;
				break;
			case Turtle.MOTION_MOVE_BACKWARD:
				if(speed == 0) speed = context.getMovementSpeed(MODULE);
				leftWheel = -speed;
				rightWheel = -speed;
				break;
			case Turtle.MOTION_TURN_LEFT:
				if(speed == 0) speed = context.getTurningSpeed(MODULE);
				leftWheel = -speed;
				rightWheel = speed;
				break;
			case Turtle.MOTION_TURN_RIGHT:
				if(speed == 0) speed = context.getTurningSpeed(MODULE);
				leftWheel = speed;
				rightWheel = -speed;
				break;
			case Turtle.MOTION_PIVOT_LEFT_HEAD:
				if(speed == 0) speed = context.getTurningSpeed(MODULE);
				leftWheel = 0;
				rightWheel = speed;
				break;
			case Turtle.MOTION_PIVOT_LEFT_TAIL:
				if(speed == 0) speed = context.getTurningSpeed(MODULE);
				leftWheel = 0;
				rightWheel = -speed;
				break;
			case Turtle.MOTION_PIVOT_RIGHT_HEAD:
				if(speed == 0) speed = context.getTurningSpeed(MODULE);
				leftWheel = speed;
				rightWheel = 0;
				break;
			case Turtle.MOTION_PIVOT_RIGHT_TAIL:
				if(speed == 0) speed = context.getTurningSpeed(MODULE);
				leftWheel = -speed;
				rightWheel = 0;
				break;
			case Turtle.MOTION_SWING_LEFT_HEAD:
				if(speed == 0) speed = context.getTurningSpeed(MODULE);
				leftWheel = self.calculateInnerSpeed(speed, motoring.motion.radius);
				rightWheel = speed;
				break;
			case Turtle.MOTION_SWING_LEFT_TAIL:
				if(speed == 0) speed = context.getTurningSpeed(MODULE);
				leftWheel = -self.calculateInnerSpeed(speed, motoring.motion.radius);
				rightWheel = -speed;
				break;
			case Turtle.MOTION_SWING_RIGHT_HEAD:
				if(speed == 0) speed = context.getTurningSpeed(MODULE);
				leftWheel = speed;
				rightWheel = self.calculateInnerSpeed(speed, motoring.motion.radius);
				break;
			case Turtle.MOTION_SWING_RIGHT_TAIL:
				if(speed == 0) speed = context.getTurningSpeed(MODULE);
				leftWheel = -speed;
				rightWheel = -self.calculateInnerSpeed(speed, motoring.motion.radius);
				break;
		}
		
		if(((motoring.map & 0x00040000) || (motoring.map & 0x04000000)) && wheel.pulse > 0) {
			wheel.nextLeftWheel = leftWheel;
			wheel.nextRightWheel = rightWheel;
			wheel.nextPulse = wheel.pulse;
			wheel.next = true;
			leftWheel = 0;
			rightWheel = 0;
			wheel.pulse = 0;
		} else if(wheel.next) {
			wheel.next = false;
			leftWheel = wheel.nextLeftWheel;
			rightWheel = wheel.nextRightWheel;
			wheel.pulse = wheel.nextPulse;
			motoring.map |= 0x04000000;
		}
		
		var str = '10';
		if(leftWheel < 0) str += self.toHex2(leftWheel * 10.68 - 0.5);
		else str += self.toHex2(leftWheel * 10.68 + 0.5);
		if(rightWheel < 0) str += self.toHex2(rightWheel * 10.68 - 0.5);
		else str += self.toHex2(rightWheel * 10.68 + 0.5);
		if((motoring.map & 0x00040000) || (motoring.map & 0x04000000)) {
			if(wheel.pulse != 0 || wheel.pulsePrev != 0) {
				wheel.id = (wheel.id % 255) + 1;
			}
			wheel.count = 0;
			if(wheel.pulse > 0) {
				wheel.event = 1;
			} else if(wheel.event != -1) {
				wheel.event = 0;
			}
			wheel.pulsePrev = wheel.pulse;
		}
		str += self.toHex(wheel.id);
		str += self.toHex2(wheel.pulse);
		
		var lineTracer = self.lineTracer;
		var tmp = 0;
		switch(motoring.lineTracerMode) {
			case Turtle.LINE_TRACER_MODE_OFF: tmp = 0x00; break;
			case Turtle.LINE_TRACER_MODE_BLACK: tmp = 0x08; break;
			case Turtle.LINE_TRACER_MODE_RED: tmp = 0x09; break;
			case Turtle.LINE_TRACER_MODE_GREEN: tmp = 0x0b; break;
			case Turtle.LINE_TRACER_MODE_BLUE: tmp = 0x0d; break;
			case Turtle.LINE_TRACER_MODE_ANY: tmp = 0x0f; break;
			case Turtle.LINE_TRACER_MODE_TURN_LEFT: tmp = 0x10; break;
			case Turtle.LINE_TRACER_MODE_TURN_RIGHT: tmp = 0x18; break;
			case Turtle.LINE_TRACER_MODE_CROSS: tmp = 0x20; break;
			case Turtle.LINE_TRACER_MODE_UTURN: tmp = 0x28; break;
			case Turtle.LINE_TRACER_MODE_BLACK_UNTIL_RED: tmp = 0x31; break;
			case Turtle.LINE_TRACER_MODE_BLACK_UNTIL_YELLOW: tmp = 0x32; break;
			case Turtle.LINE_TRACER_MODE_BLACK_UNTIL_GREEN: tmp = 0x33; break;
			case Turtle.LINE_TRACER_MODE_BLACK_UNTIL_CYAN: tmp = 0x34; break;
			case Turtle.LINE_TRACER_MODE_BLACK_UNTIL_BLUE: tmp = 0x35; break;
			case Turtle.LINE_TRACER_MODE_BLACK_UNTIL_MAGENTA: tmp = 0x36; break;
			case Turtle.LINE_TRACER_MODE_BLACK_UNTIL_ANY: tmp = 0x37; break;
			case Turtle.LINE_TRACER_MODE_RED_UNTIL_BLACK: tmp = 0x39; break;
			case Turtle.LINE_TRACER_MODE_GREEN_UNTIL_BLACK: tmp = 0x3b; break;
			case Turtle.LINE_TRACER_MODE_BLUE_UNTIL_BLACK: tmp = 0x3d; break;
			case Turtle.LINE_TRACER_MODE_ANY_UNTIL_BLACK: tmp = 0x3f; break;
		}
		if(motoring.map & 0x00800000) {
			lineTracer.count = 0;
			if(tmp > 0) {
				lineTracer.flag ^= 0x80;
				lineTracer.event = 1;
			} else {
				lineTracer.event = 0;
			}
		}
		tmp |= lineTracer.flag;
		str += self.toHex(tmp);
		if(router.isExtraButtonActive()) {
			++ self.identityCount;
			var rgb;
			if(self.identityCount <= 20) {
				rgb = self.getIdentityRgb(0);
			} else {
				rgb = self.getIdentityRgb(1);
				if(self.identityCount >= 40) self.identityCount = 0;
			}
			str += self.toHex(rgb[0]);
			str += self.toHex(rgb[1]);
			str += self.toHex(rgb[2]);
		} else {
			self.identityCount = 0;
			str += self.toHex(motoring.led.r);
			str += self.toHex(motoring.led.g);
			str += self.toHex(motoring.led.b);
		}
		tmp = (motoring.lock & 0x01) << 4;
		tmp |= (1 - motoring.lamp) & 0x01;
		str += self.toHex(tmp);
		str += self.toHex3(motoring.buzzer * 100);
		str += self.toHex(motoring.note);
		
		var sound = self.sound;
		tmp = 0;
		switch(motoring.sound) {
			case Turtle.SOUND_OFF: tmp = 0x00; break;
			case Turtle.SOUND_BEEP: tmp = 0x01; break;
			case Turtle.SOUND_RANDOM_BEEP: tmp = 0x05; break;
			case Turtle.SOUND_SIREN: tmp = 0x10; break;
			case Turtle.SOUND_ENGINE: tmp = 0x20; break;
			case Turtle.SOUND_ROBOT: tmp = 0x30; break;
			case Turtle.SOUND_MARCH: tmp = 0x40; break;
			case Turtle.SOUND_BIRTHDAY: tmp = 0x41; break;
			case Turtle.SOUND_DIBIDIBIDIP: tmp = 0x42; break;
			case Turtle.SOUND_GOOD_JOB: tmp = 0x43; break;
		}
		if(motoring.map & 0x01000000) {
			sound.count = 0;
			if(tmp > 0) {
				sound.flag ^= 0x80;
				sound.event = 1;
			} else {
				sound.event = 0;
			}
		}
		tmp |= sound.flag;
		str += self.toHex(tmp);
		tmp = ((motoring.lineTracerGain - 1) & 0x0f) << 4;
		tmp |= (motoring.lineTracerSpeed - 1) & 0x0f;
		str += self.toHex(tmp);
		str += '00-';
		str += address;
		str += '\r';
		return str;
	}
};

Controller.prototype.clear = function() {
	var motoring = this.motoring;
	motoring.map = 0;
	motoring.leftWheel = 0;
	motoring.rightWheel = 0;
	motoring.led.r = 0;
	motoring.led.g = 0;
	motoring.led.b = 0;
	motoring.buzzer = 0.0;
	var sensory = this.sensory;
	sensory.map = 0;
	sensory.button = 0;
	sensory.colorNumber = -1;
	sensory.colorPattern = -1;
	sensory.pulseCount = 0;
	sensory.wheelState = 0;
	sensory.soundState = 0;
	sensory.lineTracerState = 0;
	sensory.tilt = 0;
	sensory.batteryState = 2;
};

Controller.prototype.reset = function() {
	this.clear();
	this.cancelTimeout();
	var motoring = this.motoring;
	motoring.pulse = 0;
	motoring.note = 0;
	motoring.sound = 0;
	motoring.lineTracerMode = 0;
	motoring.lineTracerGain = 5;
	motoring.lineTracerSpeed = 5;
	motoring.lamp = 1;
	motoring.lock = 0;
	motoring.motion.type = 0;
	motoring.motion.unit = 0;
	motoring.motion.speed = 0;
	motoring.motion.value = 0;
	motoring.motion.radius = 0;
	var motion = this.motion;
	motion.type = 0;
	motion.speed = 0;
	var acc = this.acceleration;
	acc.sumx = 0.0;
	acc.sumy = 0.0;
	acc.sumz = 0.0;
	acc.index = 0;
	acc.count = 0;
	var button = this.button;
	button.clickId = -1;
	button.longPressId = -1;
	var wheel = this.wheel;
	wheel.pulse = 0;
	wheel.pulsePrev = -1;
	wheel.event = 0;
	wheel.state = 0;
	wheel.count = 0;
	wheel.next = false;
	wheel.nextLeftWheel = 0;
	wheel.nextRightWheel = 0;
	wheel.nextPulse = 0;
	var sound = this.sound;
	sound.event = 0;
	sound.state = 0;
	sound.count = 0;
	var lineTracer = this.lineTracer;
	lineTracer.event = 0;
	lineTracer.state = 0;
	lineTracer.count = 0;
	var event = this.event;
	event.button = -1;
	event.colorNumber = -2;
	event.colorPattern = -1;
	event.pulseCount = -1;
	event.tilt = -4;
	event.batteryState = -1;
	var batt = this.battery;
	batt.sum = 0.0;
	batt.index = 0;
	batt.count = 0;
};

Controller.prototype.activateAlignment = function() {
	this.alignment.active = true;
};

Controller.prototype.deactivateAlignment = function() {
	this.alignment.active = false;
};

Controller.prototype.startAlignmentMovement = function(callback) {
	var alignment = this.alignment;
	if(alignment.type == ALIGNMENT_TYPE_DISTANCE) alignment.moveState = 101;
	else alignment.moveState = 1;
	alignment.moveCallback = callback;
};

Controller.prototype.stopAlignmentMovement = function() {
	this.alignment.moveState = 0;
	var wheel = this.wheel;
	wheel.count = 0;
	wheel.event = 0;
};

Controller.prototype.setAlignmentType = function(type) {
	this.alignment.type = type;
};

Controller.prototype.setAlignmentDelta = function(delta) {
	this.alignment.delta = delta;
};

Controller.prototype.getAlignmentDistance = function() {
	return this.alignment.distance * 4;
};

Controller.prototype.resetAlignment = function() {
	var alignment = this.alignment;
	alignment.moveState = 0;
	alignment.type = 0;
	alignment.saveState = 0;
	alignment.saveCount = 0;
	alignment.delta = 0;
	var wheel = this.wheel;
	wheel.count = 0;
	wheel.event = 0;
};

Controller.prototype.saveAlignment = function(delta, distance, callback) {
	var alignment = this.alignment;
	if(alignment.type == ALIGNMENT_TYPE_DISTANCE) {
		try {
			var distance = parseFloat(distance);
			if(typeof distance == 'number') {
				var value = distance * 2500 - Turtle.WHEEL_DISTANCE_10000;
				if(value > -32000 && value < 32000) {
					alignment.delta = parseInt(value);
				}
			}
		} catch (e) {
		}
	} else {
		alignment.delta = delta;
	}
	alignment.saveCount = 0;
	alignment.saveCallback = callback;
	alignment.saveState = 1;
};

Controller.prototype.getSensorMonitor = function() {
	return this.sensorMonitor;
};

Controller.prototype.getSensorNames = function() {
	return this.sensorNames;
};

Controller.prototype.setDuration = function(duration) {
	var sensorMonitor = this.sensorMonitor;
	sensorMonitor.reset();
	sensorMonitor.duration = duration;
};

module.exports.create = function() {
	return new Controller();
};