'use strict';
var context = require('context');
var translations = require('./translations');
var Turtle = require('../turtle');
var factory = require('../turtle-roboid');
const MODULE = 'turtle';

var PARTS = {};
var DIRECTIONS = {};
var TOWARDS = {};
var UNITS = {};
var LINE_COLORS = {};
var RGB_COLORS = {};
var COLOR_NUMBERS = {};
var COLOR_PATTERNS = {};
var NOTES = {};
var BEATS = { '¼': 0.25, '½': 0.5, '¾': 0.75, '1¼': 1.25, '1½': 1.5, '1¾': 1.75 };
var SOUNDS = {};

var CLICKED = '/clicked ';
var DOUBLE_CLICKED = '/double-clicked ';
var LONG_PRESSED = '/long-pressed ';
var BOOLEAN_TILT_FORWARD = '/tilt%20forward ';
var BOOLEAN_TILT_BACKWARD = '/tilt%20backward ';
var BOOLEAN_TILT_LEFT = '/tilt%20left ';
var BOOLEAN_TILT_RIGHT = '/tilt%20right ';
var BOOLEAN_TILT_FLIP = '/tilt%20flip ';
var BOOLEAN_TILT_NONE = '/not%20tilt ';
var BOOLEAN_BATTERY_NORMAL = '/normal ';
var BOOLEAN_BATTERY_LOW = '/low ';
var BOOLEAN_BATTERY_EMPTY = '/empty ';

const LEFT = 1;
const RIGHT = 2;
const BACK = 3;
const HEAD = 1;
const SECONDS = 1;
const PULSES = 2;
const DEGREES = 3;

function Robot(index) {
	var roboid = this.roboid = factory.create();
	this.index = index;
	this.waits = [];
	this.timeouts = [];
	this.sensory = {
		floor: 0,
		accelerationX: 0,
		accelerationY: 0,
		accelerationZ: 0,
		button: 0,
		colorNumber: -1,
		colorPattern: -1,
		tilt: 0,
		batteryState: 2
	};
	this.motoring = {
		pulse: roboid.findDeviceById(Turtle.PULSE),
		note: roboid.findDeviceById(Turtle.NOTE),
		sound: roboid.findDeviceById(Turtle.SOUND),
		lineTracerMode: roboid.findDeviceById(Turtle.LINE_TRACER_MODE),
		lineTracerGain: roboid.findDeviceById(Turtle.LINE_TRACER_GAIN),
		lineTracerSpeed: roboid.findDeviceById(Turtle.LINE_TRACER_SPEED),
		motion: roboid.findDeviceById(Turtle.MOTION),
		reset: function() {
			this.leftWheel.data = 0;
			this.rightWheel.data = 0;
			this.led.data[0] = 0;
			this.led.data[1] = 0;
			this.led.data[2] = 0;
			this.buzzer.data = 0;
		}
	};
	this.register('leftWheel', Turtle.LEFT_WHEEL);
	this.register('rightWheel', Turtle.RIGHT_WHEEL);
	this.register3('led', Turtle.LED);
	this.register('buzzer', Turtle.BUZZER);
	this.tempo = 60;
	this.motionId = -1;
	this.lineTracerId = -1;
	this.soundId = -1;
	this.currentSound = 0;
	this.soundRepeat = 1;
	this.noteId = -1;
	this.noteTimer1 = undefined;
	this.noteTimer2 = undefined;
	this.clicked = false;
	this.doubleClicked = false;
	this.longPressed = false;
	this.active = true;
	this.received = false;
}

Robot.prototype.reset = function() {
	this.tempo = 60;
	this.motionId = -1;
	this.lineTracerId = -1;
	this.soundId = -1;
	this.currentSound = 0;
	this.soundRepeat = 1;
	this.noteId = -1;
	this.noteTimer1 = undefined;
	this.noteTimer2 = undefined;
	this.clicked = false;
	this.doubleClicked = false;
	this.longPressed = false;
	this.roboid.reset();
	this.motoring.reset();
};

Robot.prototype.register = function(name, deviceId) {
	var self = this;
	self.motoring[name] = {
		data: 0,
		device: self.roboid.findDeviceById(deviceId),
		write: function(value) {
			this.data = value;
			this.device.write(value);
		},
		read: function() {
			return this.data;
		}
	};
};

Robot.prototype.register3 = function(name, deviceId) {
	var self = this;
	self.motoring[name] = {
		data: [0, 0, 0],
		device: self.roboid.findDeviceById(deviceId),
		write: function(index, value) {
			this.data[index] = value;
			this.device.write(index, value);
		},
		read: function(index) {
			return this.data[index];
		}
	};
};

Robot.prototype.getWaits = function() {
	return this.waits;
};

Robot.prototype.removeWait = function(id) {
	var waits = this.waits;
	var index = waits.indexOf(id);
	if(index >= 0) {
		waits.splice(index, 1);
	}
};

Robot.prototype.clearWaits = function() {
	this.waits = [];
};

Robot.prototype.removeTimeout = function(id) {
	clearTimeout(id);
	var timeouts = this.timeouts;
	var index = timeouts.indexOf(id);
	if(index >= 0) {
		timeouts.splice(index, 1);
	}
};

Robot.prototype.clearTimeouts = function() {
	var timeouts = this.timeouts;
	for(var i in timeouts) {
		clearTimeout(timeouts[i]);
	}
	this.timeouts = [];
};

Robot.prototype.cancelLineTracer = function() {
	this.lineTracerId = -1;
};

Robot.prototype.cancelMotion = function() {
	this.motionId = -1;
};

Robot.prototype.cancelNote = function() {
	this.noteId = -1;
	if(this.noteTimer1 !== undefined) {
		this.removeTimeout(this.noteTimer1);
	}
	if(this.noteTimer2 !== undefined) {
		this.removeTimeout(this.noteTimer2);
	}
	this.noteTimer1 = undefined;
	this.noteTimer2 = undefined;
};

Robot.prototype.cancelSound = function() {
	this.soundId = -1;
};

Robot.prototype.writeSingleSensory = function(response) {
	var sensory = this.sensory;
	response.write('turtleColorNumber ' + sensory.colorNumber + '\n');
	response.write('turtleColorPattern ' + sensory.colorPattern + '\n');
	response.write('turtleFloor ' + sensory.floor + '\n');
	response.write('turtleButton ' + sensory.button + '\n');
	response.write('turtleAccelerationX ' + sensory.accelerationX + '\n');
	response.write('turtleAccelerationY ' + sensory.accelerationY + '\n');
	response.write('turtleAccelerationZ ' + sensory.accelerationZ + '\n');
	response.write('turtleButtonState' + CLICKED + this.clicked + '\n');
	response.write('turtleButtonState' + DOUBLE_CLICKED + this.doubleClicked + '\n');
	response.write('turtleButtonState' + LONG_PRESSED + this.longPressed + '\n');
	response.write('turtleTilt' + BOOLEAN_TILT_FORWARD + (sensory.tilt == 1) + '\n');
	response.write('turtleTilt' + BOOLEAN_TILT_BACKWARD + (sensory.tilt == -1) + '\n');
	response.write('turtleTilt' + BOOLEAN_TILT_LEFT + (sensory.tilt == 2) + '\n');
	response.write('turtleTilt' + BOOLEAN_TILT_RIGHT + (sensory.tilt == -2) + '\n');
	response.write('turtleTilt' + BOOLEAN_TILT_FLIP + (sensory.tilt == 3) + '\n');
	response.write('turtleTilt' + BOOLEAN_TILT_NONE + (sensory.tilt == -3) + '\n');
	response.write('turtleBattery' + BOOLEAN_BATTERY_NORMAL + (sensory.batteryState == 2) + '\n');
	response.write('turtleBattery' + BOOLEAN_BATTERY_LOW + (sensory.batteryState == 1) + '\n');
	response.write('turtleBattery' + BOOLEAN_BATTERY_EMPTY + (sensory.batteryState == 0) + '\n');
	
	var value;
	for(var key in COLOR_NUMBERS) {
		value = (COLOR_NUMBERS[key] == sensory.colorNumber) ? true : false;
		response.write('turtleTouchingColor/' + key + ' ' + value + '\n');
	}
	for(var key1 in COLOR_PATTERNS) {
		for(var key2 in COLOR_PATTERNS) {
			value = (COLOR_PATTERNS[key1] * 10 + COLOR_PATTERNS[key2] == sensory.colorPattern) ? true : false;
			response.write('turtleIsColorPattern/' + key1 + '/' + key2 + ' ' + value + '\n');
		}
	}
	
	this.clicked = false;
	this.doubleClicked = false;
	this.longPressed = false;
	sensory.colorPattern = -1;
};

Robot.prototype.writeMultiSensory = function(response) {
	var sensory = this.sensory;
	if(this.active) {
		var index = this.index;
		var tag = index + ' ';
		response.write('turtleColorNumber/' + tag + sensory.colorNumber + '\n');
		response.write('turtleColorPattern/' + tag + sensory.colorPattern + '\n');
		response.write('turtleFloor/' + tag + sensory.floor + '\n');
		response.write('turtleButton/' + tag + sensory.button + '\n');
		response.write('turtleAccelerationX/' + tag + sensory.accelerationX + '\n');
		response.write('turtleAccelerationY/' + tag + sensory.accelerationY + '\n');
		response.write('turtleAccelerationZ/' + tag + sensory.accelerationZ + '\n');
		response.write('turtleButtonState/' + index + CLICKED + this.clicked + '\n');
		response.write('turtleButtonState/' + index + DOUBLE_CLICKED + this.doubleClicked + '\n');
		response.write('turtleButtonState/' + index + LONG_PRESSED + this.longPressed + '\n');
		response.write('turtleTilt/' + index + BOOLEAN_TILT_FORWARD + (sensory.tilt == 1) + '\n');
		response.write('turtleTilt/' + index + BOOLEAN_TILT_BACKWARD + (sensory.tilt == -1) + '\n');
		response.write('turtleTilt/' + index + BOOLEAN_TILT_LEFT + (sensory.tilt == 2) + '\n');
		response.write('turtleTilt/' + index + BOOLEAN_TILT_RIGHT + (sensory.tilt == -2) + '\n');
		response.write('turtleTilt/' + index + BOOLEAN_TILT_FLIP + (sensory.tilt == 3) + '\n');
		response.write('turtleTilt/' + index + BOOLEAN_TILT_NONE + (sensory.tilt == -3) + '\n');
		response.write('turtleBattery/' + index + BOOLEAN_BATTERY_NORMAL + (sensory.batteryState == 2) + '\n');
		response.write('turtleBattery/' + index + BOOLEAN_BATTERY_LOW + (sensory.batteryState == 1) + '\n');
		response.write('turtleBattery/' + index + BOOLEAN_BATTERY_EMPTY + (sensory.batteryState == 0) + '\n');
		
		var tag2 = index + '/';
		var value;
		for(var key in COLOR_NUMBERS) {
			value = (COLOR_NUMBERS[key] == sensory.colorNumber) ? true : false;
			response.write('turtleTouchingColor/' + tag2 + key + ' ' + value + '\n');
		}
		for(var key1 in COLOR_PATTERNS) {
			for(var key2 in COLOR_PATTERNS) {
				value = (COLOR_PATTERNS[key1] * 10 + COLOR_PATTERNS[key2] == sensory.colorPattern) ? true : false;
				response.write('turtleIsColorPattern/' + tag2 + key1 + '/' + key2 + ' ' + value + '\n');
			}
		}
		
		response.write('turtleColorNumber' + tag + sensory.colorNumber + '\n');
		response.write('turtleColorPattern' + tag + sensory.colorPattern + '\n');
		response.write('turtleFloor' + tag + sensory.floor + '\n');
		response.write('turtleButton' + tag + sensory.button + '\n');
		response.write('turtleAccelerationX' + tag + sensory.accelerationX + '\n');
		response.write('turtleAccelerationY' + tag + sensory.accelerationY + '\n');
		response.write('turtleAccelerationZ' + tag + sensory.accelerationZ + '\n');
	}
	this.clicked = false;
	this.doubleClicked = false;
	this.longPressed = false;
	sensory.colorPattern = -1;
	if(this.received == false) this.active = false;
};

Robot.prototype.moveForward = function(request) {
	var motoring = this.motoring;
	this.cancelLineTracer();
	
	motoring.leftWheel.write(0);
	motoring.rightWheel.write(0);
	motoring.pulse.write(0);
	motoring.motion.write(1, Turtle.UNIT_CM_DEG);
	motoring.motion.write(2, 0);
	motoring.motion.write(3, context.getMovementDistance(MODULE));
	motoring.motion.write(0, Turtle.MOTION_MOVE_FORWARD);
	this.motionId = parseInt(request.params.id);
	this.waits.push(this.motionId);
	motoring.lineTracerMode.write(0);
};

Robot.prototype.moveBackward = function(request) {
	var motoring = this.motoring;
	this.cancelLineTracer();
	
	motoring.leftWheel.write(0);
	motoring.rightWheel.write(0);
	motoring.pulse.write(0);
	motoring.motion.write(1, Turtle.UNIT_CM_DEG);
	motoring.motion.write(2, 0);
	motoring.motion.write(3, context.getMovementDistance(MODULE));
	motoring.motion.write(0, Turtle.MOTION_MOVE_BACKWARD);
	this.motionId = parseInt(request.params.id);
	this.waits.push(this.motionId);
	motoring.lineTracerMode.write(0);
};

Robot.prototype.turn = function(request) {
	var motoring = this.motoring;
	this.cancelLineTracer();
	
	motoring.leftWheel.write(0);
	motoring.rightWheel.write(0);
	motoring.pulse.write(0);
	motoring.motion.write(1, Turtle.UNIT_CM_DEG);
	motoring.motion.write(2, 0);
	motoring.motion.write(3, context.getTurningAngle(MODULE));
	if(DIRECTIONS[request.params.direction] === LEFT) {
		motoring.motion.write(0, Turtle.MOTION_TURN_LEFT);
	} else {
		motoring.motion.write(0, Turtle.MOTION_TURN_RIGHT);
	}
	this.motionId = parseInt(request.params.id);
	this.waits.push(this.motionId);
	motoring.lineTracerMode.write(0);
};

Robot.prototype.moveForwardUnit = function(request) {
	var motoring = this.motoring;
	this.cancelLineTracer();
	this.cancelMotion();
	
	motoring.leftWheel.write(0);
	motoring.rightWheel.write(0);
	motoring.pulse.write(0);
	var value = parseFloat(request.params.value);
	if(value) {
		var unit = UNITS[request.params.unit];
		if(unit === SECONDS) motoring.motion.write(1, Turtle.UNIT_SEC);
		else if(unit === PULSES) motoring.motion.write(1, Turtle.UNIT_PULSE);
		else motoring.motion.write(1, Turtle.UNIT_CM_DEG);
		motoring.motion.write(2, 0);
		if(value < 0) {
			motoring.motion.write(3, -value);
			motoring.motion.write(0, Turtle.MOTION_MOVE_BACKWARD);
		} else {
			motoring.motion.write(3, value);
			motoring.motion.write(0, Turtle.MOTION_MOVE_FORWARD);
		}
		this.motionId = parseInt(request.params.id);
		this.waits.push(this.motionId);
		motoring.lineTracerMode.write(0);
	} else {
		motoring.motion.write(0);
		motoring.lineTracerMode.write(0);
	}
};

Robot.prototype.moveBackwardUnit = function(request) {
	var motoring = this.motoring;
	this.cancelLineTracer();
	this.cancelMotion();
	
	motoring.leftWheel.write(0);
	motoring.rightWheel.write(0);
	motoring.pulse.write(0);
	var value = parseFloat(request.params.value);
	if(value) {
		var unit = UNITS[request.params.unit];
		if(unit === SECONDS) motoring.motion.write(1, Turtle.UNIT_SEC);
		else if(unit === PULSES) motoring.motion.write(1, Turtle.UNIT_PULSE);
		else motoring.motion.write(1, Turtle.UNIT_CM_DEG);
		motoring.motion.write(2, 0);
		if(value < 0) {
			motoring.motion.write(3, -value);
			motoring.motion.write(0, Turtle.MOTION_MOVE_FORWARD);
		} else {
			motoring.motion.write(3, value);
			motoring.motion.write(0, Turtle.MOTION_MOVE_BACKWARD);
		}
		this.motionId = parseInt(request.params.id);
		this.waits.push(this.motionId);
		motoring.lineTracerMode.write(0);
	} else {
		motoring.motion.write(0);
		motoring.lineTracerMode.write(0);
	}
};

Robot.prototype.turnUnitInPlace = function(request) {
	var motoring = this.motoring;
	this.cancelLineTracer();
	this.cancelMotion();
	
	motoring.leftWheel.write(0);
	motoring.rightWheel.write(0);
	motoring.pulse.write(0);
	var value = parseFloat(request.params.value);
	if(value) {
		var unit = UNITS[request.params.unit];
		if(unit === SECONDS) motoring.motion.write(1, Turtle.UNIT_SEC);
		else if(unit === PULSES) motoring.motion.write(1, Turtle.UNIT_PULSE);
		else motoring.motion.write(1, Turtle.UNIT_CM_DEG);
		motoring.motion.write(2, 0);
		if(value < 0) {
			motoring.motion.write(3, -value);
			if(DIRECTIONS[request.params.direction] === LEFT) {
				motoring.motion.write(0, Turtle.MOTION_TURN_RIGHT);
			} else {
				motoring.motion.write(0, Turtle.MOTION_TURN_LEFT);
			}
		} else {
			motoring.motion.write(3, value);
			if(DIRECTIONS[request.params.direction] === LEFT) {
				motoring.motion.write(0, Turtle.MOTION_TURN_LEFT);
			} else {
				motoring.motion.write(0, Turtle.MOTION_TURN_RIGHT);
			}
		}
		this.motionId = parseInt(request.params.id);
		this.waits.push(this.motionId);
		motoring.lineTracerMode.write(0);
	} else {
		motoring.motion.write(0);
		motoring.lineTracerMode.write(0);
	}
};

Robot.prototype.turnUnitWithRadiusInDirection = function(request) {
	var motoring = this.motoring;
	this.cancelLineTracer();
	this.cancelMotion();
	
	motoring.leftWheel.write(0);
	motoring.rightWheel.write(0);
	motoring.pulse.write(0);
	var value = parseFloat(request.params.value);
	var radius = parseFloat(request.params.radius);
	if(value && (typeof radius == 'number') && radius >= 0) {
		var unit = UNITS[request.params.unit];
		if(unit === SECONDS) motoring.motion.write(1, Turtle.UNIT_SEC);
		else if(unit === PULSES) motoring.motion.write(1, Turtle.UNIT_PULSE);
		else motoring.motion.write(1, Turtle.UNIT_CM_DEG);
		motoring.motion.write(2, 0);
		motoring.motion.write(4, radius);
		if(value < 0) {
			motoring.motion.write(3, -value);
			if(DIRECTIONS[request.params.direction] === LEFT) {
				if(TOWARDS[request.params.head] === HEAD) motoring.motion.write(0, Turtle.MOTION_SWING_LEFT_TAIL);
				else motoring.motion.write(0, Turtle.MOTION_SWING_LEFT_HEAD);
			} else {
				if(TOWARDS[request.params.head] === HEAD) motoring.motion.write(0, Turtle.MOTION_SWING_RIGHT_TAIL);
				else motoring.motion.write(0, Turtle.MOTION_SWING_RIGHT_HEAD);
			}
		} else {
			motoring.motion.write(3, value);
			if(DIRECTIONS[request.params.direction] === LEFT) {
				if(TOWARDS[request.params.head] === HEAD) motoring.motion.write(0, Turtle.MOTION_SWING_LEFT_HEAD);
				else motoring.motion.write(0, Turtle.MOTION_SWING_LEFT_TAIL);
			} else {
				if(TOWARDS[request.params.head] === HEAD) motoring.motion.write(0, Turtle.MOTION_SWING_RIGHT_HEAD);
				else motoring.motion.write(0, Turtle.MOTION_SWING_RIGHT_TAIL);
			}
		}
		this.motionId = parseInt(request.params.id);
		this.waits.push(this.motionId);
		motoring.lineTracerMode.write(0);
	} else {
		motoring.motion.write(0);
		motoring.lineTracerMode.write(0);
	}
};

Robot.prototype.pivotAroundWheelUnitInDirection = function(request) {
	var motoring = this.motoring;
	this.cancelLineTracer();
	this.cancelMotion();
	
	motoring.leftWheel.write(0);
	motoring.rightWheel.write(0);
	motoring.pulse.write(0);
	var value = parseFloat(request.params.value);
	if(value) {
		var unit = UNITS[request.params.unit];
		if(unit === SECONDS) motoring.motion.write(1, Turtle.UNIT_SEC);
		else if(unit === PULSES) motoring.motion.write(1, Turtle.UNIT_PULSE);
		else motoring.motion.write(1, Turtle.UNIT_CM_DEG);
		motoring.motion.write(2, 0);
		if(value < 0) {
			motoring.motion.write(3, -value);
			if(PARTS[request.params.wheel] === LEFT) {
				if(TOWARDS[request.params.head] === HEAD) motoring.motion.write(0, Turtle.MOTION_PIVOT_LEFT_TAIL);
				else motoring.motion.write(0, Turtle.MOTION_PIVOT_LEFT_HEAD);
			} else {
				if(TOWARDS[request.params.head] === HEAD) motoring.motion.write(0, Turtle.MOTION_PIVOT_RIGHT_TAIL);
				else motoring.motion.write(0, Turtle.MOTION_PIVOT_RIGHT_HEAD);
			}
		} else {
			motoring.motion.write(3, value);
			if(PARTS[request.params.wheel] === LEFT) {
				if(TOWARDS[request.params.head] === HEAD) motoring.motion.write(0, Turtle.MOTION_PIVOT_LEFT_HEAD);
				else motoring.motion.write(0, Turtle.MOTION_PIVOT_LEFT_TAIL);
			} else {
				if(TOWARDS[request.params.head] === HEAD) motoring.motion.write(0, Turtle.MOTION_PIVOT_RIGHT_HEAD);
				else motoring.motion.write(0, Turtle.MOTION_PIVOT_RIGHT_TAIL);
			}
		}
		this.motionId = parseInt(request.params.id);
		this.waits.push(this.motionId);
		motoring.lineTracerMode.write(0);
	} else {
		motoring.motion.write(0);
		motoring.lineTracerMode.write(0);
	}
};

Robot.prototype.changeWheelsByLeftRight = function(request) {
	var motoring = this.motoring;
	this.cancelLineTracer();
	this.cancelMotion();
	
	var left = parseFloat(request.params.left);
	var right = parseFloat(request.params.right);
	if(typeof left == 'number') {
		left += motoring.leftWheel.read();
		motoring.leftWheel.write(left);
	}
	if(typeof right == 'number') {
		right += motoring.rightWheel.read();
		motoring.rightWheel.write(right);
	}
	motoring.pulse.write(0);
	motoring.motion.write(0);
	motoring.lineTracerMode.write(0);
};

Robot.prototype.setWheelsToLeftRight = function(request) {
	var motoring = this.motoring;
	this.cancelLineTracer();
	this.cancelMotion();
	
	var left = parseFloat(request.params.left);
	var right = parseFloat(request.params.right);
	if(typeof left == 'number') {
		motoring.leftWheel.write(left);
	}
	if(typeof right == 'number') {
		motoring.rightWheel.write(right);
	}
	motoring.pulse.write(0);
	motoring.motion.write(0);
	motoring.lineTracerMode.write(0);
};

Robot.prototype.changeWheelBy = function(request) {
	var motoring = this.motoring;
	this.cancelLineTracer();
	this.cancelMotion();
	
	var speed = parseFloat(request.params.speed);
	if(typeof speed == 'number') {
		var wheel = PARTS[request.params.wheel];
		if(wheel === LEFT) {
			speed += motoring.leftWheel.read();
			motoring.leftWheel.write(speed);
		} else if(wheel === RIGHT) {
			speed += motoring.rightWheel.read();
			motoring.rightWheel.write(speed);
		} else {
			var left = speed + motoring.leftWheel.read();
			var right = speed + motoring.rightWheel.read();
			motoring.leftWheel.write(left);
			motoring.rightWheel.write(right);
		}
	}
	motoring.pulse.write(0);
	motoring.motion.write(0);
	motoring.lineTracerMode.write(0);
};

Robot.prototype.setWheelTo = function(request) {
	var motoring = this.motoring;
	this.cancelLineTracer();
	this.cancelMotion();
	
	var speed = parseFloat(request.params.speed);
	if(typeof speed == 'number') {
		var wheel = PARTS[request.params.wheel];
		if(wheel === LEFT) {
			motoring.leftWheel.write(speed);
		} else if(wheel === RIGHT) {
			motoring.rightWheel.write(speed);
		} else {
			motoring.leftWheel.write(speed);
			motoring.rightWheel.write(speed);
		}
	}
	motoring.pulse.write(0);
	motoring.motion.write(0);
	motoring.lineTracerMode.write(0);
};

Robot.prototype.followLine = function(request) {
	var motoring = this.motoring;
	this.cancelLineTracer();
	this.cancelMotion();
	
	var mode = Turtle.LINE_TRACER_MODE_BLACK + LINE_COLORS[request.params.color];
	motoring.leftWheel.write(0);
	motoring.rightWheel.write(0);
	motoring.pulse.write(0);
	motoring.motion.write(0);
	motoring.lineTracerMode.write(mode);
};

Robot.prototype.followLineUntil = function(request) {
	var motoring = this.motoring;
	this.cancelMotion();
	
	var mode = Turtle.LINE_TRACER_MODE_BLACK_UNTIL_RED + LINE_COLORS[request.params.color] - 1;
	motoring.leftWheel.write(0);
	motoring.rightWheel.write(0);
	motoring.pulse.write(0);
	motoring.motion.write(0);
	motoring.lineTracerMode.write(mode);
	this.lineTracerId = parseInt(request.params.id);
	this.waits.push(this.lineTracerId);
};

Robot.prototype.followLineUntilBlack = function(request) {
	var motoring = this.motoring;
	this.cancelMotion();
	
	var mode = Turtle.LINE_TRACER_MODE_RED_UNTIL_BLACK + LINE_COLORS[request.params.color] - 1;
	motoring.leftWheel.write(0);
	motoring.rightWheel.write(0);
	motoring.pulse.write(0);
	motoring.motion.write(0);
	motoring.lineTracerMode.write(mode);
	this.lineTracerId = parseInt(request.params.id);
	this.waits.push(this.lineTracerId);
};

Robot.prototype.crossIntersection = function(request) {
	var motoring = this.motoring;
	this.cancelMotion();
	
	motoring.leftWheel.write(0);
	motoring.rightWheel.write(0);
	motoring.pulse.write(0);
	motoring.motion.write(0);
	motoring.lineTracerMode.write(Turtle.LINE_TRACER_MODE_CROSS);
	this.lineTracerId = parseInt(request.params.id);
	this.waits.push(this.lineTracerId);
};

Robot.prototype.turnAtIntersection = function(request) {
	var motoring = this.motoring;
	this.cancelMotion();
	
	var mode = Turtle.LINE_TRACER_MODE_TURN_LEFT;
	var direction = DIRECTIONS[request.params.direction];
	if(direction === RIGHT)
		mode = Turtle.LINE_TRACER_MODE_TURN_RIGHT;
	else if(direction === BACK)
		mode = Turtle.LINE_TRACER_MODE_UTURN;
	
	motoring.leftWheel.write(0);
	motoring.rightWheel.write(0);
	motoring.pulse.write(0);
	motoring.motion.write(0);
	motoring.lineTracerMode.write(mode);
	this.lineTracerId = parseInt(request.params.id);
	this.waits.push(this.lineTracerId);
};

Robot.prototype.setFollowingSpeedTo = function(request) {
	var motoring = this.motoring;
	var speed = parseInt(request.params.speed);
	if(typeof speed == 'number') {
		motoring.lineTracerSpeed.write(speed);
		motoring.lineTracerGain.write(speed);
	}
};

Robot.prototype.stop = function(request) {
	var motoring = this.motoring;
	this.cancelLineTracer();
	this.cancelMotion();
	
	motoring.leftWheel.write(0);
	motoring.rightWheel.write(0);
	motoring.pulse.write(0);
	motoring.motion.write(0);
	motoring.lineTracerMode.write(0);
};

Robot.prototype.setHeadLedTo = function(request) {
	var color = RGB_COLORS[request.params.color];
	if(color) {
		this.motoring.led.write(color);
	}
};

Robot.prototype.changeHeadLedByRGB = function(request) {
	var motoring = this.motoring;
	var red = parseInt(request.params.red);
	var green = parseInt(request.params.green);
	var blue = parseInt(request.params.blue);
	if(typeof red == 'number') {
		red += motoring.led.read(0);
		motoring.led.write(0, red);
	}
	if(typeof green == 'number') {
		green += motoring.led.read(1);
		motoring.led.write(1, green);
	}
	if(typeof blue == 'number') {
		blue += motoring.led.read(2);
		motoring.led.write(2, blue);
	}
};

Robot.prototype.setHeadLedToRGB = function(request) {
	var motoring = this.motoring;
	var red = parseInt(request.params.red);
	var green = parseInt(request.params.green);
	var blue = parseInt(request.params.blue);
	if(typeof red == 'number') {
		motoring.led.write(0, red);
	}
	if(typeof green == 'number') {
		motoring.led.write(1, green);
	}
	if(typeof blue == 'number') {
		motoring.led.write(2, blue);
	}
};

Robot.prototype.clearHeadLed = function(request) {
	var motoring = this.motoring;
	motoring.led.write(0, 0);
	motoring.led.write(1, 0);
	motoring.led.write(2, 0);
};

Robot.prototype.playSound = function(request) {
	var motoring = this.motoring;
	this.cancelNote();
	this.cancelSound();
	
	var sound = SOUNDS[request.params.sound];
	motoring.buzzer.write(0);
	motoring.note.write(0);
	if(sound) {
		this.runSound(sound);
	} else {
		this.runSound(0);
	}
};

Robot.prototype.playSoundTimes = function(request) {
	var motoring = this.motoring;
	this.cancelNote();
	this.cancelSound();
	
	var sound = SOUNDS[request.params.sound];
	var count = parseInt(request.params.count);
	motoring.buzzer.write(0);
	motoring.note.write(0);
	if(sound && count) {
		this.runSound(sound, count);
	} else {
		this.runSound(0);
	}
};

Robot.prototype.playSoundTimesUntilDone = function(request) {
	var motoring = this.motoring;
	this.cancelNote();
	this.cancelSound();
	
	var sound = SOUNDS[request.params.sound];
	var count = parseInt(request.params.count);
	motoring.buzzer.write(0);
	motoring.note.write(0);
	if(count) {
		if(sound) {
			this.runSound(sound, count);
		} else {
			this.runSound(0);
		}
		this.soundId = parseInt(request.params.id);
		this.waits.push(this.soundId);
	} else {
		this.runSound(0);
	}
};

Robot.prototype.changeBuzzerBy = function(request) {
	var motoring = this.motoring;
	this.cancelNote();
	this.cancelSound();
	
	var hz = parseFloat(request.params.hz);
	if(typeof hz == 'number') {
		hz += motoring.buzzer.read();
		motoring.buzzer.write(hz);
	}
	motoring.note.write(0);
	this.runSound(0);
};

Robot.prototype.setBuzzerTo = function(request) {
	var motoring = this.motoring;
	this.cancelNote();
	this.cancelSound();
	
	var hz = parseFloat(request.params.hz);
	if(typeof hz == 'number') {
		motoring.buzzer.write(hz);
	}
	motoring.note.write(0);
	this.runSound(0);
};

Robot.prototype.clearSound = function(request) {
	var motoring = this.motoring;
	this.cancelNote();
	this.cancelSound();
	
	motoring.buzzer.write(0);
	motoring.note.write(0);
	this.runSound(0);
};

Robot.prototype.playNote = function(request) {
	var motoring = this.motoring;
	this.cancelNote();
	this.cancelSound();
	
	var note = NOTES[request.params.note];
	var octave = parseInt(request.params.octave);
	motoring.buzzer.write(0);
	if(note && octave && octave > 0 && octave < 8) {
		note += (octave - 1) * 12;
		motoring.note.write(note);
	} else {
		motoring.note.write(0);
	}
	this.runSound(0);
};

Robot.prototype.playNoteForBeats = function(request) {
	var self = this;
	var motoring = self.motoring;
	self.cancelNote();
	self.cancelSound();
	
	var note = NOTES[request.params.note];
	var octave = parseInt(request.params.octave);
	var beat = BEATS[request.params.beat];
	if(!beat) {
		beat = parseFloat(request.params.beat);
	}
	motoring.buzzer.write(0);
	if(note && octave && octave > 0 && octave < 8 && beat && beat > 0 && self.tempo > 0) {
		var id = self.noteId = parseInt(request.params.id);
		self.waits.push(id);
		note += (octave - 1) * 12;
		motoring.note.write(note);
		var timeout = beat * 60 * 1000 / self.tempo;
		var tail = 0;
		if(timeout > 100) {
			tail = 100;
		}
		if(tail > 0) {
			self.noteTimer1 = setTimeout(function() {
				if(self.noteId == id) {
					motoring.note.write(0);
					if(self.noteTimer1 !== undefined) self.removeTimeout(self.noteTimer1);
					self.noteTimer1 = undefined;
				}
			}, timeout - tail);
			self.timeouts.push(self.noteTimer1);
		}
		self.noteTimer2 = setTimeout(function() {
			if(self.noteId == id) {
				motoring.note.write(0);
				self.removeWait(self.noteId);
				self.cancelNote();
			}
		}, timeout);
		self.timeouts.push(self.noteTimer2);
	} else {
		motoring.note.write(0);
	}
	self.runSound(0);
};

Robot.prototype.restForBeats = function(request) {
	var self = this;
	var motoring = self.motoring;
	self.cancelNote();
	self.cancelSound();
	
	var beat = BEATS[request.params.beat];
	if(!beat) {
		beat = parseFloat(request.params.beat);
	}
	motoring.buzzer.write(0);
	motoring.note.write(0);
	self.runSound(0);
	if(beat && beat > 0 && self.tempo > 0) {
		var id = self.noteId = parseInt(request.params.id);
		self.waits.push(id);
		self.noteTimer1 = setTimeout(function() {
			if(self.noteId == id) {
				self.removeWait(self.noteId);
				self.cancelNote();
			}
		}, beat * 60 * 1000 / self.tempo);
		self.timeouts.push(self.noteTimer1);
	}
};

Robot.prototype.changeTempoBy = function(request) {
	var bpm = parseFloat(request.params.bpm);
	if(typeof bpm == 'number') {
		this.tempo += bpm;
		if(this.tempo < 1) this.tempo = 1;
	}
};

Robot.prototype.setTempoTo = function(request) {
	var bpm = parseFloat(request.params.bpm);
	if(typeof bpm == 'number') {
		this.tempo = bpm;
		if(this.tempo < 1) this.tempo = 1;
	}
};

Robot.prototype.runSound = function(sound, count) {
	if(typeof count != 'number') count = 1;
	if(count < 0) count = -1;
	if(count) {
		this.currentSound = sound;
		this.soundRepeat = count;
		this.motoring.sound.write(sound);
	}
};

Robot.prototype.handleSensory = function(data) {
	this.received = true;
	this.active = true;
	var sensory = this.sensory;
	sensory.floor = data.floor;
	sensory.accelerationX = data.acceleration.x;
	sensory.accelerationY = data.acceleration.y;
	sensory.accelerationZ = data.acceleration.z;
	
	if(data.map & 0x00001000) {
		sensory.button = data.button;
	}
	if(data.map & 0x00000800) {
		this.clicked = true;
	}
	if(data.map & 0x00000400) {
		this.doubleClicked = true;
	}
	if(data.map & 0x00000200) {
		this.longPressed = true;
	}
	if(data.map & 0x00000100) {
		sensory.colorNumber = data.colorNumber;
	}
	if(data.map & 0x00000080) {
		sensory.colorPattern = data.colorPattern;
	}
	if(data.map & 0x00000004) {
		sensory.tilt = data.tilt;
	}
	if(data.map & 0x00000002) {
		sensory.batteryState = data.batteryState;
	}
	
	var motoring = this.motoring;
	if(this.lineTracerId >= 0 && (data.map & 0x00000008) != 0) {
		if(data.lineTracerState == 0x02) {
			motoring.lineTracerMode.write(0);
			this.removeWait(this.lineTracerId);
			this.cancelLineTracer();
		}
	}
	if(this.motionId >= 0 && (data.map & 0x00000020) != 0) {
		if(data.wheelState == 0) {
			motoring.leftWheel.write(0);
			motoring.rightWheel.write(0);
			this.removeWait(this.motionId);
			this.cancelMotion();
		}
	}
	if((data.map & 0x00000010) != 0) {
		if(data.soundState == 0) {
			if(this.currentSound > 0) {
				if(this.soundRepeat < 0) {
					this.runSound(this.currentSound, -1);
				} else if(this.soundRepeat > 1) {
					this.soundRepeat --;
					this.runSound(this.currentSound, this.soundRepeat);
				} else {
					this.currentSound = 0;
					this.soundRepeat = 1;
					if(this.soundId >= 0) this.removeWait(this.soundId);
					this.cancelSound();
				}
			} else {
				this.currentSound = 0;
				this.soundRepeat = 1;
				if(this.soundId >= 0) this.removeWait(this.soundId);
				this.cancelSound();
			}
		}
	}
};

function Component() {
	var setup = function(map, key, value) {
		var data = translations[key];
		if(data) {
			if(value === undefined) value = key;
			for(var i in data) {
				map[data[i]] = value;
			}
			map[key] = value;
		}
	};
	setup(PARTS, 'left', LEFT);
	setup(PARTS, 'right', RIGHT);
	
	setup(DIRECTIONS, 'left', LEFT);
	setup(DIRECTIONS, 'right', RIGHT);
	setup(DIRECTIONS, 'back', BACK);
	
	setup(TOWARDS, 'head', HEAD);
	
	setup(UNITS, 'seconds', SECONDS);
	setup(UNITS, 'pulses', PULSES);
	setup(UNITS, 'degrees', DEGREES);
	
	setup(LINE_COLORS, 'black', 0);
	setup(LINE_COLORS, 'red', 1);
	setup(LINE_COLORS, 'yellow', 2);
	setup(LINE_COLORS, 'green', 3);
	setup(LINE_COLORS, 'sky blue', 4);
	setup(LINE_COLORS, 'blue', 5);
	setup(LINE_COLORS, 'purple', 6);
	setup(LINE_COLORS, 'any color', 7);
	
	setup(RGB_COLORS, 'red', [255, 0, 0]);
	setup(RGB_COLORS, 'orange', [255, 63, 0]);
	setup(RGB_COLORS, 'yellow', [255, 255, 0]);
	setup(RGB_COLORS, 'green', [0, 255, 0]);
	setup(RGB_COLORS, 'sky blue', [0, 255, 255]);
	setup(RGB_COLORS, 'blue', [0, 0, 255]);
	setup(RGB_COLORS, 'violet', [63, 0, 255]);
	setup(RGB_COLORS, 'purple', [255, 0, 255]);
	setup(RGB_COLORS, 'white', [255, 255, 255]);
	
	setup(NOTES, 'C', Turtle.NOTE_C_1);
	setup(NOTES, 'C#', Turtle.NOTE_C_SHARP_1);
	setup(NOTES, 'C♯ (D♭)', Turtle.NOTE_C_SHARP_1);
	setup(NOTES, 'D', Turtle.NOTE_D_1);
	setup(NOTES, 'D♯ (E♭)', Turtle.NOTE_E_FLAT_1);
	setup(NOTES, 'Eb', Turtle.NOTE_E_FLAT_1);
	setup(NOTES, 'E', Turtle.NOTE_E_1);
	setup(NOTES, 'F', Turtle.NOTE_F_1);
	setup(NOTES, 'F#', Turtle.NOTE_F_SHARP_1);
	setup(NOTES, 'F♯ (G♭)', Turtle.NOTE_F_SHARP_1);
	setup(NOTES, 'G', Turtle.NOTE_G_1);
	setup(NOTES, 'G#', Turtle.NOTE_G_SHARP_1);
	setup(NOTES, 'G♯ (A♭)', Turtle.NOTE_G_SHARP_1);
	setup(NOTES, 'A', Turtle.NOTE_A_1);
	setup(NOTES, 'A♯ (B♭)', Turtle.NOTE_B_FLAT_1);
	setup(NOTES, 'Bb', Turtle.NOTE_B_FLAT_1);
	setup(NOTES, 'B', Turtle.NOTE_B_1);
	
	setup(SOUNDS, 'beep', Turtle.SOUND_BEEP);
	setup(SOUNDS, 'random beep', Turtle.SOUND_RANDOM_BEEP);
	setup(SOUNDS, 'siren', Turtle.SOUND_SIREN);
	setup(SOUNDS, 'engine', Turtle.SOUND_ENGINE);
	setup(SOUNDS, 'robot', Turtle.SOUND_ROBOT);
	setup(SOUNDS, 'march', Turtle.SOUND_MARCH);
	setup(SOUNDS, 'birthday', Turtle.SOUND_BIRTHDAY);
	setup(SOUNDS, 'dibidibidip', Turtle.SOUND_DIBIDIBIDIP);
	setup(SOUNDS, 'good job', Turtle.SOUND_GOOD_JOB);
}

Component.prototype.updateLang = function(translator) { // for sensory
	var lang = translator.getLanguage();
	COLOR_NUMBERS = {};
	COLOR_PATTERNS = {};
	
	var get = function(key) {
		var value = undefined;
		var data = translations[key];
		if(data) value = data[lang];
		return value || key;
	};
	var setup = function(map, key, value) {
		key = get(key);
		map[key] = value;
	};
	
	setup(COLOR_NUMBERS, 'black', Turtle.COLOR_BLACK);
	setup(COLOR_NUMBERS, 'red', Turtle.COLOR_RED);
	setup(COLOR_NUMBERS, 'orange', Turtle.COLOR_ORANGE);
	setup(COLOR_NUMBERS, 'yellow', Turtle.COLOR_YELLOW);
	setup(COLOR_NUMBERS, 'green', Turtle.COLOR_GREEN);
	setup(COLOR_NUMBERS, 'sky blue', Turtle.COLOR_CYAN);
	setup(COLOR_NUMBERS, 'blue', Turtle.COLOR_BLUE);
	setup(COLOR_NUMBERS, 'purple', Turtle.COLOR_MAGENTA);
	setup(COLOR_NUMBERS, 'white', Turtle.COLOR_WHITE);
	
	setup(COLOR_PATTERNS, 'red', Turtle.COLOR_RED);
	setup(COLOR_PATTERNS, 'yellow', Turtle.COLOR_YELLOW);
	setup(COLOR_PATTERNS, 'green', Turtle.COLOR_GREEN);
	setup(COLOR_PATTERNS, 'sky blue', Turtle.COLOR_CYAN);
	setup(COLOR_PATTERNS, 'blue', Turtle.COLOR_BLUE);
	setup(COLOR_PATTERNS, 'purple', Turtle.COLOR_MAGENTA);
	
	CLICKED = '/' + get('clicked') + ' ';
	DOUBLE_CLICKED = '/' + get('double-clicked') + ' ';
	LONG_PRESSED = '/' + get('long-pressed') + ' ';
	
	BOOLEAN_TILT_FORWARD = '/' + get('tilt%20forward') + ' ';
	BOOLEAN_TILT_BACKWARD = '/' + get('tilt%20backward') + ' ';
	BOOLEAN_TILT_LEFT = '/' + get('tilt%20left') + ' ';
	BOOLEAN_TILT_RIGHT = '/' + get('tilt%20right') + ' ';
	BOOLEAN_TILT_FLIP = '/' + get('tilt%20flip') + ' ';
	BOOLEAN_TILT_NONE = '/' + get('not%20tilt') + ' ';
	BOOLEAN_BATTERY_NORMAL = '/' + get('normal') + ' ';
	BOOLEAN_BATTERY_LOW = '/' + get('low') + ' ';
	BOOLEAN_BATTERY_EMPTY = '/' + get('empty') + ' ';
};

Component.prototype.getModule = function() {
	return MODULE;
};

Component.prototype.createRobot = function(index) {
	return new Robot(index);
};

module.exports = new Component();