'use strict';
var router = require('router');
var translator = require('translator');
var component = require('./component');
const MODULE = 'turtle';

function Extension() {
	this.robots = {};
}

Extension.prototype.getRobotByIndex = function(index) {
	if(index >= 0) {
		var robot = this.robots[index];
		if(!robot) {
			robot = component.createRobot(index);
			if(robot) {
				this.robots[index] = robot;
			}
		}
		return robot;
	}
};

Extension.prototype.getRobotByRoute = function(route) {
	return this.getRobotByIndex(route.index);
};

Extension.prototype.start = function(express) {
	var self = this;
	var robots = self.robots;
	var problemMessage = '';
	
	self.running = true;
	for(var i in robots) {
		robots[i].clearWaits();
	}
	var setup = function() {
		problemMessage = '_problem ' + translator.translate('First robot is not connected.') + '\n';
		component.updateLang(translator);
	};
	translator.on('turtle-scratch-multi', function() {
		setup();
	});
	setup();
	
	if(express) {
		const HEADER = { 'Content-Type': 'text/plain' };
		
		express.get('/reset_all', function(request, response) {
			response.writeHead(200, HEADER);
			var robot;
			for(var i in robots) {
				robot = robots[i];
				robot.clearWaits();
				robot.clearTimeouts();
				robot.reset();
			}
			router.reset();
			response.end();
		});
		
		express.get('/poll', function(request, response) {
			if(!self.running) {
				response.set('Connection', 'close');
			}
			response.writeHead(200, HEADER);
			var waits, busy = '';
			for(var i in robots) {
				waits = robots[i].getWaits();
				if(waits.length > 0) {
					busy += ' ' + waits.join(' ');
				}
			}
			if(busy.length > 0) {
				response.write('_busy' + busy + '\n');
			}
			for(var i in robots) {
				robots[i].writeMultiSensory(response);
			}
			if(self.state != 'connected') {
				response.write(problemMessage);
			}
			response.end();
		});
		
		express.get('/turtleMoveForward/:id/:index', function(request, response) {
			response.writeHead(200, HEADER);
			var robot = self.getRobotByIndex(parseInt(request.params.index));
			if(robot) {
				robot.moveForward(request);
			}
			response.end();
		});
		
		express.get('/turtleMoveBackward/:id/:index', function(request, response) {
			response.writeHead(200, HEADER);
			var robot = self.getRobotByIndex(parseInt(request.params.index));
			if(robot) {
				robot.moveBackward(request);
			}
			response.end();
		});
		
		express.get('/turtleTurn/:id/:index/:direction', function(request, response) {
			response.writeHead(200, HEADER);
			var robot = self.getRobotByIndex(parseInt(request.params.index));
			if(robot) {
				robot.turn(request);
			}
			response.end();
		});
		
		express.get('/turtleMoveForwardUnit/:id/:index/:value/:unit', function(request, response) {
			response.writeHead(200, HEADER);
			var robot = self.getRobotByIndex(parseInt(request.params.index));
			if(robot) {
				robot.moveForwardUnit(request);
			}
			response.end();
		});
		
		express.get('/turtleMoveBackwardUnit/:id/:index/:value/:unit', function(request, response) {
			response.writeHead(200, HEADER);
			var robot = self.getRobotByIndex(parseInt(request.params.index));
			if(robot) {
				robot.moveBackwardUnit(request);
			}
			response.end();
		});
		
		express.get('/turtleTurnUnitInPlace/:id/:index/:direction/:value/:unit', function(request, response) {
			response.writeHead(200, HEADER);
			var robot = self.getRobotByIndex(parseInt(request.params.index));
			if(robot) {
				robot.turnUnitInPlace(request);
			}
			response.end();
		});
		
		express.get('/turtleTurnUnitWithRadiusInDirection/:id/:index/:direction/:value/:unit/:radius/:head', function(request, response) {
			response.writeHead(200, HEADER);
			var robot = self.getRobotByIndex(parseInt(request.params.index));
			if(robot) {
				robot.turnUnitWithRadiusInDirection(request);
			}
			response.end();
		});
		
		express.get('/turtlePivotAroundWheelUnitInDirection/:id/:index/:wheel/:value/:unit/:head', function(request, response) {
			response.writeHead(200, HEADER);
			var robot = self.getRobotByIndex(parseInt(request.params.index));
			if(robot) {
				robot.pivotAroundWheelUnitInDirection(request);
			}
			response.end();
		});
		
		express.get('/turtleChangeWheelsByLeftRight/:index/:left/:right', function(request, response) {
			response.writeHead(200, HEADER);
			var robot = self.getRobotByIndex(parseInt(request.params.index));
			if(robot) {
				robot.changeWheelsByLeftRight(request);
			}
			response.end();
		});
		
		express.get('/turtleSetWheelsToLeftRight/:index/:left/:right', function(request, response) {
			response.writeHead(200, HEADER);
			var robot = self.getRobotByIndex(parseInt(request.params.index));
			if(robot) {
				robot.setWheelsToLeftRight(request);
			}
			response.end();
		});
		
		express.get('/turtleChangeWheelBy/:index/:wheel/:speed', function(request, response) {
			response.writeHead(200, HEADER);
			var robot = self.getRobotByIndex(parseInt(request.params.index));
			if(robot) {
				robot.changeWheelBy(request);
			}
			response.end();
		});
		
		express.get('/turtleSetWheelTo/:index/:wheel/:speed', function(request, response) {
			response.writeHead(200, HEADER);
			var robot = self.getRobotByIndex(parseInt(request.params.index));
			if(robot) {
				robot.setWheelTo(request);
			}
			response.end();
		});
		
		express.get('/turtleFollowLine/:index/:color', function(request, response) {
			response.writeHead(200, HEADER);
			var robot = self.getRobotByIndex(parseInt(request.params.index));
			if(robot) {
				robot.followLine(request);
			}
			response.end();
		});
		
		express.get('/turtleFollowLineUntil/:id/:index/:color', function(request, response) {
			response.writeHead(200, HEADER);
			var robot = self.getRobotByIndex(parseInt(request.params.index));
			if(robot) {
				robot.followLineUntil(request);
			}
			response.end();
		});
		
		express.get('/turtleFollowLineUntilBlack/:id/:index/:color', function(request, response) {
			response.writeHead(200, HEADER);
			var robot = self.getRobotByIndex(parseInt(request.params.index));
			if(robot) {
				robot.followLineUntilBlack(request);
			}
			response.end();
		});
		
		express.get('/turtleCrossIntersection/:id/:index', function(request, response) {
			response.writeHead(200, HEADER);
			var robot = self.getRobotByIndex(parseInt(request.params.index));
			if(robot) {
				robot.crossIntersection(request);
			}
			response.end();
		});
		
		express.get('/turtleTurnAtIntersection/:id/:index/:direction', function(request, response) {
			response.writeHead(200, HEADER);
			var robot = self.getRobotByIndex(parseInt(request.params.index));
			if(robot) {
				robot.turnAtIntersection(request);
			}
			response.end();
		});
		
		express.get('/turtleSetFollowingSpeedTo/:index/:speed', function(request, response) {
			response.writeHead(200, HEADER);
			var robot = self.getRobotByIndex(parseInt(request.params.index));
			if(robot) {
				robot.setFollowingSpeedTo(request);
			}
			response.end();
		});
		
		express.get('/turtleStop/:index', function(request, response) {
			response.writeHead(200, HEADER);
			var robot = self.getRobotByIndex(parseInt(request.params.index));
			if(robot) {
				robot.stop(request);
			}
			response.end();
		});
		
		express.get('/turtleSetHeadLedTo/:index/:color', function(request, response) {
			response.writeHead(200, HEADER);
			var robot = self.getRobotByIndex(parseInt(request.params.index));
			if(robot) {
				robot.setHeadLedTo(request);
			}
			response.end();
		});
		
		express.get('/turtleChangeHeadLedByRGB/:index/:red/:green/:blue', function(request, response) {
			response.writeHead(200, HEADER);
			var robot = self.getRobotByIndex(parseInt(request.params.index));
			if(robot) {
				robot.changeHeadLedByRGB(request);
			}
			response.end();
		});
		
		express.get('/turtleSetHeadLedToRGB/:index/:red/:green/:blue', function(request, response) {
			response.writeHead(200, HEADER);
			var robot = self.getRobotByIndex(parseInt(request.params.index));
			if(robot) {
				robot.setHeadLedToRGB(request);
			}
			response.end();
		});
		
		express.get('/turtleClearHeadLed/:index', function(request, response) {
			response.writeHead(200, HEADER);
			var robot = self.getRobotByIndex(parseInt(request.params.index));
			if(robot) {
				robot.clearHeadLed(request);
			}
			response.end();
		});
		
		express.get('/turtlePlaySound/:index/:sound', function(request, response) {
			response.writeHead(200, HEADER);
			var robot = self.getRobotByIndex(parseInt(request.params.index));
			if(robot) {
				robot.playSound(request);
			}
			response.end();
		});
		
		express.get('/turtlePlaySoundTimes/:index/:sound/:count', function(request, response) {
			response.writeHead(200, HEADER);
			var robot = self.getRobotByIndex(parseInt(request.params.index));
			if(robot) {
				robot.playSoundTimes(request);
			}
			response.end();
		});
		
		express.get('/turtlePlaySoundTimesUntilDone/:id/:index/:sound/:count', function(request, response) {
			response.writeHead(200, HEADER);
			var robot = self.getRobotByIndex(parseInt(request.params.index));
			if(robot) {
				robot.playSoundTimesUntilDone(request);
			}
			response.end();
		});
		
		express.get('/turtleChangeBuzzerBy/:index/:hz', function(request, response) {
			response.writeHead(200, HEADER);
			var robot = self.getRobotByIndex(parseInt(request.params.index));
			if(robot) {
				robot.changeBuzzerBy(request);
			}
			response.end();
		});
		
		express.get('/turtleSetBuzzerTo/:index/:hz', function(request, response) {
			response.writeHead(200, HEADER);
			var robot = self.getRobotByIndex(parseInt(request.params.index));
			if(robot) {
				robot.setBuzzerTo(request);
			}
			response.end();
		});
		
		express.get('/turtleClearSound/:index', function(request, response) {
			response.writeHead(200, HEADER);
			var robot = self.getRobotByIndex(parseInt(request.params.index));
			if(robot) {
				robot.clearSound(request);
			}
			response.end();
		});
		
		express.get('/turtlePlayNote/:index/:note/:octave', function(request, response) {
			response.writeHead(200, HEADER);
			var robot = self.getRobotByIndex(parseInt(request.params.index));
			if(robot) {
				robot.playNote(request);
			}
			response.end();
		});
		
		express.get('/turtlePlayNoteForBeats/:id/:index/:note/:octave/:beat', function(request, response) {
			response.writeHead(200, HEADER);
			var robot = self.getRobotByIndex(parseInt(request.params.index));
			if(robot) {
				robot.playNoteForBeats(request);
			}
			response.end();
		});
		
		express.get('/turtleRestForBeats/:id/:index/:beat', function(request, response) {
			response.writeHead(200, HEADER);
			var robot = self.getRobotByIndex(parseInt(request.params.index));
			if(robot) {
				robot.restForBeats(request);
			}
			response.end();
		});
		
		express.get('/turtleChangeTempoBy/:index/:bpm', function(request, response) {
			response.writeHead(200, HEADER);
			var robot = self.getRobotByIndex(parseInt(request.params.index));
			if(robot) {
				robot.changeTempoBy(request);
			}
			response.end();
		});
		
		express.get('/turtleSetTempoTo/:index/:bpm', function(request, response) {
			response.writeHead(200, HEADER);
			var robot = self.getRobotByIndex(parseInt(request.params.index));
			if(robot) {
				robot.setTempoTo(request);
			}
			response.end();
		});
	}
};

Extension.prototype.stop = function() {
	this.running = false;
	translator.remove('turtle-scratch-multi');
	router.reset();
	this.robots = {};
};

Extension.prototype.setState = function(route) {
	if(route == router.firstRoute && route.module == MODULE) {
		this.state = route.state;
	}
};

Extension.prototype.setOption = function(option) {
};

Extension.prototype.handleSensory = function(data, route) {
	if(route.module == MODULE) {
		var robot = this.getRobotByRoute(route);
		if(robot) {
			robot.handleSensory(data);
		}
	}
};

Extension.prototype.handleMotoring = function(data, route) {
	if(route.module == MODULE) {
		var robot = this.getRobotByRoute(route);
		if(robot) {
			var roboid = robot.roboid;
			roboid.handleMotoring(data);
			roboid.updateMotoring();
		}
	}
};

Extension.prototype.requestMotoring = function(route) {
	if(route.module == MODULE) {
		var robot = this.getRobotByRoute(route);
		if(robot) {
			return robot.roboid.requestMotoring();
		}
	}
};

Extension.prototype.reset = function(route) {
	if(route.module == MODULE) {
		var robot = this.getRobotByRoute(route);
		if(robot) {
			robot.roboid.reset();
			robot.motoring.reset();
		}
	}
};

module.exports.create = function() {
	return new Extension();
};