'use strict';
var DeviceType = {
	SENSOR: 0,
	EFFECTOR: 1,
	EVENT: 2,
	COMMAND: 3
};

var DataType = {
	INTEGER: 4,
	FLOAT: 5,
	STRING: 6
};

function Device(parent, uid, name, deviceType, dataType, dataSize, minValue, maxValue, initialValue) {
	var self = this;
	var data = {
		read: undefined,
		write: undefined
	};
	var deviceName = name;
	var event = false;
	var fired = false;
	var deviceDataChangedListeners = [];
	
	if(dataSize > 0) {
		data.read = new Array(dataSize);
		data.write = new Array(dataSize);
		for(var i = 0; i < dataSize; ++i) {
			data.read[i] = initialValue;
			data.write[i] = initialValue;
		}
	}
	
	parent.bind(this, deviceType, data, {
		fire: function() {
			fired = true;
		},
		update: function() {
			event = fired;
			fired = false;
		},
		notifyDeviceDataChanged: function() {
			for(var i in deviceDataChangedListeners) {
				deviceDataChangedListeners[i](self, data.read);
			}
		},
		reset: function() {
			event = false;
			fired = false;
			if(dataSize > 0) {
				for(var i = 0; i < dataSize; ++i) {
					data.read[i] = initialValue;
					data.write[i] = initialValue;
				}
			} else {
				data.read = undefined;
				data.write = undefined;
			}
		}
	});
	
	this.getId = function() {
		return uid & 0xfff00fff;
	};
	
	this.getName = function() {
		return deviceName;
	};
	
	this.setName = function(name) {
		deviceName = '' + name;
	};
	
	this.getDeviceType = function() {
		return deviceType;
	};
	
	this.getDataType = function() {
		return dataType;
	};
	
	this.getDataSize = function() {
		return dataSize;
	};
	
	this.e = function() {
		return event;
	};
	
	var checkDataType = function(value) {
		if(dataType == DataType.INTEGER || dataType == DataType.FLOAT) {
			return typeof value == 'number';
		} else if(dataType == DataType.STRING) {
			return typeof value == 'string';
		}
		return false;
	};
	
	this.read = function(arg) {
		var readData = data.read;
		if(typeof arg == 'number') { // index
			var index = parseInt(arg);
			if(index < 0) {
				if(dataType == DataType.STRING)
					return '';
				else
					return 0;
			}
			if(readData == undefined || index >= readData.length) {
				if(dataType == DataType.STRING)
					return '';
				else
					return 0;
			}
			return readData[index];
		} else if(arg && arg.constructor == Array) { //array
			var dataLen = arg.length;
			if(dataLen <= 0) {
				if(dataType == DataType.STRING)
					return '';
				else
					return 0;
			}
			if(readData == undefined) {
				if(dataType == DataType.STRING)
					return '';
				else
					return 0;
			}
			var len = Math.min(readData.length, dataLen);
			for(var i = 0; i < len; ++i) {
				arg[i] = readData[i];
			}
			return len;
		} else { // return data at index 0
			if(readData == undefined || readData.length <= 0) {
				if(dataType == DataType.STRING)
					return '';
				else
					return 0;
			}
			return readData[0];
		}
	};
	
	this.write = function(arg1, arg2) {
		if(deviceType == DeviceType.SENSOR || deviceType == DeviceType.EVENT) {
			return false;
		}
		var writeData = data.write;
		if(arguments.length == 0) {
			fired = true;
			if(parent) {
				parent.onMotoringDeviceWritten(this);
			}
			return true;
		} else if((typeof arg1 == 'number') && ((typeof arg2 == 'number') || (typeof arg2 == 'string'))) { // data at index
			var index = parseInt(arg1);
			if(index < 0) {
				return false;
			}
			if(checkDataType(arg2) == false) {
				return false;
			}
			if(writeData == undefined) {
				if(dataSize < 0) {
					writeData = data.write = new Array(index + 1);
					for(var i = 0; i < index; ++i) {
						writeData[i] = initialValue;
					}
				} else {
					return false;
				}
			} else if(index >= writeData.length) {
				if(dataSize < 0) {
					var newData = new Array(index + 1);
					var len = writeData.length;
					var i;
					for(i = 0; i < len; ++i) {
						newData[i] = writeData[i];
					}
					for(i = len; i < index; ++i) {
						newData[i] = initialValue;
					}
					writeData = data.write = newData;
				} else {
					return false;
				}
			}
			if(typeof arg2 == 'number') {
				if(arg2 < minValue) arg2 = minValue;
				else if(arg2 > maxValue) arg2 = maxValue;
			}
			writeData[index] = arg2;
			fired = true;
			if(parent) {
				parent.onMotoringDeviceWritten(this);
			}
			return true;
		} else if(arg1 && arg1.constructor == Array) { // array
			var dataLen = arg1.length;
			if(dataLen <= 0) {
				return 0;
			}
			if(dataSize < 0) {
				if(writeData == undefined || writeData.length != dataLen) {
					writeData = data.write = new Array(dataLen);
				}
			}
			if(writeData == undefined) {
				return 0;
			}
			var thisLen = writeData.length;
			if(thisLen <= 0) {
				return 0;
			}
			var len = Math.min(thisLen, dataLen);
			var i, value;
			for(i = 0; i < len; ++i) {
				value = arg1[i];
				if(checkDataType(value)) {
					if(typeof value == 'number') {
						if(value < minValue) value = minValue;
						else if(value > maxValue) value = maxValue;
					}
					writeData[i] = value;
					fired = true;
				}
			}
			if(parent) {
				parent.onMotoringDeviceWritten(this);
			}
			return len;
		} else if((typeof arg1 == 'number') || (typeof arg1 == 'string')) { // write data at index 0
			if(checkDataType(arg1) == false) {
				return false;
			}
			if(writeData == undefined || writeData.length <= 0) {
				if(dataSize < 0) {
					writeData = data.write = new Array(1);
				} else {
					return false;
				}
			}
			if(typeof arg1 == 'number') {
				if(arg1 < minValue) arg1 = minValue;
				else if(arg1 > maxValue) arg1 = maxValue;
			}
			writeData[0] = arg1;
			fired = true;
			if(parent) {
				parent.onMotoringDeviceWritten(this);
			}
			return true;
		} else {
			return false;
		}
	};
	
	this.addDeviceDataChangedListener = function(listener) {
		if(typeof listener == 'function') {
			deviceDataChangedListeners.push(listener);
		}
	};
	
	this.removeDeviceDataChangedListener = function(listener) {
		if(typeof listener == 'function') {
			var index = deviceDataChangedListeners.indexOf(listener);
			if(index >= 0) {
				deviceDataChangedListeners.splice(index, 1);
			}
		}
	};
	
	this.clearDeviceDataChangedListeners = function() {
		deviceDataChangedListeners = [];
	};
}

module.exports.Device = Device;
module.exports.DeviceType = DeviceType;
module.exports.DataType = DataType;