'use strict';
function Connector() {
	this.ports = [];
}

Connector.prototype.open = function(port, callback) {
	var serialport = require('serialport');
	var sp = new serialport.SerialPort(port, {
		baudRate: 115200, flowControl: true,
		parser: serialport.parsers.readline('\r')
	});
	this.sp = sp;
	sp.on('error', function(error) {});
	sp.on('open', function(error) {
		sp.removeAllListeners('open');
		if(callback) {
			callback(error, sp);
		}
	});
};

Connector.prototype.load = function(callback) {
	var sp = this.sp;
	if(sp && sp.isOpen()) {
		sp.on('data', function(data) {
			if(callback) {
				callback(null, data);
			}
		});
	} else {
		if(callback) {
			callback('error');
		}
	}
};

Connector.prototype.unload = function() {
	if(this.sp) {
		this.sp.removeAllListeners('data');
	}
};

Connector.prototype.connect = function(callback) {
	var self = this;
	if(self.sp) {
		self.connected = false;
		self.received = true;
		var sp = self.sp;
		sp.on('data', function(data) {
			if(data.length == 53) {
				if(self.connected == false) {
					self.connected = true;
					if(callback) {
						callback('connected', data);
					}
				}
				self.received = true;
				if(callback) {
					callback(null, data);
				}
			}
		});
		sp.on('disconnect', function() {
			self.close();
			if(callback) {
				callback('disconnected');
			}
		});
		self.timer = setInterval(function() {
			if(self.connected) {
				if(self.received == false) {
					self.connected = false;
					if(callback) {
						callback('lost');
					}
				}
				self.received = false;
			}
		}, 500);
	}
};

Connector.prototype.close = function() {
	this.connected = false;
	if(this.timer) {
		clearInterval(this.timer);
		this.timer = undefined;
	}
	this.clearReconnectTimer();
	if(this.sp) {
		this.removePort(this.sp.path);
		this.sp.removeAllListeners();
		if(this.sp.isOpen()) {
			this.sp.close();
		}
		this.sp = undefined;
	}
};

Connector.prototype.send = function(data) {
	if(this.sp && this.sp.isOpen()) {
		this.sp.write(data);
	}
};

Connector.prototype.reconnect = function(route, callback) {
	var serialport = require('serialport');
	var self = this;
	
	self.scanning = true;
	self.connectors = {};
	self.closeConnectors();
	self.clearReconnectTimer();
	self.reconnectTimer = setInterval(function() {
		serialport.list(function(error, devices) {
			if(error) {
				if(callback) {
					callback(error);
				}
				return;
			}
			
			var config = route.config;
			var vendor = config.vendor;
			devices.forEach(function(device) {
				if(!self.scanning) return;
				
				var comName = device.comName;
				if(vendor == device.manufacturer && self.ports.indexOf(comName) < 0) {
					var connector = self.connectors[comName];
					if(connector == undefined) {
						connector = require('../connector/serial').create();
						self.connectors[comName] = connector;
						
						connector.open(comName, function(error, sp) {
							if(error) {
								if(callback) {
									callback(error);
								}
							} else {
								sp.on('data', function(data) {
									if(data.slice(0, 2) == 'FF') {
										self.validate(config, data, function(result, property) {
											if(result && route.address == property.address) {
												sp.removeAllListeners('data');
												self.connectors[comName] = undefined;
												self.scanning = false;
												self.clearReconnectTimer();
												self.closeConnectors();
												self.ports.push(comName);
												if(callback) {
													property.port = comName;
													callback(null, connector, property);
												}
											}
										});
									} else {
										sp.write('FF\r');
									}
								});
							}
						});
					}
				}
			});
		});
	}, 2000);
};

Connector.prototype.removePort = function(port) {
	var index = this.ports.indexOf(port);
	if(index >= 0) {
		this.ports.splice(index, 1);
	}
};

Connector.prototype.clearReconnectTimer = function() {
	if(this.reconnectTimer) {
		clearInterval(this.reconnectTimer);
		this.reconnectTimer = undefined;
	}
};

Connector.prototype.closeConnectors = function() {
	var connectors = this.connectors;
	if(connectors) {
		var connector;
		for(var key in connectors) {
			connector = connectors[key];
			if(connector) {
				connector.close();
			}
		}
	}
	this.connectors = {};
};

Connector.prototype.validate = function(config, data, callback) {
	if(callback) {
		var info = data.split(',');
		if(info && info.length >= 5) {
			var property = this.validateInfo(info, config);
			if(property) {
				callback(true, property);
			} else {
				callback(false);
			}
		}
	}
};

Connector.prototype.validateInfo = function(info, config) {
	if((config.deviceNameChangeable || info[1] == config.deviceName) && info[2] == config.id.substring(2, 4) && info[4].length >= 12) {
		var id = config.id.substring(0, 4) + info[3];
		var address = info[4].substring(0, 12);
		return {
			id: id,
			address: address
		};
	} else {
		return null;
	}
};

module.exports.create = function() {
	return new Connector();
};