'use strict';
var util = require('util');
var EventEmitter = require('events').EventEmitter;
var logger = require('logger').get();

function Router() {
	EventEmitter.call(this);
	this.scanner = require('./scanner/scanner');
	this.routes = [];
	this.modules = {};
	this.numberOfActiveModules = 0;
	this.majorIdentityActive = true;
	this.minorIdentityActive = true;
}

util.inherits(Router, EventEmitter);

Router.prototype.startScan = function(configs) {
	var self = this;
	self.configs = configs;
	if(self.scanner) {
		self.scanner.startScan(self, configs, function(state) {
			if(state == 'done') {
				self.emit('scan-done');
			}
		});
	}
};

Router.prototype.stopScan = function() {
	if(this.scanner) {
		this.scanner.stopScan();
	}
};

Router.prototype.found = function(controller, config, property) {
	var route = this.findRouteByPort(property.port);
	if(route) {
		var oldModule = route.hwModule;
		var oldIndex = route.index;
		route.config = config;
		route.model = config.model;
		route.module = config.module;
		route.hwModule = config.module;
		route.group = config.group;
		route.id = property.id;
		route.port = property.port;
		route.address = property.address;
		route.dongle = property.dongle;
		route.index = this.findEmptyRouteIndex(config.group, route);
		route.controller = controller;
		route.state = 'connecting';
		var str = config.deviceName + '(' + config.type;
		if(property.port) {
			str += ' ' + property.port;
		}
		str += '): ' + property.address + ' found';
		logger.i(str);
		this.updateModules();
		this.emit('update', oldModule, oldIndex, route);
		return route;
	} else {
		var added = false;
		route = this.findRouteByAddress(config.module, property.address);
		if(!route) {
			route = {};
			route.model = config.model;
			route.module = config.module;
			route.hwModule = config.module;
			route.group = config.group;
			route.index = this.findEmptyRouteIndex(config.group);
			this.addRoute(route);
			added = true;
		} else if(route.address == undefined) {
			added = true;
		}
		if(added) {
			route.config = config;
			route.controller = controller;
			route.state = 'connecting';
			var str = config.deviceName + '(' + config.type;
			if(property.port) {
				str += ' ' + property.port;
			}
			str += '): ' + property.address + ' found';
			logger.i(str);
		}
		route.id = property.id;
		route.port = property.port;
		route.address = property.address;
		route.dongle = property.dongle;
		if(added) {
			this.emit('found', route);
		}
		return route;
	}
};

Router.prototype.connect = function(connector, route) {
	var self = this;

	route.connector = connector;
	connector.connect(function(state, data) {
		if(state) {
			var config = route.config;
			var str = config.deviceName + '(' + config.type;
			if(route.port) {
				str += ' ' + route.port;
			}
			str += '): ' + route.address;
			if(state == 'connected') {
				route.state = 'connected';
				var newAddress = data.slice(41, 53);
				if(route.address != newAddress) {
					if(route.connector) {
						route.connector.close();
					}
					if(self.scanner) {
						self.scanner.startScan(self, self.configs);
					}
					return;
				}
				logger.i(str + ' connected');
			} else if(state == 'disconnected') {
				route.state = 'disconnected';
				logger.w(str + ' disconnected');
				self.removeRoute(route);
				//logger.i(str + ' connecting');
				
				//connector.reconnect(route, function(error, conn, property) {
				//	if(error) {
				//		logger.e(error);
				//	} else {
				//		self.found(route.controller, config, property);
				//		self.connect(conn, route);
				//	}
				//});
			} else {
				route.state = 'lost';
				logger.w(str + ' connection lost');
			}
			self.emit('state', route);
		} else if(self.multi || (self.selectedRoute && route == self.selectedRoute) || self.isExtraButtonActive()) {
			if(route.controller) {
				var controller = route.controller;
				var extension = self.extension;
				
				var sensory = controller.handleLocalData(data, route); // json simulacrum
				if(sensory && extension) {
					extension.handleSensory(sensory, route);
				}
				
				var motoring;
				if(extension) {
					motoring = controller.write(extension.requestMotoring(route));
				}
				
				if(route.dongle && route.dongle.doReset) {
					route.dongle.doReset = false;
					connector.send('EA\r');
				} else {
					// write to hardware
					var packet = controller.requestLocalData(route.address, route);
					if(packet) {
						connector.send(packet);
					}
				}
				
				if(extension) {
					extension.handleMotoring(motoring, route);
				}
				controller.clear();
			}
		}
	});
};

Router.prototype.setExtension = function(extension) {
	this.extension = extension;
};

Router.prototype.getRepresentativeRoute = function() {
	if(this.multi) {
		return this.firstRoute;
	} else {
		if(this.selectedRoute) return this.selectedRoute;
		else return this.firstRoute;
	}
};

Router.prototype.getRoutes = function() {
	return this.routes;
};

Router.prototype.getModules = function() {
	return this.modules;
};

Router.prototype.addRoute = function(route) {
	this.routes.push(route);
	if(!this.firstRoute) {
		this.firstRoute = route;
	}
	var modules = this.modules;
	var module = route.module;
	if(module in modules) {
		var count = modules[module];
		if(count < 1) {
			this.numberOfActiveModules ++;
		}
		modules[module] = count + 1;
	} else {
		modules[module] = 1;
		this.numberOfActiveModules ++;
	}
};

Router.prototype.removeRoute = function(route) {
	var index = this.routes.indexOf(route);
	if(index >= 0) {
		this.routes.splice(index, 1);
	}
	if(this.routes.length > 0) {
		this.firstRoute = this.routes[0];
	} else {
		this.firstRoute = undefined;
	}
	var modules = this.modules;
	var module = route.module;
	if(module in modules) {
		var count = modules[module] - 1;
		if(count <= 0) {
			count = 0;
			this.numberOfActiveModules --;
		}
		modules[module] = count;
	}
};

Router.prototype.updateModules = function() {
	var routes = this.routes;
	var modules = this.modules = {};
	this.numberOfActiveModules = 0;
	
	var module, count;
	for(var i in routes) {
		module = routes[i].module;
		if(module in modules) {
			count = modules[module];
			if(count < 1) {
				this.numberOfActiveModules ++;
			}
			modules[module] = count + 1;
		} else {
			modules[module] = 1;
			this.numberOfActiveModules ++;
		}
	}
};

Router.prototype.findEmptyRouteIndex = function(group, self) {
	var indices = [];
	var routes = this.routes;
	var route, index, max = -1;
	for(var i = 0, len = routes.length; i < len; ++i) {
		route = routes[i];
		if(group == route.group && route != self) {
			index = route.index;
			indices.push(index);
			if(index > max) max = index;
		}
	}
	for(var i = 0; i <= max; ++i) {
		if(indices.indexOf(i) < 0) {
			return i;
		}
	}
	return max + 1;
};

Router.prototype.findRouteByPort = function(port) {
	var route, routes = this.routes;
	for(var i = 0, len = routes.length; i < len; ++i) {
		route = routes[i];
		if(port == route.port) {
			return route;
		}
	}
	return null;
};

Router.prototype.findRouteByAddress = function(module, address) {
	var route, routes = this.routes;
	for(var i = 0, len = routes.length; i < len; ++i) {
		route = routes[i];
		if(module == route.module) {
			if(route.address) {
				if(address == route.address) {
					return route;
				}
			} else {
				return route;
			}
		}
	}
	return null;
};

Router.prototype.findRouteByIndex = function(group, index) {
	if(index < 0) return null;
	var route, routes = this.routes;
	for(var i = 0, len = routes.length; i < len; ++i) {
		route = routes[i];
		if(group == route.group && index == route.index) {
			return route;
		}
	}
	return null;
};

Router.prototype.resetRoute = function(route) {
	if(route) {
		if(route.controller) {
			route.controller.reset();
		}
		if(this.extension) {
			this.extension.reset(route);
		}
	}
};

Router.prototype.resetOtherRoutes = function(route) {
	var routes = this.routes;
	for(var i = 0, len = routes.length; i < len; ++i) {
		if(routes[i] != route) {
			this.resetRoute(routes[i]);
		}
	}
};

Router.prototype.reset = function(route) {
	if(route) {
		this.resetRoute(route);
	} else {
		var routes = this.routes;
		for(var i = 0, len = routes.length; i < len; ++i) {
			this.resetRoute(routes[i]);
		}
	}
};

Router.prototype.dispose = function(route) {
	this.reset(route);
};

Router.prototype.close = function() {
	if(this.scanner) {
		this.scanner.stopScan();
	}
	var connector, routes = this.routes;
	for(var i = 0, len = routes.length; i < len; ++i) {
		connector = routes[i].connector;
		if(connector) {
			connector.close();
		}
	}
};

Router.prototype.checkConnectionState = function(module) {
	var route, routes = this.routes, count = 0;
	for(var i = 0, len = routes.length; i < len; ++i) {
		route = routes[i];
		if(module == route.module) {
			count ++;
			if(route.state != 'connected') return -2; // not all connected
		}
	}
	if(count > 0) return 2; // all connected
	return -3; // nothing
};

Router.prototype.getConnectionState = function(state) {
	switch(state) {
		case 'connecting': {
			return 1;
		}
		case 'connected': {
			return 2;
		}
		case 'lost': {
			return 3;
		}
		case 'disconnected': {
			return 4;
		}
	}
	return 0;
};

Router.prototype.setMajorExtraButtonActive = function(active) {
	this.majorIdentityActive = active;
};

Router.prototype.setMinorExtraButtonActive = function(active) {
	this.minorIdentityActive = active;
};

Router.prototype.isExtraButtonActive = function() {
	return this.majorIdentityActive && this.minorIdentityActive;
};

module.exports = new Router();