var DeviceType = {
	SENSOR: 0,
	EFFECTOR: 1,
	EVENT: 2,
	COMMAND: 3
};

var DataType = {
	INTEGER: 4,
	FLOAT: 5,
	STRING: 6
};

var ConnectionState = {
	CONNECTING: 1,
	CONNECTED: 2,
	CONNECTION_LOST: 3,
	DISCONNECTED: 4,
	CLOSED: 5
};

function Device(binder, uid, name, deviceType, dataType, dataSize, minValue, maxValue, initialValue) {
	var self = this;
	var data = {
		read: undefined,
		write: undefined
	};
	var deviceName = name;
	var event = false;
	var fired = false;
	var deviceDataChangedListeners = [];
	
	if(dataSize > 0) {
		data.read = new Array(dataSize);
		data.write = new Array(dataSize);
		for(var i = 0; i < dataSize; ++i) {
			data.read[i] = initialValue;
			data.write[i] = initialValue;
		}
	}
	
	binder.bind(self, deviceType, data, {
		fire: function() {
			fired = true;
		},
		update: function() {
			event = fired;
			fired = false;
		},
		notifyDeviceDataChanged: function() {
			for(var i in deviceDataChangedListeners) {
				deviceDataChangedListeners[i](self, data.read);
			}
		},
		reset: function() {
			event = false;
			fired = false;
			if(dataSize > 0) {
				for(var i = 0; i < dataSize; ++i) {
					data.read[i] = initialValue;
					data.write[i] = initialValue;
				}
			} else {
				data.read = undefined;
				data.write = undefined;
			}
		}
	});
	
	self.getId = function() {
		return uid & 0xfff00fff;
	};
	
	self.getName = function() {
		return deviceName;
	};
	
	self.setName = function(name) {
		deviceName = '' + name;
	};
	
	self.getDeviceType = function() {
		return deviceType;
	};
	
	self.getDataType = function() {
		return dataType;
	};
	
	self.getDataSize = function() {
		return dataSize;
	};
	
	self.e = function() {
		return event;
	};
	
	var checkDataType = function(value) {
		if(dataType == DataType.INTEGER || dataType == DataType.FLOAT) {
			return typeof value == 'number';
		} else if(dataType == DataType.STRING) {
			return typeof value == 'string';
		}
		return false;
	};
	
	self.read = function(arg) {
		var readData = data.read;
		if(typeof arg == 'number') { // index
			var index = parseInt(arg);
			if(index < 0) {
				if(dataType == DataType.STRING)
					return '';
				else
					return 0;
			}
			if(readData == undefined || index >= readData.length) {
				if(dataType == DataType.STRING)
					return '';
				else
					return 0;
			}
			return readData[index];
		} else if(arg && arg.constructor == Array) { //array
			var dataLen = arg.length;
			if(dataLen <= 0) {
				if(dataType == DataType.STRING)
					return '';
				else
					return 0;
			}
			if(readData == undefined) {
				if(dataType == DataType.STRING)
					return '';
				else
					return 0;
			}
			var len = Math.min(readData.length, dataLen);
			for(var i = 0; i < len; ++i) {
				arg[i] = readData[i];
			}
			return len;
		} else { // return data at index 0
			if(readData == undefined || readData.length <= 0) {
				if(dataType == DataType.STRING)
					return '';
				else
					return 0;
			}
			return readData[0];
		}
	};
	
	self.write = function(arg1, arg2) {
		if(deviceType == DeviceType.SENSOR || deviceType == DeviceType.EVENT) {
			return false;
		}
		var writeData = data.write;
		if(arguments.length == 0) {
			fired = true;
			if(binder) {
				binder.onMotoringDeviceWritten(self);
			}
			return true;
		} else if((typeof arg1 == 'number') && ((typeof arg2 == 'number') || (typeof arg2 == 'string'))) { // data at index
			var index = parseInt(arg1);
			if(index < 0) {
				return false;
			}
			if(checkDataType(arg2) == false) {
				return false;
			}
			if(writeData == undefined) {
				if(dataSize < 0) {
					writeData = data.write = new Array(index + 1);
					for(var i = 0; i < index; ++i) {
						writeData[i] = initialValue;
					}
				} else {
					return false;
				}
			} else if(index >= writeData.length) {
				if(dataSize < 0) {
					var newData = new Array(index + 1);
					var len = writeData.length;
					var i;
					for(i = 0; i < len; ++i) {
						newData[i] = writeData[i];
					}
					for(i = len; i < index; ++i) {
						newData[i] = initialValue;
					}
					writeData = data.write = newData;
				} else {
					return false;
				}
			}
			if(typeof arg2 == 'number') {
				if(arg2 < minValue) arg2 = minValue;
				else if(arg2 > maxValue) arg2 = maxValue;
			}
			writeData[index] = arg2;
			fired = true;
			if(binder) {
				binder.onMotoringDeviceWritten(self);
			}
			return true;
		} else if(arg1 && arg1.constructor == Array) { // array
			var dataLen = arg1.length;
			if(dataLen <= 0) {
				return 0;
			}
			if(dataSize < 0) {
				if(writeData == undefined || writeData.length != dataLen) {
					writeData = data.write = new Array(dataLen);
				}
			}
			if(writeData == undefined) {
				return 0;
			}
			var thisLen = writeData.length;
			if(thisLen <= 0) {
				return 0;
			}
			var len = Math.min(thisLen, dataLen);
			var i, value;
			for(i = 0; i < len; ++i) {
				value = arg1[i];
				if(checkDataType(value)) {
					if(typeof value == 'number') {
						if(value < minValue) value = minValue;
						else if(value > maxValue) value = maxValue;
					}
					writeData[i] = value;
					fired = true;
				}
			}
			if(binder) {
				binder.onMotoringDeviceWritten(self);
			}
			return len;
		} else if((typeof arg1 == 'number') || (typeof arg1 == 'string')) { // write data at index 0
			if(checkDataType(arg1) == false) {
				return false;
			}
			if(writeData == undefined || writeData.length <= 0) {
				if(dataSize < 0) {
					writeData = data.write = new Array(1);
				} else {
					return false;
				}
			}
			if(typeof arg1 == 'number') {
				if(arg1 < minValue) arg1 = minValue;
				else if(arg1 > maxValue) arg1 = maxValue;
			}
			writeData[0] = arg1;
			fired = true;
			if(binder) {
				binder.onMotoringDeviceWritten(self);
			}
			return true;
		} else {
			return false;
		}
	};
	
	self.addDeviceDataChangedListener = function(listener) {
		if(typeof listener == 'function') {
			deviceDataChangedListeners.push(listener);
		}
	};
	
	self.removeDeviceDataChangedListener = function(listener) {
		if(typeof listener == 'function') {
			var index = deviceDataChangedListeners.indexOf(listener);
			if(index >= 0) {
				deviceDataChangedListeners.splice(index, 1);
			}
		}
	};
	
	self.clearDeviceDataChangedListeners = function() {
		deviceDataChangedListeners = [];
	};
}

function __Runner() {
	var self = this;
	var robots = {};
	var packets = {};
	var binders = {};
	var socket = undefined;
	var alive = false;
	var processing = false;
	var received = false;
	var stopping = false;
	
	var countRobots = function(model) {
		var count = 0;
		for(var i in robots) {
			if(model == robots[i].getId()) {
				++ count;
			}
		}
		return count;
	};
	
	var reset = function() {
		for(var i in robots) {
			robots[i].reset();
		}
	};
	
	var clear = function() {
		robots = {};
		packets = {};
		binders = {};
		received = false;
		stopping = false;
	};
	
	self.addRobot = function(robot, motoring, binder) {
		var model = robot.getId();
		var index = countRobots(model);
		var key = model + index;
		motoring.model = model;
		motoring.index = index;
		robots[key] = robot;
		packets[key] = motoring;
		binders[key] = binder;
		return index;
	};
	
	self.open = function(url, openCallback, closeCallback) {
		if('WebSocket' in window) {
			try {
				var sock = new WebSocket(url);
				sock.binaryType = 'arraybuffer';
				socket = sock;
				sock.onopen = function() {
					if(openCallback && (typeof openCallback == 'function')) {
						openCallback();
					}
					
					sock.onmessage = function(message) { // message: MessageEvent
						try {
							var data = JSON.parse(message.data);
							var key = data.model + data.index;
							var binder = binders[key];
							if(binder) {
								binder.handleSensory(data);
								received = true;
							}
						} catch (e) {
						}
					};
					sock.onclose = function() {
						alive = false;
						if(closeCallback && (typeof closeCallback == 'function')) {
							closeCallback();
						}
					};
					
					if(!Date.now) {
						Date.now = function() {
							return new Date().getTime();
						};
					}
					
					var targetTime = Date.now();
					var run = function() {
						if(alive && socket) {
							if(Date.now() > targetTime) {
								try {
									if(processing) {
										for(var i in binders) {
											binders[i].updateSensory();
										}
										if(received && window.execute) {
											try {
												window.execute();
											} catch (ee) {
											}
										}
										for(var i in binders) {
											binders[i].updateMotoring();
											binders[i].requestMotoring();
										}
									}
									if(stopping) {
										stopping = false;
										reset();
										for(var i in binders) {
											binders[i].requestMotoring();
										}
										var json = JSON.stringify(packets);
										if(alive && socket) socket.send(json);
										clear();
									} else {
										var json = JSON.stringify(packets);
										if(alive && socket) socket.send(json);
									}
								} catch (e) {
								}
								targetTime += 20;
							}
							setTimeout(run, 5);
						}
					};
					
					alive = true;
					run();
				};
				return true;
			} catch (e) {
			}
		}
		return false;
	};
	
	self.close = function() {
		alive = false;
		if(socket) {
			socket.close();
			socket = undefined;
		}
	};
	
	self.start = function() {
		clear();
		processing = true;
	};
	
	self.stop = function() {
		processing = false;
		stopping = true;
	};
};

var __runner = new __Runner();
var execute = undefined;
var __timeouts = [];
var __intervals = [];
var __chats = [];

window.__setTimeout = function(fn, timeout) {
	var id = setTimeout(fn, timeout);
	__timeouts.push(id);
};

window.__setInterval = function(fn, interval) {
	var id = setInterval(fn, interval);
	__intervals.push(id);
};

function connect(url, openCallback, closeCallback) {
	if(url) {
		if(url.indexOf(':') < 0) {
			__runner.open('ws://' + url + ':58000', openCallback, closeCallback);
		} else {
			__runner.open('ws://' + url, openCallback, closeCallback);
		}
		__runner.start();
	}
}

function disconnect() {
	stopCode();
	__runner.close();
}

function runCode(code) {
	__runner.start();
	code = code.replace(/setTimeout/gi, '__setTimeout');
	code = code.replace(/setInterval/gi, '__setInterval');
	code += "if(execute && typeof execute == 'function') { window.execute = execute; }";
	eval(code);
}

function stopCode() {
	window.execute = undefined;
	__runner.stop();
	for(var i in __timeouts) {
		clearTimeout(__timeouts[i]);
	}
	__timeouts = [];
	for(var i in __intervals) {
		clearInterval(__intervals[i]);
	}
	__intervals = [];
	for(var i in __chats) {
		__chats[i].disconnect();
	}
	__chats = [];
}

function Hamster() {
	var self = this;
	var id = 'kr.robomation.physical.hamster';
	var robotName = 'Hamster';
	var blockId = 0;
	var wheelId = 0;
	var wheelTimer = undefined;
	var lineTracerId = 0;
	var lineTracerCallback = undefined;
	var boardId = 0;
	var boardCommand = 0;
	var boardState = 0;
	var boardCount = 0;
	var boardCallback = undefined;
	var noteId = 0;
	var noteTimer1 = undefined;
	var noteTimer2 = undefined;
	var ioId = 0;
	var ioTimer = undefined;
	var tempo = 60;
	var handFound = false;
	var timeouts = [];
	
	var sensory = {
		devices: [],
		readData: [],
		callbacks: []
	};
	var motoring = {
		devices: [],
		readData: [],
		writeData: [],
		callbacks: []
	};
	var state = ConnectionState.CONNECTING;
	var writeMotoringFlag = 0;
	
	var binder = {
		bind: function(device, type, data, callback) {
			if(type == DeviceType.SENSOR || type == DeviceType.EVENT) {
				sensory.devices.push(device);
				sensory.readData.push(data.read);
				sensory.callbacks.push(callback);
			} else {
				motoring.devices.push(device);
				motoring.readData.push(data.read);
				motoring.writeData.push(data.write);
				motoring.callbacks.push(callback);
			}
		},
		onMotoringDeviceWritten: function(device) {
			switch(device.getId()) {
				case Hamster.TOPOLOGY:
					writeMotoringFlag |= 0x02000000;
					break;
				case Hamster.LEFT_LED:
					writeMotoringFlag |= 0x01000000;
					break;
				case Hamster.RIGHT_LED:
					writeMotoringFlag |= 0x00800000;
					break;
				case Hamster.NOTE:
					writeMotoringFlag |= 0x00400000;
					break;
				case Hamster.LINE_TRACER_MODE:
					writeMotoringFlag |= 0x00200000;
					break;
				case Hamster.LINE_TRACER_SPEED:
					writeMotoringFlag |= 0x00100000;
					break;
				case Hamster.IO_MODE_A:
					writeMotoringFlag |= 0x00080000;
					break;
				case Hamster.IO_MODE_B:
					writeMotoringFlag |= 0x00040000;
					break;
				case Hamster.CONFIG_PROXIMITY:
					writeMotoringFlag |= 0x00020000;
					break;
				case Hamster.CONFIG_GRAVITY:
					writeMotoringFlag |= 0x00010000;
					break;
				case Hamster.CONFIG_BAND_WIDTH:
					writeMotoringFlag |= 0x00008000;
					break;
			}
		}
	};
	
	var devices = {};
	devices[Hamster.LEFT_WHEEL] = new Device(binder, Hamster.LEFT_WHEEL, 'LeftWheel', DeviceType.EFFECTOR, DataType.INTEGER, 1, -100, 100, 0);
	devices[Hamster.RIGHT_WHEEL] = new Device(binder, Hamster.RIGHT_WHEEL, 'RightWheel', DeviceType.EFFECTOR, DataType.INTEGER, 1, -100, 100, 0);
	devices[Hamster.BUZZER] = new Device(binder, Hamster.BUZZER, 'Buzzer', DeviceType.EFFECTOR, DataType.FLOAT, 1, 0, 167772.15, 0.0);
	devices[Hamster.OUTPUT_A] = new Device(binder, Hamster.OUTPUT_A, 'OutputA', DeviceType.EFFECTOR, DataType.INTEGER, 1, 0, 255, 0);
	devices[Hamster.OUTPUT_B] = new Device(binder, Hamster.OUTPUT_B, 'OutputB', DeviceType.EFFECTOR, DataType.INTEGER, 1, 0, 255, 0);
	devices[Hamster.TOPOLOGY] = new Device(binder, Hamster.TOPOLOGY, 'Topology', DeviceType.COMMAND, DataType.INTEGER, 1, 0, 15, 0);
	devices[Hamster.LEFT_LED] = new Device(binder, Hamster.LEFT_LED, 'LeftLed', DeviceType.COMMAND, DataType.INTEGER, 1, 0, 7, 0);
	devices[Hamster.RIGHT_LED] = new Device(binder, Hamster.RIGHT_LED, 'RightLed', DeviceType.COMMAND, DataType.INTEGER, 1, 0, 7, 0);
	devices[Hamster.NOTE] = new Device(binder, Hamster.NOTE, 'Note', DeviceType.COMMAND, DataType.INTEGER, 1, 0, 88, 0);
	devices[Hamster.LINE_TRACER_MODE] = new Device(binder, Hamster.LINE_TRACER_MODE, 'LineTracerMode', DeviceType.COMMAND, DataType.INTEGER, 1, 0, 15, 0);
	devices[Hamster.LINE_TRACER_SPEED] = new Device(binder, Hamster.LINE_TRACER_SPEED, 'LineTracerSpeed', DeviceType.COMMAND, DataType.INTEGER, 1, 1, 8, 5);
	devices[Hamster.IO_MODE_A] = new Device(binder, Hamster.IO_MODE_A, 'IoModeA', DeviceType.COMMAND, DataType.INTEGER, 1, 0, 15, 0);
	devices[Hamster.IO_MODE_B] = new Device(binder, Hamster.IO_MODE_B, 'IoModeB', DeviceType.COMMAND, DataType.INTEGER, 1, 0, 15, 0);
	devices[Hamster.CONFIG_PROXIMITY] = new Device(binder, Hamster.CONFIG_PROXIMITY, 'ConfigProximity', DeviceType.COMMAND, DataType.INTEGER, 1, 1, 7, 2);
	devices[Hamster.CONFIG_GRAVITY] = new Device(binder, Hamster.CONFIG_GRAVITY, 'ConfigGravity', DeviceType.COMMAND, DataType.INTEGER, 1, 0, 3, 0);
	devices[Hamster.CONFIG_BAND_WIDTH] = new Device(binder, Hamster.CONFIG_BAND_WIDTH, 'ConfigBandWidth', DeviceType.COMMAND, DataType.INTEGER, 1, 1, 8, 3);
	devices[Hamster.SIGNAL_STRENGTH] = new Device(binder, Hamster.SIGNAL_STRENGTH, 'SignalStrength', DeviceType.SENSOR, DataType.INTEGER, 1, -128, 0, 0);
	devices[Hamster.LEFT_PROXIMITY] = new Device(binder, Hamster.LEFT_PROXIMITY, 'LeftProximity', DeviceType.SENSOR, DataType.INTEGER, 1, 0, 255, 0);
	devices[Hamster.RIGHT_PROXIMITY] = new Device(binder, Hamster.RIGHT_PROXIMITY, 'RightProximity', DeviceType.SENSOR, DataType.INTEGER, 1, 0, 255, 0);
	devices[Hamster.LEFT_FLOOR] = new Device(binder, Hamster.LEFT_FLOOR, 'LeftFloor', DeviceType.SENSOR, DataType.INTEGER, 1, 0, 255, 0);
	devices[Hamster.RIGHT_FLOOR] = new Device(binder, Hamster.RIGHT_FLOOR, 'RightFloor', DeviceType.SENSOR, DataType.INTEGER, 1, 0, 255, 0);
	devices[Hamster.ACCELERATION] = new Device(binder, Hamster.ACCELERATION, 'Acceleration', DeviceType.SENSOR, DataType.INTEGER, 3, -32768, 32767, 0);
	devices[Hamster.LIGHT] = new Device(binder, Hamster.LIGHT, 'Light', DeviceType.SENSOR, DataType.INTEGER, 1, 0, 65535, 0);
	devices[Hamster.TEMPERATURE] = new Device(binder, Hamster.TEMPERATURE, 'Temperature', DeviceType.SENSOR, DataType.INTEGER, 1, -40, 88, 0);
	devices[Hamster.INPUT_A] = new Device(binder, Hamster.INPUT_A, 'InputA', DeviceType.SENSOR, DataType.INTEGER, 1, 0, 255, 0);
	devices[Hamster.INPUT_B] = new Device(binder, Hamster.INPUT_B, 'InputB', DeviceType.SENSOR, DataType.INTEGER, 1, 0, 255, 0);
	devices[Hamster.LINE_TRACER_STATE] = new Device(binder, Hamster.LINE_TRACER_STATE, 'LineTracerState', DeviceType.EVENT, DataType.INTEGER, 1, 0, 255, 0);
	
	var packet = {
		map: 0,
		leftWheel: 0,
		rightWheel: 0,
		buzzer: 0,
		outputA: 0,
		outputB: 0,
		topology: 0,
		leftLed: 0,
		rightLed: 0,
		note: 0,
		lineTracerMode: 0,
		lineTracerSpeed: 5,
		ioModeA: 0,
		ioModeB: 0,
		configProximity: 2,
		configGravity: 0,
		configBandWidth: 3,
		motion: 0
	};
	
	self.getId = function() {
		return id;
	};
	
	self.getName = function() {
		return robotName;
	};
	
	self.setName = function(name) {
		robotName = '' + name;
	};
	
	self.findDeviceById = function(deviceId) {
		var device = devices[deviceId];
		if(device) {
			return device;
		}
		return undefined;
	};
	
	self.e = function(deviceId) {
		var device = self.findDeviceById(deviceId);
		if(device) {
			return device.e();
		}
		return false;
	};
	
	self.read = function(deviceId, arg) {
		var device = self.findDeviceById(deviceId);
		if(device) {
			return device.read(arg);
		}
		return 0;
	};
	
	self.write = function(deviceId, arg1, arg2) {
		var device = self.findDeviceById(deviceId);
		if(device) {
			return device.write(arg1, arg2);
		}
		if(arg1 && arg1.constructor == Array) {
			return 0;
		}
		return false;
	};
	
	self.addDeviceDataChangedListener = function(listener) {
		for(var i in devices) {
			devices[i].addDeviceDataChangedListener(listener);
		}
	};
	
	self.removeDeviceDataChangedListener = function(listener) {
		for(var i in devices) {
			devices[i].removeDeviceDataChangedListener(listener);
		}
	};
	
	self.clearDeviceDataChangedListeners = function() {
		for(var i in devices) {
			devices[i].clearDeviceDataChangedListeners();
		}
	};
	
	var removeTimeout = function(id) {
		clearTimeout(id);
		var index = timeouts.indexOf(id);
		if(index >= 0) {
			timeouts.splice(index, 1);
		}
	};

	var removeAllTimeouts = function() {
		for(var i in timeouts) {
			clearTimeout(timeouts[i]);
		}
		timeouts = [];
	};
	
	var issueWheelId = function() {
		wheelId = blockId = (blockId % 65535) + 1;
		return wheelId;
	};
	
	var cancelWheel = function() {
		wheelId = 0;
		if(wheelTimer !== undefined) {
			removeTimeout(wheelTimer);
		}
		wheelTimer = undefined;
	};
	
	var issueLineTracerId = function() {
		lineTracerId = blockId = (blockId % 65535) + 1;
		return lineTracerId;
	};
	
	var cancelLineTracer = function() {
		lineTracerId = 0;
		lineTracerCallback = undefined;
	};
	
	var issueBoardId = function() {
		boardId = blockId = (blockId % 65535) + 1;
		return boardId;
	};
	
	var cancelBoard = function() {
		boardId = 0;
		boardCommand = 0;
		boardState = 0;
		boardCount = 0;
		boardCallback = undefined;
	};
	
	var issueNoteId = function() {
		noteId = blockId = (blockId % 65535) + 1;
		return noteId;
	};
	
	var cancelNote = function() {
		noteId = 0;
		if(noteTimer1 !== undefined) {
			removeTimeout(noteTimer1);
		}
		if(noteTimer2 !== undefined) {
			removeTimeout(noteTimer2);
		}
		noteTimer1 = undefined;
		noteTimer2 = undefined;
	};
	
	var issueIoId = function() {
		ioId = blockId = (blockId % 65535) + 1;
		return ioId;
	};
	
	var cancelIo = function() {
		ioId = 0;
		if(ioTimer !== undefined) {
			removeTimeout(ioTimer);
		}
		ioTimer = undefined;
	};
	
	var index = __runner.addRobot(self, packet, {
		handleSensory: function(data) {
			state = data.connectionState;
			var map = data.map;
			var readData = sensory.readData;
			readData[0][0] = data.signalStrength;
			readData[1][0] = data.leftProximity;
			readData[2][0] = data.rightProximity;
			readData[3][0] = data.leftFloor;
			readData[4][0] = data.rightFloor;
			readData[5][0] = data.accelerationX;
			readData[5][1] = data.accelerationY;
			readData[5][2] = data.accelerationZ;
			readData[6][0] = data.light;
			readData[7][0] = data.temperature;
			readData[8][0] = data.inputA;
			readData[9][0] = data.inputB;
			handFound = (data.handFound === undefined) ? (data.leftProximity > 50 || data.rightProximity > 50) : data.handFound;
			
			var callbacks = sensory.callbacks;
			var i;
			for(i = 0; i < 10; ++i) {
				callbacks[i].fire();
			}
			
			if(map & 0x00000010) {
				readData[10][0] = data.lineTracerState;
				callbacks[10].fire();
			}
			
			for(i = 0; i < 10; ++i) {
				callbacks[i].notifyDeviceDataChanged();
			}
			
			if(map & 0x00000010) {
				callbacks[10].notifyDeviceDataChanged();
				if(lineTracerId > 0 && data.lineTracerState == 0x40) {
					self.write(Hamster.LINE_TRACER_MODE, 0);
					var callback = lineTracerCallback;
					cancelLineTracer();
					if(typeof callback == 'function') callback();
				}
			}
			
			if(boardId > 0) {
				if(boardCommand == 1) {
					switch(boardState) {
						case 1: {
							if(boardCount < 2) {
								if(data.leftFloor < 50 && data.rightFloor < 50)
									boardCount ++;
								else
									boardCount = 0;
								var diff = data.leftFloor - data.rightFloor;
								self.write(Hamster.LEFT_WHEEL, 45 + diff * 0.25);
								self.write(Hamster.RIGHT_WHEEL, 45 - diff * 0.25);
							} else {
								boardCount = 0;
								boardState = 2;
							}
							break;
						}
						case 2: {
							var diff = data.leftFloor - data.rightFloor;
							self.write(Hamster.LEFT_WHEEL, 45 + diff * 0.25);
							self.write(Hamster.RIGHT_WHEEL, 45 - diff * 0.25);
							boardState = 3;
							wheelTimer = setTimeout(function() {
								self.write(Hamster.LEFT_WHEEL, 0);
								self.write(Hamster.RIGHT_WHEEL, 0);
								boardState = 4;
								if(wheelTimer !== undefined) removeTimeout(wheelTimer);
								wheelTimer = undefined;
							}, 250);
							timeouts.push(wheelTimer);
							break;
						}
						case 3: {
							var diff = data.leftFloor - data.rightFloor;
							self.write(Hamster.LEFT_WHEEL, 45 + diff * 0.25);
							self.write(Hamster.RIGHT_WHEEL, 45 - diff * 0.25);
							break;
						}
						case 4: {
							self.write(Hamster.LEFT_WHEEL, 0);
							self.write(Hamster.RIGHT_WHEEL, 0);
							var callback = boardCallback;
							cancelBoard();
							if(typeof callback == 'function') callback();
							break;
						}
					}
				} else if(boardCommand == 2) {
					switch(boardState) {
						case 1: {
							if(boardCount < 2) {
								if(data.leftFloor > 50)
									boardCount ++;
							} else {
								boardCount = 0;
								boardState = 2;
							}
							break;
						}
						case 2: {
							if(data.leftFloor < 20) {
								boardState = 3;
							}
							break;
						}
						case 3: {
							if(boardCount < 2) {
								if(data.leftFloor < 20)
									boardCount ++;
							} else {
								boardCount = 0;
								boardState = 4;
							}
							break;
						}
						case 4: {
							if(data.leftFloor > 50) {
								boardState = 5;
							}
							break;
						}
						case 5: {
							var diff = data.leftFloor - data.rightFloor;
							if(diff > -15) {
								self.write(Hamster.LEFT_WHEEL, 0);
								self.write(Hamster.RIGHT_WHEEL, 0);
								var callback = boardCallback;
								cancelBoard();
								if(typeof callback == 'function') callback();
							} else {
								self.write(Hamster.LEFT_WHEEL, diff * 0.5);
								self.write(Hamster.RIGHT_WHEEL, -diff * 0.5);
							}
							break;
						}
					}
				} else if(boardCommand == 3) {
					switch(boardState) {
						case 1: {
							if(boardCount < 2) {
								if(data.rightFloor > 50)
									boardCount ++;
							} else {
								boardCount = 0;
								boardState = 2;
							}
							break;
						}
						case 2: {
							if(data.rightFloor < 20) {
								boardState = 3;
							}
							break;
						}
						case 3: {
							if(boardCount < 2) {
								if(data.rightFloor < 20)
									boardCount ++;
							} else {
								boardCount = 0;
								boardState = 4;
							}
							break;
						}
						case 4: {
							if(data.rightFloor > 50) {
								boardState = 5;
							}
							break;
						}
						case 5: {
							var diff = data.rightFloor - data.leftFloor;
							if(diff > -15) {
								self.write(Hamster.LEFT_WHEEL, 0);
								self.write(Hamster.RIGHT_WHEEL, 0);
								var callback = boardCallback;
								cancelBoard();
								if(typeof callback == 'function') callback();
							} else {
								self.write(Hamster.LEFT_WHEEL, -diff * 0.5);
								self.write(Hamster.RIGHT_WHEEL, diff * 0.5);
							}
							break;
						}
					}
				}
			}
		},
		requestMotoring: function() {
			var map = 0xfc000000 | writeMotoringFlag;
			writeMotoringFlag = 0;
			
			var writeData = motoring.writeData;
			packet.map = map;
			packet.leftWheel = writeData[0][0];
			packet.rightWheel = writeData[1][0];
			packet.buzzer = writeData[2][0];
			packet.outputA = writeData[3][0];
			packet.outputB = writeData[4][0];
			if(map & 0x02000000) packet.topology = writeData[5][0];
			if(map & 0x01000000) packet.leftLed = writeData[6][0];
			if(map & 0x00800000) packet.rightLed = writeData[7][0];
			if(map & 0x00400000) packet.note = writeData[8][0];
			if(map & 0x00200000) packet.lineTracerMode = writeData[9][0];
			if(map & 0x00100000) packet.lineTracerSpeed = writeData[10][0];
			if(map & 0x00080000) packet.ioModeA = writeData[11][0];
			if(map & 0x00040000) packet.ioModeB = writeData[12][0];
			if(map & 0x00020000) packet.configProximity = writeData[13][0];
			if(map & 0x00010000) packet.configGravity = writeData[14][0];
			if(map & 0x00008000) packet.configBandWidth = writeData[15][0];
		},
		updateSensory: function() {
			var callbacks = sensory.callbacks;
			for(var i in callbacks) {
				callbacks[i].update();
			}
		},
		updateMotoring: function() {
			var callbacks = motoring.callbacks;
			for(var i in callbacks) {
				callbacks[i].update();
			}
		}
	});
	
	self.getIndex = function() {
		return index;
	};
	
	self.reset = function() {
		var callbacks = motoring.callbacks;
		for(var i in callbacks) {
			callbacks[i].reset();
		}
		callbacks = sensory.callbacks;
		for(var i in callbacks) {
			callbacks[i].reset();
		}
		writeMotoringFlag |= 0x03ff8000;
		
		blockId = 0;
		wheelId = 0;
		wheelTimer = undefined;
		lineTracerId = 0;
		lineTracerCallback = undefined;
		boardId = 0;
		boardCommand = 0;
		boardState = 0;
		boardCount = 0;
		boardCallback = undefined;
		noteId = 0;
		noteTimer1 = undefined;
		noteTimer2 = undefined;
		ioId = 0;
		ioTimer = undefined;
		tempo = 60;
		handFound = false;
		removeAllTimeouts();
	};
	
	self.connectionState = function() {
		return state;
	};
	
	var board = function(leftVelocity, rightVelocity, command, callback) {
		cancelWheel();
		cancelLineTracer();
		
		self.write(Hamster.LEFT_WHEEL, leftVelocity);
		self.write(Hamster.RIGHT_WHEEL, rightVelocity);
		packet.motion = 0;
		boardCommand = command;
		boardCount = 0;
		boardState = 1;
		boardCallback = callback;
		issueBoardId();
		self.write(Hamster.LINE_TRACER_MODE, 0);
	};
	
	self.boardForward = function(callback) {
		board(45, 45, 1, callback);
	};
	
	self.boardLeft = function(callback) {
		board(-45, 45, 2, callback);
	};
	
	self.boardRight = function(callback) {
		board(45, -45, 3, callback);
	};
	
	var move = function(motion, arg1, arg2, arg3) {
		cancelBoard();
		cancelWheel();
		cancelLineTracer();
		
		// hamster uses internal velocity if motion is 1 ~ 4
		var sec = 1, velocity = 30, callback = undefined, packetmotion = motion;
		if(typeof arg1 == 'number') {
			sec = arg1;
			if(typeof arg2 == 'number') {
				velocity = arg2;
				packetmotion = 0;
				if(typeof arg3 == 'function') {
					callback = arg3;
				}
			} else if(typeof arg2 == 'function') {
				callback = arg2;
			}
		} else if(typeof arg1 == 'function') {
			callback = arg1;
		}
		
		if(sec > 0) {
			var id = issueWheelId();
			switch(motion) {
				case 1:
					self.write(Hamster.LEFT_WHEEL, velocity);
					self.write(Hamster.RIGHT_WHEEL, velocity);
					break;
				case 2:
					self.write(Hamster.LEFT_WHEEL, -velocity);
					self.write(Hamster.RIGHT_WHEEL, -velocity);
					break;
				case 3:
					self.write(Hamster.LEFT_WHEEL, -velocity);
					self.write(Hamster.RIGHT_WHEEL, velocity);
					break;
				case 4:
					self.write(Hamster.LEFT_WHEEL, velocity);
					self.write(Hamster.RIGHT_WHEEL, -velocity);
					break;
			}
			packet.motion = packetmotion;
			self.write(Hamster.LINE_TRACER_MODE, 0);
			wheelTimer = setTimeout(function() {
				if(wheelId == id) {
					self.write(Hamster.LEFT_WHEEL, 0);
					self.write(Hamster.RIGHT_WHEEL, 0);
					packet.motion = 0;
					cancelWheel();
					if(callback) callback();
				}
			}, sec * 1000);
			timeouts.push(wheelTimer);
		} else { // stop
			self.write(Hamster.LEFT_WHEEL, 0);
			self.write(Hamster.RIGHT_WHEEL, 0);
			packet.motion = 0;
			self.write(Hamster.LINE_TRACER_MODE, 0);
			if(callback) callback();
		}
	};
	
	self.moveForward = function(arg1, arg2, arg3) {
		move(1, arg1, arg2, arg3);
	};
	
	self.moveBackward = function(arg1, arg2, arg3) {
		move(2, arg1, arg2, arg3);
	};
	
	self.turnLeft = function(arg1, arg2, arg3) {
		move(3, arg1, arg2, arg3);
	};
	
	self.turnRight = function(arg1, arg2, arg3) {
		move(4, arg1, arg2, arg3);
	};
	
	self.wheels = function(leftVelocity, rightVelocity) {
		cancelBoard();
		cancelWheel();
		cancelLineTracer();
		
		if(typeof leftVelocity == 'number') {
			self.write(Hamster.LEFT_WHEEL, leftVelocity);
		}
		if(typeof rightVelocity == 'number') {
			self.write(Hamster.RIGHT_WHEEL, rightVelocity);
		} else {
			if(typeof leftVelocity == 'number') {
				self.write(Hamster.RIGHT_WHEEL, leftVelocity);
			}
		}
		packet.motion = 0;
		self.write(Hamster.LINE_TRACER_MODE, 0);
	};
	
	self.leftWheel = function(velocity) {
		cancelBoard();
		cancelWheel();
		cancelLineTracer();
		
		if(typeof velocity == 'number') {
			self.write(Hamster.LEFT_WHEEL, velocity);
		}
		packet.motion = 0;
		self.write(Hamster.LINE_TRACER_MODE, 0);
	};
	
	self.rightWheel = function(velocity) {
		cancelBoard();
		cancelWheel();
		cancelLineTracer();
		
		if(typeof velocity == 'number') {
			self.write(Hamster.RIGHT_WHEEL, velocity);
		}
		packet.motion = 0;
		self.write(Hamster.LINE_TRACER_MODE, 0);
	};
	
	self.stop = function() {
		cancelBoard();
		cancelWheel();
		cancelLineTracer();
		
		self.write(Hamster.LEFT_WHEEL, 0);
		self.write(Hamster.RIGHT_WHEEL, 0);
		packet.motion = 0;
		self.write(Hamster.LINE_TRACER_MODE, 0);
	};
	
	self.lineTracerMode = function(arg1, arg2) {
		cancelBoard();
		cancelWheel();
		
		var mode = 0, callback = undefined;
		if(typeof arg1 == 'number') {
			mode = arg1;
			switch(arg1) {
				case Hamster.LINE_TRACER_MODE_BLACK_TURN_LEFT:
				case Hamster.LINE_TRACER_MODE_BLACK_TURN_RIGHT:
				case Hamster.LINE_TRACER_MODE_BLACK_MOVE_FORWARD:
				case Hamster.LINE_TRACER_MODE_BLACK_UTURN:
				case Hamster.LINE_TRACER_MODE_WHITE_TURN_LEFT:
				case Hamster.LINE_TRACER_MODE_WHITE_TURN_RIGHT:
				case Hamster.LINE_TRACER_MODE_WHITE_MOVE_FORWARD:
				case Hamster.LINE_TRACER_MODE_WHITE_UTURN:
					if(typeof arg2 == 'function') callback = arg2;
					lineTracerCallback = callback;
					issueLineTracerId();
					break;
				default:
					cancelLineTracer();
					break;
			}
		} else {
			cancelLineTracer();
		}
		self.write(Hamster.LEFT_WHEEL, 0);
		self.write(Hamster.RIGHT_WHEEL, 0);
		packet.motion = 0;
		self.write(Hamster.LINE_TRACER_MODE, mode);
	};
	
	self.lineTracerSpeed = function(speed) {
		if(typeof speed == 'number') {
			self.write(Hamster.LINE_TRACER_SPEED, speed);
		}
	};
	
	self.leds = function(leftColor, rightColor) {
		if(typeof leftColor == 'number') {
			self.write(Hamster.LEFT_LED, leftColor);
		}
		if(typeof rightColor == 'number') {
			self.write(Hamster.RIGHT_LED, rightColor);
		} else {
			if(typeof leftColor == 'number') {
				self.write(Hamster.RIGHT_LED, leftColor);
			}
		}
	};
	
	self.leftLed = function(color) {
		if(typeof color == 'number') {
			self.write(Hamster.LEFT_LED, color);
		}
	};
	
	self.rightLed = function(color) {
		if(typeof color == 'number') {
			self.write(Hamster.RIGHT_LED, color);
		}
	};
	
	self.beep = function(callback) {
		cancelNote();
		var id = issueNoteId();
		self.write(Hamster.BUZZER, 440);
		self.write(Hamster.NOTE, 0);
		noteTimer1 = setTimeout(function() {
			if(noteId == id) {
				self.write(Hamster.BUZZER, 0);
				cancelNote();
				if(typeof callback == 'function') callback();
			}
		}, 200);
		timeouts.push(noteTimer1);
	};
	
	self.buzzer = function(hz) {
		cancelNote();
		if(typeof hz == 'number') {
			self.write(Hamster.BUZZER, hz);
		}
		self.write(Hamster.NOTE, 0);
	};
	
	self.tempo = function(bpm) {
		if(typeof bpm == 'number') {
			tempo = bpm;
			if(tempo < 1) tempo = 1;
		}
	};
	
	self.note = function(arg1, arg2, arg3) {
		cancelNote();
		
		var pitch = 0, beat = undefined, callback = undefined;
		if(typeof arg1 == 'number') {
			pitch = arg1;
			if(typeof arg2 == 'number') {
				beat = arg2;
				if(typeof arg3 == 'function') {
					callback = arg3;
				}
			}
		}
		
		self.write(Hamster.BUZZER, 0);
		if(typeof beat == 'number') {
			if(beat > 0 && tempo > 0) {
				var id = issueNoteId();
				if(pitch > 0) {
					self.write(Hamster.NOTE, pitch);
					var timeout = beat * 60 * 1000 / tempo;
					var tail = 0;
					if(timeout > 100) tail = 100;
					if(tail > 0) {
						noteTimer1 = setTimeout(function() {
							if(noteId == id) {
								self.write(Hamster.NOTE, 0);
								if(noteTimer1 !== undefined) removeTimeout(noteTimer1);
								noteTimer1 = undefined;
							}
						}, timeout - tail);
						timeouts.push(noteTimer1);
					}
					noteTimer2 = setTimeout(function() {
						if(noteId == id) {
							self.write(Hamster.NOTE, 0);
							cancelNote();
							if(typeof callback == 'function') callback();
						}
					}, timeout);
					timeouts.push(noteTimer2);
				} else {
					self.write(Hamster.NOTE, 0);
					var id = issueNoteId();
					noteTimer1 = setTimeout(function() {
						if(noteId == id) {
							cancelNote();
							if(typeof callback == 'function') callback();
						}
					}, beat * 60 * 1000 / tempo);
					timeouts.push(noteTimer1);
				}
			} else { // off
				self.write(Hamster.NOTE, 0);
				if(typeof callback == 'function') callback();
			}
		} else { // continue
			self.write(Hamster.NOTE, pitch);
		}
	};
	
	self.ioModeA = function(mode) {
		cancelIo();
		if(typeof mode == 'number') {
			self.write(Hamster.IO_MODE_A, mode);
		}
	};
	
	self.ioModeB = function(mode) {
		cancelIo();
		if(typeof mode == 'number') {
			self.write(Hamster.IO_MODE_B, mode);
		}
	};
	
	self.outputA = function(value) {
		cancelIo();
		if(typeof value == 'number') {
			self.write(Hamster.OUTPUT_A, value);
		}
	};
	
	self.outputB = function(value) {
		cancelIo();
		if(typeof value == 'number') {
			self.write(Hamster.OUTPUT_B, value);
		}
	};
	
	var gripper = function(outputA, outputB, callback) {
		cancelIo();
		
		var id = issueIoId();
		self.write(Hamster.IO_MODE_A, Hamster.IO_MODE_DIGITAL_OUTPUT);
		self.write(Hamster.IO_MODE_B, Hamster.IO_MODE_DIGITAL_OUTPUT);
		self.write(Hamster.OUTPUT_A, outputA);
		self.write(Hamster.OUTPUT_B, outputB);
		ioTimer = setTimeout(function() {
			if(ioId == id) {
				cancelIo();
				if(typeof callback == 'function') callback();
			}
		}, 500);
		timeouts.push(ioTimer);
	};
	
	self.openGripper = function(callback) {
		gripper(1, 0, callback);
	};
	
	self.closeGripper = function(callback) {
		gripper(0, 1, callback);
	};
	
	self.releaseGripper = function(callback) {
		cancelIo();
		self.write(Hamster.IO_MODE_A, Hamster.IO_MODE_DIGITAL_OUTPUT);
		self.write(Hamster.IO_MODE_B, Hamster.IO_MODE_DIGITAL_OUTPUT);
		self.write(Hamster.OUTPUT_A, 0);
		self.write(Hamster.OUTPUT_B, 0);
	};
	
	self.signalStrength = function() {
		return self.read(Hamster.SIGNAL_STRENGTH);
	};
	
	self.leftProximity = function() {
		return self.read(Hamster.LEFT_PROXIMITY);
	};
	
	self.rightProximity = function() {
		return self.read(Hamster.RIGHT_PROXIMITY);
	};
	
	self.leftFloor = function() {
		return self.read(Hamster.LEFT_FLOOR);
	};
	
	self.rightFloor = function() {
		return self.read(Hamster.RIGHT_FLOOR);
	};
	
	self.accelerationX = function() {
		return self.read(Hamster.ACCELERATION, 0);
	};
	
	self.accelerationY = function() {
		return self.read(Hamster.ACCELERATION, 1);
	};
	
	self.accelerationZ = function() {
		return self.read(Hamster.ACCELERATION, 2);
	};
	
	self.light = function() {
		return self.read(Hamster.LIGHT);
	};
	
	self.temperature = function() {
		return self.read(Hamster.TEMPERATURE);
	};
	
	self.inputA = function() {
		return self.read(Hamster.INPUT_A);
	};
	
	self.inputB = function() {
		return self.read(Hamster.INPUT_B);
	};
	
	self.handFound = function() {
		return handFound;
	};
}

Hamster.ID = 'kr.robomation.physical.hamster';

Hamster.LEFT_WHEEL = 0x00400000;
Hamster.RIGHT_WHEEL = 0x00400001;
Hamster.BUZZER = 0x00400002;
Hamster.OUTPUT_A = 0x00400003;
Hamster.OUTPUT_B = 0x00400004;
Hamster.TOPOLOGY = 0x00400005;
Hamster.LEFT_LED = 0x00400006;
Hamster.RIGHT_LED = 0x00400007;
Hamster.NOTE = 0x00400008;
Hamster.LINE_TRACER_MODE = 0x00400009;
Hamster.LINE_TRACER_SPEED = 0x0040000a;
Hamster.IO_MODE_A = 0x0040000b;
Hamster.IO_MODE_B = 0x0040000c;
Hamster.CONFIG_PROXIMITY = 0x0040000d;
Hamster.CONFIG_GRAVITY = 0x0040000e;
Hamster.CONFIG_BAND_WIDTH = 0x0040000f;

Hamster.SIGNAL_STRENGTH = 0x00400010;
Hamster.LEFT_PROXIMITY = 0x00400011;
Hamster.RIGHT_PROXIMITY = 0x00400012;
Hamster.LEFT_FLOOR = 0x00400013;
Hamster.RIGHT_FLOOR = 0x00400014;
Hamster.ACCELERATION = 0x00400015;
Hamster.LIGHT = 0x00400016;
Hamster.TEMPERATURE = 0x00400017;
Hamster.INPUT_A = 0x00400018;
Hamster.INPUT_B = 0x00400019;
Hamster.LINE_TRACER_STATE = 0x0040001a;

Hamster.TOPOLOGY_NONE = 0;
Hamster.TOPOLOGY_DAISY_CHAIN = 1;
Hamster.TOPOLOGY_STAR = 2;
Hamster.TOPOLOGY_EXTENDED_STAR = 3;

Hamster.LED_OFF = 0;
Hamster.LED_BLUE = 1;
Hamster.LED_GREEN = 2;
Hamster.LED_CYAN = 3;
Hamster.LED_RED = 4;
Hamster.LED_MAGENTA = 5;
Hamster.LED_YELLOW = 6;
Hamster.LED_WHITE = 7;

Hamster.LINE_TRACER_MODE_OFF = 0;
Hamster.LINE_TRACER_MODE_BLACK_LEFT_SENSOR = 1;
Hamster.LINE_TRACER_MODE_BLACK_RIGHT_SENSOR = 2;
Hamster.LINE_TRACER_MODE_BLACK_BOTH_SENSORS = 3;
Hamster.LINE_TRACER_MODE_BLACK_TURN_LEFT = 4;
Hamster.LINE_TRACER_MODE_BLACK_TURN_RIGHT = 5;
Hamster.LINE_TRACER_MODE_BLACK_MOVE_FORWARD = 6;
Hamster.LINE_TRACER_MODE_BLACK_UTURN = 7;
Hamster.LINE_TRACER_MODE_WHITE_LEFT_SENSOR = 8;
Hamster.LINE_TRACER_MODE_WHITE_RIGHT_SENSOR = 9;
Hamster.LINE_TRACER_MODE_WHITE_BOTH_SENSORS = 10;
Hamster.LINE_TRACER_MODE_WHITE_TURN_LEFT = 11;
Hamster.LINE_TRACER_MODE_WHITE_TURN_RIGHT = 12;
Hamster.LINE_TRACER_MODE_WHITE_MOVE_FORWARD = 13;
Hamster.LINE_TRACER_MODE_WHITE_UTURN = 14;

Hamster.IO_MODE_ANALOG_INPUT = 0;
Hamster.IO_MODE_DIGITAL_INPUT = 1;
Hamster.IO_MODE_SERVO_OUTPUT = 8;
Hamster.IO_MODE_PWM_OUTPUT = 9;
Hamster.IO_MODE_DIGITAL_OUTPUT = 10;

Hamster.NOTE_OFF = 0;
Hamster.NOTE_A_0 = 1;
Hamster.NOTE_A_SHARP_0 = 2;
Hamster.NOTE_B_FLAT_0 = 2;
Hamster.NOTE_B_0 = 3;
Hamster.NOTE_C_1 = 4;
Hamster.NOTE_C_SHARP_1 = 5;
Hamster.NOTE_D_FLAT_1 = 5;
Hamster.NOTE_D_1 = 6;
Hamster.NOTE_D_SHARP_1 = 7;
Hamster.NOTE_E_FLAT_1 = 7;
Hamster.NOTE_E_1 = 8;
Hamster.NOTE_F_1 = 9;
Hamster.NOTE_F_SHARP_1 = 10;
Hamster.NOTE_G_FLAT_1 = 10;
Hamster.NOTE_G_1 = 11;
Hamster.NOTE_G_SHARP_1 = 12;
Hamster.NOTE_A_FLAT_1 = 12;
Hamster.NOTE_A_1 = 13;
Hamster.NOTE_A_SHARP_1 = 14;
Hamster.NOTE_B_FLAT_1 = 14;
Hamster.NOTE_B_1 = 15;
Hamster.NOTE_C_2 = 16;
Hamster.NOTE_C_SHARP_2 = 17;
Hamster.NOTE_D_FLAT_2 = 17;
Hamster.NOTE_D_2 = 18;
Hamster.NOTE_D_SHARP_2 = 19;
Hamster.NOTE_E_FLAT_2 = 19;
Hamster.NOTE_E_2 = 20;
Hamster.NOTE_F_2 = 21;
Hamster.NOTE_F_SHARP_2 = 22;
Hamster.NOTE_G_FLAT_2 = 22;
Hamster.NOTE_G_2 = 23;
Hamster.NOTE_G_SHARP_2 = 24;
Hamster.NOTE_A_FLAT_2 = 24;
Hamster.NOTE_A_2 = 25;
Hamster.NOTE_A_SHARP_2 = 26;
Hamster.NOTE_B_FLAT_2 = 26;
Hamster.NOTE_B_2 = 27;
Hamster.NOTE_C_3 = 28;
Hamster.NOTE_C_SHARP_3 = 29;
Hamster.NOTE_D_FLAT_3 = 29;
Hamster.NOTE_D_3 = 30;
Hamster.NOTE_D_SHARP_3 = 31;
Hamster.NOTE_E_FLAT_3 = 31;
Hamster.NOTE_E_3 = 32;
Hamster.NOTE_F_3 = 33;
Hamster.NOTE_F_SHARP_3 = 34;
Hamster.NOTE_G_FLAT_3 = 34;
Hamster.NOTE_G_3 = 35;
Hamster.NOTE_G_SHARP_3 = 36;
Hamster.NOTE_A_FLAT_3 = 36;
Hamster.NOTE_A_3 = 37;
Hamster.NOTE_A_SHARP_3 = 38;
Hamster.NOTE_B_FLAT_3 = 38;
Hamster.NOTE_B_3 = 39;
Hamster.NOTE_C_4 = 40;
Hamster.NOTE_C_SHARP_4 = 41;
Hamster.NOTE_D_FLAT_4 = 41;
Hamster.NOTE_D_4 = 42;
Hamster.NOTE_D_SHARP_4 = 43;
Hamster.NOTE_E_FLAT_4 = 43;
Hamster.NOTE_E_4 = 44;
Hamster.NOTE_F_4 = 45;
Hamster.NOTE_F_SHARP_4 = 46;
Hamster.NOTE_G_FLAT_4 = 46;
Hamster.NOTE_G_4 = 47;
Hamster.NOTE_G_SHARP_4 = 48;
Hamster.NOTE_A_FLAT_4 = 48;
Hamster.NOTE_A_4 = 49;
Hamster.NOTE_A_SHARP_4 = 50;
Hamster.NOTE_B_FLAT_4 = 50;
Hamster.NOTE_B_4 = 51;
Hamster.NOTE_C_5 = 52;
Hamster.NOTE_C_SHARP_5 = 53;
Hamster.NOTE_D_FLAT_5 = 53;
Hamster.NOTE_D_5 = 54;
Hamster.NOTE_D_SHARP_5 = 55;
Hamster.NOTE_E_FLAT_5 = 55;
Hamster.NOTE_E_5 = 56;
Hamster.NOTE_F_5 = 57;
Hamster.NOTE_F_SHARP_5 = 58;
Hamster.NOTE_G_FLAT_5 = 58;
Hamster.NOTE_G_5 = 59;
Hamster.NOTE_G_SHARP_5 = 60;
Hamster.NOTE_A_FLAT_5 = 60;
Hamster.NOTE_A_5 = 61;
Hamster.NOTE_A_SHARP_5 = 62;
Hamster.NOTE_B_FLAT_5 = 62;
Hamster.NOTE_B_5 = 63;
Hamster.NOTE_C_6 = 64;
Hamster.NOTE_C_SHARP_6 = 65;
Hamster.NOTE_D_FLAT_6 = 65;
Hamster.NOTE_D_6 = 66;
Hamster.NOTE_D_SHARP_6 = 67;
Hamster.NOTE_E_FLAT_6 = 67;
Hamster.NOTE_E_6 = 68;
Hamster.NOTE_F_6 = 69;
Hamster.NOTE_F_SHARP_6 = 70;
Hamster.NOTE_G_FLAT_6 = 70;
Hamster.NOTE_G_6 = 71;
Hamster.NOTE_G_SHARP_6 = 72;
Hamster.NOTE_A_FLAT_6 = 72;
Hamster.NOTE_A_6 = 73;
Hamster.NOTE_A_SHARP_6 = 74;
Hamster.NOTE_B_FLAT_6 = 74;
Hamster.NOTE_B_6 = 75;
Hamster.NOTE_C_7 = 76;
Hamster.NOTE_C_SHARP_7 = 77;
Hamster.NOTE_D_FLAT_7 = 77;
Hamster.NOTE_D_7 = 78;
Hamster.NOTE_D_SHARP_7 = 79;
Hamster.NOTE_E_FLAT_7 = 79;
Hamster.NOTE_E_7 = 80;
Hamster.NOTE_F_7 = 81;
Hamster.NOTE_F_SHARP_7 = 82;
Hamster.NOTE_G_FLAT_7 = 82;
Hamster.NOTE_G_7 = 83;
Hamster.NOTE_G_SHARP_7 = 84;
Hamster.NOTE_A_FLAT_7 = 84;
Hamster.NOTE_A_7 = 85;
Hamster.NOTE_A_SHARP_7 = 86;
Hamster.NOTE_B_FLAT_7 = 86;
Hamster.NOTE_B_7 = 87;
Hamster.NOTE_C_8 = 88;

Hamster.create = function() {
	return new Hamster();
};

function Turtle() {
	var self = this;
	var id = 'kr.robomation.physical.turtle';
	var robotName = 'Turtle';
	var blockId = 0;
	var motionId = 0;
	var motionCallback = undefined;
	var lineTracerId = 0;
	var lineTracerCallback = undefined;
	var currentSound = 0;
	var soundRepeat = 1;
	var soundCallback = undefined;
	var noteId = 0;
	var noteTimer1 = undefined;
	var noteTimer2 = undefined;
	var clickedCallback = undefined;
	var doubleClickedCallback = undefined;
	var longPressedCallback = undefined;
	var colorPatternCallback = undefined;
	var tempo = 60;
	var timeouts = [];
	
	var sensory = {
		devices: [],
		readData: [],
		callbacks: []
	};
	var motoring = {
		devices: [],
		readData: [],
		writeData: [],
		callbacks: []
	};
	var state = ConnectionState.CONNECTING;
	var writeMotoringFlag = 0;
	
	var binder = {
		bind: function(device, type, data, callback) {
			if(type == DeviceType.SENSOR || type == DeviceType.EVENT) {
				sensory.devices.push(device);
				sensory.readData.push(data.read);
				sensory.callbacks.push(callback);
			} else {
				motoring.devices.push(device);
				motoring.readData.push(data.read);
				motoring.writeData.push(data.write);
				motoring.callbacks.push(callback);
			}
		},
		onMotoringDeviceWritten: function(device) {
			switch(device.getId()) {
				case Turtle.PULSE:
					writeMotoringFlag |= 0x04000000;
					break;
				case Turtle.NOTE:
					writeMotoringFlag |= 0x02000000;
					break;
				case Turtle.SOUND:
					writeMotoringFlag |= 0x01000000;
					break;
				case Turtle.LINE_TRACER_MODE:
					writeMotoringFlag |= 0x00800000;
					break;
				case Turtle.LINE_TRACER_GAIN:
					writeMotoringFlag |= 0x00400000;
					break;
				case Turtle.LINE_TRACER_SPEED:
					writeMotoringFlag |= 0x00200000;
					break;
				case Turtle.LAMP:
					writeMotoringFlag |= 0x00100000;
					break;
				case Turtle.LOCK:
					writeMotoringFlag |= 0x00080000;
					break;
				case Turtle.MOTION:
					writeMotoringFlag |= 0x00040000;
					break;
			}
		}
	};
	
	var devices = {};
	devices[Turtle.LEFT_WHEEL] = new Device(binder, Turtle.LEFT_WHEEL, 'LeftWheel', DeviceType.EFFECTOR, DataType.FLOAT, 1, -400, 400, 0);
	devices[Turtle.RIGHT_WHEEL] = new Device(binder, Turtle.RIGHT_WHEEL, 'RightWheel', DeviceType.EFFECTOR, DataType.FLOAT, 1, -400, 400, 0);
	devices[Turtle.LED] = new Device(binder, Turtle.LED, 'Led', DeviceType.EFFECTOR, DataType.INTEGER, 3, 0, 255, 0);
	devices[Turtle.BUZZER] = new Device(binder, Turtle.BUZZER, 'Buzzer', DeviceType.EFFECTOR, DataType.FLOAT, 1, 0, 167772.15, 0);
	devices[Turtle.PULSE] = new Device(binder, Turtle.PULSE, 'Pulse', DeviceType.COMMAND, DataType.INTEGER, 1, 0, 65535, 0);
	devices[Turtle.NOTE] = new Device(binder, Turtle.NOTE, 'Note', DeviceType.COMMAND, DataType.INTEGER, 1, 0, 88, 0);
	devices[Turtle.SOUND] = new Device(binder, Turtle.SOUND, 'Sound', DeviceType.COMMAND, DataType.INTEGER, 1, 0, 127, 0);
	devices[Turtle.LINE_TRACER_MODE] = new Device(binder, Turtle.LINE_TRACER_MODE, 'LineTracerMode', DeviceType.COMMAND, DataType.INTEGER, 1, 0, 127, 0);
	devices[Turtle.LINE_TRACER_GAIN] = new Device(binder, Turtle.LINE_TRACER_GAIN, 'LineTracerGain', DeviceType.COMMAND, DataType.INTEGER, 1, 1, 8, 5);
	devices[Turtle.LINE_TRACER_SPEED] = new Device(binder, Turtle.LINE_TRACER_SPEED, 'LineTracerSpeed', DeviceType.COMMAND, DataType.INTEGER, 1, 1, 8, 5);
	devices[Turtle.LAMP] = new Device(binder, Turtle.LAMP, 'Lamp', DeviceType.COMMAND, DataType.INTEGER, 1, 0, 1, 1);
	devices[Turtle.LOCK] = new Device(binder, Turtle.LOCK, 'Lock', DeviceType.COMMAND, DataType.INTEGER, 1, 0, 1, 0);
	devices[Turtle.MOTION] = new Device(binder, Turtle.MOTION, 'Motion', DeviceType.COMMAND, DataType.FLOAT, 5, 0, 2147483647, 0);
	devices[Turtle.SIGNAL_STRENGTH] = new Device(binder, Turtle.SIGNAL_STRENGTH, 'SignalStrength', DeviceType.SENSOR, DataType.INTEGER, 1, -128, 0, 0);
	devices[Turtle.COLOR] = new Device(binder, Turtle.COLOR, 'Color', DeviceType.SENSOR, DataType.INTEGER, 4, 0, 255, 0);
	devices[Turtle.FLOOR] = new Device(binder, Turtle.FLOOR, 'Floor', DeviceType.SENSOR, DataType.INTEGER, 1, 0, 255, 0);
	devices[Turtle.ACCELERATION] = new Device(binder, Turtle.ACCELERATION, 'Acceleration', DeviceType.SENSOR, DataType.INTEGER, 3, -32768, 32767, 0);
	devices[Turtle.TEMPERATURE] = new Device(binder, Turtle.TEMPERATURE, 'Temperature', DeviceType.SENSOR, DataType.INTEGER, 1, -40, 88, 0);
	devices[Turtle.BUTTON] = new Device(binder, Turtle.BUTTON, 'Button', DeviceType.EVENT, DataType.INTEGER, 1, 0, 1, 0);
	devices[Turtle.CLICKED] = new Device(binder, Turtle.CLICKED, 'Clicked', DeviceType.EVENT, DataType.INTEGER, 0, 0, 0, 0);
	devices[Turtle.DOUBLE_CLICKED] = new Device(binder, Turtle.DOUBLE_CLICKED, 'DoubleClicked', DeviceType.EVENT, DataType.INTEGER, 0, 0, 0, 0);
	devices[Turtle.LONG_PRESSED] = new Device(binder, Turtle.LONG_PRESSED, 'LongPressed', DeviceType.EVENT, DataType.INTEGER, 0, 0, 0, 0);
	devices[Turtle.COLOR_NUMBER] = new Device(binder, Turtle.COLOR_NUMBER, 'ColorNumber', DeviceType.EVENT, DataType.INTEGER, 1, -1, 8, -1);
	devices[Turtle.COLOR_PATTERN] = new Device(binder, Turtle.COLOR_PATTERN, 'ColorPattern', DeviceType.EVENT, DataType.INTEGER, 1, -1, 88, -1);
	devices[Turtle.PULSE_COUNT] = new Device(binder, Turtle.PULSE_COUNT, 'PulseCount', DeviceType.EVENT, DataType.INTEGER, 1, 0, 65535, 0);
	devices[Turtle.WHEEL_STATE] = new Device(binder, Turtle.WHEEL_STATE, 'WheelState', DeviceType.EVENT, DataType.INTEGER, 1, 0, 1, 0);
	devices[Turtle.SOUND_STATE] = new Device(binder, Turtle.SOUND_STATE, 'SoundState', DeviceType.EVENT, DataType.INTEGER, 1, 0, 1, 0);
	devices[Turtle.LINE_TRACER_STATE] = new Device(binder, Turtle.LINE_TRACER_STATE, 'LineTracerState', DeviceType.EVENT, DataType.INTEGER, 1, 0, 3, 0);
	
	var packet = {
		map: 0,
		leftWheel: 0,
		rightWheel: 0,
		ledRed: 0,
		ledGreen: 0,
		ledBlue: 0,
		buzzer: 0,
		pulse: 0,
		note: 0,
		sound: 0,
		lineTracerMode: 0,
		lineTracerGain: 5,
		lineTracerSpeed: 5,
		lamp: 1,
		lock: 0,
		motionType: 0,
		motionUnit: 0,
		motionSpeed: 0,
		motionValue: 0,
		motionRadius: 0
	};
	
	self.getId = function() {
		return id;
	};
	
	self.getName = function() {
		return robotName;
	};
	
	self.setName = function(name) {
		robotName = '' + name;
	};
	
	self.findDeviceById = function(deviceId) {
		var device = devices[deviceId];
		if(device) {
			return device;
		}
		return undefined;
	};
	
	self.e = function(deviceId) {
		var device = self.findDeviceById(deviceId);
		if(device) {
			return device.e();
		}
		return false;
	};
	
	self.read = function(deviceId, arg) {
		var device = self.findDeviceById(deviceId);
		if(device) {
			return device.read(arg);
		}
		return 0;
	};
	
	self.write = function(deviceId, arg1, arg2) {
		var device = self.findDeviceById(deviceId);
		if(device) {
			return device.write(arg1, arg2);
		}
		if(arg1 && arg1.constructor == Array) {
			return 0;
		}
		return false;
	};
	
	self.addDeviceDataChangedListener = function(listener) {
		for(var i in devices) {
			devices[i].addDeviceDataChangedListener(listener);
		}
	};
	
	self.removeDeviceDataChangedListener = function(listener) {
		for(var i in devices) {
			devices[i].removeDeviceDataChangedListener(listener);
		}
	};
	
	self.clearDeviceDataChangedListeners = function() {
		for(var i in devices) {
			devices[i].clearDeviceDataChangedListeners();
		}
	};
	
	var removeTimeout = function(id) {
		clearTimeout(id);
		var index = timeouts.indexOf(id);
		if(index >= 0) {
			timeouts.splice(index, 1);
		}
	};

	var removeAllTimeouts = function() {
		for(var i in timeouts) {
			clearTimeout(timeouts[i]);
		}
		timeouts = [];
	};
	
	var setMotion = function(type, unit, speed, value, radius) {
		self.write(Turtle.MOTION, 0, type);
		self.write(Turtle.MOTION, 1, unit);
		self.write(Turtle.MOTION, 2, speed);
		self.write(Turtle.MOTION, 3, value);
		self.write(Turtle.MOTION, 4, radius);
	};
	
	var issueMotionId = function() {
		motionId = blockId = (blockId % 65535) + 1;
		return motionId;
	};
	
	var cancelMotion = function() {
		motionId = 0;
		motionCallback = undefined;
	};
	
	var issueLineTracerId = function() {
		lineTracerId = blockId = (blockId % 65535) + 1;
		return lineTracerId;
	};
	
	var cancelLineTracer = function() {
		lineTracerId = 0;
		lineTracerCallback = undefined;
	};
	
	var issueNoteId = function() {
		noteId = blockId = (blockId % 65535) + 1;
		return noteId;
	};
	
	var cancelNote = function() {
		noteId = 0;
		if(noteTimer1 !== undefined) {
			removeTimeout(noteTimer1);
		}
		if(noteTimer2 !== undefined) {
			removeTimeout(noteTimer2);
		}
		noteTimer1 = undefined;
		noteTimer2 = undefined;
	};
	
	var runSound = function(sound, count) {
		if(typeof count != 'number') count = 1;
		if(count < 0) count = -1;
		if(count) {
			currentSound = sound;
			soundRepeat = count;
			self.write(Turtle.SOUND, sound);
		}
	};
	
	var cancelSound = function() {
		soundCallback = undefined;
	};
	
	var index = __runner.addRobot(self, packet, {
		handleSensory: function(data) {
			state = data.connectionState;
			var map = data.map;
			var readData = sensory.readData;
			readData[0][0] = data.signalStrength;
			readData[1][0] = data.colorRed;
			readData[1][1] = data.colorGreen;
			readData[1][2] = data.colorBlue;
			readData[1][3] = data.colorClear;
			readData[2][0] = data.floor;
			readData[3][0] = data.accelerationX;
			readData[3][1] = data.accelerationY;
			readData[3][2] = data.accelerationZ;
			readData[4][0] = data.temperature;
			
			var callbacks = sensory.callbacks;
			var i;
			for(i = 0; i < 5; ++i) {
				callbacks[i].fire();
			}
			
			if(map & 0x00001000) {
				readData[5][0] = data.button;
				callbacks[5].fire();
			}
			if(map & 0x00000800) {
				callbacks[6].fire();
			}
			if(map & 0x00000400) {
				callbacks[7].fire();
			}
			if(map & 0x00000200) {
				callbacks[8].fire();
			}
			if(map & 0x00000100) {
				readData[9][0] = data.colorNumber;
				callbacks[9].fire();
			}
			if(map & 0x00000080) {
				readData[10][0] = data.colorPattern;
				callbacks[10].fire();
			}
			if(map & 0x00000040) {
				readData[11][0] = data.pulseCount;
				callbacks[11].fire();
			}
			if(map & 0x00000020) {
				readData[12][0] = data.wheelState;
				callbacks[12].fire();
			}
			if(map & 0x00000010) {
				readData[13][0] = data.soundState;
				callbacks[13].fire();
			}
			if(map & 0x00000008) {
				readData[14][0] = data.lineTracerState;
				callbacks[14].fire();
			}
			
			for(i = 0; i < 5; ++i) {
				callbacks[i].notifyDeviceDataChanged();
			}
			
			if(map & 0x00001000) {
				callbacks[5].notifyDeviceDataChanged();
			}
			if(map & 0x00000800) {
				callbacks[6].notifyDeviceDataChanged();
				if(clickedCallback) clickedCallback();
			}
			if(map & 0x00000400) {
				callbacks[7].notifyDeviceDataChanged();
				if(doubleClickedCallback) doubleClickedCallback();
			}
			if(map & 0x00000200) {
				callbacks[8].notifyDeviceDataChanged();
				if(longPressedCallback) longPressedCallback();
			}
			if(map & 0x00000100) {
				callbacks[9].notifyDeviceDataChanged();
			}
			if(map & 0x00000080) {
				callbacks[10].notifyDeviceDataChanged();
				if(colorPatternCallback) colorPatternCallback(data.colorPattern);
			}
			if(map & 0x00000040) {
				callbacks[11].notifyDeviceDataChanged();
			}
			if(map & 0x00000020) {
				callbacks[12].notifyDeviceDataChanged();
				if(motionId > 0 && data.wheelState == 0) {
					self.write(Turtle.LEFT_WHEEL, 0);
					self.write(Turtle.RIGHT_WHEEL, 0);
					var callback = motionCallback;
					cancelMotion();
					if(typeof callback == 'function') callback();
				}
			}
			if(map & 0x00000010) {
				callbacks[13].notifyDeviceDataChanged();
				if(data.soundState == 0) {
					if(currentSound > 0) {
						if(soundRepeat < 0) {
							runSound(currentSound, -1);
						} else if(soundRepeat > 1) {
							soundRepeat --;
							runSound(currentSound, soundRepeat);
						} else {
							currentSound = 0;
							soundRepeat = 1;
							var callback = soundCallback;
							cancelSound();
							if(typeof callback == 'function') callback();
						}
					} else {
						currentSound = 0;
						soundRepeat = 1;
						var callback = soundCallback;
						cancelSound();
						if(typeof callback == 'function') callback();
					}
				}
			}
			if(map & 0x00000008) {
				callbacks[14].notifyDeviceDataChanged();
				if(lineTracerId > 0 && data.lineTracerState == 0x02) {
					self.write(Turtle.LINE_TRACER_MODE, 0);
					var callback = lineTracerCallback;
					cancelLineTracer();
					if(typeof callback == 'function') callback();
				}
			}
		},
		requestMotoring: function() {
			var map = 0xf8000000 | writeMotoringFlag;
			writeMotoringFlag = 0;
			
			var writeData = motoring.writeData;
			packet.map = map;
			packet.leftWheel = writeData[0][0];
			packet.rightWheel = writeData[1][0];
			packet.ledRed = writeData[2][0];
			packet.ledGreen = writeData[2][1];
			packet.ledBlue = writeData[2][2];
			packet.buzzer = writeData[3][0];
			if(map & 0x04000000) packet.pulse = writeData[4][0];
			if(map & 0x02000000) packet.note = writeData[5][0];
			if(map & 0x01000000) packet.sound = writeData[6][0];
			if(map & 0x00800000) packet.lineTracerMode = writeData[7][0];
			if(map & 0x00400000) packet.lineTracerGain = writeData[8][0];
			if(map & 0x00200000) packet.lineTracerSpeed = writeData[9][0];
			if(map & 0x00100000) packet.lamp = writeData[10][0];
			if(map & 0x00080000) packet.lock = writeData[11][0];
			if(map & 0x00040000) {
				packet.motionType = parseInt(writeData[12][0]);
				packet.motionUnit = parseInt(writeData[12][1]);
				packet.motionSpeed = writeData[12][2];
				packet.motionValue = writeData[12][3];
				packet.motionRadius = writeData[12][4];
			}
		},
		updateSensory: function() {
			var callbacks = sensory.callbacks;
			for(var i in callbacks) {
				callbacks[i].update();
			}
		},
		updateMotoring: function() {
			var callbacks = motoring.callbacks;
			for(var i in callbacks) {
				callbacks[i].update();
			}
		}
	});
	
	self.getIndex = function() {
		return index;
	};
	
	self.reset = function() {
		var callbacks = motoring.callbacks;
		for(var i in callbacks) {
			callbacks[i].reset();
		}
		callbacks = sensory.callbacks;
		for(var i in callbacks) {
			callbacks[i].reset();
		}
		writeMotoringFlag |= 0x07fc0000;
		
		blockId = 0;
		motionId = 0;
		motionCallback = undefined;
		lineTracerId = 0;
		lineTracerCallback = undefined;
		currentSound = 0;
		soundRepeat = 1;
		soundCallback = undefined;
		noteId = 0;
		noteTimer1 = undefined;
		noteTimer2 = undefined;
		clickedCallback = undefined;
		doubleClickedCallback = undefined;
		longPressedCallback = undefined;
		colorPatternCallback = undefined;
		tempo = 60;
		removeAllTimeouts();
	};
	
	self.connectionState = function() {
		return state;
	};
	
	var moveSecDeg = function(motion, oppsite, arg1, arg2, arg3) {
		cancelLineTracer();
		
		// turtle uses internal distance/angle if motion is 101 ~ 104
		// turtle uses internal speed if speed is 0
		var callback = undefined;
		if(typeof arg1 == 'number') { // value
			var speed = undefined;
			cancelMotion();
			if(typeof arg2 == 'number') {
				speed = arg2;
				if(typeof arg3 == 'function') {
					callback = arg3;
				}
			} else if(typeof arg2 == 'function') {
				callback = arg2;
			}
			
			self.write(Turtle.LEFT_WHEEL, 0);
			self.write(Turtle.RIGHT_WHEEL, 0);
			self.write(Turtle.PULSE, 0);
			if(arg1 == 0) { // stop
				setMotion(0, 0, 0, 0, 0);
				self.write(Turtle.LINE_TRACER_MODE, 0);
				if(callback) callback();
			} else {
				if(typeof speed == 'number') {
					if(speed == 0) {
						setMotion(0, 0, 0, 0, 0);
					} else {
						if(arg1 < 0) {
							arg1 = -arg1;
							if(speed < 0) speed = -speed;
							else motion = oppsite;
						} else if(speed < 0) {
							speed = -speed;
							motion = oppsite;
						}
						setMotion(motion, 1, speed, arg1, 0);
						motionCallback = callback;
						issueMotionId();
					}
				} else { // internal speed
					if(arg1 < 0) {
						arg1 = -arg1;
						motion = oppsite;
					}
					setMotion(motion, 1, 0, arg1, 0);
					motionCallback = callback;
					issueMotionId();
				}
				self.write(Turtle.LINE_TRACER_MODE, 0);
			}
		} else { // internal motion
			if(typeof arg1 == 'function') {
				callback = arg1;
			}
			self.write(Turtle.LEFT_WHEEL, 0);
			self.write(Turtle.RIGHT_WHEEL, 0);
			self.write(Turtle.PULSE, 0);
			setMotion(motion + 100, 1, 0, 0, 0); // type, unit, speed, value, radius
			motionCallback = callback;
			issueMotionId();
			self.write(Turtle.LINE_TRACER_MODE, 0);
		}
	};
	
	self.moveForward = function(arg1, arg2, arg3) {
		moveSecDeg(1, 2, arg1, arg2, arg3);
	};
	
	self.moveBackward = function(arg1, arg2, arg3) {
		moveSecDeg(2, 1, arg1, arg2, arg3);
	};
	
	self.turnLeft = function(arg1, arg2, arg3) {
		moveSecDeg(3, 4, arg1, arg2, arg3);
	};
	
	self.turnRight = function(arg1, arg2, arg3) {
		moveSecDeg(4, 3, arg1, arg2, arg3);
	};
	
	var move = function(motion, oppsite, value, unit, arg1, arg2) {
		cancelLineTracer();
		cancelMotion();
		
		// turtle uses internal speed if speed is 0
		var speed = undefined, callback = undefined;
		if(typeof arg1 == 'number') {
			speed = arg1;
			if(typeof arg2 == 'function') {
				callback = arg2;
			}
		} else if(typeof arg1 == 'function') {
			callback = arg1;
		}
		
		self.write(Turtle.LEFT_WHEEL, 0);
		self.write(Turtle.RIGHT_WHEEL, 0);
		self.write(Turtle.PULSE, 0);
		if(typeof value == 'number' && value > 0) {
			if(typeof speed == 'number') {
				if(speed == 0) {
					setMotion(0, 0, 0, 0, 0);
				} else {
					if(speed < 0) {
						speed = -speed;
						motion = oppsite;
					}
					setMotion(motion, unit, speed, value, 0);
					motionCallback = callback;
					issueMotionId();
				}
			} else { // internal speed
				setMotion(motion, unit, 0, value, 0);
				motionCallback = callback;
				issueMotionId();
			}
			self.write(Turtle.LINE_TRACER_MODE, 0);
		} else { // stop
			setMotion(0, 0, 0, 0, 0);
			self.write(Turtle.LINE_TRACER_MODE, 0);
			if(callback) callback();
		}
	};
	
	self.moveForwardSec = function(sec, arg1, arg2) {
		move(1, 2, sec, 2, arg1, arg2);
	};
	
	self.moveBackwardSec = function(sec, arg1, arg2) {
		move(2, 1, sec, 2, arg1, arg2);
	};
	
	self.turnLeftSec = function(sec, arg1, arg2) {
		move(3, 4, sec, 2, arg1, arg2);
	};
	
	self.turnRightSec = function(sec, arg1, arg2) {
		move(4, 3, sec, 2, arg1, arg2);
	};
	
	self.moveForwardPulse = function(pulse, arg1, arg2) {
		move(1, 2, pulse, 3, arg1, arg2);
	};
	
	self.moveBackwardPulse = function(pulse, arg1, arg2) {
		move(2, 1, pulse, 3, arg1, arg2);
	};
	
	self.turnLeftPulse = function(pulse, arg1, arg2) {
		move(3, 4, pulse, 3, arg1, arg2);
	};
	
	self.turnRightPulse = function(pulse, arg1, arg2) {
		move(4, 3, pulse, 3, arg1, arg2);
	};
	
	var pivot = function(motion, tail, value, unit, arg1, arg2) {
		cancelLineTracer();
		cancelMotion();
		
		if(unit != 1 && value < 0) {
			value = 0;
		}
		
		// turtle uses internal speed if speed is 0
		var speed = undefined, callback = undefined;
		if(typeof arg1 == 'number') {
			speed = arg1;
			if(typeof arg2 == 'function') {
				callback = arg2;
			}
		} else if(typeof arg1 == 'function') {
			callback = arg1;
		}
		
		self.write(Turtle.LEFT_WHEEL, 0);
		self.write(Turtle.RIGHT_WHEEL, 0);
		self.write(Turtle.PULSE, 0);
		if(typeof value == 'number' && value != 0) {
			if(typeof speed == 'number') {
				if(speed == 0) {
					setMotion(0, 0, 0, 0, 0);
				} else {
					if(value < 0) {
						value = -value;
						if(speed < 0) speed = -speed;
						else motion = tail;
					} else if(speed < 0) {
						speed = -speed;
						motion = tail;
					}
					setMotion(motion, unit, speed, value, 0);
					motionCallback = callback;
					issueMotionId();
				}
			} else { // internal speed
				if(value < 0) {
					value = -value;
					motion = tail;
				}
				setMotion(motion, unit, 0, value, 0);
				motionCallback = callback;
				issueMotionId();
			}
			self.write(Turtle.LINE_TRACER_MODE, 0);
		} else { // stop
			setMotion(0, 0, 0, 0, 0);
			self.write(Turtle.LINE_TRACER_MODE, 0);
			if(callback) callback();
		}
	};
	
	self.pivotLeft = function(deg, arg1, arg2) {
		pivot(5, 6, deg, 1, arg1, arg2);
	};
	
	self.pivotRight = function(deg, arg1, arg2) {
		pivot(7, 8, deg, 1, arg1, arg2);
	};
	
	self.pivotLeftSec = function(sec, arg1, arg2) {
		pivot(5, 6, sec, 2, arg1, arg2);
	};
	
	self.pivotRightSec = function(sec, arg1, arg2) {
		pivot(7, 8, sec, 2, arg1, arg2);
	};
	
	self.pivotLeftPulse = function(pulse, arg1, arg2) {
		pivot(5, 6, pulse, 3, arg1, arg2);
	};
	
	self.pivotRightPulse = function(pulse, arg1, arg2) {
		pivot(7, 8, pulse, 3, arg1, arg2);
	};
	
	var swing = function(motion, tail, value, unit, radius, arg1, arg2) {
		cancelLineTracer();
		cancelMotion();
		
		if(unit != 1 && value < 0) {
			value = 0;
		}
		
		// turtle uses internal speed if speed is 0
		var speed = undefined, callback = undefined;
		if(typeof arg1 == 'number') {
			speed = arg1;
			if(typeof arg2 == 'function') {
				callback = arg2;
			}
		} else if(typeof arg1 == 'function') {
			callback = arg1;
		}
		
		self.write(Turtle.LEFT_WHEEL, 0);
		self.write(Turtle.RIGHT_WHEEL, 0);
		self.write(Turtle.PULSE, 0);
		if(typeof value == 'number' && value != 0 && (typeof radius == 'number') && radius >= 0) {
			if(typeof speed == 'number') {
				if(speed == 0) {
					setMotion(0, 0, 0, 0, 0);
				} else {
					if(value < 0) {
						value = -value;
						if(speed < 0) speed = -speed;
						else motion = tail;
					} else if(speed < 0) {
						speed = -speed;
						motion = tail;
					}
					setMotion(motion, unit, speed, value, radius);
					motionCallback = callback;
					issueMotionId();
				}
			} else { // internal speed
				if(value < 0) {
					value = -value;
					motion = tail;
				}
				setMotion(motion, unit, 0, value, radius);
				motionCallback = callback;
				issueMotionId();
			}
			self.write(Turtle.LINE_TRACER_MODE, 0);
		} else { // stop
			setMotion(0, 0, 0, 0, 0);
			self.write(Turtle.LINE_TRACER_MODE, 0);
			if(callback) callback();
		}
	};
	
	self.swingLeft = function(deg, radius, arg1, arg2) {
		swing(9, 10, deg, 1, radius, arg1, arg2);
	};
	
	self.swingRight = function(deg, radius, arg1, arg2) {
		swing(11, 12, deg, 1, radius, arg1, arg2);
	};
	
	self.swingLeftSec = function(sec, radius, arg1, arg2) {
		swing(9, 10, sec, 2, radius, arg1, arg2);
	};
	
	self.swingRightSec = function(sec, radius, arg1, arg2) {
		swing(11, 12, sec, 2, radius, arg1, arg2);
	};
	
	self.swingLeftPulse = function(pulse, radius, arg1, arg2) {
		swing(9, 10, pulse, 3, radius, arg1, arg2);
	};
	
	self.swingRightPulse = function(pulse, radius, arg1, arg2) {
		swing(11, 12, pulse, 3, radius, arg1, arg2);
	};
	
	self.wheels = function(leftVelocity, rightVelocity) {
		cancelLineTracer();
		cancelMotion();
		
		if(typeof leftVelocity == 'number') {
			self.write(Turtle.LEFT_WHEEL, leftVelocity);
		}
		if(typeof rightVelocity == 'number') {
			self.write(Turtle.RIGHT_WHEEL, rightVelocity);
		} else {
			if(typeof leftVelocity == 'number') {
				self.write(Turtle.RIGHT_WHEEL, leftVelocity);
			}
		}
		self.write(Turtle.PULSE, 0);
		setMotion(0, 0, 0, 0, 0);
		self.write(Turtle.LINE_TRACER_MODE, 0);
	};
	
	self.leftWheel = function(velocity) {
		cancelLineTracer();
		cancelMotion();
		
		if(typeof velocity == 'number') {
			self.write(Turtle.LEFT_WHEEL, velocity);
		}
		self.write(Turtle.PULSE, 0);
		setMotion(0, 0, 0, 0, 0);
		self.write(Turtle.LINE_TRACER_MODE, 0);
	};
	
	self.rightWheel = function(velocity) {
		cancelLineTracer();
		cancelMotion();
		
		if(typeof velocity == 'number') {
			self.write(Turtle.RIGHT_WHEEL, velocity);
		}
		self.write(Turtle.PULSE, 0);
		setMotion(0, 0, 0, 0, 0);
		self.write(Turtle.LINE_TRACER_MODE, 0);
	};
	
	self.stop = function() {
		cancelLineTracer();
		cancelMotion();
		
		self.write(Turtle.LEFT_WHEEL, 0);
		self.write(Turtle.RIGHT_WHEEL, 0);
		self.write(Turtle.PULSE, 0);
		setMotion(0, 0, 0, 0, 0);
		self.write(Turtle.LINE_TRACER_MODE, 0);
	};
	
	self.lineTracerMode = function(arg1, arg2) {
		cancelMotion();
		
		var mode = 0, callback = undefined;
		if(typeof arg1 == 'number') {
			mode = arg1;
			switch(arg1) {
				case Turtle.LINE_TRACER_MODE_TURN_LEFT:
				case Turtle.LINE_TRACER_MODE_TURN_RIGHT:
				case Turtle.LINE_TRACER_MODE_CROSS:
				case Turtle.LINE_TRACER_MODE_UTURN:
				case Turtle.LINE_TRACER_MODE_BLACK_UNTIL_RED:
				case Turtle.LINE_TRACER_MODE_BLACK_UNTIL_YELLOW:
				case Turtle.LINE_TRACER_MODE_BLACK_UNTIL_GREEN:
				case Turtle.LINE_TRACER_MODE_BLACK_UNTIL_CYAN:
				case Turtle.LINE_TRACER_MODE_BLACK_UNTIL_BLUE:
				case Turtle.LINE_TRACER_MODE_BLACK_UNTIL_MAGENTA:
				case Turtle.LINE_TRACER_MODE_BLACK_UNTIL_ANY:
				case Turtle.LINE_TRACER_MODE_RED_UNTIL_BLACK:
				case Turtle.LINE_TRACER_MODE_GREEN_UNTIL_BLACK:
				case Turtle.LINE_TRACER_MODE_BLUE_UNTIL_BLACK:
				case Turtle.LINE_TRACER_MODE_ANY_UNTIL_BLACK:
					if(typeof arg2 == 'function') callback = arg2;
					lineTracerCallback = callback;
					issueLineTracerId();
					break;
				default:
					cancelLineTracer();
					break;
			}
		} else {
			cancelLineTracer();
		}
		self.write(Turtle.LEFT_WHEEL, 0);
		self.write(Turtle.RIGHT_WHEEL, 0);
		self.write(Turtle.PULSE, 0);
		setMotion(0, 0, 0, 0, 0);
		self.write(Turtle.LINE_TRACER_MODE, mode);
	};
	
	self.lineTracerGain = function(gain) {
		if(typeof gain == 'number') {
			self.write(Turtle.LINE_TRACER_GAIN, gain);
		}
	};
	
	self.lineTracerSpeed = function(speed) {
		if(typeof speed == 'number') {
			self.write(Turtle.LINE_TRACER_SPEED, speed);
		}
	};
	
	self.headColor = function(color) {
		if(typeof color == 'string') {
			switch(color) {
				case 'red':
					self.write(Turtle.LED, 0, 255);
					self.write(Turtle.LED, 1, 0);
					self.write(Turtle.LED, 2, 0);
					break;
				case 'orange':
					self.write(Turtle.LED, 0, 255);
					self.write(Turtle.LED, 1, 63);
					self.write(Turtle.LED, 2, 0);
					break;
				case 'yellow':
					self.write(Turtle.LED, 0, 255);
					self.write(Turtle.LED, 1, 255);
					self.write(Turtle.LED, 2, 0);
					break;
				case 'green':
					self.write(Turtle.LED, 0, 0);
					self.write(Turtle.LED, 1, 255);
					self.write(Turtle.LED, 2, 0);
					break;
				case 'cyan':
				case 'sky blue':
				case 'skyblue':
					self.write(Turtle.LED, 0, 0);
					self.write(Turtle.LED, 1, 255);
					self.write(Turtle.LED, 2, 255);
					break;
				case 'blue':
					self.write(Turtle.LED, 0, 0);
					self.write(Turtle.LED, 1, 0);
					self.write(Turtle.LED, 2, 255);
					break;
				case 'violet':
					self.write(Turtle.LED, 0, 63);
					self.write(Turtle.LED, 1, 0);
					self.write(Turtle.LED, 2, 255);
					break;
				case 'purple':
				case 'magenta':
					self.write(Turtle.LED, 0, 255);
					self.write(Turtle.LED, 1, 0);
					self.write(Turtle.LED, 2, 255);
					break;
				case 'white':
					self.write(Turtle.LED, 0, 255);
					self.write(Turtle.LED, 1, 255);
					self.write(Turtle.LED, 2, 255);
					break;
				default:
					self.write(Turtle.LED, 0, 0);
					self.write(Turtle.LED, 1, 0);
					self.write(Turtle.LED, 2, 0);
					break;
			}
		}
	};
	
	self.headRgb = function(red, green, blue) {
		if(typeof red == 'number') {
			self.write(Turtle.LED, 0, red);
		}
		if((typeof green == 'number') && (typeof blue == 'number')) {
			self.write(Turtle.LED, 1, green);
			self.write(Turtle.LED, 2, blue);
		} else if(typeof red == 'number') {
			self.write(Turtle.LED, 1, red);
			self.write(Turtle.LED, 2, red);
		}
	};
	
	self.buzzer = function(hz) {
		cancelNote();
		cancelSound();
		if(typeof hz == 'number') {
			self.write(Turtle.BUZZER, hz);
		}
		self.write(Turtle.NOTE, 0);
		runSound(0);
	};
	
	self.tempo = function(bpm) {
		if(typeof bpm == 'number') {
			tempo = bpm;
			if(tempo < 1) tempo = 1;
		}
	};
	
	self.note = function(arg1, arg2, arg3) {
		cancelNote();
		cancelSound();
		
		var pitch = 0, beat = undefined, callback = undefined;
		if(typeof arg1 == 'number') {
			pitch = arg1;
			if(typeof arg2 == 'number') {
				beat = arg2;
				if(typeof arg3 == 'function') {
					callback = arg3;
				}
			}
		}
		
		self.write(Turtle.BUZZER, 0);
		if(typeof beat == 'number') {
			if(beat > 0 && tempo > 0) {
				var id = issueNoteId();
				if(pitch > 0) {
					self.write(Turtle.NOTE, pitch);
					var timeout = beat * 60 * 1000 / tempo;
					var tail = 0;
					if(timeout > 100) tail = 100;
					if(tail > 0) {
						noteTimer1 = setTimeout(function() {
							if(noteId == id) {
								self.write(Turtle.NOTE, 0);
								if(noteTimer1 !== undefined) removeTimeout(noteTimer1);
								noteTimer1 = undefined;
							}
						}, timeout - tail);
						timeouts.push(noteTimer1);
					}
					noteTimer2 = setTimeout(function() {
						if(noteId == id) {
							self.write(Turtle.NOTE, 0);
							cancelNote();
							if(typeof callback == 'function') callback();
						}
					}, timeout);
					timeouts.push(noteTimer2);
				} else {
					self.write(Turtle.NOTE, 0);
					var id = issueNoteId();
					noteTimer1 = setTimeout(function() {
						if(noteId == id) {
							cancelNote();
							if(typeof callback == 'function') callback();
						}
					}, beat * 60 * 1000 / tempo);
					timeouts.push(noteTimer1);
				}
				runSound(0);
			} else { // off
				self.write(Turtle.NOTE, 0);
				runSound(0);
				if(typeof callback == 'function') callback();
			}
		} else { // continue
			self.write(Turtle.NOTE, pitch);
			runSound(0);
		}
	};
	
	self.sound = function(sound, arg1, arg2) {
		cancelNote();
		cancelSound();
		
		var count = 1, callback = undefined;
		if(typeof arg1 == 'number') {
			count = arg1;
			if(typeof arg2 == 'function') {
				callback = arg2;
			}
		} else if(typeof arg1 == 'function') {
			callback = arg1;
		}
		
		self.write(Turtle.BUZZER, 0);
		self.write(Turtle.NOTE, 0);
		if(typeof sound == 'number' && sound > 0 && count) {
			runSound(sound, count);
			soundCallback = callback;
		} else { // off
			runSound(0);
			if(callback) callback();
		}
	};
	
	self.lamp = function(onoff) {
		if(typeof onoff == 'number') {
			self.write(Turtle.LAMP, onoff);
		}
	};
	
	self.lock = function(onoff) {
		if(typeof onoff == 'number') {
			self.write(Turtle.LOCK, onoff);
		}
	};
	
	self.signalStrength = function() {
		return self.read(Turtle.SIGNAL_STRENGTH);
	};
	
	self.floor = function() {
		return self.read(Turtle.FLOOR);
	};
	
	self.accelerationX = function() {
		return self.read(Turtle.ACCELERATION, 0);
	};
	
	self.accelerationY = function() {
		return self.read(Turtle.ACCELERATION, 1);
	};
	
	self.accelerationZ = function() {
		return self.read(Turtle.ACCELERATION, 2);
	};
	
	self.temperature = function() {
		return self.read(Turtle.TEMPERATURE);
	};
	
	self.button = function() {
		return self.read(Turtle.BUTTON);
	};
	
	self.clicked = function(callback) {
		if(typeof callback == 'function') clickedCallback = callback;
		return self.e(Turtle.CLICKED);
	};
	
	self.doubleClicked = function(callback) {
		if(typeof callback == 'function') doubleClickedCallback = callback;
		return self.e(Turtle.DOUBLE_CLICKED);
	};
	
	self.longPressed = function(callback) {
		if(typeof callback == 'function') longPressedCallback = callback;
		return self.e(Turtle.LONG_PRESSED);
	};
	
	self.colorNumber = function() {
		return self.read(Turtle.COLOR_NUMBER);
	};
	
	self.colorPattern = function(callback) {
		if(typeof callback == 'function') colorPatternCallback = callback;
		if(self.e(Turtle.COLOR_PATTERN)) return self.read(Turtle.COLOR_PATTERN);
		return -1;
	};
	
	self.pulseCount = function() {
		return self.read(Turtle.PULSE_COUNT);
	};
}

Turtle.ID = 'kr.robomation.physical.turtle';

Turtle.LEFT_WHEEL = 0x00900000;
Turtle.RIGHT_WHEEL = 0x00900001;
Turtle.LED = 0x00900002;
Turtle.BUZZER = 0x00900003;
Turtle.PULSE = 0x00900004;
Turtle.NOTE = 0x00900005;
Turtle.SOUND = 0x00900006;
Turtle.LINE_TRACER_MODE = 0x00900007;
Turtle.LINE_TRACER_GAIN = 0x00900008;
Turtle.LINE_TRACER_SPEED = 0x00900009;
Turtle.LAMP = 0x0090000a;
Turtle.LOCK = 0x0090000b;
Turtle.MOTION = 0x0090000c;

Turtle.SIGNAL_STRENGTH = 0x0090000d;
Turtle.COLOR = 0x0090000e;
Turtle.FLOOR = 0x0090000f;
Turtle.ACCELERATION = 0x00900010;
Turtle.TEMPERATURE = 0x00900011;
Turtle.BUTTON = 0x00900012;
Turtle.CLICKED = 0x00900013;
Turtle.DOUBLE_CLICKED = 0x00900014;
Turtle.LONG_PRESSED = 0x00900015;
Turtle.COLOR_NUMBER = 0x00900016;
Turtle.COLOR_PATTERN = 0x00900017;
Turtle.PULSE_COUNT = 0x00900018;
Turtle.WHEEL_STATE = 0x00900019;
Turtle.SOUND_STATE = 0x0090001a;
Turtle.LINE_TRACER_STATE = 0x0090001b;

Turtle.COLOR_NONE = -1;
Turtle.COLOR_BLACK = 0;
Turtle.COLOR_RED = 1;
Turtle.COLOR_ORANGE = 2;
Turtle.COLOR_YELLOW = 3;
Turtle.COLOR_GREEN = 4;
Turtle.COLOR_CYAN = 5;
Turtle.COLOR_BLUE = 6;
Turtle.COLOR_MAGENTA = 7;
Turtle.COLOR_WHITE = 8;

Turtle.LINE_TRACER_MODE_OFF = 0;
Turtle.LINE_TRACER_MODE_BLACK = 10;
Turtle.LINE_TRACER_MODE_RED = 11;
Turtle.LINE_TRACER_MODE_GREEN = 13;
Turtle.LINE_TRACER_MODE_BLUE = 15;
Turtle.LINE_TRACER_MODE_ANY = 17;
Turtle.LINE_TRACER_MODE_TURN_LEFT = 20;
Turtle.LINE_TRACER_MODE_TURN_RIGHT = 30;
Turtle.LINE_TRACER_MODE_CROSS = 40;
Turtle.LINE_TRACER_MODE_UTURN = 50;
Turtle.LINE_TRACER_MODE_BLACK_UNTIL_RED = 61;
Turtle.LINE_TRACER_MODE_BLACK_UNTIL_YELLOW = 62;
Turtle.LINE_TRACER_MODE_BLACK_UNTIL_GREEN = 63;
Turtle.LINE_TRACER_MODE_BLACK_UNTIL_CYAN = 64;
Turtle.LINE_TRACER_MODE_BLACK_UNTIL_BLUE = 65;
Turtle.LINE_TRACER_MODE_BLACK_UNTIL_MAGENTA = 66;
Turtle.LINE_TRACER_MODE_BLACK_UNTIL_ANY = 67;
Turtle.LINE_TRACER_MODE_RED_UNTIL_BLACK = 71;
Turtle.LINE_TRACER_MODE_GREEN_UNTIL_BLACK = 73;
Turtle.LINE_TRACER_MODE_BLUE_UNTIL_BLACK = 75;
Turtle.LINE_TRACER_MODE_ANY_UNTIL_BLACK = 77;

Turtle.NOTE_OFF = 0;
Turtle.NOTE_A_0 = 1;
Turtle.NOTE_A_SHARP_0 = 2;
Turtle.NOTE_B_FLAT_0 = 2;
Turtle.NOTE_B_0 = 3;
Turtle.NOTE_C_1 = 4;
Turtle.NOTE_C_SHARP_1 = 5;
Turtle.NOTE_D_FLAT_1 = 5;
Turtle.NOTE_D_1 = 6;
Turtle.NOTE_D_SHARP_1 = 7;
Turtle.NOTE_E_FLAT_1 = 7;
Turtle.NOTE_E_1 = 8;
Turtle.NOTE_F_1 = 9;
Turtle.NOTE_F_SHARP_1 = 10;
Turtle.NOTE_G_FLAT_1 = 10;
Turtle.NOTE_G_1 = 11;
Turtle.NOTE_G_SHARP_1 = 12;
Turtle.NOTE_A_FLAT_1 = 12;
Turtle.NOTE_A_1 = 13;
Turtle.NOTE_A_SHARP_1 = 14;
Turtle.NOTE_B_FLAT_1 = 14;
Turtle.NOTE_B_1 = 15;
Turtle.NOTE_C_2 = 16;
Turtle.NOTE_C_SHARP_2 = 17;
Turtle.NOTE_D_FLAT_2 = 17;
Turtle.NOTE_D_2 = 18;
Turtle.NOTE_D_SHARP_2 = 19;
Turtle.NOTE_E_FLAT_2 = 19;
Turtle.NOTE_E_2 = 20;
Turtle.NOTE_F_2 = 21;
Turtle.NOTE_F_SHARP_2 = 22;
Turtle.NOTE_G_FLAT_2 = 22;
Turtle.NOTE_G_2 = 23;
Turtle.NOTE_G_SHARP_2 = 24;
Turtle.NOTE_A_FLAT_2 = 24;
Turtle.NOTE_A_2 = 25;
Turtle.NOTE_A_SHARP_2 = 26;
Turtle.NOTE_B_FLAT_2 = 26;
Turtle.NOTE_B_2 = 27;
Turtle.NOTE_C_3 = 28;
Turtle.NOTE_C_SHARP_3 = 29;
Turtle.NOTE_D_FLAT_3 = 29;
Turtle.NOTE_D_3 = 30;
Turtle.NOTE_D_SHARP_3 = 31;
Turtle.NOTE_E_FLAT_3 = 31;
Turtle.NOTE_E_3 = 32;
Turtle.NOTE_F_3 = 33;
Turtle.NOTE_F_SHARP_3 = 34;
Turtle.NOTE_G_FLAT_3 = 34;
Turtle.NOTE_G_3 = 35;
Turtle.NOTE_G_SHARP_3 = 36;
Turtle.NOTE_A_FLAT_3 = 36;
Turtle.NOTE_A_3 = 37;
Turtle.NOTE_A_SHARP_3 = 38;
Turtle.NOTE_B_FLAT_3 = 38;
Turtle.NOTE_B_3 = 39;
Turtle.NOTE_C_4 = 40;
Turtle.NOTE_C_SHARP_4 = 41;
Turtle.NOTE_D_FLAT_4 = 41;
Turtle.NOTE_D_4 = 42;
Turtle.NOTE_D_SHARP_4 = 43;
Turtle.NOTE_E_FLAT_4 = 43;
Turtle.NOTE_E_4 = 44;
Turtle.NOTE_F_4 = 45;
Turtle.NOTE_F_SHARP_4 = 46;
Turtle.NOTE_G_FLAT_4 = 46;
Turtle.NOTE_G_4 = 47;
Turtle.NOTE_G_SHARP_4 = 48;
Turtle.NOTE_A_FLAT_4 = 48;
Turtle.NOTE_A_4 = 49;
Turtle.NOTE_A_SHARP_4 = 50;
Turtle.NOTE_B_FLAT_4 = 50;
Turtle.NOTE_B_4 = 51;
Turtle.NOTE_C_5 = 52;
Turtle.NOTE_C_SHARP_5 = 53;
Turtle.NOTE_D_FLAT_5 = 53;
Turtle.NOTE_D_5 = 54;
Turtle.NOTE_D_SHARP_5 = 55;
Turtle.NOTE_E_FLAT_5 = 55;
Turtle.NOTE_E_5 = 56;
Turtle.NOTE_F_5 = 57;
Turtle.NOTE_F_SHARP_5 = 58;
Turtle.NOTE_G_FLAT_5 = 58;
Turtle.NOTE_G_5 = 59;
Turtle.NOTE_G_SHARP_5 = 60;
Turtle.NOTE_A_FLAT_5 = 60;
Turtle.NOTE_A_5 = 61;
Turtle.NOTE_A_SHARP_5 = 62;
Turtle.NOTE_B_FLAT_5 = 62;
Turtle.NOTE_B_5 = 63;
Turtle.NOTE_C_6 = 64;
Turtle.NOTE_C_SHARP_6 = 65;
Turtle.NOTE_D_FLAT_6 = 65;
Turtle.NOTE_D_6 = 66;
Turtle.NOTE_D_SHARP_6 = 67;
Turtle.NOTE_E_FLAT_6 = 67;
Turtle.NOTE_E_6 = 68;
Turtle.NOTE_F_6 = 69;
Turtle.NOTE_F_SHARP_6 = 70;
Turtle.NOTE_G_FLAT_6 = 70;
Turtle.NOTE_G_6 = 71;
Turtle.NOTE_G_SHARP_6 = 72;
Turtle.NOTE_A_FLAT_6 = 72;
Turtle.NOTE_A_6 = 73;
Turtle.NOTE_A_SHARP_6 = 74;
Turtle.NOTE_B_FLAT_6 = 74;
Turtle.NOTE_B_6 = 75;
Turtle.NOTE_C_7 = 76;
Turtle.NOTE_C_SHARP_7 = 77;
Turtle.NOTE_D_FLAT_7 = 77;
Turtle.NOTE_D_7 = 78;
Turtle.NOTE_D_SHARP_7 = 79;
Turtle.NOTE_E_FLAT_7 = 79;
Turtle.NOTE_E_7 = 80;
Turtle.NOTE_F_7 = 81;
Turtle.NOTE_F_SHARP_7 = 82;
Turtle.NOTE_G_FLAT_7 = 82;
Turtle.NOTE_G_7 = 83;
Turtle.NOTE_G_SHARP_7 = 84;
Turtle.NOTE_A_FLAT_7 = 84;
Turtle.NOTE_A_7 = 85;
Turtle.NOTE_A_SHARP_7 = 86;
Turtle.NOTE_B_FLAT_7 = 86;
Turtle.NOTE_B_7 = 87;
Turtle.NOTE_C_8 = 88;

Turtle.SOUND_OFF = 0;
Turtle.SOUND_BEEP = 1;
Turtle.SOUND_RANDOM = 2;
Turtle.SOUND_SIREN = 3;
Turtle.SOUND_ENGINE = 4;
Turtle.SOUND_ROBOT = 5;
Turtle.SOUND_MARCH = 6;
Turtle.SOUND_BIRTHDAY = 7;
Turtle.SOUND_DIBIDIBIDIP = 8;
Turtle.SOUND_GOOD_JOB = 9;

Turtle.MOTION_NONE = 0;
Turtle.MOTION_MOVE_FORWARD = 1;
Turtle.MOTION_MOVE_BACKWARD = 2;
Turtle.MOTION_TURN_LEFT = 3;
Turtle.MOTION_TURN_RIGHT = 4;
Turtle.MOTION_PIVOT_LEFT_HEAD = 5;
Turtle.MOTION_PIVOT_LEFT_TAIL = 6;
Turtle.MOTION_PIVOT_RIGHT_HEAD = 7;
Turtle.MOTION_PIVOT_RIGHT_TAIL = 8;
Turtle.MOTION_SWING_LEFT_HEAD = 9;
Turtle.MOTION_SWING_LEFT_TAIL = 10;
Turtle.MOTION_SWING_RIGHT_HEAD = 11;
Turtle.MOTION_SWING_RIGHT_TAIL = 12;

Turtle.UNIT_NONE = 0;
Turtle.UNIT_CM_DEG = 1;
Turtle.UNIT_SEC = 2;
Turtle.UNIT_PULSE = 3;

Turtle.create = function() {
	return new Turtle();
};

function Chat() {
	var self = this;
	var socket = undefined;
	
	var __send = function(data) {
		if(socket) {
			try {
				socket.send(JSON.stringify(data));
			} catch (e) {
			}
		}
	};
	
	self.connect = function(url, name, receiveCallback, openCallback, closeCallback) {
		if(('WebSocket' in window) && url) {
			if(url.indexOf(':') < 0) url += ':60000';
			
			self.disconnect();
			try {
				var sock = new WebSocket('ws://' + url);
				sock.binaryType = 'arraybuffer';
				socket = sock;
				sock.onopen = function() {
					sock.onmessage = function(message) {
						try {
							var data = JSON.parse(message.data);
							if(data.command === 'send' || data.command === 'broadcast') {
								if(typeof receiveCallback == 'function') {
									receiveCallback(data.message, data.value);
								}
							}
						} catch (e) {
						}
					};
					sock.onclose = function() {
						socket = undefined;
						if(typeof closeCallback == 'function') {
							closeCallback();
						}
					};
					__send({
						command: 'register',
						name: name
					});
					if(typeof openCallback == 'function') {
						openCallback();
					}
				};
			} catch (e) {
			}
		}
	};
	
	self.disconnect = function() {
		if(socket) {
			socket.close();
			socket = undefined;
		}
	};
	
	self.send = function(target, key, value) {
		if(value === undefined) value = 0;
		__send({
			command: 'send',
			to: target,
			message: key,
			value: value
		});
	};
	
	self.broadcast = function(key, value) {
		if(value === undefined) value = 0;
		__send({
			command: 'broadcast',
			message: key,
			value: value
		});
	};
}

Chat.create = function() {
	var chat = new Chat();
	__chats.push(chat);
	return chat;
};